Package: TesiproV
Type: Package
Title: Reliability Analysis Methods for Structural Engineering
Version: 0.9.6
Date: 2026-04-22
Authors@R: c(person("Konstantin", "Nille-Hauf", role = "aut",email = "konstantin.nillehauf@googlemail.com"),
            person("Tânia", "Feiri", role = "aut",email="tania.feiri@tu-dortmund.de"),
            person("Marcus","Ricker",role="aut",email="marcus.ricker@tu-dortmund.de"),
            person("Til","Lux",role= c("aut", "cre"),email="til.lux@tu-dortmund.de")
            )
Maintainer: Til Lux <til.lux@tu-dortmund.de>
Description: Calculate the failure probability of civil engineering problems with Level I up to Level III Methods. Have fun and enjoy. References: Spaethe (1991, ISBN:3-211-82348-4) "Die Sicherheit tragender Baukonstruktionen", AU,BECK (2001) "Estimation of small failure probabilities in high dimensions by subset simulation." <doi:10.1016/S0266-8920(01)00019-4>, Breitung (1989) "Asymptotic approximations for probability integrals." <doi:10.1016/0266-8920(89)90024-6>.
URL: https://mvb.ab.tu-dortmund.de/
Encoding: UTF-8
License: MIT + file LICENSE
RoxygenNote: 7.3.3
Imports: stats, future, future.apply, digest, nloptr, methods, ggplot2,
        gridExtra, pracma
Suggests: testthat, knitr, rmarkdown, evd, brms
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2026-04-22 12:57:15 UTC; smtiluxx
Author: Konstantin Nille-Hauf [aut],
  Tânia Feiri [aut],
  Marcus Ricker [aut],
  Til Lux [aut, cre]
Repository: CRAN
Date/Publication: 2026-04-22 13:20:21 UTC
Built: R 4.5.2; ; 2026-04-22 16:58:56 UTC; unix

Package: TesiproV
Type: Package
Title: Reliability Analysis Methods for Structural Engineering
Version: 0.9.6
Date: 2026-04-22
Authors@R: c(person("Konstantin", "Nille-Hauf", role = "aut",email = "konstantin.nillehauf@googlemail.com"),
            person("Tânia", "Feiri", role = "aut",email="tania.feiri@tu-dortmund.de"),
            person("Marcus","Ricker",role="aut",email="marcus.ricker@tu-dortmund.de"),
            person("Til","Lux",role= c("aut", "cre"),email="til.lux@tu-dortmund.de")
            )
Maintainer: Til Lux <til.lux@tu-dortmund.de>
Description: Calculate the failure probability of civil engineering problems with Level I up to Level III Methods. Have fun and enjoy. References: Spaethe (1991, ISBN:3-211-82348-4) "Die Sicherheit tragender Baukonstruktionen", AU,BECK (2001) "Estimation of small failure probabilities in high dimensions by subset simulation." <doi:10.1016/S0266-8920(01)00019-4>, Breitung (1989) "Asymptotic approximations for probability integrals." <doi:10.1016/0266-8920(89)90024-6>.
URL: https://mvb.ab.tu-dortmund.de/
Encoding: UTF-8
License: MIT + file LICENSE
RoxygenNote: 7.3.3
Imports: stats, future, future.apply, digest, nloptr, methods, ggplot2,
        gridExtra, pracma
Suggests: testthat, knitr, rmarkdown, evd, brms
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2026-04-22 12:57:15 UTC; smtiluxx
Author: Konstantin Nille-Hauf [aut],
  Tânia Feiri [aut],
  Marcus Ricker [aut],
  Til Lux [aut, cre]
Repository: CRAN
Date/Publication: 2026-04-22 13:20:21 UTC
Built: R 4.6.0; ; 2026-04-26 10:48:01 UTC; unix

# Generated by roxygen2: do not edit by hand

export(FORM)
export(MC_CRUDE)
export(MC_IS)
export(MC_SubSam)
export(MVFOSM)
export(PARAM_BASEVAR)
export(PARAM_DETVAR)
export(PARAM_LSF)
export(PROB_BASEVAR)
export(PROB_DETVAR)
export(PROB_MACHINE)
export(SORM)
export(SYS_LSF)
export(SYS_PARAM)
export(SYS_PROB)
export(dlt)
export(dst)
export(plt)
export(pst)
export(qlt)
export(qst)
export(rlt)
export(rst)
exportClasses(PARAM_BASEVAR)
exportClasses(PARAM_DETVAR)
exportClasses(PARAM_LSF)
exportClasses(PROB_BASEVAR)
exportClasses(PROB_DETVAR)
exportClasses(PROB_MACHINE)
exportClasses(SYS_LSF)
exportClasses(SYS_PARAM)
exportClasses(SYS_PROB)
import(digest)
import(future)
import(future.apply)
import(ggplot2)
import(gridExtra)
import(nloptr)
import(parallel)
import(pracma)
importFrom(future,availableCores)
importFrom(future,multicore)
importFrom(future,multisession)
importFrom(future,plan)
importFrom(future,sequential)
importFrom(methods,"new")
importFrom(parallel,clusterEvalQ)
importFrom(parallel,clusterExport)
importFrom(parallel,clusterSetRNGStream)
importFrom(parallel,makeCluster)
importFrom(stats,dnorm)
importFrom(stats,pnorm)
importFrom(stats,qnorm)
importFrom(stats,rnorm)
importFrom(utils,str)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# TesiproV

<!-- badges: start -->
<!-- badges: end -->

The goal of TesiproV is to provide a flexible framework for
probabilistic reliability analysis of structural systems. It implements
several state-of-the-art algorithms (FORM, SORM, MVFOSM and Monte-Carlo
methods) and supports parametric studies through reference classes such
as `SYS_PROB`, `SYS_PARAM`, and `SYS_LSF`.

## Installation

The stable release will be available on CRAN soon. Until then you can
install directly from our GitLab server:

``` r
# install.packages("remotes")
remotes::install_gitlab("ls-massivbau/tesiprov", host = "gitlab.tu-dortmund.de")
```

You can install the development version of TesiproV via the gitLab
Server of the TU Dortmund University:
<https://gitlab.tu-dortmund.de/ls-massivbau/tesiprov>

If you need any permissions, please contact the project authors (see
below).

## Parallel computing configuration

TesiproV automatically configures parallel execution using the parallel
and future packages. By default up to 4 cores are used for outer
parallelisation (ProbMachines), while inner Monte-Carlo simulations
manage their own threads. You can change this behaviour with environment
variables before loading the package:

``` r
Sys.setenv(TesiproV.max_workers = "8") # allow up to 8 workers
Sys.setenv(TesiproV.future = "multisession") # choose backend explicitly
library(TesiproV)
```

For details see the help page of `.onLoad()` or your system’s
documentation.

## Example

This is a basic example which shows how to define a limit-state function
and run a FORM analysis:

``` r
library("TesiproV")
Var1 <- PROB_BASEVAR(Id = 1, Name = "X1", DistributionType = "norm", Mean = 0.25, Sd = 1) # kN/m²
Var2 <- PROB_BASEVAR(Id = 2, Name = "X2", DistributionType = "norm", Mean = 0.25, Sd = 1) # m

lsf <- SYS_LSF(vars = list(Var1, Var2), name = "UQLab 2d_hat")
lsf$func <- function(X1, X2) {
  return(20 - (X1 - X2)^2 - 8 * (X1 + X2 - 4)^3)
}

form <- PROB_MACHINE(name = "FORM Rack.-Fieß.", fCall = "FORM")

ps <- SYS_PROB(
  sys_input = list(lsf),
  probMachines = list(form)
)
ps$runMachines()
```

``` r
ps$beta_single
#>                     UQLab 2d_hat
#> FORM Rack.-Fieß. 3.4345652859862
```

## Parametric study example

A simple sweep over different mean values can be performed with
`PARAM_BASEVAR` and `SYS_PARAM`:

``` r
pvar <- PARAM_BASEVAR(
  Name = "E_mod",
  DistributionType = "norm",
  ParamType = "Mean",
  ParamValues = c(30e3, 35e3, 40e3)
)

lsf_param <- SYS_LSF(vars = list(pvar), name = "ParamStudy")
lsf_param$func <- function(E_mod) {
  E_mod / 1000 - 30
}

machine_form <- PROB_MACHINE(name = "FORM", fCall = "FORM")

ps_param <- SYS_PARAM(
  sys_input = list(lsf_param),
  probMachines = list(machine_form)
)
ps_param$runMachines()
print(ps_param$beta_params)
```

## Recent changes

Version 0.9.5 introduces major refactoring:

- Improved consistency between probabilistic objects (`SYS_PROB`,
  `SYS_PARAM`, …).
- Corrected implementation of the Gumbel distribution in `$prepare()`.
- Enhanced error handling in `$check()` for limit-state functions.
- Automatic parallel plan configuration via `.onLoad()` with capped
  worker pool (default 4 cores).

For a complete changelog see **NEWS.md**.

## Further information

For more, check out the vignette! There are plenty more demonstrations
and detailed explanations for each object class.

You can open it directly in R:

``` r
vignette("TesiproV")
#> Warning: vignette 'TesiproV' not found
```

# Contact & Citation

If you use TesiproV for research or teaching, please cite it as:

> Nille-Hauf K., Lux T., Feiri T., Ricker M. (2026): *TesiproV —
> Probabilistic Reliability Analysis Framework*. TU Dortmund University.

For questions or bug reports please contact
<massivbau.ab@tu-dortmund.de>.

------------------------------------------------------------------------

## License

TesiproV is licensed under the [MIT
License](https://opensource.org/licenses/MIT).  
You are free to use, modify and redistribute the software provided that
the original copyright notice and license terms are retained.

© 2021-2026 Konstantin Nille-Hauf, Til Lux, Tania Feiri, Marcus Ricker —
Hochschule Biberach / TU Dortmund University.

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----eval = FALSE-------------------------------------------------------------
# install.packages("TesiproV")

## ----eval = FALSE-------------------------------------------------------------
# remotes::install_gitlab("ls-massivbau/tesiprov", host = "gitlab.tu-dortmund.de")

## -----------------------------------------------------------------------------
library("TesiproV")

## -----------------------------------------------------------------------------
var_fck <- PROB_BASEVAR(
  Id = 1,
  Name = "f_ck",
  Description = "Concrete compressive strength",
  DistributionType = "lnorm",
  Mean = 47.35,
  Sd = 5.86
)

## -----------------------------------------------------------------------------
var.f_ck <- PROB_BASEVAR(Id = 1, Name = "f_ck", Description = "Char. Betondruckfestigkeit", Package = "TesiproV", DistributionType = "lt", DistributionParameters = c(3.85, 0.09, 3, 10))

## -----------------------------------------------------------------------------
pvar_d <- PARAM_BASEVAR(
  Id = 1,
  Name = "d",
  Description = "Effective depth",
  DistributionType = "norm",
  Sd = 10,
  ParamType = "Mean",
  ParamValues = c(seq(200, 800, 50), seq(1000, 3000, 100)) + 10
)

## -----------------------------------------------------------------------------
var.gamma_c <- PROB_DETVAR(Id = 1, Name = "gamma_c", Description = "PSF Concrete", Value = 1.5)

## -----------------------------------------------------------------------------
var.V_Ed <- PARAM_DETVAR(Id = 1, Name = "V_Ed", Description = "applied shear force", ParamValues = c(1, 2, 3, 4, seq(5, 10, 0.5)))

## -----------------------------------------------------------------------------
var1 <- PROB_BASEVAR(Name = "E", Description = "Effect", DistributionType = "norm", Mean = 10, Sd = 1)
var2 <- PROB_BASEVAR(Name = "R", Description = "Resistance", DistributionType = "norm", Mean = 15, Sd = 1.5)


lsf <- SYS_LSF(vars = list(var1, var2), name = "LSF XY")
lsf$func <- function(E, R) {
  R - E
}
# you can run this check to see if all transformations of the variables worked, all needed data is available and the set of vars fits to the limit state function
lsf$check()

## -----------------------------------------------------------------------------
machine <- PROB_MACHINE(name = "MVFOSM", fCall = "MVFOSM", options = list("isExpression" = FALSE, "h" = 0.1))

## -----------------------------------------------------------------------------
machine <- PROB_MACHINE(name = "FORM Rack.-Fies.", fCall = "FORM", options = list("n_optim" = 20, "loctol" = 0.001, "optim_type" = "rackfies"))

machine <- PROB_MACHINE(name = "FORM Lagrange.", fCall = "FORM", options = list("optim_type" = "auglag"))

## -----------------------------------------------------------------------------
machine <- PROB_MACHINE(name = "SORM", fCall = "SORM")

## -----------------------------------------------------------------------------
machine <- PROB_MACHINE(name = "MC CoV 0.05", fCall = "MC_CRUDE", options = list("n_max" = 1e6, "cov_user" = 0.05))

## ----eval = FALSE-------------------------------------------------------------
# help("MC_IS")

## -----------------------------------------------------------------------------
machine <- PROB_MACHINE(name = "MC IS", fCall = "MC_IS", options = list("cov_user" = 0.05, "n_max" = 300000))

## -----------------------------------------------------------------------------
machine <- PROB_MACHINE(name = "MC Subset Sampling", fCall = "MC_SubSam", options = list("variance" = "uniform"))

## ----results="hide"-----------------------------------------------------------
var1 <- PROB_BASEVAR(Name = "E", Description = "Effect", DistributionType = "norm", Mean = 10, Sd = 1)
var2 <- PROB_BASEVAR(Name = "R", Description = "Resistance", DistributionType = "norm", Mean = 15, Sd = 1.5)


lsf <- SYS_LSF(vars = list(var1, var2), name = "LSF XY")
lsf$func <- function(E, R) {
  R - E
}
# you can run this check to see if all transformations of the variables worked, all needed data is available and the set of vars fits to the limit state function
lsf$check()


machine <- PROB_MACHINE(name = "FORM Rack.-Fieß.", fCall = "FORM", options = list("n_optim" = 20, "loctol" = 0.001, "optim_type" = "rackfies"))

ps <- SYS_PROB(
  sys_input = list(lsf),
  probMachines = list(machine),
  debug.level = 0
)

ps$runMachines()

## -----------------------------------------------------------------------------
ps$beta_single

## -----------------------------------------------------------------------------
# Systemberechnungn (System aus zwei gleichen LSF´s)
ps_sys <- SYS_PROB(
  sys_input = list(lsf, lsf),
  probMachines = list(machine),
  sys_type = "serial"
)

ps_sys$runMachines()
ps_sys$calculateSystemProbability()

## -----------------------------------------------------------------------------
ps_sys$beta_sys

## -----------------------------------------------------------------------------
# Force in kN
pvar_F_mod <- PARAM_BASEVAR(
  Name = "Force", DistributionType = "gumbel", Sd = 2, Package = "evd",
  ParamType = "Mean", ParamValues = c(17, 18, 19, 20) 
)

# yield strength in kN, for Mean the mean of the unshifted distribution must be specified.
fy <- PROB_BASEVAR(
  Name = "fy", DistributionType = "slnorm", Package = "brms", Mean = 265000, Sd = 25000, x0 = 160000
)

# plastic moment of resistance in m³ - considered deterministic in this example
W_pl <- 2.5e-4 
# length of cantilever beam in m - considered deterministic in this example
l    <- 2     

lsf_param <- SYS_LSF(vars = list(pvar_F_mod, fy), name = "Spaethe (1992) - Beispiel 2")
lsf_param$func <- function(Force,fy) {
  return(W_pl*fy-l*Force)
}

machine_form <- PROB_MACHINE(name = "FORM", fCall = "FORM")

ps_param <- SYS_PARAM(
  sys_input = list(lsf_param),
  probMachines = list(machine_form)
)

ps_param$runMachines()
print(ps_param$beta_params)

## -----------------------------------------------------------------------------
para_values <- c(17, 18, 19, 20) 
beta_values <- ps_param$beta_params
plot(para_values, beta_values, type = "l", xlab = "Force F [kN]", ylab = "Reliability index β [-]")

## ----eval=FALSE---------------------------------------------------------------
# MC_IS(..., seed = 123, use_threads = 1)
# MC_IS(..., seed = 123, use_threads = 8)

## ----eval=FALSE---------------------------------------------------------------
# MC_IS(..., backend = "parallel", use_threads = 8)

## ----eval=FALSE---------------------------------------------------------------
# library(future)
# plan(multicore, workers = 8)
# 
# MC_IS(..., backend = "future")

## ----eval=FALSE---------------------------------------------------------------
# library(future)
# plan(multisession, workers = 8)
# 
# MC_IS(..., backend = "future")If no future plan is set, execution will be sequential.

## ----eval=FALSE---------------------------------------------------------------
# 250,000 -- 500,000

## ----eval=FALSE---------------------------------------------------------------
# library(future)
# plan(multicore, workers = 16)
# 
# res <- MC_IS(
#   lsf = my_lsf,
#   lDistr = my_distr,
#   n_batch = 300000,
#   backend = "future",
#   seed = 12345
# )

## ----eval = FALSE-------------------------------------------------------------
# ps <- SYS_PARAM(
#   sys_input = list(lsf),
#   probMachines = list(machine)
# )
# 
# ps$runMachines()
# 
# local_beta <- ps$beta
# local_runtime <- ps$res_single[[1]][[1]]$runtime
# 
# # define path with setwd("...")
# ps$printResults("result_file") # Prints a report file in .txt format with all informations about the calculation
# ps$saveProject(4, "project_file") # stores the project in a file

## ----eval = FALSE-------------------------------------------------------------
# readRDS("example_proj_res.rds")
# load("example_proj.RData")

## ----results="hide"-----------------------------------------------------------
Var1 <- PROB_BASEVAR(Id = 1, Name = "X1", DistributionType = "norm", Mean = 0.25, Sd = 1) # kN/m²
Var2 <- PROB_BASEVAR(Id = 2, Name = "X2", DistributionType = "norm", Mean = 0.25, Sd = 1) # m

lsf <- SYS_LSF(vars = list(Var1, Var2), name = "UQLab 2d_hat")
lsf$func <- function(X1, X2) {
  return(20 - (X1 - X2)^2 - 8 * (X1 + X2 - 4)^3)
}

form <- PROB_MACHINE(name = "FORM Rack.-Fieß.", fCall = "FORM")

ps <- SYS_PROB(
  sys_input = list(lsf),
  probMachines = list(form)
)
ps$runMachines()

## -----------------------------------------------------------------------------
ps$beta_single

## ----results ='hide'----------------------------------------------------------
library("evd")
X1 <- PROB_BASEVAR(Id = 1, Name = "Z", DistributionType = "norm", Mean = 100, Cov = 0.04) # kN/m²
X2 <- PROB_BASEVAR(Id = 2, Name = "Fy", DistributionType = "lnorm", Mean = 40, Cov = 0.1) # m
X3 <- PROB_BASEVAR(Id = 2, Name = "M", DistributionType = "gumbel", Package = "evd", Mean = 2000, Cov = 0.1) # m

lsf <- SYS_LSF(vars = list(X1, X2, X3), name = "Nowak & Collins Exp5.11_1")
lsf$func <- function(M, Fy, Z) {
  return(Z * Fy - M)
}

form <- PROB_MACHINE(name = "FORM Rack.-Fieß.", fCall = "FORM")

ps <- SYS_PROB(
  sys_input = list(lsf),
  probMachines = list(form)
)
ps$runMachines()

## -----------------------------------------------------------------------------
ps$beta_single

## ----results="hide"-----------------------------------------------------------
Var1 <- PROB_BASEVAR(Id = 1, Name = "X1", DistributionType = "norm", Mean = 0.25, Sd = 1) # kN/m²
Var2 <- PROB_BASEVAR(Id = 2, Name = "X2", DistributionType = "norm", Mean = 0.25, Sd = 1) # m

lsf1 <- SYS_LSF(vars = list(Var1, Var2), name = "UQLab 2d_hat")
lsf1$func <- function(X1, X2) {
  return(20 - (X1 - X2)^2 - 8 * (X1 + X2 - 4)^3)
}


X1 <- PROB_BASEVAR(Id = 4, Name = "Z", DistributionType = "norm", Mean = 100, Cov = 0.04) # kN/m²
X2 <- PROB_BASEVAR(Id = 5, Name = "Fy", DistributionType = "lnorm", Mean = 40, Cov = 0.1) # m
X3 <- PROB_BASEVAR(Id = 6, Name = "M", DistributionType = "gumbel", Package = "evd", Mean = 2000, Cov = 0.1) # m

lsf2 <- SYS_LSF(vars = list(X1, X2, X3), name = "Nowak & Collins Exp5.11_1")
lsf2$func <- function(M, Fy, Z) {
  return(Z * Fy - M)
}

form <- PROB_MACHINE(name = "FORM Rack.-Fieß.", fCall = "FORM")

ps <- SYS_PROB(
  sys_input = list(lsf1, lsf2),
  sys_type = "serial",
  probMachines = list(form)
)
ps$runMachines()

## -----------------------------------------------------------------------------
ps$beta_single

ps$calculateSystemProbability()
ps$beta_sys

## ----eval = FALSE-------------------------------------------------------------
# params <- list("cov_user" = 0.05, "n_max" = 50000)
# ps$calculateSystemProbability("MC", params)
# 
# params <- list("cov_user" = 0.05, "n_max" = 50000)
# ps$calculateSystemProbability("MCIS", params)
# 
# ps$calculateSystemProbability("MCSUS", params)

## ----results="hide", warning=FALSE--------------------------------------------
var.d <- PARAM_BASEVAR(Id = 1, Name = "d", Description = "Stat. Nutzhöhe", DistributionType = "norm", Sd = 10, ParamType = "Mean", ParamValues = c(seq(200, 800, 50), seq(1000, 3000, 100)) + 10)
pre.d <- c(seq(200, 800, 50), seq(1000, 3000, 100))

var.f_ck <- PROB_BASEVAR(Id = 2, Name = "f_ck", Description = "Char.Betondruckfestigkeit", Package = "TesiproV", DistributionType = "lt", DistributionParameters = c(3.85, 0.09, 3, 10))
pre.f_ck <- 35

var.f_ywk <- PROB_BASEVAR(Id = 3, Name = "f_ywk", Description = "Zugfestigkeit Bügelbewehrung", DistributionType = "norm", Mean = 560, Cov = 0.02)
pre.f_ywk <- 500

var.f_yk <- PROB_BASEVAR(Id = 4, Name = "f_yk", Description = "Zugfestigkeit Längsbewehrung", DistributionType = "norm", Mean = 560, Sd = 30)
pre.f_yk <- 500

var.rho_l <- PROB_BASEVAR(Id = 5, Name = "rho_l", Description = "Biegebewehrungsgrad", DistributionType = "norm", Mean = 0.01, Cov = 0.02)
pre.rho_l <- var.rho_l$Mean

var.c_D <- PROB_BASEVAR(Id = 6, Name = "c_D", Description = "Stützendurchmesser", DistributionType = "norm", Mean = 500 + 0.003 * 500, Sd = 4 + 0.006 * 500)
pre.c_D <- 500

var.theta_c <- PROB_BASEVAR(Id = 7, Name = "theta_c", Description = "Modellunsicherheit c nach DIBT", DistributionType = "lnorm", Mean = 1.0644, Cov = 0.1679)

var.gamma_c <- PROB_DETVAR(Id = 8, Name = "gamma_c", Description = "TSB Beton", Value = 1.5)
var.gamma_s <- PROB_DETVAR(Id = 9, Name = "gamma_s", Description = "TSB Stahl", Value = 1.15)
var.alpha_cc <- PROB_DETVAR(Id = 10, Name = "alpha_cc", Description = "Langzeitfaktor AlphaCC", Value = 0.85)

V_Ed <- c(
  0.6412447, 0.8760515, 1.1424185, 1.4399554, 1.7554079, 2.0188587, 2.3003489, 2.5997038, 2.9167825, 3.2514671, 3.6036567, 3.9732630,
  4.3602076, 6.0800504, 7.0424326, 8.0726216, 9.1702994, 10.3351829, 11.5670173, 12.8655715, 14.2306347, 15.6620132, 17.1595283, 18.7230145,
  20.3523175, 22.0472935, 23.8078076, 25.6337332, 27.5249511, 29.4813487, 31.5028193, 33.5892621, 35.7405811, 37.9566849
)
var.V_Ed <- PARAM_DETVAR(Id = 11, Name = "V_Ed", Description = "Einwirkende Querkraft", ParamValues = V_Ed)

lsf1 <- SYS_LSF(
  vars = list(
    var.d, var.f_ck, var.f_yk, var.f_ywk, var.rho_l, var.c_D,
    var.theta_c, var.V_Ed, var.gamma_c, var.gamma_s, var.alpha_cc
  ),
  name = "Durchstanzwiderstand ohne Bewehrung"
)
lsf1$func <- function(d, f_ck, f_yk, f_ywk, rho_l, c_D, theta_c, V_Ed, gamma_c, gamma_s, alpha_cc) {
  f_cd <- f_ck * alpha_cc / gamma_c
  f_yd <- f_yk / gamma_s
  f_ywd <- f_ywk / gamma_s

  u_0 <- c_D * pi
  u_1 <- (c_D / 2 + 2 * d) * 2 * pi
  u_0.5 <- (c_D / 2 + 0.5 * d) * 2 * pi

  k <- min(1 + sqrt(200 / d), 2)

  if ((u_0 / d) < 4) {
    C_Rd_c <- 0.18 * (0.1 * u_0 / d + 0.6)
  } else {
    C_Rd_c <- 0.18
  }
  rho_l <- min(rho_l, 0.5 * f_cd / f_yd, 0.02)
  C_Rd_c1 <- C_Rd_c * k * (100 * rho_l * f_ck)^(1 / 3)

  v_Rd_c <- C_Rd_c1
  V_Rd_c <- v_Rd_c * d * u_1 * 1e-6 * theta_c
  return(V_Rd_c - V_Ed)
}

form <- PROB_MACHINE(name = "FORM RF.", fCall = "FORM", options = list("n_optim" = 20, "loctol" = 0.0001, "optim_type" = "rackfies"))

ps <- SYS_PARAM(
  sys_input = list(lsf1),
  probMachines = list(form)
)
ps$runMachines()

## ----results="hide", warning=FALSE--------------------------------------------
d <- pre.d
beta <- unlist(unname(ps$beta_params))
plot(x = d, y = beta, type = "l")

#' @name FORM
#' @title First-Order Reliability Method (FORM)
#' @description
#' First-Order Reliability Method (FORM) for the approximation
#' of failure probabilities in structural reliability analysis.
#'
#' The FORM estimates the probability of failure by transforming
#' the basic random variables into standard normal space
#' and approximating the limit-state function by a first-order
#' (linear) Taylor expansion at the design point.
#'
#' The reliability index is defined as the minimum distance
#' from the origin to the limit-state surface in standard normal space.
#' The probability of failure is then approximated using
#' the standard normal cumulative distribution function.
#'
#' @param lsf Objective function representing the limit-state, e.g. \code{function(R,E){R-E}}.
#'   Supplied automatically by a SYS_ object - do not provide manually.
#' @param lDistr List of distribution objects created by TesiproV. Supplied automatically by
#'   a SYS_ object - do not provide manually.
#' @param n_optim Number of optimization cycles (not required for Lagrangian algorithms).
#' @param loctol Local tolerance for convergence of the solver algorithm.
#' @param optim_type Optimization type: `"auglag"` (Augmented Lagrangian) or `"rackfies"`
#'   (Rackwitz-Fiessler iterative scheme).
#' @param debug.level Verbosity level: 0 = silent, 1 = basic info, 2 = detailed output.
#'
#' @return A list containing:
#'   * `beta` - Hasofer-Lind reliability index
#'   * `pf` - probability of failure
#'   * `x_points` - design point in physical space
#'   * `dy` - gradient vector at design point
#'
#' @references
#' Hasofer, A. M., & Lind, N. C. (1974).
#' An exact and invariant first-order reliability format.
#' \emph{Journal of the Engineering Mechanics Division, ASCE},
#' 100(1), 111-121.
#'
#' Rackwitz, R., & Fieszler, B. (1978).
#' Structural reliability under combined random load sequences.
#' \emph{Computers & Structures}, 9(5), 489-494.
#'
#' @import nloptr
#'
#' @author (C) 2021-2026 K. Nille-Hauf, J.P. Schulze-Ardey, T. Feiri, M. Ricker, T. Lux -- Hochschule Biberach (until 2022),
#' TU Dortmund University - Chair of Structural Concrete (since 2023)
#' @export
#'
#'
FORM <- function(lsf, lDistr, n_optim = 10, loctol = 1e-2, optim_type = "rackfies", debug.level = 0) {
  debug.TAG <- "FORM_OP"
  debug.print(debug.level, debug.TAG, c(optim_type), msg = "FORM Algorithm started with Solvingtype:")
  tic <- proc.time()

  # options(error = recover)

  ## message("[DEBUG] typeof(lsf)=", typeof(lsf))
  ## message("[DEBUG] is.function(lsf)=", is.function(lsf))
  ## message("[DEBUG] class(lsf)=", paste(class(lsf), collapse = ","))

  # --- Default settings -------------------------------------------------
  if (is.null(loctol) || is.na(loctol)) loctol <- 1e-4
  if (is.null(n_optim) || is.na(n_optim)) n_optim <- 20L

  # Local control variables ----------------------------------------------
  n_sim <- 0L # iteration counter
  delta_x <- Inf

  n_vars <- length(lDistr)

  # --- Transformation between physical and standard Gaussian domains ------------#
  gausTrans <- function(a, toGaussian = TRUE) {
    x <- vector(length = n_vars)
    y <- vector(length = n_vars)
    b <- vector(length = n_vars)

    if (toGaussian) {
      # Entspricht Y zu X
      y <- a
      for (i in 1:n_vars) {
        p <- pmin(pmax(stats::pnorm(y[i]), 1e-12), 1 - 1e-12)
        b[i] <- lDistr[[i]][[1]]$q(p)
      }
    } else {
      # Entspricht X zu Y
      x <- a
      for (i in 1:n_vars) {
        x_ <- x[i]
        b[i] <- stats::qnorm(lDistr[[i]][[1]]$p(x_))
      }
    }
    return(b)
  }


  # lsf in gaussian domain
  h <- function(y) {
    return(lsf(gausTrans(y, TRUE)))
  }


  beta <- vector("numeric", n_optim)
  par_u <- matrix(ncol = n_vars, nrow = n_optim)
  par_x <- matrix(ncol = n_vars, nrow = n_optim)
  dy <- matrix(ncol = n_vars, nrow = n_optim)


  ###############################################################################
  ## Augmented Lagrangian variant ----------------------------------------------
  ###############################################################################
  if (optim_type == "auglag") {
    for (i in 1:n_optim) {
      # Zufallsstartvektor erzeugen
      q_rand <- stats::qnorm(stats::runif(n_vars))

      # Standart-Optimierungsfunktion
      unorm <- function(u) {
        sqrt(sum(u^2))
      }
      info.print(debug.TAG, debug.level, c("optim_run_i", "/"), c(i, n_optim))
      res <- nloptr::auglag(
        x0 = q_rand,
        fn = unorm,
        heq = h,
        localtol = loctol
      )

      beta[i] <- res$value
      par_u[i, ] <- res$par
      par_x[i, ] <- gausTrans(res$par, FALSE)
      dy[i, ] <- grad(h, res$par)
    }
    if (res$convergence < 0) {
      warning("Error, Optimization failed!")
      return()
    }

    duration <- proc.time() - tic
    output <- list(
      "beta" = min(beta),
      "pf" = stats::pnorm(-abs(min(beta))),
      "u_points" = par_u[which.min(beta), ],
      "x_points" = par_x[which.min(beta), ],
      "dy" = dy[which.min(beta), ],
      "optim_type" = optim_type,
      "runtime" = duration
    )

    ###############################################################################
    ## Rackwitz-Fiessler iterative scheme ----------------------------------------
    ###############################################################################
  } else if (identical(tolower(optim_type), "rackfies")) {
    n_sim <- 0L
    x <- numeric(n_vars)
    y <- numeric(n_vars)
    dy <- numeric(n_vars)
    m_s <- numeric(n_vars)
    sd_s <- numeric(n_vars)

    ## Safe wrappers for p/d calls ---------------------------------------
    safe_p <- function(fun, x) {
      val <- fun(x)
      if (is.na(val) || val <= 0 || val >= 1) {
        val <- min(max(val, .Machine$double.eps), 1 - .Machine$double.eps)
      }
      val
    }
    safe_d <- function(fun, x) {
      val <- fun(x)
      if (is.na(val) || val == 0) val <- .Machine$double.eps
      val
    }

    ## Numerical differentiation ----------------------------------------
    finite_diff <- function(fun_obj, x, h = 1e-5) {
      dx <- numeric(length(x))
      for (i in seq_along(x)) {
        x_lowerbound <- x
        x_upperbound <- x
        x_lowerbound[i] <- x[i] - h
        x_upperbound[i] <- x[i] + h

        # message("[DEBUG] calling lsf() at point A")
        dx_low <- fun_obj(x_lowerbound)
        # message("[DEBUG] calling lsf() at point B")
        dx_high <- fun_obj(x_upperbound)
        dx[i] <- (dx_high - dx_low) / (2 * h)
      }
      dx
    }

    ## Initial values (mean if available, otherwise median) -------------
    for (i in seq_len(n_vars)) {
      ld <- lDistr[[i]][[1]]

      if (!is.null(ld$mean) && is.finite(ld$mean)) {
        x[i] <- ld$mean
      } else {
        # fallback for heavy-tail distributions (e.g. Log-Student-t)
        x[i] <- ld$q(0.5)
        if (debug.level >= 1) {
          message(sprintf(
            "[FORM] Using median as start value for variable %d (type=%s)",
            i, ld$DistributionType
          ))
        }
      }
    }

    if (debug.level >= 2) {
      message("\n[DEBUG] Initial mean values of basic variables:")
      print(sapply(lDistr, function(ld) {
        c(
          mean = ld[[1]]$mean,
          sd = ld[[1]]$Sd,
          type = ld[[1]]$DistributionType
        )
      }))
    }
    # debug
    # print(str(lDistr[[1]][[1]]))
    # print(str(lDistr[[2]][[1]]))

    # lokale Kopie verhindern Ueberschreibung durch finite_diff()
    lsf_local <- lsf

    repeat {
      sd_s <- m_s <- y <- numeric(n_vars)

      ## Transformation: physical -> standardized -------------------------
      for (i in seq_len(n_vars)) {
        ld <- lDistr[[i]][[1]]
        x_ <- x[i]

        p_val <- safe_p(ld$p, x_)
        d_val <- safe_d(ld$d, x_)

        sd_s[i] <- stats::dnorm(stats::qnorm(p_val)) / d_val
        m_s[i] <- x[i] - sd_s[i] * stats::qnorm(p_val)

        y[i] <- (x[i] - m_s[i]) / sd_s[i]

        # Debug information
        if (debug.level >= 2) {
          message(sprintf(
            "[DEBUG][%d] p=%g d=%g sd_s=%g m_s=%g y=%g",
            i, p_val, d_val, sd_s[i], m_s[i], y[i]
          ))
        }
      }

      h_y <- lsf_local(x)

      # message("[DEBUG] calling lsf() at point D")
      dy <- finite_diff(lsf_local, x) * sd_s

      # --- Richtungscosinus alpha --------------------------------------
      alpha <- numeric(n_vars)

      for (i in seq_len(n_vars)) {
        denom <- sqrt(sum(dy^2))
        alpha[i] <- ifelse(denom == 0, -dy[i], -dy[i] / denom)
      }

      beta <- (h_y - sum(y * dy)) / sqrt(sum(dy^2))

      if (!is.finite(beta)) {
        warning(sprintf(
          "[FORM/rackfies]: Beta not finite! h_y=%s sum(y*dy)=%s",
          h_y, sum(y * dy)
        ))
        beta <- NA_real_
      }

      x_neu <- numeric(n_vars)

      for (i in seq_len(n_vars)) {
        x_neu[i] <- if (!is.na(alpha[i]) && !is.na(sd_s[i]) && !is.na(m_s[i])) {
          m_s[i] + alpha[i] * sd_s[i] * beta
        } else {
          x[i]
        }
      }

      delta_x <- sqrt(sum((x_neu - x)^2))

      x <- x_neu
      n_sim <- n_sim + 1L

      debug.print(
        debug.level, debug.TAG, c("n_sim", "beta", "delta_x", "loctol"),
        c(n_sim, beta, delta_x, loctol)
      )

      info.print(
        debug.TAG, debug.level, c("n_sim", "beta", "delta_x", "loctol"),
        c(n_sim, beta, delta_x, loctol)
      )

      if ((isTRUE(delta_x <= loctol)) || isTRUE(n_sim > n_optim)) break
    } # repeat loop end

    duration <- proc.time() - tic

    output <- list(
      method = "FORM",
      beta = beta,
      pf = stats::pnorm(-beta),
      x_points = x,
      dy = dy,
      alpha = alpha,
      diff = delta_x,
      n_sim = n_sim,
      optim_type = optim_type,
      runtime = duration[1:5]
    )
  }
  debug.print(debug.level, debug.TAG, c(duration), msg = "\nFORM Algorithm finished in  [s]: ")

  return(output)
}

#'
#' @title internal  Helper function to debug more easy
#' @param infoLevel If 0, no Output (just Errors), if 1 little output, if 2 bigger output
#'
#' @param flag Parse additonal info
#' @param values If you check variables then post this into values
#' @param msg here add some extra msg
#' @param type Type can be "INFO" or "ERROR"
#' @param step_interval Optional integer specifying
#'   the interval at which debug messages are printed.
#' @param iter Current iteration index used for
#'   conditional debug output.
#' @author (C) 2021-2026 K. Nille-Hauf, T. Feiri, M. Ricker, T. Lux -- Hochschule Biberach (until 2022),
#' TU Dortmund University - Chair of Structural Concrete (since 2023)
#' @keywords internal
debug.print <- function(infoLevel,
                        flag = "",
                        values = character(0),
                        msg = "",
                        type = "INFO",
                        step_interval = 10,
                        iter = NA) {
  if (infoLevel == 0 && type != "ERROR") {
    return(invisible(NULL))
  }

  # nur alle n Schritte loggen bei Level=1
  if (infoLevel == 1 && is.numeric(iter) && !is.na(iter) && (iter %% step_interval != 0)) {
    return(invisible(NULL))
  }

  # Zeitstempel fuer jede Ausgabe
  t <- format(Sys.time(), "%H:%M:%OS3")

  # Ausgabe vorbereiten
  out_msg <- paste0(
    "#", type,
    if (flag != "") paste(" FROM ", flag),
    " - ", t,
    "\t", msg,
    if (length(values) > 0) paste0(" ", paste(values, collapse = " ")) else "",
    "\n"
  )

  # Ausgabeart: INFO oder ERROR
  if (type == "ERROR") {
    warning(out_msg)
  } else {
    message(out_msg)
  }
}

debug.print.getDebugLevel <- function() {
  return(0)
}


info.print <- function(tag, verbose, varnames, values) {
  if (verbose == 2) {
    cat("\r#", tag, ": ", paste(varnames, ":", values, "\t"))
  }
}


#' Test whether an object is *empty*
#'
#' This lightweight helper mimics the behaviour of `pracma::isempty()` but
#' depends only on base R.  An object is considered *empty* when it contains
#' no data:
#'   * `NULL`
#'   * an atomic vector, matrix, array or factor of length 0
#'   * a list (including named lists) of length 0
#'   * a data frame with zero rows (`nrow == 0`)
#'
#' The function is deliberately strict - environments, functions, S4/R6
#' objects, etc. are **not** treated as empty.  If you need additional
#' classes to be regarded as empty you can extend the internal checks.
#'
#' @param x Any R object.
#' @return A logical scalar (`TRUE` if `x` is empty, `FALSE` otherwise).
#' @keywords internal
#' @examples
#' \dontrun{
#' is_empty(NULL) # TRUE
#' is_empty(numeric(0)) # TRUE
#' is_empty(list()) # TRUE
#' is_empty(data.frame()) # TRUE
#' is_empty(1:5) # FALSE
#' is_empty(list(a = 1)) # FALSE
#' is_empty(data.frame(x = 1)) # FALSE
#' }

is_empty <- function(x) {
  ## 1) NULL -------------------------------------------------------------
  if (is.null(x)) {
    return(TRUE)
  }

  ## 2) Character vectors - empty strings count as empty -----------------
  if (is.character(x)) {
    return(length(x) == 0L || all(nchar(x) == 0L))
  }

  ## 3) Atomic, factor, matrix, array ------------------------------------
  if (is.atomic(x) || is.factor(x) || is.matrix(x) || is.array(x)) {
    return(length(x) == 0L)
  }

  ## 4) data.frame -------------------------------------------------------
  if (inherits(x, "data.frame")) {
    return(nrow(x) == 0L)
  }

  ## 5) Lists (including named lists) ------------------------------------
  if (is.list(x)) {
    return(length(x) == 0L)
  }

  ## 6) Anything else (functions, environments, R6/S4 objects,...) ------
  FALSE
}

#' Helper: is a value missing (NULL, NA, zero-length, or empty string)
#' Used inside validateArgs() to avoid the missing value where TRUE/FALSE needed
#' problem.
#' @param x any R object
#' @return TRUE if x should be treated as missing, FALSE otherwise
#'
#' @keywords internal
is_missing <- function(x) {
  if (is.null(x)) {
    return(TRUE)
  }
  if (length(x) == 0L) {
    return(TRUE)
  }
  if (is.atomic(x) && any(is.na(x))) {
    return(TRUE)
  }
  if (is.character(x) && any(nchar(x) == 0L)) {
    return(TRUE)
  }
  FALSE
}

#' Log-Student t Distribution
#'
#' Density, distribution function, quantile function and random generation
#' for the log-Student t distribution.
#'
#' The cumulative distribution function is given by
#' \deqn{
#' F_X(x) =
#' F_{t_{\nu}}\left(
#' \frac{\ln(X/m)}{s}
#' \sqrt{\frac{n}{n+1}}
#' \right),
#' }
#' where \eqn{F_{t_{\nu}}(\cdot)} denotes the cumulative distribution
#' function of a log-Student t distribution with \eqn{\nu} degrees of freedom.
#'
#' The log-Student t distribution arises as the marginal
#' distribution of a log-normal distributed variable with conjugate
#' log-normal-gamma prior uncertainty in \eqn{m} and \eqn{s}.
#'
#' The hyperparameter vector \code{hyper.param} is defined as
#' \deqn{(m, s, n, \nu)}
#' where
#' \itemize{
#'   \item \eqn{m} is the mean value of an equivalent sample of size \eqn{n},
#'   \item \eqn{s} is the empirical standard deviation of an equivalent sample
#'         of size \eqn{\nu + 1},
#'   \item \eqn{\nu} denotes the degrees of freedom of the Student t distribution.
#' }
#'
#' For details see:
#' \url{https://www.jcss-lc.org/publications/jcsspmc/concrete.pdf}
#'
#' @name LogStudentT
#' @aliases logStudentT lt
#'
#' @param x,q Vector of quantiles.
#' @param p Vector of probabilities.
#' @param n_vals Number of random values to generate.
#' @param hyper.param Numeric vector \eqn{(m, s, n, \nu)}.
#'
#' @return
#' \itemize{
#'   \item \code{dlt()} returns the density.
#'   \item \code{plt()} returns the distribution function.
#'   \item \code{qlt()} returns the quantile function.
#'   \item \code{rlt()} generates random deviates.
#' }
#'
#' @examples
#' hp <- c(3.4, 0.14, 3, 10)
#' #' dlt(0.5, hp)
#'
#' @author
#' (C) 2021-2026 K. Nille-Hauf, T. Feiri, M. Ricker, T. Lux --
#' Hochschule Biberach (until 2022), TU Dortmund University - Chair of Structural Concrete (since 2023)

#' @rdname LogStudentT
#' @export
dlt <- function(x, hyper.param){
  m <- hyper.param[1]
  s <- hyper.param[2]
  n <- hyper.param[3]
  v <- hyper.param[4]

  s_trans <- s/sqrt(n/(n+1))
  x_trans <- (log(x)-m)/s_trans

  d <- 1/(x*s_trans)*stats::dt(df=v, x=x_trans)
  return(d)
}

#' @rdname LogStudentT
#' @export
plt <- function(q, hyper.param){
  m <- hyper.param[1]
  s <- hyper.param[2]
  n <- hyper.param[3]
  v <- hyper.param[4]

  s_trans <- s/sqrt(n/(n+1))
  q_trans <- (log(q)-m)/s_trans
  p <- stats::pt(q_trans, df=v)
  return(p)
}

#' @rdname LogStudentT
#' @export
qlt <- function(p, hyper.param){
  m <- hyper.param[1]
  s <- hyper.param[2]
  n <- hyper.param[3]
  v <- hyper.param[4]

  s_trans <- s/sqrt(n/(n+1))
  q <- exp(stats::qt(df=v, p=p)*s_trans+m)
  return(q)
}

#' @rdname LogStudentT
#' @export
rlt  <- function(n_vals, hyper.param){
  m <- hyper.param[1]
  s <- hyper.param[2]
  n <- hyper.param[3]
  v <- hyper.param[4]

  q <- stats::runif(n_vals)
  r <- qlt(q, hyper.param)
  return(r)
}

test_that("Basis var, transformation", {
  a <- PROB_BASEVAR(Id = 1, Name = "a", DistributionType = "norm", Mean = 1, Sd = 0.1)
  a$prepare()
  mean_ <- a$DistributionParameters[1]
  sd_ <- a$DistributionParameters[2]
  expect_equal(mean_, 1)
  expect_equal(sd_, 0.1)

  b <- PROB_BASEVAR(Id = 1, Name = "b", DistributionType = "lnorm", Mean = 1, Sd = 0.1)
  b$prepare()
  mean_ <- b$DistributionParameters[1]
  sd_ <- b$DistributionParameters[2]
  expect_equal(mean_, -0.004975165)
  expect_equal(sd_, 0.099751345)

  b <- PROB_BASEVAR(Id = 1, Name = "c", DistributionType = "lnorm", DistributionParameters = c(mean_, sd_))
  b$prepare()
  expect_equal(b$Mean, 1)
  expect_equal(b$Sd, 0.1)
})


test_that("Basis var, parameters", {
  params <- c(2, 3, 4, 5)
  means <- round(c(0.69189874, 1.098057, 1.385982, 1.609238, 0.6918987), 2)
  sds <- round(c(0.04996879, 0.03332408, 0.0249961, 0.019998, 0.04996879), 4)
  a <- PARAM_BASEVAR(Id = 1, Name = "a", DistributionType = "lnorm", ParamType = "Mean", Sd = 0.1, ParamValues = params)

  for (i in 1:5) {
    mean_ <- round(a$DistributionParameters[1], 2)
    sd_ <- round(a$DistributionParameters[2], 4)
    expect_equal(mean_, means[i])
    expect_equal(sd_, sds[i])
    a$nextParam()
  }
})


test_that("Gumbel distribution forward and backward transformation are consistent", {
  # Beispielwerte für Mittelwert und Standardabweichung
  mean_val <- 2000
  sd_val <- 200

  # Vorwärtstransformation mit PROB_BASEVAR
  gumbel_var <- PROB_BASEVAR(
    Id = 1, Name = "M",
    DistributionType = "gumbel",
    Package = "evd",
    Mean = mean_val,
    Sd = sd_val
  )

  gumbel_var$prepare()

  loc_param <- gumbel_var$DistributionParameters[1]
  scale_param <- gumbel_var$DistributionParameters[2]

  # Rücktransformation aus Parametern prüfen
  gumbel_back <- PROB_BASEVAR(
    Id = 2, Name = "M_back",
    DistributionType = "gumbel",
    Package = "evd",
    DistributionParameters = c(loc_param, scale_param)
  )

  gumbel_back$prepare()

  expect_equal(gumbel_back$Mean, mean_val, tolerance = 1e-8)
  expect_equal(gumbel_back$Sd, sd_val, tolerance = 1e-8)

  # Zusatzprüfung: Location liegt unterhalb des Mittelwerts (wegen positiver Gamma-Konstante)
  expect_lt(loc_param, mean_val)
})


test_that("Weibull distribution forward and backward transformation are consistent", {
  mean_val <- 100
  sd_val <- 15

  weibull_var <- PROB_BASEVAR(
    Id = 3, Name = "W",
    DistributionType = "weibull",
    Mean = mean_val,
    Sd = sd_val
  )

  weibull_var$prepare()

  shape_param <- weibull_var$DistributionParameters[1]
  scale_param <- weibull_var$DistributionParameters[2]

  # Rückrichtung prüfen:
  weibull_back <- PROB_BASEVAR(
    Id = 4, Name = "W_back",
    DistributionType = "weibull",
    DistributionParameters = c(shape_param, scale_param)
  )

  weibull_back$prepare()

  # Wegen der Näherungsfunktion für k ist eine Toleranz erforderlich
  rel_err_mean <- abs(weibull_back$Mean - mean_val) / mean_val
  rel_err_sd <- abs(weibull_back$Sd - sd_val) / sd_val

  expect_lt(rel_err_mean, 0.01) # < 1 % Abweichung erlaubt
  expect_lt(rel_err_sd, 0.01)

  expect_true(shape_param > .Machine$double.eps)
  expect_true(scale_param > .Machine$double.eps)
})

test_that("Weibull shape reacts correctly to different CoV values", {
  low_cov_obj <- PROB_BASEVAR(
    Id = 5, Name = "W_lowCov",
    DistributionType = "weibull", Mean = 1000, Sd = 80
  ) # Cov≈0.05

  high_cov_obj <- PROB_BASEVAR(
    Id = 6, Name = "W_highCov",
    DistributionType = "weibull", Mean = 1000, Sd = 300
  ) # Cov≈0.5

  low_cov_obj$prepare()
  high_cov_obj$prepare()

  k_lowCov <- low_cov_obj$DistributionParameters[1]
  k_highCov <- high_cov_obj$DistributionParameters[1]

  # Bei größerer Streuung (hoher COV) wird der Shapeparameter kleiner.
  expect_gt(k_lowCov, k_highCov)
})


test_that("Old incorrect Gumbel implementation would fail consistency check", {
  mean_test <- 2000
  sd_test <- 200

  correct_scale <- (sd_test * sqrt(6)) / pi
  correct_location <- mean_test + digamma(1) * correct_scale

  wrong_location_old_code <-
    mean_test + digamma(1) / correct_scale # alte fehlerhafte Formel

  expect_gt(abs(correct_location - wrong_location_old_code), .Machine$double.eps)
})


test_that("Exponential forward and backward transformation are consistent", {
  evar <- PROB_BASEVAR(
    Id = 10, Name = "ExpVar",
    DistributionType = "exp", Mean = 5
  )
  evar$prepare()
  expect_equal(evar$DistributionParameters[1], 0.2, tolerance = 5e-8)

  eback <- PROB_BASEVAR(
    Id = 11, Name = "ExpBack",
    DistributionType = "exp",
    DistributionParameters = c(evar$DistributionParameters)
  )
  eback$prepare()
  expect_equal(eback$Mean, 5, tolerance = 5e-8)
  expect_equal(eback$Sd, 5, tolerance = 5e-8)
})


test_that("Beta forward and backward transformation are consistent", {
  bvar <- PROB_BASEVAR(
    Id = 12, Name = "BetaVar",
    DistributionType = "beta", Mean = 0.6, Sd = 0.05
  )
  bvar$prepare()

  a_param <- bvar$DistributionParameters[1]
  b_param <- bvar$DistributionParameters[2]

  bback <- PROB_BASEVAR(
    Id = 13, Name = "BetaBack",
    DistributionType = "beta",
    DistributionParameters = c(a_param, b_param)
  )
  bback$prepare()

  expect_equal(bback$Mean, 0.6, tolerance = 5e-3)
  expect_equal(bback$Sd, 0.05, tolerance = 5e-3)
})


test_that("is_empty works as intended", {
  expect_true(is_empty(NULL))
  expect_true(is_empty(numeric(0)))
  expect_true(is_empty(integer(0)))
  expect_true(is_empty(list()))
  expect_true(is_empty(data.frame()))
  expect_true(is_empty(matrix(numeric(0), nrow = 0, ncol = 2)))

  expect_false(is_empty(1))
  expect_false(is_empty(c(1, 2)))
  expect_false(is_empty(list(a = 1)))
  expect_false(is_empty(data.frame(x = 1)))
})


test_that("PROB_BASEVAR rejects unknown distribution types", {
  expect_error(
    PROB_BASEVAR(
      Name = "x",
      DistributionType = "unknown",
      Mean = 1, Sd = 0.1
    ),
    regexp = "Distribution type 'unknown' is not known"
  )
})


test_that("PROB_MACHINE rejects non-existent fCall", {
  mach <- PROB_MACHINE(name = "Bad", fCall = "NON_EXISTENT")
  ps <- SYS_PROB(
    sys_input = list(SYS_LSF(vars = list(
      PROB_BASEVAR(Name = "a", Mean = 1, Sd = 0.1)
    ))),
    probMachines = list(mach)
  )
  expect_error(
    ps$runMachines(),
    regexp = "Method 'NON_EXISTENT' is not loaded"
  )
})


test_that("Distribution parameter round-trip (norm, lnorm, gumbel, gamma, beta, weibull)", {
  # Helper, prüft relative Abweichung
  rel_err <- function(ref, got) abs(ref - got) / max(abs(ref), .Machine$double.eps)

  # 1. Normal
  nb <- PROB_BASEVAR(
    Name = "norm", DistributionType = "norm",
    Mean = 12.3, Sd = 4.5
  )
  nb$prepare()
  expect_equal(nb$DistributionParameters, c(12.3, 4.5))

  # 2. Lognormal
  ln <- PROB_BASEVAR(
    Name = "lnorm", DistributionType = "lnorm",
    Mean = 5, Sd = 2
  )
  ln$prepare()
  # round‑trip
  mn <- ln$DistributionParameters[1]
  sn <- ln$DistributionParameters[2]
  # back‑transformation (inside prepare) was already executed,
  # check that Mean/Sd ≈ original
  expect_lt(rel_err(ln$Mean, 5), 1e-12)
  expect_lt(rel_err(ln$Sd, 2), 1e-12)

  # 3. Gumbel
  gb <- PROB_BASEVAR(
    Name = "gumbel", DistributionType = "gumbel",
    Mean = 2000, Sd = 200
  )
  gb$prepare()
  loc <- gb$DistributionParameters[1]
  scale <- gb$DistributionParameters[2]
  # back‑transform
  expect_lt(rel_err(gb$Mean, loc - digamma(1) * scale), 1e-12)
  expect_lt(rel_err(gb$Sd, pi / sqrt(6) * scale), 1e-12)

  # 4. Gamma
  ga <- PROB_BASEVAR(
    Name = "gamma", DistributionType = "gamma",
    Mean = 15, Sd = 5
  )
  ga$prepare()
  shape <- ga$DistributionParameters[1]
  scale <- ga$DistributionParameters[2]
  expect_lt(rel_err(ga$Mean, shape * scale), 1e-12)
  expect_lt(rel_err(ga$Sd, sqrt(shape) * scale), 1e-12)

  # 5. Beta (bound to (0,1))
  be <- PROB_BASEVAR(
    Name = "beta", DistributionType = "beta",
    Mean = 0.6, Sd = 0.05
  )
  be$prepare()
  alpha <- be$DistributionParameters[1]
  beta_ <- be$DistributionParameters[2]
  mean.back <- alpha / (alpha + beta_)
  var.back <- alpha * beta_ / ((alpha + beta_)^2 * (alpha + beta_ + 1))
  expect_lt(rel_err(be$Mean, mean.back), 1e-12)
  expect_lt(rel_err(be$Sd, sqrt(var.back)), 1e-12)

  # 6. Weibull – hier nutzen wir die numerische Approximation,
  #    die wir bereits implementiert haben
  mean_val <- 100
  sd_val <- 15

  we <- PROB_BASEVAR(
    Name = "weibull", DistributionType = "weibull",
    Mean = mean_val,
    Sd = sd_val
  )

  we$prepare()
  shape.w <- we$DistributionParameters[1]
  scale.w <- we$DistributionParameters[2]
  # back‑transform
  mean.w <- scale.w * base::gamma(1 + 1 / shape.w)
  sd.w <- sqrt(scale.w^2 *
    (base::gamma(1 + 2 / shape.w) - base::gamma(1 + 1 / shape.w)^2))
  # Toleranz 1 % ist für die empirische Approximation realistisch
  rel_err_mean <- abs(we$Mean - mean_val) / mean_val
  rel_err_sd <- abs(we$Sd - sd_val) / sd_val

  expect_lt(rel_err_mean, 0.01) # < 1 %
  expect_lt(rel_err_sd, 0.01)
})


test_that("PROB_BASEVAR handles zero or near-zero mean correctly", {
  vb <- PROB_BASEVAR(
    Name = "zeroMean",
    DistributionType = "norm",
    Mean = 0,
    Sd = 0.1
  )

  # The prepare() method should run silently (no warning expected)
  expect_silent(vb$prepare())

  # Coefficient of variation must be infinite for mean ≈ 0
  expect_true(is.infinite(vb$Cov))

  # Numerical tolerance: near-zero means also yield Inf CoV
  v_small <- PROB_BASEVAR(Name = "tinyMean", DistributionType = "norm", Mean = 1e-12, Sd = 1e-4)

  v_small$prepare()

  # Check that Cov is still treated as infinite due to small mean value
  expect_true(is.infinite(v_small$Cov))

  # For normal finite means → finite CoV value expected
  v_normal <- PROB_BASEVAR(Name = "normalMean", DistributionType = "norm", Mean = 10, Sd = 2)

  v_normal$prepare()

  expect_true(is.finite(v_normal$Cov))

  expect_equal(round(v_normal$Cov, 3), round(2 / 10, 3))
})


# Test for SYS_LSF$check() behaviour in different function states
# ---------------------------------------------------------------

test_that("SYS_LSF$check() correctly distinguishes missing, empty and valid functions", {
  # Create two basic random variables
  var_E <- PROB_BASEVAR(Name = "E", DistributionType = "norm", Mean = 10, Sd = 0.2)
  var_R <- PROB_BASEVAR(Name = "R", DistributionType = "norm", Mean = 35, Sd = 0.5)

  ## Case 1: Missing function (func == NULL) ---------------------
  lsf_missing <- SYS_LSF(vars = list(var_E, var_R), name = "LSF_missing")
  expect_true(is.null(lsf_missing$func))
  expect_error(lsf_missing$check(),
    regexp = "Limit-state function 'LSF_missing' is missing",
    fixed = TRUE
  )

  ## Case 2: Empty function without arguments --------------------
  lsf_empty <- SYS_LSF(vars = list(var_E, var_R), name = "LSF_empty")
  lsf_empty$func <- function() NULL
  expect_true(is.function(lsf_empty$func))
  expect_equal(length(formals(lsf_empty$func)), 0L)
  expect_error(lsf_empty$check(),
    regexp = "has no formal arguments",
    fixed = FALSE
  )

  ## Case 3: Valid function with arguments -----------------------
  lsf_valid <- SYS_LSF(vars = list(var_E, var_R), name = "LSF_valid")
  lsf_valid$func <- function(E, R) {
    R - E
  }

  # Should run silently (no error)
  expect_silent(lsf_valid$check())
})


test_that("PROB_BASEVAR catches negative standard deviation", {
  expect_error(
    PROB_BASEVAR(
      Name = "negSd",
      DistributionType = "norm",
      Mean = 10, Sd = -1
    ),
    regexp = "Standard deviation (Sd) must be non-negative.",
    fixed = TRUE
  )
})

test_that("PARAM_BASEVAR works with a single value", {
  pb <- PARAM_BASEVAR(
    Name = "single",
    DistributionType = "norm",
    ParamType = "Mean",
    ParamValues = 5
  )
  # Erstes `nextParam()` wird beim Initialisieren aufgerufen
  expect_equal(pb$Mean, 5)
  # Weiteres `nextParam()` muss denselben Wert zurückgeben (wrap‑around)
  pb$nextParam()
  expect_equal(pb$Mean, 5)
})

test_that("PARAM_DETVAR works with one deterministic value", {
  pd <- PARAM_DETVAR(
    Name = "det_single",
    ParamValues = 42
  )
  expect_equal(pd$Mean, 42)
  expect_equal(pd$Sd, 42 / 1e7)
  pd$nextParam()
  expect_equal(pd$Mean, 42) # wrap‑around, gleiche Zahl
})


test_that("PROB_BASEVAR caching improves runtime", {
  # Prepare a simple normal variable
  v <- PROB_BASEVAR(
    Name = "cacheTest",
    DistributionType = "norm",
    Mean = 0,
    Sd = 1
  )
  v$prepare()

  # Skip test if microbenchmark is not available
  skip_if_not_installed("microbenchmark")
  library(microbenchmark)

  # Benchmark both calls: first (creates cache) vs. second (uses cache)
  bench <- microbenchmark(
    first_call = {
      v$getlDistr()
    },
    second_call = {
      v$getlDistr()
    },
    times = 10L # number of repetitions per expression
  )

  # Print benchmark results for inspection (optional)
  print(bench)

  # Extract median execution times in milliseconds
  t_first_ms <- median(bench$time[bench$expr == "first_call"]) / 1e6
  t_second_ms <- median(bench$time[bench$expr == "second_call"]) / 1e6

  message(sprintf("Median time: first=%.3f ms, second=%.3f ms", t_first_ms, t_second_ms))

  # Compare performance: cached call should be equal or faster
  expect_true(t_second_ms <= t_first_ms * 1.2,
    info = sprintf(
      "Second call should be faster or within +20%% tolerance (first=%.3f ms, second=%.3f ms)",
      t_first_ms, t_second_ms
    )
  )

  # Additional functional check: both calls should return identical output values
  res_first <- v$getlDistr()
  res_second <- v$getlDistr()

  expect_equal(res_first[[1]]$d(0), res_second[[1]]$d(0))
})


test_that("runMachines() yields identical β-values for identical seeds", {
  set.seed(1234) # Master‑Seed – hat keinen Einfluss auf die Worker‑Seeds
  # 2 Maschinen: FORM (deterministisch) + MC_CRUDE (stochastic)
  m_form <- PROB_MACHINE(name = "FORM", fCall = "FORM")
  m_mc <- PROB_MACHINE(
    name = "MC", fCall = "MC_CRUDE",
    options = list(
      cov_user = 0.05,
      n_max = 5e5,
      seed = 777
    )
  ) # fester Seed

  lsf_obj <- SYS_LSF(vars = list(
    PROB_BASEVAR(Name = "E", Mean = 10, Sd = 0.2),
    PROB_BASEVAR(Name = "R", Mean = 35, Sd = 0.5)
  ))
  lsf_obj$func <- function(E, R) {
    R - E
  } # einfache Funktion definieren

  ps <- SYS_PROB(
    sys_input = list(lsf_obj),
    probMachines = list(m_form, m_mc)
  )

  # Erster Durchlauf
  ps$runMachines()
  beta1 <- ps$beta_single

  # zweiter Durchlauf (identischer Seed!)
  ps$runMachines()
  beta2 <- ps$beta_single

  expect_equal(beta1, beta2)
})


test_that("MC_IS throws informative error if n_batch <= 0", {
  expect_error(
    MC_IS(
      lsf = function(x) 1,
      lDistr = list(list(
        d = function() NULL,
        p = function() NULL,
        q = function() NULL,
        r = function() NULL
      )),
      n_batch = 0, n_max = 1e5, cov_user = 0.05,
      seed = 123
    ),
    regexp = "n_batch.*positive"
  )
})


test_that("SYS_PARAM aborts cleanly when a PARAM variable loses its ParamValues", {
  pvar <- PARAM_BASEVAR(
    Name = "bad",
    DistributionType = "norm",
    ParamType = "Mean",
    ParamValues = c(10) # valid initialization first
  )
  pvar$ParamValues <- numeric(0) # now remove all values manually

  lsf <- SYS_LSF(vars = list(pvar), name = "badLSF")
  ps <- SYS_PARAM(
    sys_input = list(lsf),
    probMachines = list(PROB_MACHINE(name = "FORM", fCall = "FORM"))
  )

  expect_error(ps$runMachines(),
    regexp = "must define a non-empty ParamValues",
    fixed = FALSE
  )
})


test_that("SYS_LSF throws an informative error when no limit-state function is defined", {
  # Create a valid set of base variables
  var_E <- PROB_BASEVAR(Name = "E", DistributionType = "norm", Mean = 10, Sd = 0.2)
  var_R <- PROB_BASEVAR(Name = "R", DistributionType = "norm", Mean = 35, Sd = 0.5)

  # Case 1: No function assigned -> should trigger 'missing' message
  lsf_missing <- SYS_LSF(vars = list(var_E, var_R), name = "LSF_missing")

  expect_error(
    lsf_missing$check(),
    regexp = "Limit-state function 'LSF_missing' is missing",
    fixed = FALSE
  )

  # Case 2: Function exists but has no arguments -> should trigger 'no formal arguments' message
  lsf_empty <- SYS_LSF(vars = list(var_E, var_R), name = "LSF_empty")
  lsf_empty$func <- function() NULL

  expect_error(
    lsf_empty$check(),
    regexp = "has no formal arguments",
    fixed = FALSE
  )

  # Case 3: Properly defined function -> check() should pass silently
  lsf_valid <- SYS_LSF(vars = list(var_E, var_R), name = "LSF_valid")
  lsf_valid$func <- function(E, R) {
    R - E
  }

  expect_silent(lsf_valid$check())
})


## ------------------------------------------------------------------------
## 1) PROB_BASEVAR – check missing or invalid inputs
## ------------------------------------------------------------------------

test_that("PROB_BASEVAR throws error when Mean/Sd and parameters are missing", {
  expect_error(
    PROB_BASEVAR(Name = "X1"),
    regexp = "Mean/Sd|distribution parameters",
    fixed = FALSE
  )
})

test_that("PROB_BASEVAR accepts valid Mean/Sd combination", {
  var_ok <- PROB_BASEVAR(Name = "X2", DistributionType = "norm", Mean = 10, Sd = 2)
  expect_s4_class(var_ok, "PROB_BASEVAR")
  expect_equal(var_ok$Mean, 10)
})

test_that("PROB_BASEVAR rejects unknown distribution type", {
  expect_error(
    PROB_BASEVAR(Name = "X3", DistributionType = "unknown", Mean = 10, Sd = 2),
    regexp = "Distribution type 'unknown' is not known",
    fixed = FALSE
  )
})

test_that("PROB_BASEVAR rejects negative standard deviation", {
  expect_error(
    PROB_BASEVAR(Name = "X4", DistributionType = "norm", Mean = 5, Sd = -1),
    regexp = "Standard deviation (Sd) must be non-negative",
    fixed = TRUE
  )
})


## ------------------------------------------------------------------------
## 2) PARAM_BASEVAR – check mandatory fields and parameter sweep setup
## ------------------------------------------------------------------------

test_that("PARAM_BASEVAR throws error when ParamValues are missing", {
  expect_error(
    PARAM_BASEVAR(Name = "P1", ParamType = "Mean"),
    regexp = "PARAM_BASEVAR .*must define a non-empty ParamValues vector.",
    fixed = FALSE
  )
})

# "PARAM_BASEVAR '%s' must define a non-empty ParamValues vector."
test_that("PARAM_BASEVAR throws error when ParamType is missing", {
  expect_error(
    PARAM_BASEVAR(Name = "P2", ParamValues = c(10, 20)),
    regexp = "requires a valid 'ParamType'",
    fixed = FALSE
  )
})

test_that("PARAM_BASEVAR throws error when both Mean/Sd and ParamValues are missing", {
  expect_error(
    PARAM_BASEVAR(Name = "P3"),
    regexp = "PARAM_BASEVAR .*must define a non-empty ParamValues vector.",
    fixed = FALSE
  )
})

test_that("PARAM_BASEVAR initializes correctly with valid inputs (Mean sweep)", {
  pvar <- PARAM_BASEVAR(
    Name = "E_mod",
    DistributionType = "norm",
    ParamType = "Mean",
    ParamValues = c(30000, 35000),
    Sd = 1000
  )

  # Object should be created successfully and first parameter value assigned:
  expect_s4_class(pvar, "PARAM_BASEVAR")

  # The mean should correspond to the first value in ParamValues:
  expect_equal(round(pvar$Mean), 30000)
})


## ------------------------------------------------------------------------
## Optional: combined test for prepare() / getlDistr()
## ------------------------------------------------------------------------

test_that("getlDistr returns list of functions for valid PROB_BASEVAR object", {
  v <- PROB_BASEVAR(Name = "f_ck", DistributionType = "norm", Mean = 30, Sd = 1)
  distr <- v$getlDistr()

  # Density function must exist (closure expected):
  expect_type(distr[[1]]$d, "closure")

  # Parameter vector must be numeric:
  expect_true(is.numeric(distr[[2]]))
})

## ------------------------------------------------------------------------
## Empirical function
## ------------------------------------------------------------------------

test_that("Empirical distribution (emp) is implemented correctly", {
  skip_if_not_installed("EnvStats")
  set.seed(123)
  # x <- rnorm(50)
  x <- c(2, 3, 4, 4.000001, 4.000002, 6, 7, 200)
  v <- PROB_BASEVAR(
    Name = "data_sequence",
    DistributionType = "emp",
    DistributionParameters = x
  )
  lDistr <- v$getlDistr()
  funs <- lDistr[[1]]
  # --- 1) Functions exist ----------------------------------------------------
  expect_type(funs$d, "closure")
  expect_type(funs$p, "closure")
  expect_type(funs$q, "closure")
  expect_type(funs$r, "closure")
  # --- 2) Density equals EnvStats reference ---------------------------------
  grid <- seq(min(x) - 1, max(x) + 1, length.out = 2000)
  expect_equal(
    funs$d(grid),
    EnvStats::demp(grid, obs = x)
  )
  # --- 4) Distribution function consistency ---------------------------------
  ref_p <- suppressWarnings(
    EnvStats::pemp(grid, obs = x)
  )
  expect_equal(
    funs$p(grid),
    ref_p
  )
  # --- 5) Quantile function consistency -------------------------------------
  probs <- seq(0.1, 0.9, by = 0.2)

  expect_equal(
    funs$q(probs),
    EnvStats::qemp(probs, obs = x)
  )
  # --- 6) Random generator only returns observed values ---------------------
  r_vals <- funs$r(200)
  expect_true(all(r_vals %in% x))
})



## ------------------------------------------------------------------------
## Empirical function
## ------------------------------------------------------------------------
test_that("Empirical distribution functions behave correctly", {
  skip_if_not_installed("EnvStats")

  set.seed(1)

  sample_data <- rnorm(1000, mean = 100, sd = 4)

  X <- PROB_BASEVAR(
    Id = 1,
    Name = "Z",
    DistributionType = "emp",
    Package = "EnvStats",
    DistributionParameters = sample_data
  )

  distr <- X$getlDistr()[[1]]

  # CDF monotonicity
  x_vals <- sort(sample_data[1:100])
  p_vals <- distr$p(x_vals)

  expect_true(all(diff(p_vals) >= 0))

  # Quantile inversion check
  u_vals <- seq(0.1, 0.9, length.out = 10)
  q_vals <- distr$q(u_vals)
  p_back <- distr$p(q_vals)

  expect_equal(p_back, u_vals, tolerance = 1e-2)
})

test_that("Student-t distribution is implemented correctly", {

  hyper.param <- c(38, 7, 4, 6)

  m <- hyper.param[1]
  s <- hyper.param[2]
  n <- hyper.param[3]
  v <- hyper.param[4]

  # Density integrates to 1
  f <- function(x) dst(x, hyper.param)
  result <- integrate(f, -Inf, Inf)
  expect_equal(result$value, 1, tolerance = 1e-6)

  # Density matches analytical dt transformation
  s_trans <- s/sqrt(n/(n+1))
  x <- seq(-100, 100, length.out = 1000)
  expected <- dt((x - m)/s_trans, df = v) / s_trans
  actual   <- dst(x, hyper.param)

  # CDF and quantile are inverse
  expect_equal(actual, expected, tolerance = 1e-12)
  p <- seq(0.01, 0.99, length.out = 100)
  q <- qst(p, hyper.param)
  p_back <- pst(q, hyper.param)
  expect_equal(p_back, p, tolerance = 1e-8)

  # Random generation has correct mean and sd
  sim <- rst(1e6, hyper.param)
  # Mean exists for v > 1
  expect_equal(mean(sim), m, tolerance = 0.05)
  # Variance exists for v > 2
  sd_theoretical <- s_trans * sqrt(v/(v-2))
  expect_equal(sd(sim), sd_theoretical, tolerance = 0.05)
})

test_that("Log-Student-t distribution is implemented correctly", {

  set.seed(123)

  hyper.param <- c(3.4, 0.14, 3, 10)

  m <- hyper.param[1]
  s <- hyper.param[2]
  n <- hyper.param[3]
  v <- hyper.param[4]

  s_trans <- s/sqrt(n/(n+1))

  # Log-t density integrates to 1
    f <- function(x) dlt(x, hyper.param)
    result <- integrate(f, 0, Inf)
    expect_equal(result$value, 1, tolerance = 1e-6)

  # Log-t density matches analytical transformation
    x <- seq(0.001, 100, length.out = 1000)
    expected <- (1/x) *
      dt((log(x) - m)/s_trans, df = v) /
      s_trans
    actual <- dlt(x, hyper.param)
    expect_equal(actual, expected, tolerance = 1e-12)

  # Log-t CDF and quantile are inverse
    p <- seq(0.01, 0.99, length.out = 100)
    q <- qlt(p, hyper.param)
    p_back <- plt(q, hyper.param)
    expect_equal(p_back, p, tolerance = 1e-8)

  # Log-t random generation matches exp(t-simulation)
    sim1 <- rlt(1e5, hyper.param)
    # direkte Konstruktion
    sim2 <- exp(m + s_trans * rt(1e5, df = v))
    expect_equal(mean(log(sim1)),
                 mean(log(sim2)),
                 tolerance = 0.02)
    expect_equal(sd(log(sim1)),
                 sd(log(sim2)),
                 tolerance = 0.02)
})

#' @title Package startup - register default parallelisation preferences
#' @description
#' This internal function is executed automatically when the
#' **TesiproV** package is loaded (via `library(TesiproV)`).
#'
#' Setting package-specific default options that can later be used by TesiproV
#' functions to configure parallel
#' execution in a controlled and CRAN-compliant way.
#'
#' Two global options are defined:
#'
#' * **`TesiproV.future.default`** - the default parallel backend.
#'   The default is `"multisession"`, which works on all operating systems
#'   including Windows.
#'
#' * **`TesiproV.max_workers.default`** - the default upper limit for the
#'   number of parallel workers. The default is `4L`.
#'
#' These values serve as internal fallback settings and may be overridden
#' explicitly by user-level function arguments or environment variables
#' within TesiproV functions.
#'
#' The function returns `invisible(NULL)` and is **not exported**. It is
#' intended solely for internal package initialisation.
#'
#' @param libname Ignored. Path to the library where the package is installed
#'   (standard argument for `.onLoad`).
#' @param pkgname Ignored. Name of the package (standard argument for `.onLoad`).
#'
#' @details
#' The function performs two steps:
#'
#' 1. It checks whether the namespace of the package `future` is available
#'    and aborts with a clear error message if it is missing.
#' 2. It registers internal default options via `options()` without modifying
#'    the active `future` plan.
#'
#' By not calling `future::plan()` during package startup, TesiproV avoids
#' altering the user's global parallel configuration. This follows CRAN
#' recommendations that packages should not change global execution plans
#' on load.
#'
#' @examples
#' ## Load the package (initialises default options)
#' library(TesiproV)
#'
#' @keywords internal
#' @author (C) 2021-2026 K. Nille-Hauf, T. Feiri, M. Ricker, T. Lux -- Hochschule Biberach (until 2022),
#' TU Dortmund University - Chair of Structural Concrete (since 2023)
#' @importFrom future availableCores plan sequential multisession multicore
#' @import future.apply
#' @import digest
#'
#' @keywords internal
#'
.onLoad <- function(libname, pkgname) {
  if (!requireNamespace("future", quietly = TRUE)) {
    stop("Package \"future\" is required but not installed.",
      call. = FALSE
    )
  }

  # DO NOT set a plan here
  # Only store default preferences

  options(
    TesiproV.future.default = "multisession",
    TesiproV.max_workers.default = 4L
  )

  invisible()
}

#' @name MC_CRUDE
#' @title Crude Monte Carlo Simulation
#' @description
#' Crude Monte Carlo simulation (MC_CRUDE) method for the estimation of failure
#' probabilities in structural reliability analysis.
#'
#' The crude Monte Carlo method estimates the probability of failure
#' by direct sampling of the basic random variables and evaluation
#' of the limit-state function.
#'
#' The failure probability is estimated as
#' \deqn{
#'   \hat{p}_f = \frac{1}{N}
#'   \sum_{i=1}^{N}
#'   \mathbf{1}_{\{g(\mathbf{X}_i) \le 0\}}.
#' }
#'
#' Although conceptually simple and unbiased, the method becomes
#' computationally inefficient for very small failure probabilities,
#' as a large number of samples is required to obtain stable estimates.
#'
#' @param lsf Limit-state function. Must follow the LSF interface
#'   specification described in the "Limit-State Function (LSF) Interface"
#'   section below.
#'
#' @param lDistr List of distribution objects as returned by \code{PROB_BASEVAR$getlDistr()}.
#'
#' @param cov_user Target coefficient of variation to be achieved.
#'
#' @param n_batch Batch size per iteration (used for parallel computation).
#'
#' @param n_max Maximum number of Monte Carlo samples (stopping criterion).
#'
#' @param use_threads Number of parallel threads. Set to 1 for single-core
#'   execution. Parallel execution is not supported on Windows.
#'
#' @param backend Parallel backend, either \code{"future"} (default) or \code{"parallel"}.
#'
#' @param dataRecord Logical; if \code{TRUE}, intermediate results are recorded.
#'
#' @param debug.level Integer controlling verbosity (0 = silent, 2 = detailed output).
#'
#' @param seed Optional integer value for reproducible random numbers.
#'
#' @return MC_CRUDE returns an object containing the following elements:
#' \itemize{
#'   \item \code{method}: Character string identifying the method ("MCC").
#'   \item \code{beta}: Estimated reliability index \eqn{\beta = -\Phi^{-1}(p_f)}.
#'   \item \code{pf}: Estimated probability of failure.
#'   \item \code{var}: Estimated variance of the Monte Carlo estimator.
#'   \item \code{cov_mc}: Estimated coefficient of variation of \code{pf}.
#'   \item \code{cov_user}: Target coefficient of variation specified by the user.
#'   \item \code{n_mc}: Total number of Monte Carlo samples generated.
#'   \item \code{n_max}: Maximum allowed number of samples.
#'   \item \code{n_batch}: Batch size per iteration.
#'   \item \code{n_threads}: Number of threads used.
#'   \item \code{runtime}: CPU time returned by \code{proc.time()}.
#'   \item \code{data}: Optional data frame containing intermediate results
#'         (only if \code{dataRecord = TRUE}).
#' }
#'
#' @section Limit-State Function (LSF) Interface:
#'
#' The argument \code{lsf} must define a valid limit-state function
#' using one of the following supported signatures:
#'
#' \describe{
#'   \item{\code{function(x)}}{
#'     A single numeric vector \code{x} containing all basic variables
#'     in the order defined in \code{lDistr}.
#'
#'     Example:
#'     \preformatted{
#'     lsf <- function(x) {
#'       x[1] - sum((x[-1]^2) / seq_along(x)[-1])
#'     }
#'     }
#'   }
#'
#'   \item{\code{function(Z, Fy, M, ...)}}{
#'     Explicitly named scalar arguments corresponding to the variable
#'     names defined in the probabilistic model.
#'
#'     Example:
#'     \preformatted{
#'     lsf <- function(Z, Fy, M) {
#'       Z * Fy - M
#'     }
#'     }
#'   }
#' }
#'
#' The following form is \strong{not supported}:
#'
#' \preformatted{
#' function(...)
#' }
#'
#' Using \code{...} as the sole argument may lead to inconsistent
#' behaviour across different reliability algorithms and is therefore
#' prohibited.
#'
#' For maximum robustness and cross-method compatibility,
#' the vector form \code{function(x)} is recommended.
#'
#' @section Windows users:
#'
#' On Windows systems, it is recommended to use
#' \code{backend = "future"} for multi-core parallelization.
#' The \code{"parallel"} backend does not support forking on Windows.
#'
#' @references
#' Spaethe, G. (1991).
#' \emph{Die Sicherheit tragender Baukonstruktionen}.
#' Springer.'
#'
#' @import parallel
#'
#' @author (C) 2021-2026 K. Nille-Hauf, T. Feiri, M. Ricker, T. Lux -- Hochschule Biberach (until 2022),
#' TU Dortmund University - Chair of Structural Concrete (since 2023)
#'
#' @export
#'

MC_CRUDE <- function(lsf,
                     lDistr,
                     cov_user = 0.05,
                     n_batch = 10000,
                     n_max = 1e7,
                     use_threads = parallel::detectCores(),
                     backend = c("future", "parallel"),
                     dataRecord = TRUE,
                     debug.level = 0,
                     seed = NULL) {
  backend <- match.arg(backend)

  # Warning if parallel was chosen under Windows
  if (.Platform$OS.type == "windows" && backend == "parallel") {
    warning(
      "For Windows systems, backend = 'future' is recommended ",
      "for robust multi-core parallelization."
    )
  }

  ## Initialize Future plan
  if (backend == "future") {
    current_plan <- future::plan()

    is_multicore <- inherits(current_plan, "multicore")
    is_multisession <- inherits(current_plan, "multisession")

    if (.Platform$OS.type == "windows") {
      if (!is_multisession) {
        future::plan(future::multisession, workers = use_threads)
      }
    } else {
      if (!is_multicore) {
        future::plan(future::multicore, workers = use_threads)
      }
    }
  }

  debug.TAG <- "MC_Crude"
  debug.print(debug.level, debug.TAG, c(TRUE), msg = "Crude Monte-Carlo Simulation started...")

  tic <- proc.time()

  pf <- 1
  I_n <- 0L
  # pf_i  <- 0
  # var_i <- 0
  n_sim <- 0L
  cov_mc <- +Inf
  var_mc <- NA_real_
  k <- 0L

  n_vars <- length(lDistr)

  v <- matrix(nrow = n_batch, ncol = n_vars)

  # Data record
  # If dataRecord = TRUE, record the following data (frequency saveEvery)

  if (dataRecord) {
    # create empty DataFraame
    df <- data.frame(
      n_sim = numeric(),
      pf = numeric(),
      var = numeric(),
      cov = numeric(),
      time.user = numeric(),
      time.sys = numeric(),
      time.elapsed = numeric(),
      time.user.child = numeric(),
      time.sys.child = numeric()
    )
  } else {
    df <- data.frame()
  }

  # Create empty list
  records <- list()

  # Definition of update frequency of dataRecord
  saveEvery <- max(10L, (use_threads * n_batch) / 250L)

  # Generate random numbers
  # Reproducible streams for parallel random numbers
  RNGkind(kind = "L'Ecuyer-CMRG")

  if (!is.null(seed)) {
    set.seed(seed)
    message(sprintf("Fixed RNG seed set to %d", seed))
  }

  # # mc_local is used for parallelisation purpose
  mc_local <- function(x) {
    v_local <- matrix(nrow = n_batch, ncol = n_vars)

    for (i in seq_len(n_vars)) {
      v_local[, i] <- lDistr[[i]][[1]]$r(n_batch)
    }
    if (is.function(lsf)) {
      I <- sum(apply(v_local, 1, lsf) < 0)
    } else if (!is.null(lsf$func) && is.function(lsf$func)) {
      I <- sum(apply(v_local, 1, lsf$func) < 0)
    } else {
      stop("Parameter 'lsf' must be a function or SYS_LSF object with field $func.")
    }

    return(I)
  }

  while (TRUE) {
    # # create Realisations
    # if (Sys.info()[[1]]=="Windows"){ #Windows is not able to fork - parallelisation under windows os not efficient
    #   I_n <- I_n + sum(unlist(parallel::mclapply(seq(1,use_threads),mc_local, mc.cores=1)))
    # }else{ # parallelisation for unix based platforms (macOS, Linux etc.)
    #   I_n <- I_n + sum(unlist(parallel::mclapply(seq(1,use_threads),mc_local, mc.set.seed = TRUE, mc.cores=use_threads)))
    #   stats::runif(1)
    # }
    # n_sim <- n_sim + use_threads*n_batch

    # k <- k+1
    # For only one thread, mclapply is slower than lapply
    if (use_threads == 1) {
      I_list <- lapply(seq_len(use_threads), mc_local)
    } else {
      if (backend == "parallel") {
        I_list <- parallel::mclapply(
          seq_len(use_threads),
          mc_local,
          mc.set.seed = TRUE,
          mc.cores = use_threads
        )
      } else if (backend == "future") {
        I_list <- future.apply::future_lapply(
          seq_len(use_threads),
          mc_local,
          future.seed = TRUE
        )
      }
    }

    I_n <- I_n + sum(unlist(I_list))
    n_sim <- n_sim + use_threads * n_batch

    k <- k + 1

    # time measurement
    p_stamp <- proc.time() - tic

    # Calculate stochastics
    debug.print(debug.level, debug.TAG, I_n, "I_n: ")

    if (I_n > 0) {
      pf <- I_n / n_sim
      # add more robust equation for the CoV
      # var_mc <- 1/(n_sim-1)*((1/n_sim)*I_n - pf^2)
      var_mc <- pf * (1 - pf) / n_sim
      # direct calculation of CoV is faster
      cov_mc <- sqrt((1 - pf) / (pf * n_sim)) # sqrt(var_mc)/pf

      # additional information if option debug = 2
      info.print(debug.TAG, debug.level, c("I_n", "pf", "cov", "nsim"), c(I_n, pf, cov_mc, n_sim))
      debug.print(debug.level, debug.TAG, pf, "Pf: ")
      debug.print(debug.level, debug.TAG, var_mc, "var_mc: ")
      debug.print(debug.level, debug.TAG, cov_mc, "CoV_mc: ")
      debug.print(debug.level, debug.TAG, n_sim, "n_sim: ")
      debug.print(debug.level, debug.TAG, n_max, "n_max: ")

      if (cov_mc < cov_user || n_sim > n_max) break # break loop, if target CoV is reached
    } else {
      info.print(debug.TAG, debug.level, c("I_n", "pf", "cov", "nsim"), c(I_n, "NA", "NA", n_sim))

      if (n_sim > n_max) break # break loop, if n_sim > n_max
    }


    # Save data only every 10th step and use a list to increase performance
    if (dataRecord && k %% saveEvery == 0) {
      records[[length(records) + 1]] <- list(
        n_sim = n_sim,
        pf = pf,
        var = var_mc,
        cov = cov_mc,
        time.user = p_stamp[1],
        time.sys = p_stamp[2],
        time.elapsed = p_stamp[3],
        time.user.child = p_stamp[4],
        time.sys.child = p_stamp[5]
      )
    }

    # # Allocated RAM
    #   if (debug.level > 0 && k %% saveEvery == 0) {
    #   mem_now <- pryr::mem_used()
    #   msg <- sprintf("[%s] [Iter %d] RAM: %.2f MB\n",
    #                  format(Sys.time(), "%H:%M:%S"),
    #                  n_sim,
    #                  mem_now / (1024^2))
    #   cat(msg, file="MC_CRUDE_log.txt", append=TRUE)
    # }
  } # end of while loop


  # Convert list to DataFrame
  if (dataRecord && length(records) > 0) {
    df <- as.data.frame(do.call(rbind, lapply(records, as.data.frame)))
  } else {
    df <- data.frame()
  }

  cat("\n")
  duration <- proc.time() - tic

  # calculate reliabilty index beta
  beta_val <- ifelse(pf > 0 && !is.na(pf), -stats::qnorm(pf), Inf)

  output <- list(
    method = "MC_CRUDE",
    beta = beta_val, #-stats::qnorm(pf),
    pf = pf,
    var = var_mc,
    cov_mc = cov_mc,
    cov_user = cov_user,
    n_mc = n_sim,
    n_max = n_max,
    n_batch = n_batch,
    n_threads = use_threads,
    runtime = duration[1:5],
    data = df
  )

  if (dataRecord) {
    output$data <- df
  } else {
    output$data <- data.frame()
  }

  debug.print(debug.level, debug.TAG, c(duration), msg = "Crude Monte-Carlo Simulation finished in [s]:  ")

  return(output)
}

#' @title Object for parametric Studies (SYS_PARAM)
#'
#' @description Extends `SYS_PROB` to perform a **parametric sweep** over one or more
#' `PARAM_BASEVAR` objects.  For each combination of parameter values a full
#' reliability analysis is executed and the results are stored in
#' `beta_params` and `res_params`.
#'
#' @field beta_params   List of \eqn{\beta}-matrices (one matrix per parameter run).
#' @field res_params    List of detailed result objects (one per run).
#' @field nParams       Numeric vector: number of parameter values for each
#'                       `PARAM_BASEVAR`.
#' @field nLsfs         Number of limit-state functions in the system.
#' @field nMachines     Number of analysis machines (methods) used.
#'
#' @details
#' The method `runMachines()` is almost identical to the one in `SYS_PROB`,
#' but it loops over **all** parameter values, creates a readable run name,
#' and stores the whole result hierarchy.
#'
#' @examples
#' # Simple 2-parameter sweep ------------------------------------------------
#' # two random variables that will be varied
#' pv1 <- PARAM_BASEVAR(
#'   Name = "E_mod",
#'   DistributionType = "norm",
#'   ParamType = "Mean",
#'   ParamValues = c(30e3, 35e3, 40e3)
#' )
#' pv2 <- PARAM_BASEVAR(
#'   Name = "f_ck",
#'   DistributionType = "norm",
#'   ParamType = "Mean",
#'   ParamValues = c(30, 35, 40)
#' )
#'
#' lsf <- SYS_LSF(vars = list(pv1, pv2), name = "Bending")
#' ps <- SYS_PARAM(
#'   sys_input = list(lsf),
#'   probMachines = list(PROB_MACHINE(name = "FORM", fCall = "FORM")),
#'   sys_type = "serial"
#' )
#'
#' \dontrun{
#' ps$runMachines()
#' ps$beta_params # matrix of \eqn{\beta}-values for every run
#' }
#'
#' @author (C) 2021-2026 K. Nille-Hauf, T. Feiri, M. Ricker, T. Lux -- Hochschule Biberach (until 2022),
#' TU Dortmund University - Chair of Structural Concrete (since 2023)
#'
#' @import future.apply
#' @import future
#' @export SYS_PARAM
#' @exportClass SYS_PARAM

SYS_PARAM <- setRefClass(
  Class = "SYS_PARAM",
  contains = "SYS_PROB",
  fields = list(
    beta_params = "list", # list of beta-matrices (one per parameter run)
    res_params = "list", # list of full result objects (one per run)
    nParams = "vector", # vector: number of parameter values per PARAM_BASEVAR
    nLsfs = "numeric", # total number of limit-state functions in the system
    nMachines = "numeric" # total number of analysis machines (methods)
  ),
  methods = list(
    ## -------------------------------------------------
    ##  INITIALIZER - create empty containers
    ## -------------------------------------------------
    initialize = function(...) {
      # --- DEBUG: capture all user-supplied arguments ---
      args <- list(...)

      # --- Step 1: manually assign inherited fields from parent class ---
      # We do not use initFields(), because it fails during reference-class creation.
      if (!is.null(args$sys_input)) sys_input <<- args$sys_input
      if (!is.null(args$probMachines)) probMachines <<- args$probMachines
      if (!is.null(args$sys_type)) sys_type <<- args$sys_type

      # --- Step 2: call parent initializer to set up default values ---
      callSuper(...)

      # --- Step 3: initialize own fields for parametric study bookkeeping ---
      beta_params <<- list()
      res_params <<- list()
      nParams <<- numeric()
      nLsfs <<- length(sys_input)
      nMachines <<- length(probMachines)
    },

    ## -------------------------------------------------
    ##  runMachines - perform the parametric study
    ## -------------------------------------------------
    runMachines = function() {
      ## ---------------------------------------------------------------
      ## Guard-Checks - verhindern subscript out of bounds
      ## ---------------------------------------------------------------
      if (length(sys_input) == 0L) {
        stop("No limit-state functions ('sys_input') defined", call. = FALSE)
      }
      if (length(probMachines) == 0L) {
        stop("No reliabilty machines ('probMachines') defined", call. = FALSE)
      }

      ## ---------------------------------------------------------------
      ## Additional integrity check: ensure PARAM_BASEVAR objects still have ParamValues
      lapply(sys_input, function(lsf) {
        lapply(lsf$vars, function(v) {
          if (inherits(v, "PARAM_BASEVAR") && length(v$ParamValues) == 0L) {
            stop(sprintf(
              "PARAM_BASEVAR '%s' must define a non-empty ParamValues vector.",
              v$Name
            ), call. = FALSE)
          }
        })
      })


      # #Check integrity of basisvars first
      # for(i in 1:length(sys_input)){
      #   sys_input[[i]]$check()
      # }

      ## 1. Verify the integrity of all basic variables (faster)
      lapply(sys_input, function(ls) ls$check())

      res_full_loc <- list()
      p <- 0
      paramVar_i <- 0
      # runNames <- vector()

      nLsfs <<- length(sys_input)
      nMachines <<- length(probMachines)

      ## -----------------------------------------------------------------
      ## 1)  Compute total number of runs (needed for pre-allocation)
      ## -----------------------------------------------------------------
      totalRuns <- 0L
      for (m in seq_along(sys_input)) {
        for (v in sys_input[[m]]$vars) {
          if (inherits(v, "PARAM_BASEVAR")) {
            if (length(v$ParamValues) == 0L) {
              stop(sprintf(
                "Parametric variable \"%s\" (ID=%s) has no `ParamValues`.",
                v$Name, as.character(v$Id)
              ), call. = FALSE)
            }
            totalRuns <- totalRuns + length(v$ParamValues)
          }
        }
      }

      runNames <- character(totalRuns) # pre-allocated vector
      p <- 0L

      ## Ergebnis-Container (je Run ein beta-Matrix-Objekt und ein Listen-Objekt)
      beta_params <<- list()
      res_params <<- list()

      ## Hilfsvektor, der fuer jede Param-Variable die Anzahl ihrer Werte speichert
      nParams <<- numeric()

      ## 3. Loop over every LSF and then over every variable belonging to it
      for (m in seq_along(sys_input)) {
        for (n in 1:length(sys_input[[m]]$vars)) {
          var <- sys_input[[m]]$vars[[n]]

          ## Only handle PARAM_BASEVAR objects
          if (inherits(var, "PARAM_BASEVAR")) {
            paramVar_i <- paramVar_i + 1L
            nParams[paramVar_i] <<- length(var$ParamValues)

            ## 4. Iterate over all values of the current parameter variable
            for (u in seq_along(var$ParamValues)) {
              p <- p + 1L
              runNames[p] <- sprintf(
                "Paramrun No. %d. Paramname %s with value(Type): %f (%s)",
                p, var$Name, var$getCurrentParam(), var$ParamType
              )
              ## -------------------------------------------------
              ## Containers for results of this single run
              ## -------------------------------------------------
              beta_loc <- matrix(
                NA_real_,
                nrow = length(probMachines),
                ncol = length(sys_input),
                dimnames = list(
                  NULL, # Zeilennamen (werden spaeter gesetzt)
                  sapply(sys_input, function(ls) {
                    if (is_empty(ls$name)) ls$name <- "Unknown Problem Name"
                    ls$name
                  })
                )
              )

              colnames(beta_loc) <- vector("character", length(sys_input))
              rownames(beta_loc) <- vector("character", length(probMachines))
              # Loop through each Problem in the System

              ## Store the results of each LSF for this run
              res_full_loc <- list()

              for (i in 1:length(sys_input)) {
                ## Optional: use the symbolic expression instead of the compiled function
                lsf_expr <- NULL
                if (!is_empty(sys_input[[i]]$expr)) {
                  lsf_expr <- sys_input[[i]]$expr
                }

                ## Compiled objective function (always available)
                lsf <- sys_input[[i]]$getLSF()
                # lsf<-sys_input[[i]]$func

                ## Ensure every LSF has a name (required for matrix column names)
                if (is_empty(sys_input[[i]]$name)) {
                  sys_input[[i]]$name <<- "Unknown Problem Name"
                }
                colnames(beta_loc)[i] <- sys_input[[i]]$name

                ## Build a list of distribution objects for all variables used in this LSF (Faster)
                distr <- lapply(sys_input[[i]]$vars, function(v) v$getlDistr())

                ## -------------------------------------------------
                ## Parallel execution of all machines for this LSF
                ## -------------------------------------------------
                ## library(future.apply)   # make sure it is loaded

                #  A) Vor future_lapply:
                res_machine <- vector("list", length(probMachines))

                res_machine_raw <- future.apply::future_lapply(
                  seq_along(probMachines),
                  function(j) {
                    pm <- probMachines[[j]]

                    ## ---- Build the argument list (identical to the original code) ----
                    if (identical(pm$fCall, "MVFOSM") &&
                      !is_empty(pm$options) &&
                      pm$options$isExpression == "TRUE") {
                      args <- list(lsf = lsf_expr, lDistr = distr)
                    } else {
                      args <- list(lsf = lsf, lDistr = distr)
                    }

                    ## ---- Append user-defined options (if any) ------------------------
                    if (!is_empty(pm$options)) args <- c(args, pm$options)

                    ## ---- Existence check - **must be done before** `do.call()` -----
                    pm$checkMethodExists()

                    ## Run the algorithm
                    res <- do.call(pm$fCall, args)

                    ## ---- Optional debug output (only when requested) ---------------
                    if (debug.level >= 2) {
                      message(sprintf(
                        "[SYS_PARAM/runMachines] %s -> beta = %g",
                        pm$fCall, res$beta
                      ))
                    }

                    ## Return a small list that contains everything we need later
                    list(
                      name    = pm$name,
                      beta    = res$beta,
                      result  = res
                    )
                  },
                  future.seed = TRUE
                )

                ## -------------------------------------------------
                ## Copy results back into the beta matrix and the list
                ## -------------------------------------------------

                if (!exists("res_machine")) res_machine <- vector("list", length(probMachines))

                for (j in seq_along(res_machine_raw)) {
                  mech <- res_machine_raw[[j]]
                  beta_loc[j, i] <- mech$beta
                  res_machine[[j]] <- mech$result
                  rownames(beta_loc)[j] <- mech$name
                }

                ## Save the results of this LSF
                res_full_loc[[i]] <- res_machine
                # end of LSF loop
              }

              ## -------------------------------------------------
              ## Save the complete results of this parameter run
              ## -------------------------------------------------
              beta_params[[p]] <<- beta_loc
              res_params[[p]] <<- res_full_loc

              ## set the parametric basisvariable one step further
              var$nextParam()

              ## check if there are determenistic vars that also needs to be shifted
              for (z in 1:length(sys_input[[m]]$vars)) {
                if (class(sys_input[[m]]$vars[[z]])[1] == "PARAM_DETVAR") {
                  sys_input[[m]]$vars[[z]]$nextParam()
                }
              }
            } # end of u-loop (all values of the current parameter variable)
          }
        } # end of n-loop (variables inside a single LSF)
      } # end of m-loop (over all LSFs)

      ## -------------------------------------------------
      ## Attach the run names to the result lists for easier post-processing
      ## -------------------------------------------------
      names(beta_params) <<- runNames
      names(res_params) <<- runNames
    },

    ## -------------------------------------------------
    ##  printResults - adapted to the parametric study output
    ## -------------------------------------------------
    printResults = function(path = "") {
      if (!path == "") {
        sink(paste(path, ".txt", sep = ""), type = "output")
        cat("Ergebnisausdruck TesiproV Berechnung\n")
        cat(date())
        cat("\n")
      }

      n_sys <- length(sys_input)
      n_machines <- length(probMachines)
      cat("\n -----------  1. Summary of Results  ---------\n\n")

      cat("\n1.1 Ergebnisse je Loesungsalgorithmus\n")
      for (i in 1:n_machines) {
        cat("\n____________________________________________\n")
        for (j in 1:nParams) {
          loc_res <- res_params[[j]][[1]][[i]]
          cat(sprintf("Run %d\tBeta: %.4f\tPf: %f\tMethod: %s\tRuntime: %s\n", j, loc_res$beta, loc_res$pf, loc_res$method, loc_res$runtime))
        }
      }

      cat("\n\n-------------------------  2. SYSTEM BESCHREIBUNG ------------------\n")

      cat(sprintf("2.1 Das System umfasst eine Gleichung.\n\n"))
      for (i in 1:n_sys) {
        cat("\n___________________________________\n")
        cat(sprintf("2.1.%d Gleichung Nr: %d \t-%s\n", i, i, sys_input[[i]]$name))
        if (!is_empty(sys_input[[i]]$expr)) {
          cat("Folgender symbolischer Ausdruck wurde definiert:\n")
          print(sys_input[[i]]$expr)
        }
        if (!is_empty(sys_input[[i]]$func)) {
          cat("Folgende Funktion wurde definiert:\n")
          print(sys_input[[i]]$func)
        }

        n_vars <- length(sys_input[[i]]$vars)
        cat(sprintf("\nBasisvariablen (%d gegeben):\n", n_vars))
        for (k in 1:n_vars) {
          v <- sys_input[[i]]$vars[[k]]

          if (class(v)[1] == "PARAM_BASEVAR") {
            cat("\n____________________")
            cat("\nParametervariable:\n")
            cat(sprintf(
              "%d.\tName (ID): %s (%d)\tPackage::Verteilungstyp: %s::%s\tMean: %.2f\tSd: %.2f\tCov: %.2f\tx0: %.2f\tVerteilungsparameter: %.5f\t%.5f\n",
              k, v$Name, v$Id, v$Package, v$DistributionType, v$Mean, v$Sd, v$Cov, v$x0, v$DistributionParameters[1], v$DistributionParameters[2]
            ))
            cat(sprintf("ParamType: %s\n", v$ParamType))
            cat("ParamValues:")
            cat(v$ParamValues)
            cat("\n____________________\n\n")
          } else if (inherits(var, "PARAM_DETVAR")) {
            cat("\n____________________")
            cat("\nDeterministische, von Parametervariable anhaengige Variable \n")
            cat(sprintf("%d.\tName (ID): %s (%d)\n", k, v$Name, v$Id))
            cat("ParamValues:")
            cat(v$ParamValues)
            cat("\n____________________\n\n")
          } else {
            cat(sprintf(
              "%d.\tName (ID): %s (%d)\tPackage::Verteilungstyp: %s::%s\tMean: %.3f\tSd: %.3f\tCov: %.3f\tx0: %.3f\tVerteilungsparameter: %.5f\t%.5f\n",
              k, v$Name, v$Id, v$Package, v$DistributionType, v$Mean, v$Sd, v$Cov, v$x0, v$DistributionParameters[1], v$DistributionParameters[2]
            ))
          }
        }
      }
      cat("\n\n ")

      cat("\n ------------------------  3. METHODEN BESCHREIBUNG ------------------\n")
      if (length(n_machines) > 0) {
        cat(sprintf("\n\n3.1 Die Berechnung wurde von %d Methoden analysiert.\nIm folgenden werden die Eingangsparameter der Methoden beschrieben.", n_sys))
        for (i in 1:n_machines) {
          cat("\n______________________________________\n")
          cat(sprintf("3.%d Machine Name (Type): %s (%s)\n", i, probMachines[[i]]$name, probMachines[[i]]$fCall))
          if (!is.null(probMachines[[i]]$options)) {
            cat("Die folgenden Optionen waren von den Standartwerten abweichend:\n")
            for (j in 1:length(probMachines[[i]]$options)) {
              cat(sprintf("\t-%s\t=\t%s\n", names(probMachines[[i]]$options[j]), probMachines[[i]]$options[j]))
            }
          } else {
            cat("Ausschliesslich Standartoptionen verwendet!\n")
          }
        }
      }


      cat("--------------------- ENDE -----------------")


      if (!path == "") {
        sink()
      }
    }
  )
)


#' @title Parametric Limit-State Function (PARAM_LSF)
#'
#' @description Identical to `SYS_LSF` but inherits from it to make the class hierarchy clear
#' when a parametric study is performed.  No additional fields or methods are
#' required.
#'
#' @author (C) 2021-2026 K. Nille-Hauf, T. Feiri, M. Ricker, T. Lux -- Hochschule Biberach (until 2022),
#' TU Dortmund University - Chair of Structural Concrete (since 2023)
#'
#' @export PARAM_LSF
#' @exportClass PARAM_LSF
#'
PARAM_LSF <- setRefClass(
  Class = "PARAM_LSF",
  contains = "SYS_LSF"
)


#' @title Parametric Random Variable (PARAM_BASEVAR)
#'
#' @description Extends `PROB_BASEVAR` so that one of the distribution parameters (Mean,
#' Sd or DistributionType) can be **swept** over a user-defined vector of values.
#'
#' @field ParamValues  Numeric vector holding the values that will be assigned
#'                     during the sweep.
#' @field ParamType    Which parameter is varied: `"Mean"`, `"Sd"` or
#'                     `"DistributionType"`.
#' @field pos          Current index inside `ParamValues` (used internally).
#'
#' @details
#' The method `nextParam()` advances `pos` (circularly) and updates the basic
#' fields (`Mean`, `Sd`, `DistributionType`, `DistributionParameters`) before
#' calling `prepare()` again.  `getCurrentParam()` returns the *last* value that
#' was used (convenient for readable run names).
#'
#' @examples
#' pvar <- PARAM_BASEVAR(
#'   Name = "E_mod",
#'   DistributionType = "norm",
#'   ParamType = "Mean",
#'   ParamValues = c(30e3, 35e3, 40e3)
#' )
#' pvar$nextParam() # sets Mean = 30.000 and prepares the distribution
#' pvar$getCurrentParam()
#'
#' @author (C) 2021-2026 K. Nille-Hauf, T. Feiri, M. Ricker, T. Lux -- Hochschule Biberach (until 2022),
#' TU Dortmund University - Chair of Structural Concrete (since 2023)
#'
#' @export PARAM_BASEVAR
#' @exportClass PARAM_BASEVAR

PARAM_BASEVAR <- setRefClass(
  Class = "PARAM_BASEVAR",
  contains = "PROB_BASEVAR",
  fields = list(
    ParamValues = "vector",
    ParamType = "character",
    pos = "numeric"
  )
)

PARAM_BASEVAR$methods(
  list(
    initialize = function(...) {
      ## ---------------------------------------------------------------
      ## 1) Felder uebernehmen und Grundinitialisierung
      ## ---------------------------------------------------------------
      initFields(...)

      ## ---- Numerische Felder sicherstellen ---------------------------
      if (is.null(Id)) Id <<- NA_real_
      if (is.null(Mean)) Mean <<- NA_real_
      if (is.null(Sd)) Sd <<- NA_real_
      if (is.null(Cov)) Cov <<- NA_real_
      if (is.null(x0)) x0 <<- NA_real_

      ## ---- Zeichenfelder --------------------------------------------
      if (is.null(Name)) Name <<- ""
      if (is.null(Description)) Description <<- ""

      ## ---- DistributionType / Package mit Defaults ------------------
      if (is_missing(DistributionType) || is.na(DistributionType) || DistributionType == "") {
        DistributionType <<- "norm"
      }
      if (is_missing(Package) || is.na(Package) || Package == "") {
        Package <<- "stats"
      }

      ## ---- Parametervektor immer Laenge > 2 --------------------------
      if (is.null(DistributionParameters) || length(DistributionParameters) < 2) {
        DistributionParameters <<- c(NA_real_, NA_real_)
      }

      .cache <<- new.env(parent = emptyenv())

      ## ---------------------------------------------------------------
      ## 2) Check integrity of parametric variable before validation
      ## ---------------------------------------------------------------

      # Ensure position counter exists
      if (is.null(pos) || length(pos) == 0L) {
        pos <<- 1L
      }

      # --- Early consistency check ------------------------------------
      # Case 1: ParamValues completely missing or empty
      if (is.null(ParamValues) || length(ParamValues) == 0L) {
        stop(sprintf(
          "PARAM_BASEVAR '%s' must define a non-empty ParamValues vector.",
          ifelse(is_empty(Name), "<unnamed>", Name)
        ), call. = FALSE)
      }

      # Case 2: Neither Mean nor Sd provided and no ParamValues -> abort
      if (all(is.na(c(Mean, Sd))) && length(ParamValues) == 0L) {
        stop(sprintf(
          "PARAM_BASEVAR '%s' must define either Mean/Sd or valid ParamValues.",
          ifelse(is_empty(Name), "<unnamed>", Name)
        ), call. = FALSE)
      }

      # Case 3: ParamType not specified -> meaningless sweep
      if (is_missing(ParamType) || is.na(ParamType) || ParamType == "") {
        stop(sprintf(
          "PARAM_BASEVAR '%s' requires a valid 'ParamType' (e.g., 'Mean', 'Sd', 'DistributionType').",
          ifelse(is_empty(Name), "<unnamed>", Name)
        ), call. = FALSE)
      }

      ## ---------------------------------------------------------------
      ## Proceed with validation and initial parameter setup
      ## ---------------------------------------------------------------
      validateArgs() # safe now - all fields exist and make sense

      if (length(ParamValues) == 0L || all(is.na(ParamValues))) {
        stop(sprintf("PARAM_BASEVAR '%s' requires non-empty ParamValues.", Name),
          call. = FALSE
        )
      }

      # Set first parameter value immediately to initialise Mean/Sd etc.
      nextParam()

      invisible(.self)
    },
    nextParam = function() {
      ## --- Sicherstellen, dass ParamValues existieren -----------------
      if (length(ParamValues) == 0L || all(is.na(ParamValues))) {
        stop(sprintf("No parameter values defined for PARAM_BASEVAR '%s'.", Name),
          call. = FALSE
        )
      }

      ## --- Numerische Felder vorbereiten -------------------------------
      if (is_missing(Mean)) Mean <<- NA_real_
      if (is_missing(Sd)) Sd <<- NA_real_
      if (is_missing(Cov)) Cov <<- NA_real_
      if (is_missing(x0)) x0 <<- NA_real_

      ## ensure DistributionParameters always numeric vector of length 2
      if (is.null(DistributionParameters) || length(DistributionParameters) < 2) {
        DistributionParameters <<- c(NA_real_, NA_real_)
      }

      ## --- Wrap-around Logik -----------------------------------------
      pos <<- (pos %% length(ParamValues)) + 1L
      i <- if (pos == 1L) length(ParamValues) else pos - 1L

      DistributionParameters <<- c(NA_real_, NA_real_) # Reset

      ## --- Neuen Parameter setzen je nach Typ ------------------------
      if (tolower(trimws(ParamType)) == "mean") {
        Mean <<- ParamValues[i]
        if (!is_missing(Sd) && !is_missing(Mean) && Mean != 0) {
          Cov <<- as.numeric(Sd / Mean)
        } else if (!is_missing(Mean) && Mean == 0) {
          Cov <<- 0
        } else {
          Cov <<- NA_real_
        }
      } else if (tolower(trimws(ParamType)) == "sd") {
        Sd <<- ParamValues[i]
        if (!is_missing(Sd) && !is_missing(Mean) && Mean != 0) {
          Cov <<- as.numeric(Sd / Mean)
        } else if (!is_missing(Mean) && Mean == 0) {
          Cov <<- 0
        } else {
          Cov <<- NA_real_
        }
      } else if (tolower(trimws(ParamType)) == "distributiontype") {
        DistributionType <<- as.character(ParamValues[i])
      } else {
        warning(sprintf(
          "Unknown ParamType '%s' in PARAM_BASEVAR '%s'.",
          ParamType, Name
        ))
      }

      # Safety check before prepare()
      stopifnot(is.numeric(Cov), is.numeric(Sd), is.numeric(Mean))

      prepare()

      # ## Cache leeren - Sicherheit gegen veraltete Funktionen ----------
      # if (!is.null(.cache)) {
      #   rm(list = ls(envir = .cache), envir = .cache)
      # }

      invisible(.self)
    },
    getCurrentParam = function() {
      idx <- if (pos == 1L) length(ParamValues) else pos - 1L
      ParamValues[idx]
    }
  )
)


#' @title Parametric Deterministic Variable (PARAM_DETVAR)
#'
#' @description Like `PROB_DETVAR` but the deterministic value itself can be swept.
#' The class inherits from `PROB_BASEVAR` and therefore re-uses the same
#' transformation machinery.
#'
#' @field ParamValues  Vector of deterministic values that will be assigned
#'                     sequentially in a parametric sweep.
#' @field pos          Current index (used internally).
#'
#' @examples
#' pdet <- PARAM_DETVAR(
#'   Name = "E_mod",
#'   ParamValues = c(30e3, 35e3, 40e3)
#' )
#' pdet$nextParam() # sets Mean = 30.000, Sd = Mean/1e7 and prepares
#' pdet$getCurrentParam()
#'
#' @author (C) 2021-2026 K. Nille-Hauf, T. Feiri, M. Ricker, T. Lux -- Hochschule Biberach (until 2022),
#' TU Dortmund University - Chair of Structural Concrete (since 2023)
#'
#' @export PARAM_DETVAR
#' @exportClass PARAM_DETVAR

PARAM_DETVAR <- setRefClass(
  Class = "PARAM_DETVAR",
  contains = "PROB_BASEVAR",
  fields = list(
    ParamValues = "vector",
    pos = "numeric"
  )
)

PARAM_DETVAR$methods(
  list(
    initialize = function(...) {
      initFields(...)
      DistributionType <<- "norm" # Eigentlich nicht notwendig, da PROB_BASE var bei DistributionTyp = "" auf "norm" setzt
      pos <<- 1L
      if (!is_empty(ParamValues)) nextParam()
    },
    nextParam = function() {
      ## 1) Naechsten Index bestimmen - Wrap-Around
      pos <<- (pos %% length(ParamValues)) + 1L # <- neuer, kompakter Code

      ## 2) Der tatsaechlich zu nutzende Index ist nun pos-1 (der vorherige Wert)
      i <- if (pos == 1L) length(ParamValues) else pos - 1L

      ## 3) Reset & Einsetzen der Parameter
      DistributionParameters <<- c(0, 0)


      Mean <<- ParamValues[i]
      Sd <<- Mean / 1e7
      Cov <<- ifelse(Mean == 0, 0, Sd / Mean)

      ## 4) Abschluss - interne Vorbereitung
      prepare()
      if (!is.null(.cache)) rm(list = ls(envir = .cache), envir = .cache)
    },
    getCurrentParam = function() {
      idx <- if (pos == 1L) length(ParamValues) else pos - 1L
      ParamValues[idx]
    }
  )
)

#' @name MC_IS
#' @title Monte Carlo Simulation with Importance Sampling
#' @description
#' Performs Monte Carlo simulation with Importance Sampling (IS) to estimate
#' structural failure probabilities for single limit state functions or
#' systems of limit state functions.
#'
#' The method reduces the variance of crude Monte Carlo by sampling from a
#' shifted (or multimodal) density in standard normal space and correcting
#' via likelihood ratios. For systems, a multimodal sampling density based on
#' FORM design points is used (cf. Melchers & Beck, Chapter 5).
#'
#' The implementation supports:
#' \itemize{
#'   \item Single limit state functions (FORM-based IS shift)
#'   \item Serial and parallel systems (union / intersection events)
#'   \item Multimodal sampling densities for system reliability
#'   \item Optional adaptive mixture weights (adaptive \eqn{\alpha_i})
#'   \item Parallel execution (via \pkg{future} or \pkg{parallel})
#'   \item Deterministic reproducibility via fixed seeds
#'   \item Automatic detection of vectorized limit-state functions
#' }
#'
#' @details
#' \strong{Single limit state case:}
#' The importance sampling density is constructed in standard normal space
#' using the FORM design point \eqn{u^*}. Samples are generated from
#' \eqn{u \sim \mathcal{N}(u^*, I)} and transformed to physical space.
#'
#'
#' \strong{System case:}
#' For a system with limit states \eqn{g_i}, the failure event is
#' \eqn{F = \cup_i F_i} (serial system) or
#' \eqn{F = \cap_i F_i} (parallel system).
#'
#' The sampling density follows Melchers & Beck (Eq. 5.26):
#' \deqn{
#'   h(u) = \sum_{i=1}^m \alpha_i \, \phi(u - u_i^*)
#' }
#' where \eqn{u_i^*} are the FORM design points in standard normal space and
#' \eqn{\alpha_i} are mixture weights.
#'
#' By default, mixture weights are proportional to individual failure
#' probabilities:
#' \deqn{
#'   \alpha_i \propto P(F_i).
#' }
#'
#' If \code{adaptive_alpha = TRUE}, mixture weights are updated iteratively
#' based on weighted failure frequency estimates:
#' \deqn{
#'   \alpha_i^{new} =
#'   (1-\lambda)\alpha_i^{old} +
#'   \lambda \frac{\hat{P}(F_i)}{\sum_k \hat{P}(F_k)}
#' }
#' where \eqn{\lambda} is given by \code{alpha_update_rate}.
#'
#' The failure probability is estimated using a self-normalized IS estimator:
#' \deqn{
#'   \hat{P}_f =
#'   \frac{\sum I(u) w(u)}{\sum w(u)}
#' }
#' with likelihood ratio
#' \deqn{
#'   w(u) = \frac{\phi(u)}{h(u)}.
#' }
#'
#' Numerical stability is ensured via global log-weight stabilization.
#'
#'
#' \strong{Stopping Criterion}
#'
#' The simulation terminates when both:
#' \itemize{
#'   \item The estimated coefficient of variation (CoV) falls below \code{cov_user}, and
#'   \item The effective sample size (ESS) exceeds a minimum threshold.
#' }
#'
#' For parallel systems with very small failure probabilities, the ESS-based
#' safeguard prevents premature termination due to unstable variance estimates.
#'
#'
#' \strong{Numerical stability}
#'
#' The default \code{"robust"} mode performs global log-domain
#' accumulation of importance sampling weights using a
#' log-sum-exp formulation. This avoids overflow and underflow
#' effects in high-dimensional or rare-event settings.
#'
#' The alternative \code{"fast"} mode uses shifted exponentiation
#' at the batch level and is computationally equivalent for most
#' engineering reliability problems.
#'
#'
#' \strong{Automatic detection of vectorized limit-state functions}
#'
#' For performance reasons, the algorithm automatically checks whether the
#' supplied limit-state function (LSF) supports vectorized evaluation.
#'
#' A short internal test is performed before the simulation starts. If the LSF
#' accepts a matrix of input samples and returns a numeric vector of matching
#' length, it is evaluated in fully vectorized form:
#'
#' \deqn{g(X_{1:n}) \rightarrow \{g(x_1), \dots, g(x_n)\}}
#'
#' Otherwise, the algorithm falls back to row-wise evaluation using
#' \code{apply()}.
#'
#' This mechanism is fully automatic and backward compatible.
#' Users do not need to modify existing scalar LSF definitions.
#'
#' For computationally expensive LSFs (e.g., nonlinear models or surrogate FEM
#' models), vectorized evaluation can significantly improve performance.
#'
#'
#' \strong{Parallel execution and future plan handling}
#'
#' When \code{backend = "future"} is used, the function respects the currently
#' active \pkg{future} plan.
#'
#' If no parallel plan has been set by the user (i.e., the active plan is
#' sequential), a temporary plan is created internally using:
#'
#' \itemize{
#'   \item \code{multicore} on Unix-like systems,
#'   \item \code{multisession} on Windows.
#' }
#'
#' The original plan is automatically restored after completion of the
#' simulation.
#'
#' If the user has already defined a parallel strategy via
#' \code{future::plan()}, it will not be modified.
#'
#' This design ensures CRAN compliance while preserving full flexibility
#' for advanced users.
#'
#'
#' \strong{Reproducibility}
#'
#' Reproducible results across different numbers of cores are ensured through:
#'
#' \itemize{
#'   \item L'Ecuyer-CMRG random number streams,
#'   \item deterministic master sampling,
#'   \item worker-invariant parallel chunking,
#'   \item controlled seed propagation.
#' }
#'
#' If \code{seed} is supplied, results are reproducible regardless of the
#' number of worker threads.
#'
#'
#' @param lsf A single limit-state function \code{function(x)} or a list of
#'   such functions for system reliability. Must follow the LSF interface
#'   specification described below.
#'
#' @param lDistr List of marginal distribution objects (funlists with \code{$d},
#'   \code{$p}, \code{$q}, \code{$r}) corresponding to input variables.
#'
#' @param cov_user Target coefficient of variation (CoV) for the Monte Carlo
#'   estimator. Simulation stops once this threshold is reached.
#'
#' @param n_batch Number of samples generated per iteration.
#'
#' @param n_max Maximum total number of samples (upper stopping limit).
#'
#' @param use_threads Number of worker threads (for parallel execution).
#'
#' @param backend Parallel backend, either \code{"future"} or \code{"parallel"}.
#'
#' @param sys_type Character string, either \code{"serial"} or \code{"parallel"},
#'   specifying system configuration.
#'
#' @param dataRecord Logical; if \code{TRUE}, intermediate results per iteration
#'   are stored and returned.
#'
#' @param beta_l Optional threshold; limit states with \eqn{\beta > beta_l}
#'   may be excluded in system analysis.
#'
#' @param densityType Sampling density type (currently \code{"norm"} supported).
#'
#' @param dps Optional vector of design points in physical space; if supplied,
#'   FORM analysis is skipped.
#'
#' @param debug.level Integer verbosity level (0 = silent, 1 = summary, 2 = detailed).
#'
#' @param seed Optional integer seed for reproducible random numbers.
#'
#' @param adaptive_alpha Logical; if \code{TRUE}, adaptive mixture weights are used
#'   (system case only).
#'
#' @param alpha_update_rate Damping factor \eqn{\lambda \in [0,1]} controlling
#'   adaptation speed of mixture weights.
#'
#' @param adaptive_batch Logical; if TRUE, the batch size is adapted dynamically
#'   based on the current effective sample size (ESS) and target RSE.
#'
#' @param batch_control List controlling adaptive batch behaviour:
#'   \describe{
#'     \item{n_min}{Minimum batch size}
#'     \item{n_max}{Maximum batch size}
#'     \item{K_future}{Desired number of remaining iterations}
#'   }
#'
#' @param min_adapt_samples Integer. Minimum total number of MC samples required before adaptive mixtures are updated.
#'    The adaption mechanism is activated only after this threshold has been exceeded to ensure statistically stable estimates.
#'
#' @param alpha_min Numeric scalar in \eqn{(0,1)}.
#'   Lower bound for mixture weights \eqn{\alpha_i} during adaptive
#'   updating. This prevents individual components of the importance
#'   sampling density from collapsing to zero and ensures numerical
#'   stability in multimodal system analyses.
#'
#' @param stability_mode Character string specifying the numerical
#'   stabilization strategy used for importance sampling weight
#'   accumulation.
#'
#'   \describe{
#'     \item{"robust"}{
#'       Default. Uses full log-domain accumulation of weights
#'       across Monte Carlo iterations. This improves numerical
#'       stability for high-dimensional systems and very small
#'       failure probabilities.
#'     }
#'     \item{"fast"}{
#'       Uses shifted exponentiation with batch-level scaling.
#'       Computationally equivalent under normal engineering
#'       conditions but may be slightly less stable in extreme
#'       rare-event scenarios.
#'     }
#'   }
#'
#'   Both modes are mathematically equivalent for typical
#'   reliability problems. The \code{"robust"} mode is recommended
#'   for research applications and extreme reliability levels.
#'
#' @return MC_IS returns an object containing the following elements:
#' \itemize{
#'   \item \code{method} - method identifier
#'   \item \code{pf} - estimated failure probability
#'   \item \code{beta} - reliability index \eqn{\beta = -\Phi^{-1}(P_f)}
#'   \item \code{var} - variance estimate
#'   \item \code{cov_mc} - coefficient of variation
#'   \item \code{n_mc} - number of samples used
#'   \item \code{ESS} - effective sample size
#'   \item \code{ESS_modes} - effective sample size per mixture component (system case)
#'   \item \code{data} - optional iteration history (if \code{dataRecord = TRUE})
#' }
#'
#' @section Limit-State Function (LSF) Interface:
#'
#' The argument \code{lsf} must define a valid limit-state function
#' using one of the following supported signatures:
#'
#' \describe{
#'   \item{\code{function(x)}}{
#'     A single numeric vector \code{x} containing all basic variables
#'     in the order defined in \code{lDistr}.
#'
#'     This form is fully compatible with vectorized evaluation
#'     in Monte Carlo and importance sampling algorithms.
#'   }
#'
#'   \item{\code{function(Z, Fy, M, ...)}}{
#'     Explicitly named scalar arguments corresponding to the variable
#'     names defined in the probabilistic model.
#'   }
#' }
#'
#' The use of \code{function(...)} as the sole argument is not supported.
#' Such usage may cause inconsistent behaviour between FORM-based and
#' Monte Carlo-based reliability methods.
#'
#' The vector form \code{function(x)} is recommended for
#' best performance and consistency.
#'
#' @section Windows users:
#'
#' On Windows systems, it is recommended to use
#' \code{backend = "future"} for multi-core parallelization.
#' The \code{"parallel"} backend does not support forking on Windows.
#'
#' @references
#' Ditlevsen, O., & Madsen, H. O. (1996).
#' \emph{Structural Reliability Methods}.
#' Wiley.
#'
#' Melchers, R. E., & Beck, A. T. (2018).
#' \emph{Structural Reliability Analysis and Prediction}.
#' Wiley.
#'
#' Spaethe, G. (1991).
#' \emph{Die Sicherheit tragender Baukonstruktionen}.
#' Springer.
#'
#' @author (C) 2021-2026 K. Nille-Hauf, T. Feiri, M. Ricker, T. Lux -- Hochschule Biberach (until 2022),
#' TU Dortmund University - Chair of Structural Concrete (since 2023)
#'
#' @import parallel
#' @import future
#' @import future.apply
#' @export
MC_IS <- function(lsf,
                  lDistr,
                  cov_user = 0.05,
                  n_batch = 5000,
                  n_max = 1e6,
                  use_threads = 4,
                  backend = NULL,
                  sys_type = "serial",
                  dataRecord = TRUE,
                  beta_l = 100,
                  densityType = "norm",
                  dps = NULL,
                  debug.level = 0,
                  seed = NULL,
                  adaptive_alpha = FALSE,
                  alpha_update_rate = 0.1,
                  adaptive_batch = FALSE,
                  batch_control = list(),
                  min_adapt_samples = 10000,
                  alpha_min = 0.02,
                  stability_mode = c("robust", "fast")) {
  stability_mode <- match.arg(stability_mode)

  ## -------------------------------------------------------------------------
  ## Backend selection with fallback and validation
  ## -------------------------------------------------------------------------
  valid_backends <- c("parallel", "future")

  ## message("Debug.level 0 %d", debug.level)

  if (is.null(backend)) {
    message("[TesiproV::MC_IS] No backend specified -- using 'future' by default.")
    backend <- "future"
  } else if (!(backend %in% valid_backends)) {
    stop(sprintf(
      "[TesiproV::MC_IS] Invalid backend '%s'. Valid options are: %s.",
      backend, paste(valid_backends, collapse = ", ")
    ), call. = FALSE)
  }

  ## Warning if parallel was chosen under Windows
  if (.Platform$OS.type == "windows" && backend == "parallel") {
    warning(
      "For Windows systems, backend = 'future' is recommended ",
      "for robust multi-core parallelization."
    )
  }

  ## Warning if no future plan was set by the user
  if (backend == "future") {
    current_plan <- future::plan()

    if (inherits(current_plan, "sequential")) {
      message(
        "[TesiproV::MC_IS] No active future plan detected.\n",
        "Computation will run sequentially.\n",
        "To enable parallel execution, set for example:\n",
        if (.Platform$OS.type == "windows") {
          "  future::plan(multisession, workers = 4)"
        } else {
          "  future::plan(multicore, workers = 4)"
        }
      )
    }
  }

  # ## Initialize Future plan
  # if (backend == "future") {
  #   current_plan <- future::plan()

  #   is_multicore <- inherits(current_plan, "multicore")
  #   is_multisession <- inherits(current_plan, "multisession")

  #   if (.Platform$OS.type == "windows") {
  #     if (!is_multisession) {
  #       future::plan(future::multisession, workers = use_threads)
  #     }
  #   } else {
  #     if (!is_multicore) {
  #       future::plan(future::multicore, workers = use_threads)
  #     }
  #   }
  # }

  # -------------------------------------------------------------------------
  ## RNG initialization depending on backend
  ## -------------------------------------------------------------------------
  if (backend == "future") {
    if (is.null(seed)) {
      # No fixed seed provided -> create and set a random global seed
      message("[TesiproV::MC_IS] No seed provided -- using random streams for 'future' backend.")

      rng_state <- init_rng_master(NULL, debug.level = debug.level) # <--- call the helper to draw and set a random seed!

      # The first integer element is sufficient as base seed for future.apply()
      seed_used <- NULL # must remain NULL for random seeding

      # --- Optional debug output -----------------------------------------------
      if (debug.level >= 1) {
        message("[TesiproV::MC_IS] Future backend running with automatic random substreams.")
        message(sprintf(
          "[TesiproV::MC_IS] Global RNG state head: %s",
          paste(.Random.seed[1:5], collapse = ", ")
        ))
      }
    } else {
      # Fixed deterministic seed -> reproducible run
      message("[TesiproV::MC_IS] Fixed seed provided -- using reproducible streams for 'future' backend.")

      rng_state <- init_rng_master(seed, debug.level = debug.level)

      # The first integer element of the returned stream is sufficient as base seed
      # seed_used <- as.integer(rng_state[1])
      seed_used <- seed
    }
  } else if (backend == "parallel") {
    if (!is.null(seed)) {
      message("[TesiproV::MC_IS] Fixed seed provided -- using reproducible streams for 'parallel' backend.")
      rng_state <- init_rng_master(seed, debug.level = debug.level)
      seed_used <- rng_state
    } else {
      message("[TesiproV::MC_IS] No seed provided -- using random streams for 'parallel' backend.")
      rng_state <- init_rng_master(seed, debug.level = debug.level)
      seed_used <- NULL
    }
  }

  ## -------------------------------------------------------------------------
  ## -------------------------------------------------------------------------
  ## Windows fallback for threads
  ## -------------------------------------------------------------------------
  if (.Platform$OS.type == "windows") {
    message("[MC_IS] Windows detected - PSOCK cluster will be used (if backend == 'parallel').")
  }
  # if (.Platform$OS.type == "windows") use_threads <- 1L

  ## -------------------------------------------------------------------------
  ## Cluster library paths setup
  ## -------------------------------------------------------------------------
  libPaths_local <- .libPaths()
  # assign("libPaths_local", libPaths_local, envir = .GlobalEnv)

  ## -------------------------------------------------------------------------
  ## Decide which mode to run: system or single limit state function
  ## -------------------------------------------------------------------------

  if (is.list(lsf)) {
    return(
      MC_IS_system(
        lsf = lsf, lDistr = lDistr,
        cov_user = cov_user, n_batch = n_batch, n_max = n_max,
        use_threads = use_threads, sys_type = sys_type, dataRecord = dataRecord,
        beta_l = beta_l, densityType = densityType, dps = dps,
        debug.level = debug.level, streams = NULL, libPaths_local = libPaths_local,
        seed = seed_used, backend = backend, adaptive_alpha = adaptive_alpha,
        alpha_update_rate = alpha_update_rate,
        adaptive_batch = adaptive_batch,
        batch_control = batch_control,
        min_adapt_samples = min_adapt_samples,
        alpha_min = alpha_min,
        stability_mode = stability_mode
      )
    )
  } else if (is.function(lsf)) {
    return(
      MC_IS_single(
        lsf = lsf, lDistr = lDistr, cov_user = cov_user, n_batch = n_batch,
        n_max = n_max, use_threads = use_threads, dataRecord = dataRecord,
        densityType = densityType, dps = dps,
        debug.level = debug.level, streams = NULL,
        libPaths_local = libPaths_local,
        seed = seed_used, backend = backend,
        adaptive_alpha = adaptive_alpha,
        alpha_update_rate = alpha_update_rate,
        min_adapt_samples = min_adapt_samples,
        alpha_min = alpha_min,
        stability_mode = stability_mode
      )
    )
  } else {
    stop("[TesiproV::MC_IS] Argument 'lsf' must be either a function (single LSF) or a list of functions (system).",
      call. = FALSE
    )
  }
}

# File: R/parallel_helpers.R

#' @title Create a parallel cluster (fork on *nix, PSOCK on Windows)
#' @description Returns a `parallel` cluster object that already has the
#'   RNG streams (L'Ecuyer‑CMRG) and required packages loaded.
#' @param use_threads Integer \>=1 – number of workers.
#' @param seed Optional integer or L'Ecuyer‑CMRG stream (length 7) for reproducibility.
#' @return A cluster object (or NULL when the backend is "future").
#' @keywords internal
make_parallel_cluster <- function(use_threads, seed = NULL) {
    ## -------------------------------------------------
    ## 1) Determine cluster type (fork vs. PSOCK)
    ## -------------------------------------------------
    cl_type <- if (.Platform$OS.type == "windows") "PSOCK" else "FORK"

    ## -------------------------------------------------
    ## 2) Create the cluster
    ## -------------------------------------------------
    cl <- parallel::makeCluster(use_threads, type = cl_type)

    ## -------------------------------------------------
    ## 3) Initialise RNG streams
    ## -------------------------------------------------
    if (!is.null(seed)) {
        ## User supplied seed (integer or full CMRG stream)
        parallel::clusterSetRNGStream(cl, iseed = seed)
    } else {
        ## Random sub‑streams – reproducible later via future.seed = TRUE
        parallel::clusterSetRNGStream(cl)
    }

    ## -------------------------------------------------
    ## 4) Load required packages on the workers
    ## -------------------------------------------------
    parallel::clusterEvalQ(cl, {
        suppressMessages({
            if (!requireNamespace("TesiproV", quietly = TRUE)) {
                stop("Package 'TesiproV' not found on worker.", call. = FALSE)
            }
            ## Add additional packages that LS‑functions may need, e.g.:
            ## library(evd); library(EnvStats); …
            invisible(NULL)
        })
    })

    return(cl)
}


#' @title Deterministic RNG Block Manager (L'Ecuyer-CMRG)
#' @description
#' Provides deterministic generation of random number blocks that are
#' independent of the number of worker threads.
#'
#' The RNG state is advanced only by the requested block size.
#'
#' @param seed Optional integer for reproducibility.
#' @return An RNG manager object (environment).
#' @keywords internal
create_rng_manager <- function(seed = NULL) {
    RNGkind("L'Ecuyer-CMRG")

    if (!is.null(seed)) {
        set.seed(seed)
    }

    rng_env <- new.env(parent = emptyenv())
    rng_env$current_state <- .Random.seed

    rng_env$generate_norm_block <- function(n) {
        assign(".Random.seed", rng_env$current_state, envir = .GlobalEnv)
        values <- rnorm(n)
        rng_env$current_state <- .Random.seed
        return(values)
    }

    return(rng_env)
}


# File: R/parallel_helpers.R
#' @title Run a parallel computation using either "parallel" or "future"
#' @description Small wrapper that hides the branching between the two back‑ends.
#'   It always returns a **named list** whose names are identical to the names
#'   of the input split list. This guarantees that `unlist()` produces the same
#'   named vector for both back‑ends, which is required by the unit tests.
#' @param X List (or vector) of inputs that will be split among the workers.
#' @param FUN Function that will be executed on each chunk.
#' @param backend Either "parallel" or "future".
#' @param use_threads Number of workers (relevant for both back‑ends).
#' @param cl Optional cluster object created with \code{parallel::makeCluster()}.
#'   Only required when using the \code{"parallel"} backend on Windows, where
#'   \code{parallel::parLapply()} is used. Ignored on Unix-like systems where
#'   \code{parallel::mclapply()} is available.
#' @return List of results (identical to the output of `parLapply` /
#'   `future_lapply`). The list is **named** – the names are taken from the
#'   split index (`split_idx`).
#' @keywords internal
#'
run_parallel <- function(X,
                         FUN,
                         backend = c("parallel", "future"),
                         use_threads = NULL,
                         cl = NULL) {
    backend <- match.arg(backend)

    if (backend == "parallel") {
        if (use_threads <= 1L) {
            return(lapply(X, FUN))
        }

        ## ---- Linux / macOS: use fork directly ----
        if (.Platform$OS.type != "windows") {
            return(parallel::mclapply(
                X,
                FUN,
                mc.cores = use_threads,
                mc.preschedule = TRUE
            ))
        }

        ## ---- Windows: use PSOCK cluster ----
        if (is.null(cl)) {
            stop("Cluster object 'cl' must be provided for Windows parallel backend.")
        }

        return(parallel::parLapply(cl, X, FUN))
    }

    ## ---- future backend ----
    return(future.apply::future_lapply(
        X = X,
        FUN = FUN,
        future.seed = FALSE, # deterministic master RNG
        future.globals = FALSE # prevent unnecessary global export
    ))
}

library(testthat)

test_that("make_parallel_cluster works on current OS", {
    cl <- make_parallel_cluster(use_threads = 2)
    expect_true(inherits(cl, "cluster"))
    parallel::stopCluster(cl)
})


library(testthat)

test_that("run_parallel works with future (default plan)", {
    chunks <- split(1:10, cut(1:10, 2))

    out <- run_parallel(
        X = chunks,
        FUN = function(v) sum(v),
        backend = "future",
        use_threads = 2
    )

    expect_equal(
        unlist(out),
        unlist(lapply(chunks, sum))
    )
})


test_that("RNG manager reproducibility", {
    rng1 <- create_rng_manager(seed = 123)
    rng2 <- create_rng_manager(seed = 123)

    block1 <- rng1$generate_norm_block(1000)
    block2 <- rng2$generate_norm_block(1000)

    expect_equal(block1, block2)
})


test_that("RNG manager advances state correctly", {
    rng <- create_rng_manager(seed = 123)

    block1 <- rng$generate_norm_block(10)
    block2 <- rng$generate_norm_block(10)

    set.seed(123)
    full <- rnorm(20)

    expect_equal(block1, full[1:10])
    expect_equal(block2, full[11:20])
})

#' @name SORM
#' @title Second-Order Reliability Method (SORM)
#' @description
#' Second-Order Reliability Method (SORM) for the approximation
#' of failure probabilities in structural reliability analysis.
#'
#' The SORM extends the First-Order Reliability Method (FORM)
#' by incorporating second-order information of the limit-state function
#' at the design point. By accounting for local curvature effects,
#' SORM generally provides more accurate failure probability estimates
#' for nonlinear limit-state functions.
#'
#' The method evaluates the principal curvatures of the limit-state surface
#' in standard normal space and applies an asymptotic correction
#' to the FORM probability of failure.
#'
#' @param lsf objective function with limit state function in form of
#'   \code{function(x) \{ x[1] + x[2] + ... \}}.
#' @param lDistr list ob distribiutions regarding the distribution object of TesiproV
#' @param debug.level If 0 no additional info if 2 high output during calculation
#'
#' @return SORM returns an object containing the following elements:
#' \itemize{
#'   \item \code{beta}: Hasofer-Lind reliability index.
#'   \item \code{pf}: Estimated probability of failure.
#'   \item \code{u_points}: Design point(s) in the standard normal space (\eqn{u}-space).
#'   \item \code{dy}: Gradient(s) of the limit-state function at the design point.
#' }
#' @import pracma
#' @author (C) 2021-2026 K. Nille-Hauf, T. Feiri, M. Ricker, T. Lux -- Hochschule Biberach (until 2022),
#' TU Dortmund University - Chair of Structural Concrete (since 2023)
#' @references
#' Breitung, K. (1989).
#' Asymptotic approximations for probability integrals.
#' \emph{Probabilistic Engineering Mechanics}, 4(4), 187-190.
#'
#' Cai, G. Q., & Elishakoff, I. (1994).
#' Refined second-order reliability analysis.
#' \emph{Structural Safety}, 14(4), 267-276.
#'
#' Hohenbichler, M., Gollwitzer, S., Kruse, W., & Rackwitz, R. (1987).
#' New light on first- and second-order reliability methods.
#' \emph{Structural Safety}, 4, 267-284.
#'
#' Tvedt, L. (1990).
#' Distribution of quadratic forms in normal space -
#' Applications to structural reliability.
#' \emph{Journal of Engineering Mechanics}, 116(6), 1183-1197.
#'
#' Marelli, S., & Sudret, B. (2014).
#' UQLab: A framework for uncertainty quantification in Matlab.
#' In \emph{Proceedings of the 2nd International Conference on
#' Vulnerability, Risk Analysis and Management (ICVRAM2014)},
#' Liverpool, United Kingdom, 2554-2563.
#'
#' Lacaze, S., & Missoum, S. (2015).
#' CODES: A toolbox for computational design (Version 1.0).
#' Retrieved from \url{http://www.codes.arizona.edu/toolbox/}
#'
#' Wu, X. Z. (2017).
#' Implementing statistical fitting and reliability analysis
#' for geotechnical engineering problems in R.
#' \emph{Georisk: Assessment and Management of Risk for Engineered Systems and Geohazards},
#' 11(2), 173-188.
#' @export

SORM <- function(lsf, # the limit-state function.
                 lDistr,
                 debug.level=0){ #  description of distribution of random variables

  debug.TAG <- "SORM_RABU"
  debug.print(debug.level,debug.TAG,c(TRUE), msg="SORM started...")
  tic <- proc.time()

  # Isoprobabilistic transformation to the standard Gaussian space U
  # Tlsf, convenient for users and required by mistral::FORM. AZC-style Tinv could be an alternative, e.g. Tinv <- function(X) { lsf(qnorm(pnorm(X),0.25,1)) }
  Tlsf <- function(X) {
    for(i in 1:length(X)){
      XU <- stats::pnorm(X[i])
      if(XU == 1){ XU = 1-1E-9}
      X[i] <- lDistr[[i]][[1]]$q(XU)
    }
    return(lsf(X))
  }


  # Orthonormal rotation matrix (implementation inspired by uq_gram_schmidt)
  GRAM_RABU <- function(alphas) {
    dimension <- length(alphas)
    Base<-diag(dimension)
    Base[,dimension] <- alphas
    NewBase <- Base
    Rot <- matrix( rep( 0, dimension*dimension), nrow = dimension)
    Rot[dimension,dimension] <- 1
    for (k in dimension:1){
      Z <- matrix( rep( 0, dimension), nrow = dimension)
      j <- k+1
      while(j<=dimension) {
        Z <- Z + pracma::dot(Base[,k],NewBase[,j]) * NewBase[,j]
        j <- j+1
      }
      NewBase[,k] <- Base[,k] - Z;
      Rot[k,k] <- pracma::Norm(NewBase[,k],2);
      NewBase[,k] <- NewBase[,k] / Rot[k,k];
      if (k<dimension) {
        Rot[k,(k+1):dimension] <- t(Base[,k]) %*% NewBase[,(k+1):dimension];
      }
    }
    return(t(NewBase))
  }




  # FORM
  res <- TesiproV::FORM(lsf, lDistr,loctol = 0.01,n_optim=10)

  # Alpha's (direction cosines) as components of the unit gradient vector:
  # The alphas can be determined through the gradient vector in the U space at the design-point u*
  gradient <- res$dy
  alphas <- gradient/pracma::Norm(gradient)

  # Matrix R
  R <- GRAM_RABU(alphas)

  # Hessian Matrix H
  H <- pracma::hessian(Tlsf, res$x_points) # Second-derivative Hessian matrix at the design-point u*


  # Matrix A as a product of R,H,R'divided by the lenght of the gradient vector at the design-point u*
  A <- R %*% H %*% t(R)
  A <- A/pracma::Norm(gradient)

  # Curvatures (found through the eigenvalues of the reduced matrix A
  # Fixed by MR 19-01-2026: older R versions silently accepted vector conditions in 'if()',
  # but newer versions require a single logical value (length == 1).
  # !any() and any() added inside the if() conditions.
  if(!any(!is.nan(A))){
    curvatures <- eigen(A[1:nrow(A)-1,1:nrow(A)-1])$values
    warning("No corvature found. Maybe bad designpoints of FORM.")
  }else{
    curvatures <- rep(0,length(gradient))
  }
  if (any(curvatures >= 1)) {warning("FORM did not converge to the design point")}
  debug.print(0,debug.TAG,c("curvatures"),curvatures)

  # SORM: Reliability Index (Beta) and Probability of failure (Pf)
  # Breitung's formula
  SORMpfBreitung <- res$pf/sqrt(prod(1+res$beta*curvatures)) # Breitung's Formula
  SORMbetaBreitung <- -stats::qnorm(SORMpfBreitung) # Beta_SORM   norminv() in AZC, icdf() in UQL
  # NOTE: Check if the starting point was negative or not based on the origin point from FORM (as in the UQLab)

  # HohenBichler's formula
  CDFminusBeta <- stats::pnorm(-res$beta, 0, 1);
  RatioCDFPDF <- stats::dnorm(-res$beta, 0, 1)/CDFminusBeta;
  HohenbichlerVec <- 1./sqrt(1 + RatioCDFPDF*curvatures);
  SORMpfHB <- CDFminusBeta*prod(HohenbichlerVec);
  SORMbetaHB <- (-1)*stats::qnorm(SORMpfHB)

  # Tvedt's formula
  A11 <- 1/sqrt(prod(1+res$beta*curvatures))
  A12 <- res$beta*res$pf-stats::dnorm(-res$beta)
  A1 <- res$pf*A11
  A2 <- A12*(A11-1/sqrt(prod(1+(res$beta+1)*curvatures)))
  A3 <- (res$beta+1)*A12*(A11-Re(1/sqrt(prod(1+(res$beta+1i)*curvatures))));
  SORMpfTvedt <- A1+A2+A3;
  SORMbetaTvedt <- -stats::qnorm(SORMpfTvedt)


  cat("\n")
  duration<-proc.time()-tic

  # Return the results
  output <- list(
            "method"="SORM",
            "beta"=SORMbetaBreitung,
            "pf"=SORMpfBreitung,
            "FORM_pf"= res$pf,
            "FORMbeta"=res$beta,
            "design.point_X"=res$x_points,
            "SORMpfTvedt"=SORMpfTvedt,
            "SORMbetaTvedt"=SORMbetaTvedt,
            "SORMpfHB"=SORMpfHB,
            "SORMbetaHB"=SORMbetaHB,
            "runtime"=duration[1:5]
            )

  debug.print(debug.level,debug.TAG,c(duration), msg="SORM finished in [s]: ")
  return(output)
  }

#' @name MC_SubSam
#' @title Monte Carlo Simulation with Subset-Sampling
#' @description
#' MonteCarlo with Subset-Sampling (MC_SubSam) method for the
#' estimation of small failure probabilities in structural reliability analysis.
#'
#' Subset Simulation is an advanced Monte Carlo technique that
#' expresses a rare failure event as a sequence of intermediate
#' conditional events with higher probabilities.
#'
#' The method combines conditional sampling with Markov Chain Monte Carlo (MCMC)
#' techniques to efficiently estimate very small probabilities of failure,
#' which would be computationally expensive to obtain using crude Monte Carlo simulation.
#'
#'
#' @param lsf limit-state function
#' @param lDistr list of basevariables in input space
#' @param Nsubset number of samples in each simulation level
#' @param p0 level probability or conditional probability
#' @param MaxSubsets maximum number of simulation levels that are used to terminate the simulation procedure to avoid infinite loop when the target domain cannot be reached
#' @param Alpha  confidence level
#' @param variance gaussian, uniform
#' @param debug.level If 0 no additional info if 2 high output during calculation
#' @param seed Optional integer value. If specified, sets a fixed seed for reproducible random numbers (useful for testing).
#'
#' @return MC_SubSam returns an object containing the following elements:
#' \itemize{
#'   \item \code{beta}: Estimated reliability index.
#'   \item \code{pf}: Estimated probability of failure.
#'   \item \code{betaCI}: Confidence interval of the reliability index.
#'   \item \code{pfCI}: Confidence interval of the probability of failure.
#'   \item \code{CoV}: Coefficient of variation of the estimator.
#'   \item \code{NumOfSubsets}: Number of Markov chains (subsets).
#'   \item \code{NumOfEvalLSF_nom}: Nominal number of limit-state function evaluations
#'         (Markov chains / iterations).
#'   \item \code{NumOfEvalLSF_eff}: Effective number of limit-state function evaluations.
#'   \item \code{runtime}: Total runtime of the function.
#' }
#'
#' @references
#' Au, S. K., & Beck, J. L. (2001).
#' Estimation of small failure probabilities in high dimensions
#' by subset simulation.
#' \emph{Probabilistic Engineering Mechanics}, 16(4), 263-277.
#'
#' Marelli, S., & Sudret, B. (2014).
#' UQLab: A framework for uncertainty quantification in Matlab.
#' In \emph{Proceedings of the 2nd International Conference on
#' Vulnerability, Risk Analysis and Management (ICVRAM2014)},
#' Liverpool, United Kingdom, 2554-2563.
#'
#' @author (C) 2021-2026 K. Nille-Hauf, T. Feiri, M. Ricker, T. Lux -- Hochschule Biberach (until 2022),
#' TU Dortmund University - Chair of Structural Concrete (since 2023)
#' @import pracma
#' @export

MC_SubSam <- function(lsf, # limit-state function
                        lDistr, # dimension of the input space
                        Nsubset=1e5, # number of samples in each simulation level
                        p0=0.1, # level probability or conditional probability
                        MaxSubsets=10, # maximum number of simulation levels that are used to terminate the simulation procedure to avoid infinite loop when the target domain cannot be reached
                        Alpha=0.05, # confidence level
                        variance="uniform",
                        debug.level=0, # gaussian, uniform
                        seed = NULL) # seed for random number generator
{

  # set seed for random number generator Mersenne Twister
  if (!is.null(seed)) {
    set.seed(seed)
    message(sprintf("Fixed RNG seed set to %d", seed))
  }

  RNGkind("Mersenne-Twister", "Inversion")

  debug.TAG <- "MC_SUS"
  debug.print(debug.level,debug.TAG,c(TRUE), msg="Monte-Carlo Simulation with Subset Sampling started...")
  tic<-proc.time()
  # === Transformation to the standard Gaussian space U and initial algorithm parameters definition ==

  # Isoprobabilistic transformation to the standard Gaussian space U
  Tlsf <- function(X) {
    XU <- stats::pnorm(X)
    for(i in 1:length(X)){
      X[i] <- lDistr[[i]][[1]]$q(XU[i, drop = FALSE])
    }
    return(lsf(X))
  }

  # Algorithm parameters definition (they are used for the "minimum" length of the Markov Chains)
  chainlength <- floor(1/p0)
  chainnumber <- floor(Nsubset*p0)
  n <- length(lDistr)

  # ====  SUBSET SIMULATION WITH MONTE CARLO SIMULATION (MCS) ====

  # === Initial MCS Stage: generate samples and read the function values ==
  ii <- 1
  U <- matrix(ncol=n,stats::rnorm(Nsubset*n,0,1))  # Generate samples from a standard normal distribution
  LSF <- matrix(apply(U,1,Tlsf))  # Compute the model responses for the samples (in standard normal space)
  LFScounter = nrow(U)  # Indicator to count the number of times the LSF is called
  PDFeval <- matrix(apply(stats::dnorm(U, 0, 1),1,prod)) # Compute the PDF values in the standard normal space
  q <- c()
  Pf <- c()
  Pfcond <- c()
  LSFhistory <- c()

  # === MCS stage: iterations among the subsets ==
  while (1) {
    Pf[ii] <- sum((LSF[]<=0))/Nsubset   # Estimate the failure probability
    LSFhistory[[ii]] <- matrix(LSF)     # Store the history

    # Terminate the iteration due to convergence or maximum number of subsets
    if ((Pf[ii] > p0) || (ii >= MaxSubsets)) {
      Pfcond[ii] <- Pf[ii]
      q[ii] <- 0
      break
    }

        # Estimate the quantile
    q[ii] <- stats::quantile(LSF, p0, type=5)
    Pfcond[ii] <- p0

    # Estimate the seeds for the next MCMC
    idxsort <- sort(LSF,index.return=TRUE)$ix
    Q <- U[idxsort[1:chainnumber],]
    QPDF <- matrix(PDFeval[idxsort[1:chainnumber]])

    # Update the MCMC options
    AcceptCritGi <- q[ii]
    AcceptCritLSF <- matrix(LSF[idxsort[1:chainnumber]]) # Acceptance criterion for the next subset
    RuntimeLSF = AcceptCritLSF;

    Xseed <- Q
    Yseed <- matrix(apply(stats::dnorm(Xseed),1,prod))
    Xsample <- matrix(ncol=ncol(Xseed),nrow=0)
    Ysample <- matrix(ncol=ncol(Yseed),nrow=0)

    while (1) {

      if (variance == "gaussian") {
        Xprop = Xseed + (matrix(stats::rnorm(prod(dim(Xseed))),nrow(Xseed),ncol(Xseed))*1)
      } else { # if (variance == "uniform") {
        Xprop = Xseed + ((matrix(stats::runif(prod(dim(Xseed))),nrow(Xseed),ncol(Xseed))*2-1)*1)
      }

      a = size(Xseed);
      b = size(Xprop);
      c = size(AcceptCritLSF);

      # Metropolies-Hasting MCMC component-wise

      UU <- Xseed  # repmat(Xseed,a(2),1);
      for (i in 2:a[2]) { UU <- rbind(UU,Xseed) }
      UPROP <- Xprop # repmat(Xprop, a(2), 1);
      for (i in 2:a[2]) { UPROP <- rbind(UPROP,Xprop) }
      jj = pracma::kron(pracma::eye(a[2]), pracma::ones(a[1],1));
      UU[jj==1] = UPROP[jj==1];
      YPROP <- apply(stats::dnorm(UU),1,prod) # prod(normpdf(UU),2);
      YY <- Yseed # repmat(Yseed, a(2), 1);
      for (i in 2:a[2]) { YY <- rbind(YY,Yseed) }
      idx <- (matrix(stats::runif(b[1]*b[2])) < YPROP/YY); # rand(b(1)*b(2),1)
      IDX <- matrix(nrow=a[1], data=idx)

      # Write an updated proposed candidate sample
      Uprop2 <- Xseed;
      Uprop2[IDX] = Xprop[IDX];
      Yprop <- apply(stats::dnorm(Uprop2),1,prod)
      Xprop <- Uprop2;

      idx <- rowSums(IDX)
      idx[rowSums(IDX) != 0] = 1;

      # When X is in a range of PDF=0
      idx[Yseed == 0] = 1;

      # Assigning the new samples of the Markov Chain
      LSFacc = NaN*zeros(a[1], c[2]);
      for (i in 1:a[1]) {
        if (idx[i]) {
          LSFacc[i] <- Tlsf(Xprop[i,])
          LFScounter = LFScounter + 1
        }
      }

      for (i in 1:nrow(Xseed)) {
        if (idx[i] && LSFacc[i]<AcceptCritGi) {
          Xseed[i,] <- Uprop2[i,]
        }
      }
      for (i in 1:nrow(Yseed)) {
        if (idx[i] && LSFacc[i]<AcceptCritGi) {
          Yseed[i] <- Yprop[i]
        }
      }

      # Assign the new samples according to the threshold
      for (i in 1:nrow(AcceptCritLSF)) {
        if (idx[i] && LSFacc[i]<AcceptCritGi) {
          AcceptCritLSF[i,] <- LSFacc[i,]
        }
      }
      RuntimeLSF <- rbind(RuntimeLSF,AcceptCritLSF)

      # Define the new sample vector
      Xsample <- rbind(Xsample, Xseed) # assign LSF to the mcmcopts to have it somewhere here
      Ysample <- rbind(Ysample, Yseed) # Also store the (calculated) limit-state function evaluations
      if (nrow(Xsample) >= round((1 - p0)*Nsubset)) {break;} #Stopping criterion

      loc_pf <- p0^(ii-1)*Pf[length(Pf)] # Probability of failure
      loc_beta <- -stats::qnorm(loc_pf)
      info.print(debug.TAG,debug.level,c("ii","Beta","pf"),c(ii,loc_beta, loc_pf))
    }

    U <- rbind(Q,Xsample[1:(Nsubset-nrow(Q)),])
    PDFeval <- rbind(QPDF,matrix(ncol=1,Ysample[1:(Nsubset-nrow(Q)),]))
    LSF <- RuntimeLSF[1:Nsubset,]
    ii <- ii + 1 #  Increase the counter
  }


  # === Final MCS stage: collect the results ==
  ResPf <- p0^(ii-1)*Pf[length(Pf)] # Probability of failure
  ResBeta <- -stats::qnorm(ResPf) # Reliability index
  debug.print(debug.level,debug.TAG,ResPf,"Results Pf:")
  debug.print(debug.level,debug.TAG,ResBeta,"Results Beta:")


  # Coefficient of Variation
  delta2 <- c((1-Pfcond[1])/Pfcond[1]/Nsubset)

  # Computation of gamma
  for (jj in 1:ii) {
    # rearrange the LSFhistory vector to a matrix in order to have each chain separate
    LSFH = matrix(nrow=chainnumber, data=LSFhistory[[jj]][1:(chainlength * chainnumber)])
    Indicator = pracma::zeros(nrow(LSFH),ncol(LSFH));
    Indicator[LSFH <= q[jj]] = 1;

    # Compute the rho (i.e. the auto-correlation between the Markov Chain samples in the i-th subset)
    # and consider every chain independently
    Rik <- vector() #= [];
    for (k in 1:chainlength - 1) {
      sums <- 0;
      for (l in 1 : ((Nsubset/chainnumber)-k)) {
        sums <- sums + sum(Indicator[,l] * Indicator[,l+k]);
      }
      Rik[k] <- sums * 1/(Nsubset - k * chainnumber) - Pfcond[jj]^2;
    }
    # Compute gamma for each each iteration
    rho <- Rik / (Pfcond[jj]*(1-Pfcond[jj]));
    gamma <- 2 * sum( rho*(1-(1:length(rho))*chainnumber / Nsubset) );

    delta2[jj] = (1-Pfcond[jj])/Pfcond[jj]/Nsubset*(1+gamma)
  }

  debug.print(debug.level,debug.TAG,ii,"Number of subsets")
  debug.print(debug.level,debug.TAG,ii*Nsubset,"Number of model evaluations (nominally)")
  debug.print(debug.level,debug.TAG,LFScounter,"Number of model evaluations (effectively)")


  # Compute the confidence bounds
  CoV <- sqrt(sum(delta2))
  ResPfCI <- ResPf * matrix(c(1 + stats::qnorm(Alpha/2)*CoV, 1 + stats::qnorm(1-Alpha/2)*CoV), nrow=1, byrow=TRUE)

  debug.print(debug.level,debug.TAG,CoV,"CoV: ")
  debug.print(debug.level,debug.TAG,ResPfCI,"ResPfCI: ")
  debug.print(debug.level,debug.TAG,t(apply(-stats::qnorm(ResPfCI),1,rev)),"ResBetaCI: ")


  cat("\n")
  duration<-proc.time()-tic

  output<-list(
    "method"="MCS",
    "beta"=ResBeta,
    "pf"=ResPf,
    "betaCI"=t(apply(-stats::qnorm(ResPfCI),1,rev)),
    "pfCI"=ResPfCI,
    "CoV"=CoV,
    "NumOfSubsets"=ii,
    "NumOfEvalLSF_nom"=ii*Nsubset,
    "NumOfEvalLSF_eff"=LFScounter,
    "runtime"=duration[1:5]
  )
  info.print(debug.TAG,debug.level,c("Beta","pf","CoV","NumOfEvalLSF_eff"),c(ResBeta,ResPf,CoV,LFScounter))

  debug.print(debug.level,debug.TAG,c(duration), msg="Monte-Carlo Simulation with SubsetSampling finished in: ")
  return(output)
}

# ====  END OF SUBSET SIMULATION WITH MCS ====

#' TesiproV: A package for the calculation of reliability and failure probability in civil engineering
#'
#' The Package provides three main types of objects:
#' \enumerate{
#' \item Objects for modeling base variables
#' \item Objects for modeling limit state functions and systems of them
#' \item Objects for modeling solving algorithms
#' }
#'
#'
#' By creating and combining those objects, one is able to model quite complex problems in terms of structural reliablity calculation.
#' For normally distributed variables there might be an workflow to calculate correlated problems (but no systems then).
#' There is also implemented a new distribution (logStudentT, often used for conrete compression strength) to show how one can implement
#' your very own or maybe combined multi modal distribution and use it with TesiproV.
#'
#'
#' @section Objects for base variables:
#' \code{PROB_BASEVAR}, \code{PROB_DETVAR}, \code{PARAM_BASEVAR}, \code{PARAM_DETVAR}
#'
#' @section Limit state functions:
#' \code{SYS_LSF}, \code{PROB_SYS}, \code{PARAM_SYS}
#'
#' @section Solving algorithms:
#' \code{PROB_MACHINE}
#'
#' @author (C) 2021-2026 K. Nille-Hauf, T. Feiri, M. Ricker, T. Lux -- Hochschule Biberach (until 2022),
#' TU Dortmund University - Chair of Structural Concrete (since 2023)
#'
"_PACKAGE"

#' @title System Probability Solution Object (SYS_PROB)
#'
#' @description
#' The \code{SYS_PROB} class represents a probabilistic system
#' consisting of one or more limit-state functions (\code{SYS_LSF})
#' and a set of reliability algorithms (\code{PROB_MACHINE}).
#'
#' It allows the execution of reliability analyses for
#' serial or parallel systems, aggregation of single
#' limit-state results, and system-level probability evaluation.
#'
#' Each limit-state function may include multiple random variables
#' (\code{PROB_BASEVAR}) with different probability distributions.
#' The transformation between mean/standard deviation and the
#' native distribution parameters is handled automatically.
#'
#' \itemize{
#'
#'   \item Normal ("norm"):
#'   mean = \eqn{\mu}, standard deviation = \eqn{\sigma}.
#'   Parameters: mean, sd.
#'
#'   \item Lognormal ("lnorm"):
#'   defined by mean \eqn{m} and sd \eqn{s}.
#'   Parameters:
#'   \eqn{\mu = \log\left( \frac{m}{\sqrt{1 + (s/m)^2}} \right)},
#'   \eqn{\sigma = \sqrt{ \log\left( 1 + (s/m)^2 \right) }}.
#'
#'   \item shifted Lognormal ("slnorm"):
#'   defined by Mean \eqn{m}, Sd \eqn{s} and \eqn{x0}.
#'   Parameters:
#'   \eqn{\mu = \log\left( \frac{m-x0}{\sqrt{1 + (s/(m-x0))^2}} \right)},
#'   \eqn{\sigma = \sqrt{ \log\left( 1 + (s/(m-x0))^2 \right) }}.
#'
#'   \item Gumbel ("gumbel", package "evd"):
#'   location-scale type I extreme value distribution.
#'   Parameters:
#'   location = \eqn{\mu - \gamma \cdot \beta},
#'   where \eqn{\gamma} is the Euler-Mascheroni constant,
#'   and
#'   \eqn{\beta = \frac{\sigma \sqrt{6}}{\pi}}.
#'
#'   \item Gamma ("gamma"):
#'   shape-scale parameterization.
#'   Parameters:
#'   \eqn{k = \frac{\mu^2}{\sigma^2}},
#'   \eqn{\theta = \frac{\sigma^2}{\mu}}.
#'
#'   \item Exponential ("exp"):
#'   rate \eqn{\lambda} or scale \eqn{\theta = 1/\lambda}.
#'   Parameter:
#'   \eqn{\lambda = 1/\mu}.
#'
#'   \item Beta ("beta"):
#'   bounded on \eqn{[0,1]}.
#'   Parameters:
#'   \eqn{\alpha = \mu t},
#'   \eqn{\beta = (1 - \mu) t},
#'   with
#'   \eqn{t = \frac{\mu (1-\mu)}{\sigma^2} - 1}.
#'
#'   \item Weibull ("weibull"):
#'   shape \eqn{k} and scale \eqn{\lambda}.
#'   Parameters:
#'   empirical approximation
#'   \eqn{k \approx \left( \frac{\sigma}{\mu} \right)^{-1.086}},
#'   \eqn{\lambda = \frac{\mu}{\Gamma\left(1 + \frac{1}{k}\right)}}.
#'
#'   \item Empirical ("emp"):
#'   non-parametric distribution defined by observed sample.
#'   Parameter: \code{obs} = numeric vector of observations.
#'   The empirical CDF is constructed from the ordered sample.
#'   Duplicate values are allowed but may trigger a warning.
#'
#'   \item Student t ("st"):
#'   marginal density of a normal-gamma model describing uncertainty
#'   in \eqn{\mu} and \eqn{\sigma}.
#'   Hyperparameters:
#'   \eqn{(m, s, n, \nu)}.
#'
#'   \item Log-Student t ("lt"):
#'   marginal density of a log-normal-gamma model describing uncertainty
#'   in \eqn{\mu} and \eqn{\sigma}.
#'   Hyperparameters:
#'   \eqn{(m, s, n, \nu)}.
#' }
#'
#' The types `"binom"` and `"unif"` are reserved for future extensions.
#' The package automatically checks consistency between mean, standard deviation and coefficient of variation.
#'
#' Methods available in this class:
#'
#' * `$runMachines()` - executes all solution algorithms for each limit state function.
#' * `$calculateSystemProbability()` - computes system reliability using bounds or Monte Carlo based methods (`MC_IS`, `MC_CRUDE`, `MC_SubSam`).
#' * `$printResults()` - prints a detailed report to console or file.
#' * `$saveProject(level)` - saves results at different detail levels.
#' * `$plotGraph(plotType)` - experimental plotting routine for simulation performance.
#'
#'
#' @field sys_input      List of `SYS_LSF` objects - the individual limit-state functions.
#' @field sys_type       Character string that defines whether the system
#'                       is *serial* or *parallel* (not yet implemented).
#' @field probMachines   List of `PROB_MACHINE` objects - the analysis methods.
#' @field res_single     List of results per machine for each limit-state function.
#' @field res_sys        List of results for the system-level reliability calculation.
#' @field beta_single    Matrix of \eqn{\beta}-values obtained from the single-problem analyses.
#' @field beta_sys       Matrix of \eqn{\beta}-values obtained from the system reliability calculations.
#' @field params_sys     List that stores the parameter sets used in Monte-Carlo runs.
#' @field debug.level    Numeric verbosity level (0 = silent, 1 = basic, 2 = detailed).
#'
#' @details
#' The class is built on \code{setRefClass} and therefore uses mutable fields.
#' All objects (`PROB_BASEVAR`, `SYS_LSF`, ...) are expected to be already
#' **checked** (via the `check()` method) before they are passed to `SYS_PROB`.
#'
#' @examples
#' ps <- SYS_PROB(
#'   sys_input = list(SYS_LSF(), SYS_LSF()),
#'   probMachines = list(PROB_MACHINE()),
#'   sys_type = "serial"
#' )
#' \dontrun{
#' ps$runMachines()
#' ps$beta_sys
#' ps$res_sys
#' ps$printResults("example_1")
#' ps$saveProject(4, "example_1")
#' }
#'
#' @author (C) 2021-2026 K. Nille-Hauf, T. Feiri, M. Ricker, T. Lux -- Hochschule Biberach (until 2022),
#' TU Dortmund University - Chair of Structural Concrete (since 2023)
#'
#' @import gridExtra
#' @import ggplot2
#' @importFrom methods "new"
#' @importFrom stats dnorm pnorm qnorm rnorm
#' @importFrom utils str
#' @export SYS_PROB
#' @exportClass SYS_PROB


SYS_PROB <- setRefClass(
  Class = "SYS_PROB",
  fields = list(
    sys_input = "list", # List of SYS_LSFs
    sys_type = "character", # determining serial or parallel system, not implemented yet
    probMachines = "list", # list of PROB_MACHINES
    res_single = "list",
    res_sys = "list",
    beta_single = "matrix",
    beta_sys = "matrix",
    params_sys = "list",
    sys_structure = "ANY",
    smooth_kappa = "numeric",
    debug.level = "numeric"
  )
)


SYS_PROB$methods(
  list(
    initialize = function(...) {
      initFields(...)
      beta_single <<- matrix(, 0, 0)
      beta_sys <<- matrix(, 0, 0)

      # Default smooth parameter
      if (is_empty(smooth_kappa)) {
        smooth_kappa <<- 20
      }

      # Default system structure
      if (is_empty(sys_structure)) {
        sys_structure <<- sys_type # fallback to existing behaviour
      }
      # Only set default if not provided by user
      if (is_empty(debug.level)) {
        debug.level <<- 0
      }
    },
    runMachines = function() {
      "Starts solving all given problems (sys_input) with all given algorithms (probMachines).
       First checks that all analysis methods are loaded before validating limit-state functions."

      # --- Step 1: check existence of reliability machines first ---
      for (j in seq_along(probMachines)) {
        probMachines[[j]]$checkMethodExists()
      }

      # --- Step 2: validate each limit-state function afterwards ---
      for (i in 1:length(sys_input)) {
        sys_input[[i]]$check()
      }

      if (is_empty(debug.level)) {
        debug.level <<- 0
      }

      # --- Safety check: all limit-state functions must be defined ---
      for (i in seq_along(sys_input)) {
        lsf_obj <- sys_input[[i]]

        # Case 1: no function object at all
        if (is_empty(lsf_obj$func)) {
          stop(sprintf(
            "Limit-state function for SYS_LSF[%d] ('%s') is not defined.\n",
            i,
            ifelse(is_empty(lsf_obj$name), "<unnamed>", lsf_obj$name)
          ), call. = FALSE)
        }

        # Case 2: function exists but has no arguments
        f_args <- formals(lsf_obj$func)
        if (is.null(f_args) || length(f_args) == 0L) {
          stop(sprintf(
            "Limit-state function for SYS_LSF[%d] ('%s') has no arguments.\n",
            i,
            ifelse(is_empty(lsf_obj$name), "<unnamed>", lsf_obj$name)
          ), call. = FALSE)
        }
      }

      # Helper: normalize getlDistr() output to a funlist (list with $d/$p/$q/$r)
      normalize_to_funlist <- function(dd) {
        # case 1: already funlist
        if (is.list(dd) && !is.null(dd$p) && is.function(dd$p)) {
          return(dd)
        }
        # case 2: wrapper -> first element is funlist
        if (is.list(dd) && length(dd) >= 1 && is.list(dd[[1]]) && !is.null(dd[[1]]$p) && is.function(dd[[1]]$p)) {
          return(dd[[1]])
        }
        # fallback: find any element that looks like a funlist
        if (is.list(dd)) {
          for (el in dd) {
            if (is.list(el) && !is.null(el$p) && is.function(el$p)) {
              return(el)
            }
          }
        }
        stop(sprintf("SYS_PROB$runMachines: unexpected getlDistr() format for variable (type=%s).", typeof(dd)), call. = FALSE)
      }

      beta_single <<- matrix(nrow = length(probMachines), ncol = length(sys_input))
      colnames(beta_single) <<- vector("character", length(sys_input))
      rownames(beta_single) <<- vector("character", length(probMachines))
      # Loop through each Problem in the System
      for (i in 1:length(sys_input)) {
        if (!is_empty(sys_input[[i]]$expr)) {
          lsf_expr <- sys_input[[i]]$expr
        }

        lsf <- sys_input[[i]]$getLSF()
        # lsf<-sys_input[[i]]$func
        if (is_empty(sys_input[[i]]$name)) {
          sys_input[[i]]$name <<- "Unknown problem name"
        }
        colnames(beta_single)[i] <<- sys_input[[i]]$name

        # Build basic distribution list for probMachines (normalized funlists)
        distr <- lapply(sys_input[[i]]$vars, function(v) {
          normalize_to_funlist(v$getlDistr())
        })

        lDistr_funlist <- distr

        # lDistr_for_FORM: wrapper list(funlist, params) (used only when calling FORM)
        lDistr_for_FORM <- lapply(lDistr_funlist, function(fun) {
          params <- c(NA_real_, NA_real_)
          if (!is.null(fun$mean)) params[1] <- as.numeric(fun$mean)
          if (!is.null(fun$Sd)) params[2] <- as.numeric(fun$Sd)
          list(fun, params)
        })

        # Loop through each probMachine
        res_machine <- list()
        for (j in 1:length(probMachines)) {
          if (probMachines[[j]]$fCall == "MVFOSM") {
            options <- probMachines[[j]]$options
            if (options$isExpression == "TRUE") {
              params <- list("lsf" = lsf_expr, "lDistr" = distr, "debug.level" = debug.level)
            } else {
              params <- list("lsf" = lsf, "lDistr" = distr, "debug.level" = debug.level)
            }
          } else {
            params <- list("lsf" = lsf, "lDistr" = distr, "debug.level" = debug.level)
          }

          if (!is_empty(probMachines[[j]]$options)) {
            params <- c(params, probMachines[[j]]$options)
          }

          # Later, when building params for the machine call:
          # decide which lDistr variant to pass
          # to be clarified if necessary
          if (probMachines[[j]]$fCall %in% c("FORM", "SORM") ||
            grepl("FORM", probMachines[[j]]$fCall, ignore.case = TRUE)) {
            params[["lDistr"]] <- lDistr_for_FORM
          } else {
            params[["lDistr"]] <- lDistr_for_FORM
          }

          probMachines[[j]]$checkMethodExists()

          safe_params <- params[names(params) %in% names(formals(get(probMachines[[j]]$fCall)))]
          res <- do.call(probMachines[[j]]$fCall, safe_params)

          # --- Debug: check structure of returned result -------------------
          if (debug.level >= 2) {
            message(sprintf("[SYS_PROB/runMachines] Structure of result from '%s':", probMachines[[j]]$fCall))
            str(res)
            str(res$beta)
          }
          # -----------------------------------------------------------------
          rownames(beta_single)[j] <<- probMachines[[j]]$name
          beta_single[j, i] <<- res$beta
          res_machine[[j]] <- res
        }
        res_single[[i]] <<- res_machine
      }
    },
    calculateSystemProbability = function(calcType = "simpleBounds", params = list()) {
      "Calculates the system probablity if more than one lsf is given and a system_type (serial or parallel) is set.
      If calcType is empty (or simpleBounds), only simpleBounds are applied to further calculation of single soultions.
      If calcType is MCIS, than a Monte Carlo Importance Sampling Method is used (only for parallel systems available).
      If calcType is MCC, than a Crude Monte Carlo Simulation is used.
      If calcType is MCSUS, than the Subset Sampling Algorithm ll be used.
      You can pass arguments to methods via the params field, while the argument has to be a named list (for example check the vignette)."
      params_sys[[length(params_sys) + 1]] <<- list("calcType" = calcType, params)

      ## ----- 1. ensure a debug level exists ---------------------------------
      if (is_empty(debug.level)) {
        debug.level <<- 0
      }

      ## make sure the debug level that you set globally is also forwarded
      ## (the MC_IS routine expects an argument called `debug.level`)
      # if (!is.null(debug.level)) {
      #   params$debug.level <- debug.level
      # }

      # debug-Message
      if (debug.level >= 2) {
        message("calcType = ", calcType)
        str(params)
      }

      # -------------------- Read and set parameters
      if (exists("cov_user", where = params)) {
        cov_user <- params$cov_user
      } else {
        cov_user <- 0.05
      }
      if (exists("n_batch", where = params)) {
        n_batch <- params$n_batch
      } else {
        n_batch <- 64
      }
      if (exists("n_max", where = params)) {
        n_max <- params$n_max
      } else {
        n_max <- 5e6
      }
      if (exists("use_threads", where = params)) {
        use_threads <- params$use_threads
      } else {
        use_threads <- 8
      }
      if (exists("beta_l", where = params)) {
        beta_l <- params$beta_l
      } else {
        beta_l <- 8
      }

      if (exists("Nsubset", where = params)) {
        Nsubset <- params$Nsubset
      } else {
        Nsubset <- 1e5
      }

      if (exists("variance", where = params)) {
        variance <- params$variance
      } else {
        variance <- "uniform"
      }

      if (exists("p0", where = params)) {
        p0 <- params$p0
      } else {
        p0 <- 0.1
      }

      if (exists("densityType", where = params)) {
        densityType <- params$densityType
      } else {
        densityType <- "norm"
      }

      if (exists("dps", where = params)) {
        dps <- params$dps
      } else {
        dps <- NULL
      }


      if (calcType == "simpleBounds") {
        min_beta <- vector("numeric", nrow(beta_single))
        max_beta <- vector("numeric", nrow(beta_single))

        if (sys_type == "parallel") {
          for (m in 1:nrow(beta_single)) { # m are the machines used
            min_beta[m] <- max(beta_single[m, ])
            max_beta[m] <- -1 * qnorm(pnorm(sum(-beta_single[m, ])))
          }
        } else if (sys_type == "serial") {
          for (m in 1:nrow(beta_single)) { # m are the machines used
            min_beta[m] <- -1 * qnorm(1 - prod(pnorm(beta_single[m, ])))
            max_beta[m] <- min(beta_single[m, ])
          }
        }

        if (nrow(beta_sys) < 1) {
          beta_sys <<- matrix(nrow = 2, ncol = 1)
          beta_sys[1, 1] <<- min_beta
          beta_sys[2, 1] <<- max_beta
          rownames(beta_sys) <<- c("SB min", "SB max")
        } else {
          beta_sys <<- rbind(beta_sys, "SB min" = min_beta)
          beta_sys <<- rbind(beta_sys, "SB max" = max_beta)
        }

        res <- list(
          "method" = "Simple Bounds",
          "system_mode" = sys_type,
          "single_betas" = beta_single,
          "min_beta" = min_beta,
          "max_beta" = max_beta
        )

        res_sys[[length(res_sys) + 1]] <<- res
        info.print("SimpleBounds", debug.level, c("system_mode", "beta_min", "beta_max"), c(sys_type, min_beta, max_beta))
      } else if (calcType == "MCIS") {
        # -------------------------------------------------
        # 1) Collect limit-state functions and distributions
        # -------------------------------------------------

        lsfs <- list()
        distr <- list()

        for (i in seq_along(sys_input)) {
          sys_input[[i]]$check()

          lsfs[[i]] <- sys_input[[i]]$getLSF()

          distr[[i]] <- lapply(sys_input[[i]]$vars, function(v) {
            d <- v$getlDistr()

            if (is.list(d) && length(d) >= 1 &&
              is.list(d[[1]]) && !is.null(d[[1]]$p)) {
              d[[1]]
            } else if (is.list(d) && !is.null(d$p)) {
              d
            } else {
              stop("Unexpected distribution object format returned by getlDistr()")
            }
          })
        }

        is_system <- length(lsfs) > 1

        # -------------------------------------------------
        # 2) Determine system method
        # -------------------------------------------------

        system_method <- if (!is.null(params$system_method)) {
          params$system_method
        } else {
          "multimodal"
        }

        # -------------------------------------------------
        # 3) Flatten distributions (needed for smooth)
        # -------------------------------------------------

        distr_flat <- unlist(distr, recursive = FALSE)

        # -------------------------------------------------
        # 4) Build global smooth LSF if requested
        # -------------------------------------------------

        if (is_system && system_method == "smooth") {
          g_sys <- function(x) {
            g_vals <- numeric(length(lsfs))

            k <- 1

            for (i in seq_along(lsfs)) {
              n_var <- length(sys_input[[i]]$vars)

              g_vals[i] <- lsfs[[i]](x[k:(k + n_var - 1)])

              k <- k + n_var
            }

            if (.self$sys_structure == "parallel") {
              smooth_max(g_vals, .self$smooth_kappa)
            } else {
              smooth_min(g_vals, .self$smooth_kappa)
            }
          }
        }

        # -------------------------------------------------
        # 5) Select MC_IS input
        # -------------------------------------------------

        if (!is_system) {
          lsf_arg <- lsfs[[1]]
          lDistr_arg <- distr[[1]]
        } else if (system_method == "multimodal") {
          lsf_arg <- lsfs
          lDistr_arg <- distr
        } else if (system_method == "smooth") {
          lsf_arg <- g_sys
          lDistr_arg <- distr_flat
        } else {
          stop("system_method must be 'multimodal' or 'smooth'", call. = FALSE)
        }

        # -------------------------------------------------
        # 6) Prepare MC_IS arguments
        # -------------------------------------------------

        seed_val <- params$seed %||% NULL
        backend_val <- params$backend %||% "future"
        dataRecord <- params$dataRecord %||% FALSE

        mc_args <- list(
          lsf = lsf_arg,
          lDistr = lDistr_arg,
          cov_user = cov_user,
          n_batch = n_batch,
          n_max = n_max,
          use_threads = use_threads,
          beta_l = beta_l,
          densityType = densityType,
          dps = dps,
          debug.level = debug.level,
          seed = seed_val,
          backend = backend_val,
          dataRecord = dataRecord,
          sys_type = sys_type
        )

        # -------------------------------------------------
        # 7) Run MC_IS
        # -------------------------------------------------

        res <- do.call(TesiproV::MC_IS, mc_args)

        # -------------------------------------------------
        # 8) Store result
        # -------------------------------------------------

        if (nrow(beta_sys) < 1) {
          beta_sys <<- matrix(nrow = 1, ncol = 1)

          rownames(beta_sys) <<- "MC IS"
        } else {
          beta_sys <<- rbind(beta_sys, "MC IS" = NA)
        }

        beta_sys[nrow(beta_sys), 1] <<- res$beta

        res_sys[[length(res_sys) + 1]] <<- res


        # } else if (calcType == "MCIS") {
        #   ## -------------------------------------------------
        #   ## 1) Build the list of limit-state functions and
        #   ##    marginal distribution objects (unchanged)
        #   ## -------------------------------------------------
        #   lsfs <- list()
        #   distr <- list()

        #   for (i in 1:length(sys_input)) {
        #     sys_input[[i]]$check()
        #     lsfs[[i]] <- sys_input[[i]]$getLSF()
        #     # subdistr <- lapply(sys_input[[i]]$vars, function(v) v$getlDistr())
        #     subdistr <- lapply(sys_input[[i]]$vars, function(v) {
        #       d_wrapped <- v$getlDistr()
        #       # If getlDistr() returned the wrapper list (funlist, params), extract funlist
        #       if (is.list(d_wrapped) && length(d_wrapped) >= 1 && is.list(d_wrapped[[1]]) && !is.null(d_wrapped[[1]]$p)) {
        #         return(d_wrapped[[1]])
        #       }
        #       # otherwise assume d_wrapped already is the funlist
        #       if (is.list(d_wrapped) && !is.null(d_wrapped$p)) {
        #         return(d_wrapped)
        #       }
        #       stop("Unexpected distribution object format returned by getlDistr()")
        #     })

        #     # subdistr <- list()
        #     # for (k in 1:length(sys_input[[i]]$vars)) {
        #     #   v <- sys_input[[i]]$vars[[k]]
        #     #   subdistr[[k]] <- v$getlDistr()
        #     # }
        #     distr[[i]] <- subdistr
        #   }

        #   is_system <- length(lsfs) > 1

        #   system_method <- if (!is.null(params$system_method)) {
        #     params$system_method
        #   } else {
        #     "multimodal"
        #   }

        #   # --- DEBUG: print structure to verify what is passed to MC_IS_system
        #   if (debug.level >= 2) {
        #     message("DEBUG before calling MC_IS_system: lsfs and distr structures:")
        #     str(lsfs)
        #     str(distr)
        #   }


        #   # optional sanity check - will stop early if something is wrong
        #   for (i in seq_along(distr)) {
        #     for (j in seq_along(distr[[i]])) {
        #       if (!is.list(distr[[i]][[j]]) || is.null(distr[[i]][[j]]$p)) {
        #         stop(sprintf("Sanity check failed: distr[[%d]][[%d]] not a funlist", i, j), call. = FALSE)
        #       }
        #     }
        #   }

        #   # -------------------------------------------------
        #   # Build global system LSF g_sys
        #   # -------------------------------------------------
        #   if (is_system) {
        #     g_sys <- function(x) {
        #       # -------------------------------------------------
        #       # Evaluate all individual limit-state functions
        #       # -------------------------------------------------
        #       g_vals <- numeric(length(lsfs))

        #       k <- 1
        #       for (i in seq_along(lsfs)) {
        #         n_var_i <- length(sys_input[[i]]$vars)
        #         g_vals[i] <- lsfs[[i]](x[k:(k + n_var_i - 1)])
        #         k <- k + n_var_i
        #       }

        #       # -------------------------------------------------
        #       # Determine system aggregation structure
        #       # -------------------------------------------------

        #       structure_type <- .self$sys_structure

        #       # -------------------------------------------------
        #       # Case 1: Custom function
        #       # -------------------------------------------------
        #       if (is.function(structure_type)) {
        #         return(structure_type(g_vals))
        #       }

        #       # -------------------------------------------------
        #       # Case 2: Character system structure
        #       # -------------------------------------------------
        #       if (is.character(structure_type) && length(structure_type) == 1L) {
        #         if (structure_type == "serial") {
        #           return(smooth_min(g_vals, .self$smooth_kappa))
        #         }

        #         if (structure_type == "parallel") {
        #           return(smooth_max(g_vals, .self$smooth_kappa))
        #         }

        #         stop(sprintf(
        #           "Unsupported character system structure: '%s'",
        #           structure_type
        #         ))
        #       }

        #       # -------------------------------------------------
        #       # Case 3: Fallback to sys_type
        #       # -------------------------------------------------
        #       sys_type_local <- .self$sys_type

        #       if (is.character(sys_type_local) && length(sys_type_local) == 1L) {
        #         if (sys_type_local == "serial") {
        #           return(smooth_min(g_vals, .self$smooth_kappa))
        #         }

        #         if (sys_type_local == "parallel") {
        #           return(smooth_max(g_vals, .self$smooth_kappa))
        #         }
        #       }

        #       # -------------------------------------------------
        #       # Final fallback (legacy default)
        #       # -------------------------------------------------
        #       return(smooth_min(g_vals, .self$smooth_kappa))
        #     }
        #   }

        #   # -------------------------------------------------
        #   # Flatten distribution list for global LSF
        #   # -------------------------------------------------

        #   distr_flat <- list()

        #   for (i in seq_along(distr)) {
        #     for (j in seq_along(distr[[i]])) {
        #       distr_flat[[length(distr_flat) + 1]] <- distr[[i]][[j]]
        #     }
        #   }

        #   # ---------------------------------------------------------------
        #   # 2) Pull the optional entries from `params` (if they exist)
        #   # ---------------------------------------------------------------
        #   seed_val <- if (!is.null(params$seed)) params$seed else NULL
        #   backend_val <- if (!is.null(params$backend)) params$backend else "future"
        #   dataRecord_val <- if (!is.null(params$dataRecord)) params$dataRecord else FALSE

        #   ## -------------------------------------------------
        #   ## 3) Gather *all* arguments for MC_IS
        #   ## -------------------------------------------------
        #   ## a) The arguments that are always required (hard-coded)
        #   # base_args <- list(
        #   #   lsf = lsfs,
        #   #   lDistr = distr,
        #   #   sys_type = sys_type,
        #   #   cov_user = cov_user,
        #   #   n_batch = n_batch,
        #   #   n_max = n_max,
        #   #   use_threads = use_threads,
        #   #   beta_l = beta_l,
        #   #   densityType = densityType,
        #   #   dps = dps,
        #   #   debug.level = debug.level,
        #   #   seed = seed_val,
        #   #   backend = backend_val, # fallback, wenn nicht angegeben
        #   #   dataRecord = dataRecord_val
        #   # )

        #   # if (is_system) {
        #   #   lsf_arg <- g_sys
        #   #   lDistr_arg <- distr_flat
        #   # } else {
        #   #   lsf_arg <- lsfs[[1]]
        #   #   lDistr_arg <- distr[[1]]
        #   # }

        #   if (is_system) {
        #     if (system_method == "multimodal") {
        #       # true system reliability
        #       lsf_arg <- lsfs
        #       lDistr_arg <- distr
        #     } else if (system_method == "smooth") {
        #       # smooth aggregation
        #       lsf_arg <- g_sys
        #       lDistr_arg <- distr_flat
        #     } else {
        #       stop("Unknown system_method. Use 'multimodal' or 'smooth'.", call. = FALSE)
        #     }
        #   } else {
        #     lsf_arg <- lsfs[[1]]
        #     lDistr_arg <- distr[[1]]
        #   }

        #   base_args <- list(
        #     lsf = lsf_arg,
        #     lDistr = lDistr_arg,
        #     cov_user = cov_user,
        #     n_batch = n_batch,
        #     n_max = n_max,
        #     use_threads = use_threads,
        #     beta_l = beta_l,
        #     densityType = densityType,
        #     dps = dps,
        #     debug.level = .self$debug.level,
        #     seed = seed_val,
        #     backend = backend_val,
        #     dataRecord = dataRecord_val
        #   )

        #   ## b) Append the user-supplied options (the content of `params`)
        #   ##    - but only keep those names that really exist in MC_IS
        #   ##    (otherwise `do.call` would fail with unused argument).
        #   ##    `formals(MC_IS)` returns the vector of formal argument names.
        #   mc_formals <- names(formals(TesiproV::MC_IS))

        #   ## keep only entries whose name appears in the formal list
        #   user_args <- params[
        #     (names(params) %in% mc_formals) & # a) valid for MC_IS
        #       !(names(params) %in% names(base_args)) # b) not duplicate
        #   ]

        #   ## c) Final argument list -> base args + (filtered) user args
        #   mc_args <- c(base_args, user_args)

        #   ## -------------------------------------------------
        #   ## 3) Call the Monte-Carlo routine
        #   ## -------------------------------------------------
        #   if (debug.level >= 2) {
        #     message("Calling MC_IS_system with:")
        #     print(names(mc_args))
        #   }

        #   # Ensure debug.level is always forwarded explicitly
        #   mc_args$debug.level <- debug.level

        #   res <- do.call(TesiproV::MC_IS, mc_args)

        #   ## -------------------------------------------------
        #   ## 4) Store the result (unchanged)
        #   ## -------------------------------------------------
        #   if (nrow(beta_sys) < 1) {
        #     beta_sys <<- matrix(nrow = 1, ncol = 1)
        #     beta_sys[1, 1] <<- res$beta
        #     rownames(beta_sys) <<- "MC IS"
        #   } else {
        #     beta_sys <<- rbind(beta_sys, "MC IS" = res$beta)
        #   }
        #   res_sys[[length(res_sys) + 1]] <<- res
        #   ## MC_CRUDE
      } else if (calcType == "MCC" | calcType == "MCSUS") {
        lsfs <- list()
        distr <- list()
        n <- 1
        n_vars <- vector("numeric", length(sys_input))
        for (i in 1:length(sys_input)) {
          sys_input[[i]]$check()
          lsfs[[i]] <- sys_input[[i]]$getLSF()
          n_vars[i] <- length(sys_input[[i]]$vars)
          for (k in 1:n_vars[i]) {
            v <- sys_input[[i]]$vars[[k]]
            distr[[n]] <- v$getlDistr()
            n <- n + 1
          }
        }
        if (sys_type == "parallel") {
          sys_lsf <- function(x) {
            n_lsfs <- length(lsfs)
            res <- vector("numeric", n_lsfs)
            k <- 1
            for (i in 1:n_lsfs) {
              res[i] <- lsfs[[i]](x[k:(k + n_vars[i] - 1)])
              k <- k + n_vars[i]
            }
            return(max(res))
          }
        } else if (sys_type == "serial") {
          sys_lsf <- function(x) {
            n_lsfs <- length(lsfs)
            res <- vector("numeric", n_lsfs)
            k <- 1
            for (i in 1:n_lsfs) {
              res[i] <- lsfs[[i]](x[k:(k + n_vars[i] - 1)])
              k <- k + n_vars[i]
            }
            return(min(res))
          }
        }
        if (calcType == "MCC") {
          res <- TesiproV::MC_CRUDE(
            lsf = sys_lsf,
            lDistr = distr,
            cov_user = cov_user,
            n_batch = n_batch,
            n_max = n_max,
            use_threads = use_threads,
            debug.level = debug.level
          )
          if (nrow(beta_sys) < 1) {
            beta_sys <<- matrix(nrow = 1, ncol = 1)
            beta_sys[1, 1] <<- res$beta
            rownames(beta_sys) <<- "MC Crude"
          } else {
            beta_sys <<- rbind(beta_sys, "MC Crude" = res$beta)
          }
          res_sys[[length(res_sys) + 1]] <<- res
        } else {
          res <- TesiproV::MC_SubSam(
            lsf = sys_lsf,
            lDistr = distr,
            Nsubset = Nsubset,
            variance = variance,
            p0 = p0,
            debug.level = debug.level
          )
          if (nrow(beta_sys) < 1) {
            beta_sys <<- matrix(nrow = 1, ncol = 1)
            beta_sys[1, 1] <<- res$beta
            rownames(beta_sys) <<- "MC SubSam"
          } else {
            beta_sys <<- rbind(beta_sys, "MC SubSam" = res$beta)
          }
          res_sys[[length(res_sys) + 1]] <<- res
        }
      }
    },
    printResults = function(path = "") {
      "TesiproV can create a report file with all the necessary data for you. If you provide a path (or filename, without ending) it will store
      the data there, otherwise it will report to the console. Set the path via setwd() or check it via getwd()."
      if (!path == "") {
        sink(paste(path, ".txt", sep = ""), type = "output")
        cat("Ergebnisausdruck TesiproV Berechnung\n")
        cat(date())
        cat("\n")
      }

      n_sys <- length(sys_input)
      n_machines <- length(probMachines)
      cat("\n -----------  1. Berechnungsergebnisse (Zusammenfassung)    ---------\n\n")

      cat("1.1 Versagensindicies:\n")
      if (nrow(beta_single) > 0) {
        cat("\nSicherheitsindices bei Einzelversagen:")
        print(beta_single)
      }

      if (nrow(beta_sys) > 0) {
        cat("\nSicherheitsindices bei Systemversagen:")
        print(beta_sys)
      }

      cat("\n1.2.1 Ergebnisse je Problemstellung und Algorithmus\n")

      for (i in 1:n_sys) {
        cat("\n____________________________________________\n")
        cat(sprintf("1.2.1.%d Ergebnisse fuer Problem Nr: %d \t-%s\n", i, i, sys_input[[i]]$name))
        for (j in 1:n_machines) {
          cat(sprintf("Beta: %.4f\tPf: %f\t", res_single[[i]][[j]]$beta, res_single[[i]][[j]]$pf))
          cat(sprintf("%s (%s)\n", probMachines[[j]]$name, probMachines[[j]]$fCall))
          print(res_single[[i]][[j]]$runtime)
        }
      }

      cat("\n1.2.2 Ergebnisse je Systemberechnung und Simulationsmethode\n")
      if (nrow(beta_sys) > 0) {
        for (i in 1:length(res_sys)) {
          cat("\n____________________________________________\n")
          cat(sprintf("1.2.2.%d Ergebnisse Systemsimulation %d \t %s\n", i, i, res_sys[[i]]$method))
          if (res_sys[[i]]$method == "Simple Bounds") {
            cat(sprintf("Min Beta: %.4f\t Max Beta: %f\t", res_sys[[i]]$min_beta, res_sys[[i]]$max_beta))
          } else {
            cat(sprintf("Beta: %.4f\tPf: %f\t", res_sys[[i]]$beta, res_sys[[i]]$pf))
          }
        }
      }

      cat("\n\n-------------------------  2. SYSTEM BESCHREIBUNG ------------------\n")

      cat(sprintf("2.1 Das System umfasste %d Gleichungen die in %s Beziehung stehen.\n\n", n_sys, sys_type))
      for (i in 1:n_sys) {
        cat("\n___________________________________\n")
        cat(sprintf("2.1.%d Gleichung Nr: %d \t-%s\n", i, i, sys_input[[i]]$name))
        if (!is_empty(sys_input[[i]]$expr)) {
          cat("Folgender symbolischer Ausdruck wurde definiert:\n")
          print(sys_input[[i]]$expr)
        }
        if (!is_empty(sys_input[[i]]$func)) {
          cat("Folgende Funktion wurde definiert:\n")
          print(sys_input[[i]]$func)
        }

        n_vars <- length(sys_input[[i]]$vars)
        cat(sprintf("\nBasisvariablen (%d gegeben):\n", n_vars))
        for (k in 1:n_vars) {
          v <- sys_input[[i]]$vars[[k]]
          # cat(sprintf("Name (ID): %s (%d)\tBeschreibung",v$Name,v$Id))

          cat(sprintf(
            "%d.\tName (ID): %s (%d)\tPackage::Verteilungstyp: %s::%s\tMean: %.3f\tSd: %.3f\tCov: %.3f\tx0: %.3f\tVerteilungsparameter: %.5f\t%.5f\n",
            k, v$Name, v$Id, v$Package, v$DistributionType, v$Mean, v$Sd, v$Cov, v$x0, v$DistributionParameters[1], v$DistributionParameters[2]
          ))
        }
      }
      cat("\n\n ")

      cat("\n ------------------------  3. METHODEN BESCHREIBUNG ------------------\n")
      if (length(n_machines) > 0) {
        cat(sprintf("\n\n3.1 Die Berechnung wurde von %d Methoden analysiert.\nIm folgenden werden die Eingangsparameter der Methoden beschrieben.", n_sys))
        for (i in 1:n_machines) {
          cat("\n______________________________________\n")
          cat(sprintf("3.%d Machine Name (Type): %s (%s)\n", i, probMachines[[i]]$name, probMachines[[i]]$fCall))
          if (!is.null(probMachines[[i]]$options)) {
            cat("Die folgenden Optionen waren von den Standartwerten abweichend:\n")
            for (j in 1:length(probMachines[[i]]$options)) {
              cat(sprintf("\t-%s\t=\t%s\n", names(probMachines[[i]]$options[j]), probMachines[[i]]$options[j]))
            }
          } else {
            cat("Ausschliesslich Standartoptionen verwendet!\n")
          }
        }
      }

      if (length(params_sys) > 0) {
        cat(sprintf("\n\n3.2 Parameter der Systemsimulationen."))
        for (i in 1:length(params_sys)) {
          cat("\n______________________________________\n")
          cat(sprintf("3.%d Simulation Method: %s\n", i, params_sys[[i]]$calcType))
          print(params_sys)
        }
      }


      cat("\n -------------------------  AUSFUEHRLICHE ERGEBNISSE ------------------\n")
      cat("\n4.1 Einzelprobleme\n")
      if (length(n_sys) > 0) {
        for (i in 1:n_sys) {
          cat("\n_____________________________________\n")
          cat(sprintf("Ergebnisse fuer Problem Nr: %d \t-%s\n", i, sys_input[[i]]$name))
          for (j in 1:n_machines) {
            cat(sprintf("Machine Name (Type): %s (%s)\n", probMachines[[j]]$name, probMachines[[j]]$fCall))

            n_erg <- length(res_single[[i]][[j]])
            for (k in 1:n_erg) {
              if (names(res_single[[i]][[j]][k]) == "data") {} else if (names(res_single[[i]][[j]][k]) == "runtime") {
                cat("\t -runtime:")
                print(res_single[[i]][[j]][k])
              } else {
                cat(sprintf("\t-%s\t=\t%s\n", names(res_single[[i]][[j]][k]), res_single[[i]][[j]][k]))
              }
            }
          }
        }
      }

      if (length(res_sys) > 0) {
        cat("\n\n4.2 Systemprobleme\n")
        for (i in 1:length(res_sys)) {
          cat("\n_____________________________________\n")
          cat(sprintf("Simulationmethode: %s\n", res_sys[[i]]$method))

          n_erg <- length(res_sys[[i]])
          for (k in 1:n_erg) {
            if (names(res_sys[[i]][k]) == "data") {} else if (names(res_sys[[i]][k]) == "runtime") {
              cat("\t -runtime:")
              print(res_sys[[i]][k])
            } else {
              cat(sprintf("\t-%s\t=\t%s\n", names(res_sys[[i]][k]), res_sys[[i]][k]))
            }
          }
        }
      }


      cat("--------------------- ENDE -----------------")


      if (!path == "") {
        sink()
      }
    },
    saveProject = function(level, filename = "tesiprov_project") {
      "You can save your calculation project with saveProject().
      There are four different levels of detail to save
      1st Level: Only the beta values
      2nd Level: The result Objects of single or systemcalculation
      3th Level: All The Probablity System Object, including limit state functions, machines and solutions
      4th Level: An image of your entire workspace"
      if (level == 1) {
        beta <- list(
          "beta_single" = beta_single,
          "beta_sys" = beta_sys
        )
        filename_loc <- paste(filename, "_beta.rds", sep = "")
        saveRDS(beta, filename_loc)
        cat(paste("Beta values successfully stored in\n", getwd(), "/", filename_loc, ".\nOpen with readRDS().", sep = ""))
      } else if (level == 2) {
        res <- list(
          "res_single" = res_single,
          "res_sys" = res_sys
        )
        filename_loc <- paste(filename, "_res.rds", sep = "")
        saveRDS(res, filename_loc)
        cat(paste("Results successfully stored in\n", getwd(), "/", filename_loc, ".\nOpen with readRDS().", sep = ""))
      } else if (level == 3) {
        ps <- list(
          "sys_input" = sys_input,
          "sys_type" = sys_type,
          "probMachines" = probMachines,
          "res_single" = res_single,
          "res_sys" = res_sys,
          "beta_single" = beta_single,
          "beta_sys" = beta_sys
        )
        filename_loc <- paste(filename, "_ps.rds", sep = "")
        saveRDS(ps, filename_loc)
        cat(paste("Prob. Sys. object successfully saved in\n", getwd(), "/", filename_loc, ".\nOpen with readRDS().", sep = ""))
      } else if (level == 4) {
        filename_loc <- paste(filename, ".RData", sep = "")
        save.image(filename_loc)
        cat(paste("Entire workspace successfully saved in\n", getwd(), "/", filename_loc, ".\nOpen with load().", sep = ""))
      } else {
        cat(paste("Level needs to be between 1 (less output) and 4 (entire workspace).", sep = ""))
      }
    },
    plotGraph = function(plotType = "sim.performance") {
      "not finally implemented. Do not use."
      for (i in 1:length(sys_input)) {
        sysName <- sys_input[[i]]$name
        for (j in 1:length(probMachines)) {
          pmName <- probMachines[[j]]$name
          pmfCall <- probMachines[[j]]$fCall
          if (pmfCall == "MC_IS") {
            if (plotType == "sim.performance") {
              df <- res_single[[i]][[j]]$data

              g1 <- ggplot(df, aes(x = n_sim)) +
                geom_line(aes(y = cov)) +
                labs(
                  x = "Simulationen (-)",
                  y = "CoV (-)",
                  title = paste(pmName, "fuer", sysName)
                ) +
                theme_bw()

              g2 <- ggplot(df, aes(x = n_sim)) +
                geom_line(aes(y = time)) +
                labs(
                  x = "Simulationen (-)",
                  y = "Zeit (s)"
                ) +
                theme_bw()

              ga <- grid.arrange(g1, g2, ncol = 1)
              return(ga)
            } else if (plotType == "sim.beta") {
              df <- res_single[[i]][[j]]$data

              g1 <- ggplot(df, aes(x = n_sim)) +
                geom_line(aes(y = -qnorm(pf)), size = 1) +
                geom_hline(yintercept = res_single[[i]][[j]]$beta, linetype = 2) +
                geom_text(x = 0, y = res_single[[i]][[j]]$beta, vjust = "bottom", hjust = "left", label = paste("Beta =", round(res_single[[i]][[j]]$beta, 4))) +
                labs(
                  x = "Simulationen (-)",
                  y = "Beta (-)",
                  title = paste(pmName, "fuer", sysName)
                ) +
                theme_bw()
              return(g1)
            }
          }
        }
      }

      if (plotType == "sim.hv") {
        hv_func <- res_single[[1]]$hv_func
        plots <- list()
        for (i in 1:length(hv_func)) {
          rval <- 0.0001
          samplerate <- 1e5
          x.min <- hv_func[[i]]$qfun(rval)
          x.max <- hv_func[[i]]$qfun(1 - rval)
          x <- seq(x.min, x.max, length.out = samplerate)
          dhv <- hv_func[[i]]$dfun(x)
          rhv <- hv_func[[i]]$rfun(samplerate)
          df <- data.frame(x = x, d = dhv)
          plots[[i]] <- ggplot(data = df, aes(x = x, y = d)) +
            geom_line() +
            labs(
              x = "Groesse",
              y = "Wahrscheinlichkeit (-)",
              title = paste("Verbunddichte hv_", i, sep = "")
            ) +
            theme_bw()
        }
        return(plots)
      }
    }
  )
)

#' @title System Limit-State Function (SYS_LSF)
#'
#' @description A reference-class that stores a user-defined limit-state function.
#' The function can be written either as an **expression** (`expr`) or
#' as a **regular R function** (`func`).  All random variables that appear
#' in the function must be supplied in the `vars` list as `PROB_BASEVAR`
#' objects.
#'
#' @field name  Optional human-readable name of the limit-state function.
#' @field expr  Symbolic expression (e.g. `expression(f_ck - d_nom)`) that is
#'               later converted to a callable function via `ExpressionToFunction()`.
#' @field func  Actual R function implementing the limit-state equation.
#'              May be `NULL` until defined by user or converted from `expr`.
#' @field vars  List of `PROB_BASEVAR` objects that provide the random variables.
#'
#' @details
#' The method `getLSF()` returns a **closure** that expects a numeric vector
#' `x` (ordered exactly like `vars`) and evaluates the limit-state function.
#'
#' @details
#' For consistency across FORM and Monte Carlo reliability methods,
#' limit-state functions must either accept a single numeric vector
#' argument or explicitly named scalar arguments.
#' The use of \code{function(...)} is not supported.
#'
#' @examples
#' vars <- list(
#'   PROB_BASEVAR(Name = "f_ck", DistributionType = "norm", Mean = 30, Sd = 1),
#'   PROB_BASEVAR(Name = "d_nom", DistributionType = "norm", Mean = 0.2, Sd = 0.01)
#' )
#' lsf <- SYS_LSF(name = "Bending resistance", vars = vars)
#' lsf$ExpressionToFunction()
#'
#' @section Limit-State Function (LSF) API:
#'
#' A limit-state function must follow one of the two supported
#' interface patterns:
#'
#' \describe{
#'   \item{\code{function(x)}}{
#'     A single numeric vector argument \code{x} containing all
#'     basic variables in the order specified in \code{vars}.
#'
#'     Example:
#'     \preformatted{
#'     lsf$func <- function(x) {
#'       x[1] - sum((x[-1]^2) / seq_along(x)[-1])
#'     }
#'     }
#'   }
#'
#'   \item{\code{function(Z, Fy, M, ...)}}{
#'     Explicitly named scalar arguments corresponding to the
#'     variable names defined in \code{vars}.
#'
#'     Example:
#'     \preformatted{
#'     lsf$func <- function(Z, Fy, M) {
#'       Z * Fy - M
#'     }
#'     }
#'   }
#' }
#'
#' The following form is \strong{not supported}:
#'
#' \preformatted{
#' function(...)
#' }
#'
#' Using \code{...} as the sole argument leads to inconsistent
#' behaviour between FORM-based and Monte Carlo-based reliability
#' methods and is therefore prohibited.
#'
#' For maximum robustness and compatibility across all reliability
#' algorithms, the use of the vector form \code{function(x)} is
#' recommended.
#'
#' @author (C) 2021-2026 K. Nille-Hauf, T. Feiri, M. Ricker, T. Lux -- Hochschule Biberach (until 2022),
#' TU Dortmund University - Chair of Structural Concrete (since 2023)
#'
#' @export SYS_LSF
#' @exportClass SYS_LSF
#'
SYS_LSF <- setRefClass(
  Class = "SYS_LSF",
  fields = list(
    name = "character",
    expr = "expression",
    # func = "function",
    func = "ANY",
    vars = "list"
  ),
  methods = list(
    initialize = function(...) {
      initFields(...)
      # Explicitly set func to NULL if not provided or invalid
      if (!is.function(func)) func <<- NULL
      invisible(.self)
    },
    ExpressionToFunction = function() {
      "Transforms a valid expression into a objective function. Need the set of Variables with correct spelled names and IDs"
      s <- deparse(expr)
      s <- substr(s, 12, nchar(s) - 1)
      for (i in 1:length(vars)) {
        symb <- vars[[i]]$Name
        vec <- paste(" x[", vars[[i]]$Id, "] ", sep = "")
        s <- gsub(symb, vec, s)
      }
      func <<- eval(parse(text = paste("f <- function(x) {", s, "}", sep = "")))
    },
    FunctionToExpression = function() {
      # Conversion back is not implemented yet
    },
    check = function() {
      "Checks all variables. You dont need to execute this, since the system object will do anyway."

      # --- Validate that a limit-state function exists and is callable -------

      # Case 1: completely missing or default dummy ---------------------------
      if (is.null(func)) {
        stop(sprintf(
          "Limit-state function '%s' is missing.",
          ifelse(is_empty(name), "<unnamed>", name)
        ), call. = FALSE)
      }

      if (is.function(func) &&
        length(formals(func)) == 0L &&
        identical(body(func), quote(NULL)) &&
        is_empty(expr)) {
        stop(sprintf(
          "Limit-state function '%s' is missing.",
          ifelse(is_empty(name), "<unnamed>", name)
        ), call. = FALSE)
      }

      # Case 2: not callable at all -------------------------------------------
      if (!is.function(func)) {
        stop(sprintf(
          "Limit-state function '%s' is not callable or invalid.",
          ifelse(is_empty(name), "<unnamed>", name)
        ), call. = FALSE)
      }

      # Case 3: formally defined but has no arguments -------------------------
      f_args <- tryCatch(formals(func), error = function(e) NULL)
      if (is.null(f_args) || length(f_args) == 0L) {
        stop(sprintf(
          "Limit-state function '%s' has no formal arguments.",
          ifelse(is_empty(name), "<unnamed>", name)
        ), call. = FALSE)
      }

      # Case 4: function defined only with '...' is not supported ----------
      # Using 'function(...)' leads to inconsistent behavior between
      # FORM-based and Monte Carlo-based reliability methods.
      # Therefore, limit-state functions must either accept:
      #   (a) a single vector argument 'function(x)', or
      #   (b) explicitly named scalar arguments.
      if (length(f_args) == 1L && identical(names(f_args), "...")) {
        stop(sprintf(
          "Limit-state function '%s' must not use '...' as sole argument.\n",
          ifelse(is_empty(name), "<unnamed>", name)
        ), call. = FALSE)
      }

      # --- Check all associated base variables -------------------------------
      for (i in seq_along(vars)) {
        vars[[i]]$prepare()
      }

      invisible(TRUE)
    },
    getLSF = function() {
      # Cache variable names once
      var_names <- vapply(vars, function(v) v$Name, character(1))
      n_vars <- length(var_names)

      lsf.local <- function(x) {
        fml <- formals(func)

        # ------------------------------------------------------------
        # Case 1: function(x)
        # ------------------------------------------------------------
        if (length(fml) == 1) {
          return(func(x))
        }

        # ------------------------------------------------------------
        # Case 2: function(Z, Fy, M, ...)
        # ------------------------------------------------------------

        # Vector case (FORM)
        if (is.vector(x)) {
          args <- as.list(x)
          names(args) <- var_names
          return(do.call(func, args))
        }

        # Matrix case (MC_IS)
        if (is.matrix(x)) {
          # Fast column extraction (no as.list, no do.call)
          cols <- lapply(seq_len(n_vars), function(i) x[, i])
          names(cols) <- var_names

          return(do.call(func, cols))
        }

        stop("Unsupported input type for LSF")
      }

      return(lsf.local)
    }
  )
)


#' @title Object to store the distribution model for basic variables (PROB_BASEVAR)
#'
#' @description Stores the distribution model for a basic random variable that can be used
#' in any limit-state function.  The class knows how to transform between
#' *mean / standard deviation* and the native distribution parameters.
#'
#' @field Id                     Integer identifier (position inside the LSF vector).
#' @field Name                   Human-readable name (e.g. `"f_ck"`).
#' @field Description            Short textual description.
#' @field DistributionType       Distribution identifier (`"norm"`, `"lnorm"`, `"slnorm"`, `"gumbel"`,
#'                               `"gamma"`, `"exp"`, `"beta"`, `"st"`, `"binom"`,`"emp"`,
#'                               `"unif"`, `"weibull"`, `"lt"`).
#' @field Mean                   Mean value of the variable.
#' @field Sd                     Standard deviation.
#' @field Cov                    Coefficient of variation (`Sd / Mean`).
#' @field x0                     Shifting parameter (used for shifted log-normal distribution `"slnorm"`).
#' @field Package                Name of the R package that provides the density
#'                               functions (`"stats"`, `"evd"`, `"EnvStats"` or `"brms"`).
#' @field DistributionParameters Numeric vector with the *native* distribution
#'                               parameters (e.g. `c(mu, sigma)` for a normal).
#' @field .cache                 Private environment that stores cached density
#'                               functions (used internally, **do not** access
#'                               from user code).
#'
#' @details
#' The method `prepare()` calculates missing parameters and checks for consistency.
#' `getlDistr()` returns a list containing *four* functions (`d`, `p`, `q`, `r`)
#' together with meta-information; the result is cached to avoid repeated
#' `loadNamespace()` calls.
#'
#' @examples
#' ## Normal random variable --------------------------------------------------
#' var1 <- PROB_BASEVAR(
#'   Name = "f_ck", DistributionType = "norm",
#'   Mean = 30, Sd = 1
#' )
#' var1$prepare()
#'
#' ## Gumbel random variable (package `evd`) ----------------------------------
#' var2 <- PROB_BASEVAR(
#'   Name = "M", DistributionType = "gumbel",
#'   Package = "evd", Mean = 2000, Sd = 200
#' )
#' var2$prepare()
#'
#' @author (C) 2021-2026 K. Nille-Hauf, T. Feiri, M. Ricker, T. Lux -- Hochschule Biberach (until 2022),
#' TU Dortmund University - Chair of Structural Concrete (since 2023)
#'
#' @export PROB_BASEVAR
#' @exportClass PROB_BASEVAR
#'
PROB_BASEVAR <- setRefClass(
  Class = "PROB_BASEVAR",
  fields = list(
    Id = "numeric",
    Name = "character",
    Description = "character",
    DistributionType = "character",
    Mean = "numeric",
    Sd = "numeric",
    Cov = "numeric",
    x0 = "numeric",
    Package = "character",
    DistributionParameters = "vector",
    .cache = "environment"
  ),
  methods = list(
    ## -----------------------------------------------------------------
    ##  Initialise the *private* cache (called once per object)
    ## -----------------------------------------------------------------
    .initCache = function() {
      .cache <<- new.env(parent = emptyenv())
    },

    ## -----------------------------------------------------------------
    ## INITIALIZE - constructor: assign fields - defaults - validation
    ## -----------------------------------------------------------------
    initialize = function(...) {
      ## Copy user-supplied arguments into object fields
      initFields(...)

      ## --- Early integrity checks ---------------------------------------
      # Case 1: completely empty object (no Mean, no Sd, no parameters)
      if ((is.null(Mean) || length(Mean) == 0L) &&
        (is.null(Sd) || length(Sd) == 0L) &&
        (is.null(DistributionParameters) ||
          all(is.na(DistributionParameters)))) {
        stop(sprintf(
          "PROB_BASEVAR '%s' must define at least Mean/Sd or valid distribution parameters.",
          ifelse(is_empty(Name), "<unnamed>", Name)
        ), call. = FALSE)
      }

      ## ---- Numeric fields: never leave them NULL --------------------
      if (is_missing(Id)) Id <<- NA_real_
      if (is_missing(Mean)) Mean <<- NA_real_
      if (is_missing(Sd)) Sd <<- NA_real_
      if (is_missing(Cov)) Cov <<- NA_real_
      if (is_missing(x0)) x0 <<- NA_real_

      ## 3) character slots - empty strings
      if (is_missing(Name)) Name <<- ""
      if (is_missing(Description)) Description <<- ""

      ## 4) distribution type / package defaults (the only change)
      if (is_missing(DistributionType) || DistributionType == "") {
        DistributionType <<- "norm"
      }
      if (is_missing(Package) || Package == "") {
        Package <<- "stats"
      }
      ## 5) DistributionParameters - always length 2
      if (is.null(DistributionParameters) ||
        length(DistributionParameters) < 2) {
        DistributionParameters <<- c(NA_real_, NA_real_)
      }

      ## ---- Create private cache environment --------------------------------
      .cache <<- new.env(parent = emptyenv())

      ## ---- Validate consistency --------------------------------------------
      validateArgs()

      prepare()

      invisible(.self)
    },


    ## -------------------------------------------------
    ##  validateArgs() - pure sanity checks (no calculations)
    ## -------------------------------------------------
    validateArgs = function() {
      ## ---- distribution type -----------------------------------------------
      allowed <- c(
        "norm", "lnorm", "slnorm", "gumbel", "gamma", "exp",
        "beta", "weibull", "st", "binom", "unif", "lt", "emp"
      )
      if (!(DistributionType %in% allowed)) {
        stop(sprintf("Distribution type '%s' is not known.", DistributionType),
          call. = FALSE
        )
      }

      ## ---- if Sd is missing but Mean & Cov are present ---------------------
      if (is_missing(Sd) && !is_missing(Mean) && !is_missing(Cov)) {
        Sd <<- Mean * Cov
      }

      ## ---- exponential - missing Sd - set equal to Mean --------------------
      if (tolower(trimws(DistributionType)) == "exp" &&
        is_missing(Sd) && !is.na(Mean)) {
        Sd <<- Mean
      }

      ## ---- Sd must be non-negative -----------------------------------------
      if (!is_missing(Sd) && Sd < 0) {
        stop("Standard deviation (Sd) must be non-negative.", call. = FALSE)
      }

      invisible(TRUE)
    },
    getlDistr = function() {
      ## 1)  make sure the object is prepared (sets defaults)
      prepare()

      if (is_missing(x0) || is.na(x0)) x0 <<- 0

      ## 2)  guarantee that Name and Package are not empty
      if (is_empty(Name)) Name <<- "Unknown Basevar Name"
      if (is_empty(Package)) Package <<- "stats"

      ## 3)  load the required package
      ensure_pkg(Package)

      ## 4)  create a private cache if it does not exist yet
      # if (!exists(".cache", inherits = FALSE)) .initCache()

      ## 5)  unique key based on distribution type + parameters
      if (!requireNamespace("digest", quietly = TRUE)) {
        stop("Package 'digest' is required for caching distribution functions.",
          call. = FALSE
        )
      }
      cache_key <- digest::digest(list(
        DistributionType,
        DistributionParameters
      ), algo = "xxhash64")

      ## 6)  return cached functions if they already exist
      if (exists(cache_key, envir = .cache, inherits = FALSE)) {
        cached <- get(cache_key, envir = .cache, inherits = FALSE)
        return(list(cached, DistributionParameters))
      }

      ## 7)  otherwise build the four functions
      if (DistributionType == "emp") {
        d_fun <- function(x) {
          do.call("demp",
            envir = loadNamespace("EnvStats"),
            args = list(x = x, obs = DistributionParameters)
          )
        }
        p_fun <- function(x) {
          do.call("pemp",
            envir = loadNamespace("EnvStats"),
            args = list(q = x, obs = DistributionParameters)
          )
        }
        q_fun <- function(x) {
          do.call("qemp",
            envir = loadNamespace("EnvStats"),
            args = list(p = x, obs = DistributionParameters)
          )
        }
        r_fun <- function(x) {
          do.call("remp",
            envir = loadNamespace("EnvStats"),
            args = list(n = x, obs = DistributionParameters)
          )
        }
      } else if (DistributionType == "slnorm") {
        d_fun <- function(x) {
          do.call("dshifted_lnorm",
            envir = loadNamespace("brms"),
            c(list(x), as.list(DistributionParameters))
          )
        }
        p_fun <- function(x) {
          do.call("pshifted_lnorm",
            envir = loadNamespace("brms"),
            c(list(x), as.list(DistributionParameters))
          )
        }
        q_fun <- function(x) {
          do.call("qshifted_lnorm",
            envir = loadNamespace("brms"),
            c(list(x), as.list(DistributionParameters))
          )
        }
        r_fun <- function(x) {
          do.call("rshifted_lnorm",
            envir = loadNamespace("brms"),
            c(list(x), as.list(DistributionParameters))
          )
        }
      } else if (DistributionType == "st" || DistributionType == "lt") {
        d_fun <- function(x) {
          do.call(paste0("d", DistributionType),
            envir = loadNamespace(Package),
            args = list(x = x, hyper.param = DistributionParameters)
          )
        }
        p_fun <- function(x) {
          do.call(paste0("p", DistributionType),
            envir = loadNamespace(Package),
            args = list(q = x, hyper.param = DistributionParameters)
          )
        }
        q_fun <- function(x) {
          do.call(paste0("q", DistributionType),
            envir = loadNamespace(Package),
            args = list(p = x, hyper.param = DistributionParameters)
          )
        }
        r_fun <- function(x) {
          do.call(paste0("r", DistributionType),
            envir = loadNamespace(Package),
            args = list(n_vals = x, hyper.param = DistributionParameters)
          )
        }
      } else {
        d_fun <- function(x) {
          do.call(paste0("d", DistributionType),
            envir = loadNamespace(Package),
            c(list(x), as.list(DistributionParameters))
          )
        }
        p_fun <- function(x) {
          do.call(paste0("p", DistributionType),
            envir = loadNamespace(Package),
            c(list(x), as.list(DistributionParameters))
          )
        }
        q_fun <- function(x) {
          do.call(paste0("q", DistributionType),
            envir = loadNamespace(Package),
            c(list(x), as.list(DistributionParameters))
          )
        }
        r_fun <- function(x) {
          do.call(paste0("r", DistributionType),
            envir = loadNamespace(Package),
            c(list(x), as.list(DistributionParameters))
          )
        }
      }

      ## 8)  store in cache
      funlist <- list(
        d = d_fun, p = p_fun, q = q_fun, r = r_fun,
        name = Name, mean = Mean, Sd = Sd, X0 = x0,
        Cov = Cov, DistributionType = DistributionType,
        DistributionPackage = Package
      )
      assign(cache_key, funlist, envir = .cache)

      ## 9)  return the structure expected by the rest of the code
      list(funlist, DistributionParameters)
    },

    ## -----------------------------------------------------------------------
    ## Performs transformations between mean/sd and distribution parameters.
    ## Ensures that Mean, Sd and Cov are consistent and valid.
    ## -----------------------------------------------------------------------
    prepare = function() {
      "Performs transformations between mean/sd and distribution parameters.
      Ensures that Mean, Sd and Cov are consistent and valid."

      ## ---- compute missing Sd from Mean & Cov -------------------------------
      if (is.na(Sd) && !is.na(Mean) && !is.na(Cov)) {
        Sd <<- Mean * Cov
      }

      ## ---- exponential shortcut --------------------------------------------
      if (tolower(DistributionType) == "exp" && is.na(Sd) && !is.na(Mean)) {
        Sd <<- Mean
      }
      ## ---------------------------------------------------------------
      ##  Forward: Mean & Sd (Cov darf NA sein) -> native Parameter
      ## ---------------------------------------------------------------
      if (!any(is.na(c(Mean, Sd)))) { # <- Cov is not required here
        switch(tolower(DistributionType),
          norm = {
            DistributionParameters[1] <<- Mean
            DistributionParameters[2] <<- Sd
          },
          lnorm = {
            sn <- sqrt(log(1 + (Sd / Mean)^2))
            mn <- log(Mean / sqrt(1 + (Sd / Mean)^2))
            DistributionParameters[1] <<- mn
            DistributionParameters[2] <<- sn
          },
          slnorm = {
            if (any(is.na(x0))) {
              x0 <<- 0
            } else if (x0 > Mean) {
              stop("For shifted lognormal distribution: Mean must be larger than x0.")
            }
            mx <- Mean
            sn <- sqrt(log(1 + (Sd / (mx - x0))^2))
            mn <- log((mx - x0) / sqrt(1 + (Sd / (mx - x0))^2))
            DistributionParameters[1] <<- mn
            DistributionParameters[2] <<- sn
            DistributionParameters[3] <<- x0
          },
          gumbel = {
            scale <- (Sd * sqrt(6)) / pi
            location <- Mean + digamma(1) * scale
            DistributionParameters[1] <<- location
            DistributionParameters[2] <<- scale
          },
          exp = {
            rate <- ifelse(is.na(Mean), NA_real_, 1 / Mean)
            scale <- Mean
            DistributionParameters[1] <<- rate
            DistributionParameters[2] <<- scale
          },
          beta = {
            var_val <- Sd^2
            t_val <- Mean * (1 - Mean) / var_val - 1
            alpha <- Mean * t_val
            beta_ <- (1 - Mean) * t_val
            DistributionParameters[1] <<- alpha
            DistributionParameters[2] <<- beta_
          },
          gamma = {
            shape <- Mean^2 / Sd^2
            scale <- Sd^2 / Mean
            DistributionParameters[1] <<- shape
            DistributionParameters[2] <<- scale
          },
          weibull = {
            ## 1) Versuche numerische Loesung fuer k (Shape)
            root_fun <- function(k) {
              if (k <= 0) {
                return(Inf)
              }
              cov_k <- sqrt(exp((base::gamma(1 + 2 / k) / base::gamma(1 + 1 / k)^2)) - 1)
              cov_target <- Sd / Mean
              cov_k - cov_target
            }
            ## 2) Intervall waehlen (typisch 0.1 - 100)
            k_est <- tryCatch(
              uniroot(root_fun, interval = c(0.1, 100))$root,
              error = function(e) {
                ## Falls die Root-Suche scheitert - gehe zurueck zur empirischen Formel
                (Sd / Mean)^(-1.086)
              }
            )
            lam_est <- Mean / base::gamma(1 + 1 / k_est)

            DistributionParameters[1] <<- k_est
            DistributionParameters[2] <<- lam_est
          },
          emp = {
            # not required
          },
          st = {
            # not required
          },
          lt = {
            # not required
          },
          {
            warning(sprintf("Forward transformation for '%s' not implemented", DistributionType))
          }
        )
      } else if ((all(is.na(c(Mean, Sd))) && any(!is.na(DistributionParameters)))) {
        if (DistributionType == "lt" && length(DistributionParameters) != 4) {
          stop("Log-Student-t requires 4 hyperparameters: m', s', n', v'")
        }
        switch(tolower(DistributionType),
          norm = {
            Mean <<- DistributionParameters[1]
            Sd <<- DistributionParameters[2]
            Cov <<- ifelse(DistributionParameters[1] == 0, 0, DistributionParameters[2] / DistributionParameters[1])
          },
          lnorm = {
            mn <- DistributionParameters[1]
            sn <- DistributionParameters[2]
            m <- exp(mn + (sn^2) / 2)
            s <- m * sqrt(exp(sn^2) - 1)
            ## use "<<-" for field assignment
            Cov <<- ifelse(m == 0, 0, s / m)
            Sd <<- s
            Mean <<- m
          },
          slnorm = {
            mn <- DistributionParameters[1]
            sn <- DistributionParameters[2]
            x0 <<- DistributionParameters[3]
            m <- x0 + exp(mn + (sn^2) / 2)
            s <- exp(mn + (sn^2) / 2) * sqrt(exp(sn^2) - 1)
            Cov <<- ifelse(m == 0, 0, s / m)
            Sd <<- s
            Mean <<- m
          },
          gumbel = {
            loc <- DistributionParameters[1]
            sc <- DistributionParameters[2]
            m <- loc - digamma(1) * sc
            s <- (pi / sqrt(6)) * sc
            Cov <<- ifelse(m == 0, 0, s / m)
            Sd <<- s
            Mean <<- m
          },
          exp = {
            rate <- DistributionParameters[1]
            m <- ifelse(rate == 0 || is.na(rate), NA_real_, 1 / rate)
            s <- m
            covv <- ifelse(m == 0 || !is.finite(s) || !is.finite(m),
              NA_real_, s / m
            )
            Cov <<- covv
            Sd <<- s
            Mean <<- m
          },
          gamma = {
            shape <- DistributionParameters[1]
            scale <- DistributionParameters[2]
            m <- shape * scale
            s <- sqrt(shape) * scale
            covv <- ifelse(m == 0 || !is.finite(s) || !is.finite(m),
              NA_real_, s / m
            )
            Cov <<- covv
            Sd <<- s
            Mean <<- m
          },
          beta = {
            ## Extract alpha and beta parameters from DistributionParameters -------------------
            alpha <- DistributionParameters[1]
            beta_ <- DistributionParameters[2]

            ## Validate parameters -----------------------------------------
            if (is.na(alpha) || is.na(beta_) || alpha <= 0 || beta_ <= 0) {
              Mean <<- NA_real_
              Sd <<- NA_real_
              Cov <<- NA_real_
            } else {
              ## Compute mean and standard deviation from parameters -------
              m <- alpha / (alpha + beta_)
              var_val <- (alpha * beta_) /
                ((alpha + beta_)^2 * (alpha + beta_ + 1))

              s <- sqrt(var_val)
              covv <- ifelse(m == 0, 0, s / m)

              ## Assign results to object fields ---------------------------
              Mean <<- m
              Sd <<- s
              Cov <<- covv
            }
          },
          weibull = {
            k <- DistributionParameters[1]
            lam <- DistributionParameters[2]
            Mean <<- lam * base::gamma(1 + 1 / k)
            Sd <<- sqrt(lam^2 *
              (base::gamma(1 + 2 / k) -
                base::gamma(1 + 1 / k)^2))
          },
          emp = {
            # https://en.wikipedia.org/wiki/Empirical_distribution_function
            # obs (sample) is given via DistributionParameters
            # https://rdrr.io/cran/EnvStats/man/Empirical.html

            Mean <<- mean(DistributionParameters)
            Sd <<- sd(DistributionParameters)
            Cov <<- Sd / Mean
          },
          st = {
            m <- DistributionParameters[1]
            s <- DistributionParameters[2]
            n <- DistributionParameters[3]
            nue <- DistributionParameters[4]

            Mean <<- if (nue > 1) m else NA
            Sd <<- if (nue > 2) s / sqrt(n / (n + 1)) * sqrt(nue / (nue - 2)) else if (nue > 1 && nue <= 2) Inf else NA
            # Sd <<- s/sqrt(n/(n+1))/sqrt(nue/(nue-2))
            Cov <<- if (nue > 2) Sd / Mean else NA
          },
          lt = {
            Mean <<- Inf # Median will be used as Start Value for FORM
            Sd <<- Inf
            Cov <<- NA_real_
          },
          {
            warning(sprintf("Backward transformation for '%s' not implemented", DistributionType))
          }
        )
      }

      # --- Recalculate CoV safely -------------------------------------------
      eps <- .Machine$double.eps^0.5 # ~1e-8

      if (is.finite(Mean) && is.finite(Sd) && abs(Mean) > eps) {
        Cov <<- Sd / Mean
      } else if (is.finite(Sd) && abs(Mean) <= eps) {
        Cov <<- Inf
      } else {
        Cov <<- NA_real_
      }

      invisible(TRUE)
    }
  )
)


#' @title Object to store a deterministic parameters (PROB_DETVAR)
#'
#' @description A convenience subclass of `PROB_BASEVAR` that represents a deterministic
#' value (i.e. a variable with practically zero variance).  The object is
#' automatically converted to a normal distribution with an infinitesimal
#' standard deviation (`Sd = Mean / 1e7`).
#'
#' @section Inherited fields:
#' \describe{
#' Inherits all fields of `PROB_BASEVAR` (see that documentation for a full list).
#' }
#'
#' @section Additional fields:
#' \describe{
#'   \item{\code{Value}}{
#'     The deterministic value used internally as \code{Mean}.
#'   }
#' }
#'
#' @examples
#' det <- PROB_DETVAR(Name = "E_mod", Value = 30e3)
#' det$prepare()
#'
#' @author (C) 2021-2026 K. Nille-Hauf, T. Feiri, M. Ricker, T. Lux -- Hochschule Biberach (until 2022),
#' TU Dortmund University - Chair of Structural Concrete (since 2023)
#'
#' @export PROB_DETVAR
#' @exportClass PROB_DETVAR
PROB_DETVAR <- setRefClass(
  Class = "PROB_DETVAR",
  contains = "PROB_BASEVAR",
  fields = c(
    Value = "numeric"
  ),
  methods = list(
    initialize = function(...) {
      initFields(...)
      DistributionType <<- "norm" # Eigentlich nicht notwendig, da PROB_BASE var bei DistributionTyp = "" auf "norm" setzt
      if (length(Value) != 1) {
        stop("PROB_DETVAR: 'Value' must be a scalar numeric.")
      }

      Value <<- as.numeric(Value)[1]

      Mean <<- Value
      Sd <<- Mean / 1e7

      Mean <<- as.numeric(Mean)[1]
      Sd <<- as.numeric(Sd)[1]
      Cov <<- if (Mean == 0) 0 else Sd / Mean

      prepare()
    }
  )
)


#' @title Reliability Algorithm (PROB_MACHINE)
#'
#' @description Stores the definition of a reliability method (FORM, SORM, Monte-Carlo,
#' etc.) that can be executed by `SYS_PROB` or `SYS_PARAM`.  The actual
#' algorithm is called via the character string `fCall` (e.g. `"FORM"` or
#' `"MC_IS"`).  Optional arguments are passed through the list `options`.
#'
#' @field name    Descriptive name of the machine (appears in reports).
#' @field fCall   Function name that implements the algorithm (must be
#'                available in the package namespace, e.g. `FORM`,
#'                `SORM`, `MC_CRUDE`, `MC_IS`, `MC_SubSam`).
#' @field options Additional options for the algorithm (list, e.g.
#'                `list(n_max = 1e6, cov_user = 0.05)`).
#'
#' @details
#' The helper method `getMethodLevel()` returns an integer that indicates
#' the *complexity* of the method (1 = first-order, 2 = second-order, 3 = Monte-Carlo).
#'
#' @examples
#' ## FORM machine ------------------------------------------------------------
#' form_rf <- PROB_MACHINE(
#'   name = "FORM - Rack-Fies",
#'   fCall = "FORM",
#'   options = list(n_optim = 20, loctol = 1e-3, optim_type = "rackfies")
#' )
#'
#' ## Monte-Carlo importance sampling -----------------------------------------
#' mcis <- PROB_MACHINE(
#'   name = "MC-IS",
#'   fCall = "MC_IS",
#'   options = list(cov_user = 0.05, n_max = 3e5, seed = 1234)
#' )
#'
#' @author (C) 2021-2026 K. Nille-Hauf, T. Feiri, M. Ricker, T. Lux -- Hochschule Biberach (until 2022),
#' TU Dortmund University - Chair of Structural Concrete (since 2023)
#'
#' @export PROB_MACHINE
#' @exportClass PROB_MACHINE

PROB_MACHINE <- setRefClass(
  Class = "PROB_MACHINE",
  fields = list(
    name    = "character",
    fCall   = "character",
    options = "list" # <-- changed from "vector"
  ),
  methods = list(
    initialize = function(name = "", fCall = "", options = list()) {
      .self$name <<- name
      .self$fCall <<- fCall
      .self$options <<- options

      ## ---- FORM / SORM need n_optim and loctol -------------------------
      ## If the user did not provide them we inject default values.
      if (grepl("^FORM$", fCall, ignore.case = TRUE) ||
        grepl("^SORM$", fCall, ignore.case = TRUE)) {
        if (is.null(.self$options$n_optim)) .self$options$n_optim <- 20L
        if (is.null(.self$options$loctol)) .self$options$loctol <- 1e-4
      }
    },
    getMethodLevel = function() {
      if (grepl("MC", fCall)) {
        return(3L)
      } else if (grepl("ORM", fCall)) {
        return(2L)
      } else if (identical(fCall, "MVFOSM")) {
        return(2L)
      } else {
        return(1L)
      }
    },

    ## -------------------------------------------------
    ##  Helper used by SYS_PROB$runMachines() and SYS_PARAM$runMachines()
    ##  to produce a **uniform** error message.
    ## -------------------------------------------------
    checkMethodExists = function() {
      ## `exists()` looks for a *function* in the global namespace.
      ## If it is not there we abort with the exact wording the tests expect.
      if (!exists(fCall, mode = "function")) {
        stop(sprintf("Method '%s' is not loaded", fCall), call. = FALSE)
      }
    }
  )
)

# utils.R (oder am Anfang einer deiner .R-Dateien)
#' Ensure that a required package is available
#'
#' @param pkg Character string with the package name
#' @return Invisible TRUE (used only for its side-effect)
#' @note The function stops with an informative error if the required
#'   package is not installed.
#' @keywords internal
ensure_pkg <- function(pkg) {
  if (!requireNamespace(pkg, quietly = TRUE)) {
    stop(
      sprintf(
        "Package \"%s\" is required but not installed.\n",
        pkg
      ),
      call. = FALSE
    )
  }
  invisible(TRUE)
}


#' Smooth approximation of the minimum function
#'
#' Computes a differentiable approximation of \code{min(x)}
#' using a log-sum-exp formulation:
#'
#' \deqn{
#'   \min(x) \approx -\frac{1}{\kappa} \log\left( \sum_i e^{-\kappa x_i} \right)
#' }
#'
#' The approximation becomes exact as \code{kappa → ∞}.
#'
#' @param x Numeric vector.
#' @param kappa Positive smoothing parameter controlling
#'   the sharpness of the approximation (default = 20).
#'
#' @return Numeric scalar.
#'
#' @details
#' This function is used internally for smooth system
#' limit-state aggregation (serial systems) to improve
#' numerical stability and differentiability.
#'
#' @keywords internal
#' @noRd
smooth_min <- function(x, kappa = 20) {
  -log(sum(exp(-kappa * x))) / kappa
}

#' Smooth approximation of the maximum function
#'
#' Computes a differentiable approximation of \code{max(x)}
#' using a log-sum-exp formulation:
#'
#' \deqn{
#'   \max(x) \approx \frac{1}{\kappa} \log\left( \sum_i e^{\kappa x_i} \right)
#' }
#'
#' The approximation becomes exact as \code{kappa → ∞}.
#'
#' @param x Numeric vector.
#' @param kappa Positive smoothing parameter controlling
#'   the sharpness of the approximation (default = 20).
#'
#' @return Numeric scalar.
#'
#' @details
#' This function is used internally for smooth system
#' limit-state aggregation (parallel systems) to improve
#' numerical stability and differentiability.
#'
#' @keywords internal
#' @noRd
smooth_max <- function(x, kappa = 20) {
  log(sum(exp(kappa * x))) / kappa
}

library(testthat)

test_that("MC_IS_single reproduces FORM for linear LSF", {
    skip_on_cran()

    # -----------------------------
    # Basisvariablen
    # -----------------------------
    R <- PROB_BASEVAR(
        Id = 1, Name = "R",
        DistributionType = "norm",
        Mean = 10, Sd = 2
    )

    S <- PROB_BASEVAR(
        Id = 2, Name = "S",
        DistributionType = "norm",
        Mean = 8, Sd = 2
    )

    # -----------------------------
    # LSF
    # -----------------------------
    lsf <- SYS_LSF(vars = list(R, S), name = "Linear test")

    lsf$func <- function(R, S) {
        R - S
    }

    lsf$check()

    # -----------------------------
    # MC_IS Maschine
    # -----------------------------
    machine_MC_IS <- PROB_MACHINE(
        name = "MC_IS",
        fCall = "MC_IS",
        options = list(
            n_batch = 2000,
            n_max = 5e4,
            cov_user = 0.05,
            seed = 123,
            use_threads = 1
        )
    )

    # -----------------------------
    # FORM Maschine
    # -----------------------------
    machine_FORM <- PROB_MACHINE(
        name = "FORM",
        fCall = "FORM"
    )

    # -----------------------------
    # Problem aufsetzen
    # -----------------------------
    ps <- SYS_PROB(
        sys_input = list(lsf),
        probMachines = list(machine_FORM, machine_MC_IS),
        debug.level = 0
    )

    ps$runMachines()

    beta_FORM <- ps$beta_single["FORM", 1]
    beta_MCIS <- ps$beta_single["MC_IS", 1]

    # -----------------------------
    # Test
    # -----------------------------
    expect_lt(abs(beta_MCIS - beta_FORM), 0.05)
})


test_that("MC_IS_single works with non-normal distributions", {
    skip_on_cran()

    # -----------------------------
    # Basisvariablen
    # -----------------------------
    Z <- PROB_BASEVAR(
        Id = 1, Name = "Z",
        DistributionType = "norm",
        Mean = 100, Cov = 0.04
    )

    Fy <- PROB_BASEVAR(
        Id = 2, Name = "Fy",
        DistributionType = "lnorm",
        Mean = 40, Cov = 0.1
    )

    M <- PROB_BASEVAR(
        Id = 3, Name = "M",
        DistributionType = "gumbel",
        Package = "evd",
        Mean = 2000, Cov = 0.1
    )

    # -----------------------------
    # LSF
    # -----------------------------
    lsf <- SYS_LSF(vars = list(Z, Fy, M), name = "Non-normal test")

    lsf$func <- function(Z, Fy, M) {
        Z * Fy - M
    }

    lsf$check()

    # -----------------------------
    # Maschinen
    # -----------------------------
    machine_FORM <- PROB_MACHINE(
        name = "FORM",
        fCall = "FORM"
    )

    machine_MC_IS <- PROB_MACHINE(
        name = "MC_IS",
        fCall = "MC_IS",
        options = list(
            n_batch = 5000,
            n_max = 1e5,
            cov_user = 0.05,
            seed = 123,
            use_threads = 1
        )
    )

    # -----------------------------
    # Problem aufsetzen
    # -----------------------------
    ps <- SYS_PROB(
        sys_input = list(lsf),
        probMachines = list(machine_FORM, machine_MC_IS),
        debug.level = 0
    )

    ps$runMachines()

    beta_FORM <- ps$beta_single["FORM", 1]
    beta_MCIS <- ps$beta_single["MC_IS", 1]

    # -----------------------------
    # Test
    # -----------------------------
    expect_lt(abs(beta_MCIS - beta_FORM), 0.1)
})


test_that("MC_IS_single estimates rare-event probability correctly", {
    skip_on_cran()

    # -----------------------------
    # Basisvariablen
    # -----------------------------
    R <- PROB_BASEVAR(
        Id = 1, Name = "R",
        DistributionType = "norm",
        Mean = 100, Sd = 10
    )

    S <- PROB_BASEVAR(
        Id = 2, Name = "S",
        DistributionType = "norm",
        Mean = 70, Sd = 10
    )

    # -----------------------------
    # LSF
    # -----------------------------
    lsf <- SYS_LSF(vars = list(R, S), name = "Rare event test")

    lsf$func <- function(R, S) {
        R - S
    }

    lsf$check()

    # -----------------------------
    # Maschinen
    # -----------------------------
    machine_FORM <- PROB_MACHINE(
        name = "FORM",
        fCall = "FORM"
    )

    machine_MC_IS <- PROB_MACHINE(
        name = "MC_IS",
        fCall = "MC_IS",
        options = list(
            n_batch = 5000,
            n_max = 1e5,
            cov_user = 0.05,
            seed = 123,
            use_threads = 1
        )
    )

    # -----------------------------
    # Problem
    # -----------------------------
    ps <- SYS_PROB(
        sys_input = list(lsf),
        probMachines = list(machine_FORM, machine_MC_IS),
        debug.level = 0
    )

    ps$runMachines()

    beta_FORM <- ps$beta_single["FORM", 1]
    beta_MCIS <- ps$beta_single["MC_IS", 1]

    # -----------------------------
    # Test
    # -----------------------------
    expect_lt(abs(beta_MCIS - beta_FORM), 0.1)
})

library(testthat)

test_that("PROB_BASEVAR: prepare and getlDistr produce funlist", {
    v <- PROB_BASEVAR(Id = 1, Name = "f_ck", DistributionType = "norm", Mean = 30, Sd = 1)

    # Accept both S3 and S4
    expect_true(inherits(v, "PROB_BASEVAR"))

    dd <- v$getlDistr()
    # dd kann wrapper (list(funlist, params)) oder funlist sein -> normalize
    funlist <- if (is.list(dd) && length(dd) >= 1 && is.list(dd[[1]]) && !is.null(dd[[1]]$p)) dd[[1]] else dd

    expect_true(is.list(funlist))
    expect_true(is.function(funlist$d))
    expect_true(is.function(funlist$p))
    expect_true(is.function(funlist$q))
    expect_true(is.function(funlist$r))
})


test_that("SYS_LSF getLSF evaluates correctly", {
    v1 <- PROB_BASEVAR(Id = 1, Name = "X1", DistributionType = "norm", Mean = 0, Sd = 1)

    lsf_obj <- SYS_LSF(name = "g1", vars = list(v1))
    lsf_obj$func <- function(X1) 3 - X1

    f <- lsf_obj$getLSF()
    expect_true(is.function(f))

    expect_equal(f(c(2)), 1)
    expect_equal(f(c(4)), -1)
})


library(testthat)

test_that("MC_IS_system: deterministic with fixed seed and dps (quick smoke test)", {
    if (.Platform$OS.type == "windows") {
        backend <- "future"
    } else {
        backend <- "parallel"
    }

    v1 <- PROB_BASEVAR(Id = 1, Name = "X1", DistributionType = "norm", Mean = 0, Sd = 1)
    v2 <- PROB_BASEVAR(Id = 2, Name = "X2", DistributionType = "norm", Mean = 0, Sd = 1)

    lsf1 <- function(x) 3 - x[1]
    lsf2 <- function(x) 3 - x[1]

    # get funlist (normalize)
    d1 <- v1$getlDistr()
    d1 <- if (is.list(d1) && length(d1) >= 1 && is.list(d1[[1]]) && !is.null(d1[[1]]$p)) d1[[1]] else d1
    d2 <- v2$getlDistr()
    d2 <- if (is.list(d2) && length(d2) >= 1 && is.list(d2[[1]]) && !is.null(d2[[1]]$p)) d2[[1]] else d2

    lDistr <- list(list(d1), list(d2))

    run_once <- function(seed_val) {
        MC_IS(
            lsf = list(lsf1, lsf2),
            lDistr = lDistr,
            cov_user = 0.5,
            n_batch = 1000,
            n_max = 2000,
            use_threads = 1,
            backend = backend,
            seed = seed_val,
            dps = c(3, 3),
            dataRecord = TRUE,
            debug.level = 0
        )
    }

    r1 <- run_once(1234)
    r2 <- run_once(1234)

    expect_equal(r1$pf, r2$pf)
    expect_equal(r1$n_mc, r2$n_mc)

    expect_true(is.numeric(r1$pf))
    expect_true(r1$pf >= 0 && r1$pf <= 1)

    # Test SYS_PROB wrapper
    lsf1obj <- SYS_LSF(name = "g1", vars = list(v1))
    lsf1obj$func <- function(X1) 3 - X1
    lsf2obj <- SYS_LSF(name = "g2", vars = list(v2))
    lsf2obj$func <- function(X2) 3 - X2
    ps <- SYS_PROB(sys_input = list(lsf1obj, lsf2obj), probMachines = list(), sys_type = "serial")
    ps$calculateSystemProbability(calcType = "MCIS", params = list(
        cov_user = 0.5, n_batch = 1000, n_max = 2000, use_threads = 1,
        seed = 1234, backend = "future", dps = c(3, 3), dataRecord = TRUE
    ))
    expect_true(length(ps$res_sys) >= 1)
    last <- ps$res_sys[[length(ps$res_sys)]]
    expect_true(!is.null(last$pf))
})


test_that("MC_IS_single is reproducible for same seed", {
    if (.Platform$OS.type == "windows") {
        backend <- "future"
    } else {
        backend <- "parallel"
    }
    v <- PROB_BASEVAR(
        Id = 1, Name = "X",
        DistributionType = "norm",
        Mean = 0, Sd = 1
    )

    lsf <- SYS_LSF(name = "g", vars = list(v))
    lsf$func <- function(X) 3 - X

    ps1 <- SYS_PROB(
        sys_input = list(lsf),
        probMachines = list(),
        sys_type = "serial"
    )

    ps1$calculateSystemProbability(
        calcType = "MCIS",
        params = list(
            cov_user = 0.01,
            n_batch = 5000,
            n_max = 1e5,
            seed = 1234,
            backend = backend
        )
    )

    beta1 <- ps1$beta_sys[1, 1]

    ps2 <- SYS_PROB(
        sys_input = list(lsf),
        probMachines = list(),
        sys_type = "serial"
    )

    ps2$calculateSystemProbability(
        calcType = "MCIS",
        params = list(
            cov_user = 0.01,
            n_batch = 5000,
            n_max = 1e5,
            seed = 1234,
            backend = backend
        )
    )

    beta2 <- ps2$beta_sys[1, 1]

    expect_equal(beta1, beta2)
})


test_that("MC_IS_system is reproducible for same seed", {
    if (.Platform$OS.type == "windows") {
        backend <- "future"
    } else {
        backend <- "parallel"
    }

    v1 <- PROB_BASEVAR(
        Id = 1, Name = "X1",
        DistributionType = "norm",
        Mean = 0, Sd = 1
    )

    v2 <- PROB_BASEVAR(
        Id = 2, Name = "X2",
        DistributionType = "norm",
        Mean = 0, Sd = 1
    )

    lsf1 <- function(x) 3 - x[1]
    lsf2 <- function(x) 3 - x[1]

    d1 <- v1$getlDistr()[[1]]
    d2 <- v2$getlDistr()[[1]]

    lDistr <- list(list(d1), list(d2))

    run_once <- function(seed_val) {
        MC_IS(
            lsf = list(lsf1, lsf2),
            lDistr = lDistr,
            cov_user = 0.5,
            n_batch = 1000,
            n_max = 2000,
            use_threads = 1,
            backend = backend,
            seed = seed_val,
            dps = c(3, 3),
            dataRecord = TRUE
        )
    }

    r1 <- run_once(1234)
    r2 <- run_once(1234)

    expect_equal(r1$pf, r2$pf)
    expect_equal(r1$n_mc, r2$n_mc)
})


test_that("ESS is positive and bounded", {
    v <- PROB_BASEVAR(
        Id = 1, Name = "X",
        DistributionType = "norm",
        Mean = 0, Sd = 1
    )

    lsf <- function(x) 3 - x[1]

    d1 <- v$getlDistr()[[1]]

    res <- MC_IS(
        lsf = list(lsf),
        lDistr = list(list(d1)),
        cov_user = 0.2,
        n_batch = 2000,
        n_max = 5000,
        use_threads = 1,
        backend = "future",
        seed = 1234
    )

    expect_true(res$ESS > 0)
    expect_true(res$ESS <= res$n_mc)
})


test_that("MC_IS_system handles intersecting limit states", {
    if (.Platform$OS.type == "windows") {
        backend <- "future"
    } else {
        backend <- "parallel"
    }

    v1 <- PROB_BASEVAR(
        Id = 1, Name = "X1",
        DistributionType = "norm",
        Mean = 0, Sd = 1
    )

    v2 <- PROB_BASEVAR(
        Id = 2, Name = "X2",
        DistributionType = "norm",
        Mean = 0, Sd = 1
    )

    lsf1 <- function(x) 3 - x[1]
    lsf2 <- function(x) 3 - (0.7 * x[1] + 0.7 * x[2])

    d1 <- v1$getlDistr()[[1]]
    d2 <- v2$getlDistr()[[1]]

    res <- MC_IS(
        lsf = list(lsf1, lsf2),
        lDistr = list(list(d1), list(d1, d2)),
        cov_user = 0.05,
        n_batch = 5000,
        n_max = 1e5,
        use_threads = 1,
        backend = backend,
        seed = 1234
    )

    expect_true(res$pf > 0)
    expect_true(res$pf < 1)
})


test_that("MC_IS serial is reproducible for fixed seed", {
    if (.Platform$OS.type == "windows") {
        backend <- "future"
    } else {
        backend <- "parallel"
    }

    v <- PROB_BASEVAR(
        Id = 1, Name = "X",
        DistributionType = "norm",
        Mean = 0, Sd = 1
    )

    lsf <- SYS_LSF(name = "g", vars = list(v))
    lsf$func <- function(X) 3 - X

    ps1 <- SYS_PROB(
        sys_input = list(lsf),
        probMachines = list(),
        sys_type = "serial"
    )

    ps2 <- SYS_PROB(
        sys_input = list(lsf),
        probMachines = list(),
        sys_type = "serial"
    )

    ps1$calculateSystemProbability(
        calcType = "MCIS",
        params = list(
            cov_user = 0.02,
            n_batch = 5000,
            n_max = 1e5,
            seed = 1234,
            backend = backend
        )
    )

    ps2$calculateSystemProbability(
        calcType = "MCIS",
        params = list(
            cov_user = 0.02,
            n_batch = 5000,
            n_max = 1e5,
            seed = 1234,
            backend = backend
        )
    )

    expect_equal(
        ps1$res_sys[[1]]$pf,
        ps2$res_sys[[1]]$pf
    )
})


# test_that("adaptive alpha is reproducible for fixed seed", {
#     skip_if_not_installed("future")
#     old_plan <- future::plan()
#     future::plan(future::sequential)
#     on.exit(future::plan(old_plan), add = TRUE)

#     v1 <- PROB_BASEVAR(
#         Id = 1, Name = "X1",
#         DistributionType = "norm",
#         Mean = 0, Sd = 1
#     )

#     v2 <- PROB_BASEVAR(
#         Id = 2, Name = "X2",
#         DistributionType = "norm",
#         Mean = 0, Sd = 1
#     )

#     lsf1 <- function(x) 3 - x[1]
#     lsf2 <- function(x) 3 - (0.7 * x[1] + 0.7 * x[2])

#     d1 <- v1$getlDistr()[[1]]
#     d2 <- v2$getlDistr()[[1]]

#     lDistr <- list(list(d1), list(d1, d2))

#     run_once <- function(seed_val) {
#         MC_IS(
#             lsf = list(lsf1, lsf2),
#             lDistr = lDistr,
#             cov_user = 0.02,
#             n_batch = 10000,
#             n_max = 2e5,
#             use_threads = 1,
#             backend = "future",
#             seed = seed_val,
#             adaptive_alpha = TRUE
#         )
#     }

#     r1 <- run_once(1234)
#     r2 <- run_once(1234)

#     expect_equal(r1$pf, r2$pf)
# })


# test_that("adaptive alpha does not catastrophically worsen Pf", {
#     v1 <- PROB_BASEVAR(
#         Id = 1, Name = "X1",
#         DistributionType = "norm",
#         Mean = 0, Sd = 1
#     )

#     v2 <- PROB_BASEVAR(
#         Id = 2, Name = "X2",
#         DistributionType = "norm",
#         Mean = 0, Sd = 1
#     )

#     lsf1 <- function(x) 3 - x[1]
#     lsf2 <- function(x) 3 - (0.7 * x[1] + 0.7 * x[2])

#     d1 <- v1$getlDistr()[[1]]
#     d2 <- v2$getlDistr()[[1]]

#     lDistr <- list(list(d1), list(d1, d2))

#     static_run <- MC_IS(
#         lsf = list(lsf1, lsf2),
#         lDistr = lDistr,
#         cov_user = 0.02,
#         n_batch = 10000,
#         n_max = 2e5,
#         use_threads = 1,
#         backend = "future",
#         seed = 1234,
#         adaptive_alpha = FALSE
#     )

#     adaptive_run <- MC_IS(
#         lsf = list(lsf1, lsf2),
#         lDistr = lDistr,
#         cov_user = 0.02,
#         n_batch = 10000,
#         n_max = 2e5,
#         use_threads = 1,
#         backend = "future",
#         seed = 1234,
#         adaptive_alpha = TRUE
#     )

#     expect_lt(
#         abs(adaptive_run$pf - static_run$pf),
#         0.5 * static_run$pf
#     )
# })


# test_that("ESS is finite and positive for adaptive alpha", {
#     if (.Platform$OS.type == "windows") {
#         backend <- "future"
#     } else {
#         backend <- "parallel"
#     }

#     v1 <- PROB_BASEVAR(
#         Id = 1, Name = "X1",
#         DistributionType = "norm",
#         Mean = 0, Sd = 1
#     )

#     v2 <- PROB_BASEVAR(
#         Id = 2, Name = "X2",
#         DistributionType = "norm",
#         Mean = 0, Sd = 1
#     )

#     lsf1 <- function(x) 3 - x[1]
#     lsf2 <- function(x) 3 - (0.7 * x[1] + 0.7 * x[2])

#     d1 <- v1$getlDistr()[[1]]
#     d2 <- v2$getlDistr()[[1]]

#     lDistr <- list(list(d1), list(d1, d2))

#     res <- MC_IS(
#         lsf = list(lsf1, lsf2),
#         lDistr = lDistr,
#         cov_user = 0.02,
#         n_batch = 10000,
#         n_max = 2e5,
#         use_threads = 1,
#         backend = backend,
#         seed = 1234,
#         adaptive_alpha = TRUE
#     )

#     expect_true(is.finite(res$ESS))
#     expect_gt(res$ESS, 0)
# })


test_that("compute_weight_update fast mode accumulates correctly", {
    state <- list(
        total_sum_w = 0,
        total_sum_Iw = 0,
        total_sum_w2 = 0,
        log_total_sum_w = NULL,
        log_total_sum_Iw = NULL,
        log_total_sum_w2 = NULL,
        n_sim = 0
    )

    all_log_w <- c(log(2), log(1), log(3))
    all_I_vals <- c(1, 0, 1)

    res <- compute_weight_update(
        all_log_w = all_log_w,
        all_I_vals = all_I_vals,
        stability_mode = "fast",
        sys_type = "serial",
        state = state
    )

    expect_true(res$total_sum_w > 0)
    expect_equal(res$n_sim, 3)
    expect_true(res$total_sum_Iw <= res$total_sum_w)
    expect_true(res$total_sum_w2 > 0)
})


test_that("compute_weight_update robust mode accumulates log-sums", {
    state <- list(
        total_sum_w = 0,
        total_sum_Iw = 0,
        total_sum_w2 = 0,
        log_total_sum_w = -Inf,
        log_total_sum_Iw = -Inf,
        log_total_sum_w2 = -Inf,
        n_sim = 0
    )

    all_log_w <- c(-1, -2, -0.5)
    all_I_vals <- c(1, 0, 1)

    res <- compute_weight_update(
        all_log_w = all_log_w,
        all_I_vals = all_I_vals,
        stability_mode = "robust",
        sys_type = "parallel",
        state = state
    )

    expect_true(is.finite(res$log_total_sum_w))
    expect_equal(res$n_sim, 3)
})


context("MC_IS_system: geometry, determinism, and debug propagation")

# ------------------------------------------------------------
# Helper: create simple base variables
# ------------------------------------------------------------

create_vars <- function() {
    X1 <- PROB_BASEVAR(
        Id = 1, Name = "X1",
        DistributionType = "norm",
        Mean = 10, Sd = 1
    )

    X2 <- PROB_BASEVAR(
        Id = 2, Name = "X2",
        DistributionType = "norm",
        Mean = 5, Sd = 1
    )

    list(X1, X2)
}

# ------------------------------------------------------------
# Helper: wrap functions into SYS_LSF
# ------------------------------------------------------------

build_lsf_objects <- function(lsf_list, vars) {
    lapply(seq_along(lsf_list), function(i) {
        obj <- SYS_LSF(vars = vars, name = paste0("LSF", i))
        obj$func <- lsf_list[[i]]
        obj$check()
        obj
    })
}

# ============================================================
# TEST 1 — Debug level propagation
# ============================================================

test_that("debug.level is correctly propagated to MC_IS_system", {
    vars <- create_vars()

    lsf <- function(x) x[1] - x[2]

    ps <- SYS_PROB(
        sys_input = build_lsf_objects(list(lsf, lsf), vars),
        probMachines = list(),
        debug.level = 1,
        sys_type = "serial"
    )

    expect_equal(ps$debug.level, 1)
})

# ============================================================
# TEST 2 — Geometry detection: identical LSFs → single shift
# ============================================================

test_that("Identical LSFs are classified as non-multimodal", {
    vars <- create_vars()

    lsf <- function(x) x[1] - x[2]

    ps <- SYS_PROB(
        sys_input = build_lsf_objects(list(lsf, lsf), vars),
        probMachines = list(),
        debug.level = 0,
        sys_type = "serial"
    )

    res <- ps$calculateSystemProbability(
        calcType = "MCIS",
        params = list(
            n_batch = 2000,
            n_max = 5000,
            cov_user = 0.5,
            use_threads = 1,
            backend = "parallel",
            seed = 123
        )
    )

    expect_true(TRUE) # no crash = pass
})

# ============================================================
# TEST 3 — Geometry detection: orthogonal LSFs → multimodal
# ============================================================

test_that("Orthogonal LSFs are classified as multimodal", {
    vars <- create_vars()

    lsf1 <- function(x) x[1] - 8
    lsf2 <- function(x) x[2] - 3

    ps <- SYS_PROB(
        sys_input = build_lsf_objects(list(lsf1, lsf2), vars),
        probMachines = list(),
        debug.level = 0,
        sys_type = "serial"
    )

    res <- ps$calculateSystemProbability(
        calcType = "MCIS",
        params = list(
            n_batch = 2000,
            n_max = 5000,
            cov_user = 0.5,
            use_threads = 1,
            backend = "parallel",
            seed = 123
        )
    )

    expect_true(is.numeric(ps$beta_sys[1, 1]))
})

# ============================================================
# TEST 4 — Determinism across thread counts
# ============================================================

test_that("MC_IS_system is deterministic across thread counts", {
    vars <- create_vars()

    lsf1 <- function(x) x[1] - x[2]
    lsf2 <- function(x) x[1] + x[2] - 12

    run_test <- function(threads) {
        ps_local <- SYS_PROB(
            sys_input = build_lsf_objects(list(lsf1, lsf2), vars),
            probMachines = list(),
            debug.level = 0,
            sys_type = "serial"
        )

        ps_local$calculateSystemProbability(
            calcType = "MCIS",
            params = list(
                n_batch = 3000,
                n_max = 8000,
                cov_user = 0.5,
                use_threads = threads,
                backend = "parallel",
                seed = 42
            )
        )

        ps_local$beta_sys
    }

    res1 <- run_test(1)
    res2 <- run_test(2)

    expect_equal(res1, res2)
})

# ============================================================
# TEST 5 — adaptive_alpha does not break single shift
# ============================================================

test_that("Adaptive alpha does not crash in single-shift case", {
    vars <- create_vars()

    lsf <- function(x) x[1] - x[2]

    ps <- SYS_PROB(
        sys_input = build_lsf_objects(list(lsf, lsf), vars),
        probMachines = list(),
        debug.level = 0,
        sys_type = "serial"
    )

    ps$calculateSystemProbability(
        calcType = "MCIS",
        params = list(
            n_batch = 2000,
            n_max = 5000,
            cov_user = 0.5,
            use_threads = 1,
            backend = "parallel",
            seed = 123,
            adaptive_alpha = TRUE
        )
    )

    expect_true(is.finite(ps$beta_sys[1, 1]))
})


test_that("build_gsys constructs valid system object", {
    v1 <- PROB_BASEVAR(Name = "X1", DistributionType = "norm", Mean = 5, Sd = 1)
    v2 <- PROB_BASEVAR(Name = "X2", DistributionType = "norm", Mean = 3, Sd = 1)

    lsf1 <- SYS_LSF(
        name = "g1",
        vars = list(v1)
    )
    lsf1$func <- function(x) x[1] - 4

    lsf2 <- SYS_LSF(
        name = "g2",
        vars = list(v2)
    )
    lsf2$func <- function(x) x[1] - 2

    sys <- build_gsys(list(lsf1, lsf2))

    expect_true(is.function(sys$g_sys))
    expect_equal(sys$n_lsfs, 2)
    expect_equal(sys$n_vars, 2)
    expect_length(sys$distr_flat, 2)
})


test_that("g_sys serial approximates minimum", {
    v1 <- PROB_BASEVAR(Name = "X1", DistributionType = "norm", Mean = 0, Sd = 1)
    v2 <- PROB_BASEVAR(Name = "X2", DistributionType = "norm", Mean = 0, Sd = 1)

    lsf1 <- SYS_LSF(vars = list(v1))
    lsf1$func <- function(x) x[1] - 1

    lsf2 <- SYS_LSF(vars = list(v2))
    lsf2$func <- function(x) x[1] - 2

    sys <- build_gsys(
        sys_input = list(lsf1, lsf2),
        sys_structure = "serial",
        smooth_kappa = 100
    )

    x <- c(1.5, 3)

    g_vals <- c(
        lsf1$getLSF()(x[1]),
        lsf2$getLSF()(x[2])
    )

    g_expected <- min(g_vals)

    g_sys <- sys$g_sys(x)

    expect_equal(g_sys, g_expected, tolerance = 1e-3)
})


test_that("smooth serial aggregation behaves correctly", {
    v1 <- PROB_BASEVAR(Name = "X1", DistributionType = "norm", Mean = 0, Sd = 1)
    v2 <- PROB_BASEVAR(Name = "X2", DistributionType = "norm", Mean = 0, Sd = 1)

    lsf1 <- SYS_LSF(vars = list(v1))
    lsf1$func <- function(x) x[1] - 1

    lsf2 <- SYS_LSF(vars = list(v2))
    lsf2$func <- function(x) x[1] - 2

    x <- c(1.5, 3)

    g_vals <- c(
        lsf1$getLSF()(c(x[1])),
        lsf2$getLSF()(c(x[2]))
    )

    g_min <- min(g_vals)

    kappa <- 20

    sys <- build_gsys(
        sys_input = list(lsf1, lsf2),
        sys_structure = "serial",
        smooth_kappa = kappa
    )

    g_sys <- sys$g_sys(x)

    # print(sys$g_sys(x))
    # print(smooth_min(c(0.5, 1), 100))

    # print(sys$g_sys(x))
    # print(g_vals)
    # print(lsf1$getLSF()(c(1.5)))

    # ----------------------------
    # Property 1: smooth ≤ min
    # ----------------------------

    expect_true(g_sys <= g_min)

    # ----------------------------
    # Property 2: error bound
    # min − log(m)/kappa ≤ g_sys
    # ----------------------------

    m <- length(g_vals)

    lower_bound <- g_min - log(m) / kappa

    expect_true(g_sys >= lower_bound)
})


test_that("smooth aggregation converges to minimum for large kappa", {
    v1 <- PROB_BASEVAR(Name = "X1", DistributionType = "norm", Mean = 0, Sd = 1)
    v2 <- PROB_BASEVAR(Name = "X2", DistributionType = "norm", Mean = 0, Sd = 1)

    lsf1 <- SYS_LSF(vars = list(v1))
    lsf1$func <- function(x) x[1] - 1

    lsf2 <- SYS_LSF(vars = list(v2))
    lsf2$func <- function(x) x[1] - 2

    x <- c(1.5, 3)

    g_vals <- c(
        lsf1$getLSF()(c(x[1])),
        lsf2$getLSF()(c(x[2]))
    )

    g_min <- min(g_vals)

    sys_small <- build_gsys(list(lsf1, lsf2), smooth_kappa = 5)
    sys_large <- build_gsys(list(lsf1, lsf2), smooth_kappa = 1000)

    err_small <- abs(sys_small$g_sys(x) - g_min)
    err_large <- abs(sys_large$g_sys(x) - g_min)

    expect_true(err_large < err_small)
})


test_that("smooth aggregation remains numerically stable for large kappa", {
    v1 <- PROB_BASEVAR(Name = "X1", DistributionType = "norm", Mean = 0, Sd = 1)
    v2 <- PROB_BASEVAR(Name = "X2", DistributionType = "norm", Mean = 0, Sd = 1)

    lsf1 <- SYS_LSF(vars = list(v1))
    lsf1$func <- function(x) x[1] - 1

    lsf2 <- SYS_LSF(vars = list(v2))
    lsf2$func <- function(x) x[1] - 2

    sys <- build_gsys(
        sys_input = list(lsf1, lsf2),
        sys_structure = "serial",
        smooth_kappa = 1e6
    )

    x <- c(1.5, 3)

    g <- sys$g_sys(x)

    expect_true(is.finite(g))
})


test_that("g_sys parallel approximates maximum", {
    v1 <- PROB_BASEVAR(Name = "X1", DistributionType = "norm", Mean = 0, Sd = 1)
    v2 <- PROB_BASEVAR(Name = "X2", DistributionType = "norm", Mean = 0, Sd = 1)

    lsf1 <- SYS_LSF(vars = list(v1))
    lsf1$func <- function(x) x[1] - 1

    lsf2 <- SYS_LSF(vars = list(v2))
    lsf2$func <- function(x) x[1] - 2

    sys <- build_gsys(
        sys_input = list(lsf1, lsf2),
        sys_structure = "parallel",
        smooth_kappa = 100
    )

    x <- c(1.5, 3)

    g_vals <- c(
        lsf1$getLSF()(c(x[1])),
        lsf2$getLSF()(c(x[2]))
    )

    g_max <- max(g_vals)

    g_sys <- sys$g_sys(x)

    expect_equal(g_sys, g_max, tolerance = 1e-3)
})


test_that("g_sys works with matrix input", {
    v1 <- PROB_BASEVAR(Name = "X1", DistributionType = "norm", Mean = 0, Sd = 1)

    lsf <- SYS_LSF(vars = list(v1))
    lsf$func <- function(x) x[1] - 1

    sys <- build_gsys(list(lsf))

    X <- matrix(c(0.5, 1.5, 2), ncol = 1)

    g <- sys$g_sys(X)

    expect_equal(length(g), nrow(X))
    expect_true(is.numeric(g))
})


test_that("stable smooth aggregation matches classical formulation", {
    v1 <- PROB_BASEVAR(Name = "X1", DistributionType = "norm", Mean = 0, Sd = 1)
    v2 <- PROB_BASEVAR(Name = "X2", DistributionType = "norm", Mean = 0, Sd = 1)

    lsf1 <- SYS_LSF(vars = list(v1))
    lsf1$func <- function(x) x[1] - 1

    lsf2 <- SYS_LSF(vars = list(v2))
    lsf2$func <- function(x) x[1] - 2

    sys <- build_gsys(
        sys_input = list(lsf1, lsf2),
        sys_structure = "serial",
        smooth_kappa = 20
    )

    x <- c(1.5, 3)

    g_vals <- c(
        lsf1$getLSF()(c(x[1])),
        lsf2$getLSF()(c(x[2]))
    )

    # klassische Formel
    g_classic <- -log(sum(exp(-20 * g_vals))) / 20

    g_sys <- sys$g_sys(x)

    expect_equal(g_sys, g_classic, tolerance = 1e-8)
})


test_that("MC_IS_gsys produces finite reliability estimates", {
    v1 <- PROB_BASEVAR(Name = "X1", DistributionType = "norm", Mean = 5, Sd = 1)
    v2 <- PROB_BASEVAR(Name = "X2", DistributionType = "norm", Mean = 3, Sd = 1)

    lsf1 <- SYS_LSF(vars = list(v1))
    lsf1$func <- function(x) x[1] - 4

    lsf2 <- SYS_LSF(vars = list(v2))
    lsf2$func <- function(x) x[2] - 2

    sys_input <- list(lsf1, lsf2)

    res <- MC_IS_gsys(
        sys_input = sys_input,
        sys_structure = "serial",
        smooth_kappa = 100,
        n_batch = 200,
        n_max = 2000,
        use_threads = 1,
        seed = 123
    )

    expect_true(is.finite(res$beta))
    expect_true(is.finite(res$pf))
})


test_that("MC_IS_gsys runs quickly", {
    v1 <- PROB_BASEVAR(Name = "X1", DistributionType = "norm", Mean = 5, Sd = 1)
    v2 <- PROB_BASEVAR(Name = "X2", DistributionType = "norm", Mean = 3, Sd = 1)

    lsf1 <- SYS_LSF(vars = list(v1))
    lsf1$func <- function(x) x[1] - 4

    lsf2 <- SYS_LSF(vars = list(v2))
    lsf2$func <- function(x) x[2] - 2

    sys_input <- list(lsf1, lsf2)

    t <- system.time(
        MC_IS_gsys(
            sys_input = sys_input,
            sys_structure = "serial",
            n_batch = 200,
            n_max = 2000,
            use_threads = 1,
            seed = 123
        )
    )[3]

    expect_true(t < 5)
})


test_that("MC_IS_system matches MC_CRUDE for small system", {
    skip_on_cran()

    # -----------------------------
    # Basisvariablen
    # -----------------------------
    R1 <- PROB_BASEVAR(
        Id = 1, Name = "R1",
        DistributionType = "norm",
        Mean = 10, Sd = 2
    )

    R2 <- PROB_BASEVAR(
        Id = 2, Name = "R2",
        DistributionType = "norm",
        Mean = 10, Sd = 2
    )

    S <- PROB_BASEVAR(
        Id = 3, Name = "S",
        DistributionType = "norm",
        Mean = 9, Sd = 2
    )

    # -----------------------------
    # LSFs
    # -----------------------------
    lsf1 <- SYS_LSF(vars = list(R1, S), name = "LSF1")
    lsf1$func <- function(R1, S) R1 - S
    lsf1$check()

    lsf2 <- SYS_LSF(vars = list(R2, S), name = "LSF2")
    lsf2$func <- function(R2, S) R2 - S
    lsf2$check()

    # -----------------------------
    # Maschinen
    # -----------------------------
    machine_MC_IS <- PROB_MACHINE(
        name = "MC_IS",
        fCall = "MC_IS",
        options = list(
            n_batch = 2000,
            n_max = 5e4,
            cov_user = 0.05,
            seed = 123,
            use_threads = 1
        )
    )

    # -----------------------------
    # Serial System (OR)
    # -----------------------------
    ps <- SYS_PROB(
        sys_input = list(lsf1, lsf2),
        probMachines = list(machine_MC_IS),
        sys_type = "serial"
    )

    ps$runMachines()
    ps$calculateSystemProbability("MCIS")

    pf_MCIS <- ps$res_sys[[1]]$pf

    # reference via crude Monte Carlo
    distr <- list(
        R1$getlDistr(),
        R2$getlDistr(),
        S$getlDistr()
    )

    lsf_sys <- function(x) {
        g1 <- x[1] - x[3]
        g2 <- x[2] - x[3]
        min(g1, g2)
    }

    res_mcc <- MC_CRUDE(
        lsf = lsf_sys,
        lDistr = distr,
        cov_user = 0.05,
        n_batch = 2000,
        n_max = 5e4,
        use_threads = 1,
        seed = 123
    )

    pf_MCC <- res_mcc$pf

    # -----------------------------
    # Test
    # -----------------------------
    expect_lt(abs(pf_MCIS - pf_MCC), 0.05)
})

# ======================================================================
# File: tests/testthat/test_rng_reproducibility.R
# ======================================================================
# library(testthat)
# library(TesiproV)   # das zu testende Paket

test_that("gleiche RNG-Seeds -> identische MC-Ergebnisse (reproduzierbar)", {

  ## ---------------------------------------------------------------
  ## 1) Plattform‑abhängige Thread‑Anzahl (Windows → 1, sonst 2)
  ## ---------------------------------------------------------------
  if (.Platform$OS.type == "windows") {
    future::plan(future::multisession)
    use_threads <- 2L
  } else {
    use_threads <- 2L # klein, damit der Test schnell läuft
  }

  ## ---------------------------------------------------------------
  ## 2) Basis‑Variablen (ohne DistributionPackage)
  ## ---------------------------------------------------------------
  var1 <- PROB_BASEVAR(
    Name             = "E",
    Description      = "Effect",
    DistributionType = "norm",
    Mean             = 10,
    Sd               = 0.2
  )
  var2 <- PROB_BASEVAR(
    Name             = "R",
    Description      = "Resistance",
    DistributionType = "norm",
    Mean             = 35,
    Sd               = 0.5
  )

  ## ---------------------------------------------------------------
  ## 3) Limit‑State‑Funktion
  ## ---------------------------------------------------------------
  lsf <- SYS_LSF(vars = list(var1, var2), name = "LSF XY")
  lsf$func <- function(E, R) {
    R - E^1.5
  }
  lsf$check() # stellt sicher, dass alles konsistent ist

  ## ---------------------------------------------------------------
  ## 4) Hilfsfunktion: erstellt ein SYS_PROB‑Objekt, führt die
  ##    Maschinen aus und gibt die Ergebnis‑Liste zurück.
  ## ---------------------------------------------------------------
  run_one <- function(seed_val) {
    ## – FORM – (nur für den Vergleich, kein Einfluss auf RNG)
    machine_form <- PROB_MACHINE(
      name = "FORM Rack.-Fieß.",
      fCall = "FORM",
      options = list(
        n_optim = 20,
        loctol = 0.001,
        optim_type = "rackfies"
      )
    )

    ## – MC‑CRUDE – das ist die Maschine, deren RNG wir testen
    machine_mc <- PROB_MACHINE(
      name = "MC CoV 0.05",
      fCall = "MC_CRUDE",
      options = list(
        n_max = 5e5, # genug, damit die Simulation fertig wird
        cov_user = 0.05,
        use_threads = use_threads,
        seed = seed_val # <-- entscheidender Seed
      )
    )

    ## – SORM – (nur für Vollständigkeit)
    machine_sorm <- PROB_MACHINE(
      name   = "SORM",
      fCall  = "SORM"
    )

    ## Gesamtsystem zusammenbauen
    prob_system <- SYS_PROB(
      sys_input    = list(lsf),
      probMachines = list(machine_form, machine_mc, machine_sorm),
      debug.level  = 0
    )

    ## Maschinen ausführen → Ergebnis‑Liste (benannt nach den Maschinen)
    prob_system$runMachines()
  }

  ## ---------------------------------------------------------------
  ## 5) Zwei unabhängige Läufe mit exakt demselben Seed
  ## ---------------------------------------------------------------
  set.seed(99) # Master‑Seed – hat keinen Einfluss,
  # weil jede Maschine ihren eigenen Seed nutzt
  res_a <- run_one(seed_val = 1234) # erster Durchlauf
  res_b <- run_one(seed_val = 1234) # zweiter Durchlauf (gleicher Seed)

  ## ---------------------------------------------------------------
  ## 6) Ergebnis‑Auswertung
  ## ---------------------------------------------------------------
  # Wir prüfen nur die MC‑Maschine (der Rest ist nicht von RNG‑Repro-
  # duzierbarkeit betroffen).  Das Ergebnis‑Objekt hat das Element $pf.
  pf_a <- res_a[["MC CoV 0.05"]][["pf"]]
  pf_b <- res_b[["MC CoV 0.05"]][["pf"]]

  # Beide Schätzungen müssen exakt übereinstimmen (bis auf Rundungs‑Fehler
  # des R‑Doubles).  Ein sehr kleine Toleranz von 1e‑14 ist sicher.
  expect_equal(pf_a, pf_b,
    tolerance = 1e-14,
    info = "Mit identischem Seed und gleicher Thread‑Zahl muss das MC‑Ergebnis reproduzierbar sein"
  )
})


# ------------------------------------------------------------------
# 0) Hilfsfunktion – Minimal‑Beispiel (ohne DistributionPackage)
# ------------------------------------------------------------------
make_demo_objects <- function() {
  var1 <- PROB_BASEVAR(
    Name             = "E",
    Description      = "Effect",
    DistributionType = "norm",
    Mean             = 10,
    Sd               = 0.2
  )
  var2 <- PROB_BASEVAR(
    Name             = "R",
    Description      = "Resistance",
    DistributionType = "norm",
    Mean             = 35,
    Sd               = 0.5
  )

  lsf_obj <- SYS_LSF(vars = list(var1, var2), name = "LSF XY")
  lsf_obj$func <- function(E, R) {
    R - E^1.5
  }
  lsf_obj$check()

  lsf_fun <- function(x) lsf_obj$func(x[1], x[2])
  lDistr_list <- list(list(var1), list(var2))
  dp <- c(var1$Mean, var2$Mean)

  list(
    lsf_fun = lsf_fun,
    lDistr = lDistr_list,
    dp = dp,
    lsf_obj = lsf_obj
  )
}

# ------------------------------------------------------------------
# 2) Ungültiger densityType → erklärender Fehler
# ------------------------------------------------------------------
test_that("invalid densityType erzeugt einen erklärenden Fehler", {
  demo <- make_demo_objects()
  if (.Platform$OS.type == "windows") {
    backend <- "future"
  } else {
    backend <- "parallel"
  }

  expect_error(
    MC_IS(
      lsf = demo$lsf_fun,
      lDistr = demo$lDistr,
      dps = demo$dp,
      cov_user = 0.10,
      n_batch = 1000,
      n_max = 1e5,
      use_threads = 2,
      backend = backend,
      densityType = "not_a_distribution", # <‑ falscher Typ
      seed = 111,
      dataRecord = FALSE
    ),
    regexp = "unsupported|unknown|invalid|densityType",
    info = "Ein unbekannter densityType muss mit einer klaren Fehlermeldung abbrechen"
  )
})


# ------------------------------------------------------------------
# 3) Edge‑Case n_batch = 0 -> sinnvolle Fehlermeldung
# ------------------------------------------------------------------
test_that("n_batch = 0 führt zu einem Fehler", {
  demo <- make_demo_objects()
  if (.Platform$OS.type == "windows") {
    backend <- "future"
  } else {
    backend <- "parallel"
  }

  expect_error(
    MC_IS(
      lsf = demo$lsf_fun,
      lDistr = demo$lDistr,
      dps = demo$dp,
      cov_user = 0.10,
      n_batch = 0,
      n_max = 1e5,
      use_threads = 2,
      backend = backend,
      seed = 111,
      dataRecord = FALSE
    ),
    regexp = "n_batch.*positive|must be > 0"
  )
})

library(testthat)
library(microbenchmark)

testthat::local_edition(3)


test_that("Spaethe (1991) - Example 2", {

  Force <- PROB_BASEVAR(
    Name = "Force", DistributionType = "gumbel", Package = "evd", Mean = 18, Sd = 2)

  # yield strength in kN
  fy <- PROB_BASEVAR(
    Name = "fy", DistributionType = "slnorm", Package = "brms", Mean = 265000, Sd = 25000, x0 = 160000)

  # plastic moment of resistance in m³ - considered deterministic in this example
  W_pl <- 2.5e-4
  # length of cantilever beam in m - considered deterministic in this example
  l    <- 2

  lsf_param <- SYS_LSF(vars = list(Force, fy), name = "Spaethe (1992) - Beispiel 2")
  lsf_param$func <- function(Force,fy) {
    return(W_pl*fy-l*Force)
  }
  lsf_param$check()

  machine_form <- PROB_MACHINE(name = "FORM", fCall = "FORM")
  machine_sorm <- PROB_MACHINE(name = "SORM", fCall = "SORM")
  machine_MCIS <- PROB_MACHINE("MCIS", "MC_IS",
                               options = list(
                                 cov_user = 0.025,
                                 n_max = 200000, # statt Millionen
                                 n_batch = 20000,
                                 use_threads = 1L,
                                 backend = "future",
                                 seed = 1234
                               ))
  machine_SubSam <- PROB_MACHINE("MC_SubSam", "MC_SubSam")
  machine_MCCrude <- PROB_MACHINE("MC_Crude", "MC_CRUDE")


  ps_param <- SYS_PROB(
    sys_input = list(lsf_param),
    probMachines = list(machine_form,machine_sorm,machine_MCIS,machine_SubSam,machine_MCCrude)
  )

  ps_param$runMachines()

  pf_expected   <- 1.398e-4
  beta_expected <- -qnorm(pf_expected)

  rel_error <- c(1,2,3,4,5)
  for (i in rel_error) {
    rel_error[i] <-abs(ps_param$beta_single[i] - beta_expected) / beta_expected
    expect_lt(rel_error[i], 0.025) # Abweichung 2.5% erlaubt
  }
})


test_that("lnorm with x0 reduces to standard case when x0 = 0", {
  fy1 <- PROB_BASEVAR(Name="fy", DistributionType="slnorm", Package = "brms", Mean=100, Sd=10, x0=0)
  fy2 <- PROB_BASEVAR(Name="fy", DistributionType="lnorm", Mean=100, Sd=10)

  expect_equal(fy1$DistributionParameters[1], fy2$DistributionParameters[1])
  expect_equal(fy1$DistributionParameters[2], fy2$DistributionParameters[2])
  expect_equal(fy1$DistributionParameters[3], 0)
})

# tests/testthat/test-seed.R -----------------------------------------

#library(testthat)

# Load the development version of the package (only needed when the
# tests are executed in the source tree, not when the package is
# installed).
# if (requireNamespace("devtools", quietly = TRUE)) {
#   devtools::load_all("..")    # path to the package root
# }

# --------------------------------------------------------------
# Simple 3-RV limit-state function (the one you already use)
# --------------------------------------------------------------
make_demo_lsf <- function() {
  X1 <- PROB_BASEVAR(Id = 1, Name = "Z",
                     DistributionType = "norm", Mean = 100, Cov = 0.04)
  X2 <- PROB_BASEVAR(Id = 2, Name = "Fy",
                     DistributionType = "lnorm", Mean = 40, Cov = 0.10)
  X3 <- PROB_BASEVAR(Id = 3, Name = "M",
                     DistributionType = "gumbel", Package = "evd",
                     Mean = 2000, Cov = 0.10)

  lsf <- SYS_LSF(vars = list(X1, X2, X3),
                 name = "Demo LSF")
  lsf$func <- function(M, Fy, Z) Z * Fy - M
  lsf$check()
  lsf
}

# --------------------------------------------------------------
# Wrapper that starts one MC-IS run with the *parallel* backend.
# --------------------------------------------------------------
run_one_mc_is <- function(use_threads = 2,
                          seed = NULL,
                          n_batch = 2e3,
                          n_max   = 5e4,
                          backend = "parallel") {

  # 1) build the machine (no fixed seed if seed == NULL)
  machine_mc_is <- PROB_MACHINE(
    name    = "MC IS (test)",
    fCall   = "MC_IS",
    options = list(
      n_max       = n_max,
      cov_user    = 0.02,
      use_threads = use_threads,
      n_batch     = n_batch,
      backend     = backend,
      seed        = seed,          # <- the argument we are testing
      dataRecord  = FALSE,
      densityType = "norm"
    )
  )

  # 2) build the SYS_PROB object that holds everything
  ps <- SYS_PROB(
    sys_input    = list(make_demo_lsf()),
    probMachines = list(machine_mc_is),
    debug.level  = 0
  )

  # 3) run the machines and return the *single* beta value
  ps$runMachines()
  # `beta_single` is a matrix, we pick the first (and only) entry
  as.numeric(ps$beta_single)
}



# -----------------------------------------------------------------
# 1)  Random seed (seed = NULL) - beta values must *vary*
# -----------------------------------------------------------------
test_that("seed = NULL gives non‑identical beta values", {
  if (.Platform$OS.type == "windows") {
    backend <- "future"
  } else {
    backend <- "parallel"
  }
  # Run the simulation twice with the SAME arguments but with seed = NULL
  beta1 <- run_one_mc_is(seed = NULL, backend = backend)
  beta2 <- run_one_mc_is(seed = NULL, backend = backend)

  # The two results must be *different* (within a reasonable tolerance)
  # – because a new master RNG stream is created for every call.
  expect_false(isTRUE(all.equal(beta1, beta2, tolerance = .Machine$double.eps)))

  # As an additional sanity check we run the simulation a few more times
  # and verify that the empirical variance is > 0.
  betas <- replicate(5, run_one_mc_is(seed = NULL, backend = backend ))
  expect_gt(var(betas), 0)
})

# -----------------------------------------------------------------
# 2)  Fixed seed (seed = 1234) - beta values must be reproducible
# -----------------------------------------------------------------
test_that("fixed seed yields identical beta values", {
  if (.Platform$OS.type == "windows") {
    backend <- "future"
  } else {
    backend <- "parallel"
  }
  fixed_seed <- 1234L

  # Two independent calls with the *same* fixed seed
  beta_a <- run_one_mc_is(seed = fixed_seed, backend = backend)
  beta_b <- run_one_mc_is(seed = fixed_seed, backend = backend)

  # They have to be exactly the same (up to machine precision)
  expect_equal(beta_a, beta_b, tolerance = .Machine$double.eps)

  # And a third call should also match – this catches accidental
  # side‑effects (e.g. the seed being consumed somewhere else).
  beta_c <- run_one_mc_is(seed = fixed_seed, backend = backend)
  expect_equal(beta_a, beta_c, tolerance = .Machine$double.eps)
})

library(testthat)
library(microbenchmark)

testthat::local_edition(3)

# Das Beispiel ist zu umfangreich für die testthat()-Umgebung

# -----------------------------------------------------------------
# Example Ex.5.11 from Nowak & Collins (2000): reliability analysis with student-t and log-student-t
# -----------------------------------------------------------------
test_that("Ex. 5.11 Nowak & Collins (2000)", {
  testthat::local_edition(3)
  # Platform-dependent thread count
  if (.Platform$OS.type == "windows") {
    backend <- "future"
  } else {
    backend <- "parallel"
  }

  # Base variables ------------------------------------------------
  X1 <- PROB_BASEVAR(
    Id = 1, Name = "Z",
    DistributionType = "st",
    DistributionParameters = c(100,4,1000000000,1000000000)
  ) # kN/m²


  sig_fy <- sqrt(log(1+0.1^2))
  mue_fy <- log(40) - sig_fy^2/2
  # data_fy <- rlnorm(100000,mue_fy,sig_fy)
  # X2 <- PROB_BASEVAR(
  #     Id = 2, Name = "Fy",
  #     DistributionType = "emp",
  #     Package = "EnvStats",
  #     DistributionParameters = data_fy)
      #Mean = 40, Cov = 0.1
  X2 <- PROB_BASEVAR(
    Id = 2, Name = "Fy",
    DistributionType = "lt",
    DistributionParameters = c(mue_fy,sig_fy,1000000000,1000000000)
  ) # m
  X3 <- PROB_BASEVAR(
    Id = 3, Name = "M",
    DistributionType = "gumbel",
    Package = "evd",
    Mean = 2000, Cov = 0.1
  ) # m

  # Limit State Function -------------------------------------------
  lsf1 <- SYS_LSF(
    vars = list(X1, X2, X3),
    name = "Nowak & Collins Exp5.11_1 (beta=4.03)"
  )

   lsf1$func <- function(M, Fy, Z) Z * Fy - M

   lsf1$check()

  # Machines -------------------------------------------------------
  form_rf <- PROB_MACHINE(
    name = "FORM Rack.-Fiess.",
    fCall = "FORM",
    options = list(
      n_optim = 20,
      loctol = 1e-4,
      optim_type = "rackfies"
    )
  )

   sorm <- PROB_MACHINE(name = "SORM", fCall = "SORM")

  mcis <- PROB_MACHINE(
    name = "MC IS",
    fCall = "MC_IS",
    options = list(
      cov_user = 0.02,
      n_max = 1e7,
      n_batch = 100000,
      use_threads = 1L,
      seed = 1234,
      backend = backend
    )
  )

  mcc <- PROB_MACHINE(
    name = "MCC",
    fCall = "MC_CRUDE",
    options = list(
      cov_user = 0.05,
      n_max = 1e8,
      use_threads = 1L,
      seed = 161268,
      backend = backend
    )
  )

  mcsus <- PROB_MACHINE(
    name = "MC SuS",
    fCall = "MC_SubSam",
    options = list(seed = 1234)
  )

  # System ---------------------------------------------------------
  ps <- SYS_PROB(
    sys_input = list(lsf1),
    probMachines = list(form_rf, sorm, mcis, mcc, mcsus),
    debug.level = 0
  )

   ps$runMachines()

  # # --------------------------------------------------------------
  # # Basic plausibility checks
  # # --------------------------------------------------------------
  expect_length(ps$beta_single, 5)
  expect_false(any(is.na(ps$beta_single)))

  ## FORM vs SORM – nearly identical results -----------------------
  rel_diff_form_sorm <- abs(ps$beta_single[1] - ps$beta_single[2]) /
    max(abs(ps$beta_single[1]), 1e-12)

  message("FORM and SORM should be identical for linear LSFs")
  expect_lt(rel_diff_form_sorm, 1e-8)

  ## Literature reference values -----------------------------------
  message("FORM beta must match literature value within 1e-8")
  expect_equal(ps$beta_single[1],
               4.02211455568222,
               tolerance = 1e-8
  )

  message("SORM beta must match literature value within 1e-8")
  expect_equal(ps$beta_single[2],
               4.02211455568236,
               tolerance = 1e-8
  )

  # MC‑IS - maximum 2% relative deviation allowed
  rel_err_1 <- abs(ps$beta_single[3] - 4.03) / 4.03
  message("MC-IS must not deviate more than 2 % from the reference beta")
  expect_lt(rel_err_1, 0.021)

  ## MC‑CRUDE - maximum 2% relative deviation allowed
  ref_mc_crude <- 4.02211455568236 # 3.9160810558165
  tol_mc_crude <- 0.02 #  ≈ 2%

   rel_err_2 <- abs(ps$beta_single[4] - ref_mc_crude) / ref_mc_crude

  message(sprintf(
    "MC-CRUDE (%s) must lie within %.1f%% tolerance range",
    .Platform$OS.type, tol_mc_crude * 100
  ))

  expect_lt(rel_err_2, tol_mc_crude)
  info=sprintf("MC-CRUDE (%s) must lie within %.1f%% tolerance range",
             .Platform$OS.type,tol_mc_crude*100)

  ## MC‑SuS – exact literature value -------------------------------
  message("MC-SuS beta must match literature value within 1e-8")
  expect_equal(ps$beta_single[5],
               4.01123480350189,
               tolerance = 1e-3)

})

#' @title Internal helper functions for Monte Carlo Importance Sampling (MC_IS)
#' @description
#' This file contains internal helper functions used by the main function
#' [`MC_IS()`] to perform Monte Carlo simulations with importance sampling.
#'
#' The helpers defined here support:
#' * random number generation and cluster setup,
#' * creation of recorder objects for iterative logging,
#' * single limit state simulation routine (`MC_IS_single()`),
#' * conversion of recorded data to data frames after completion.
#'
#' Additional helper sets will be provided for system calculations
#' (serial and parallel systems) in separate files.
#'
#' @details
#' These functions are not intended to be called directly by users.
#' They are exported internally to allow modular access within the package.
#'
#' @seealso [TesiproV::MC_IS()] for the public interface that calls these helpers.
#'
#' @author (C) 2021-2026 K. Nille-Hauf, T. Feiri, M. Ricker, T. Lux -- Hochschule Biberach (until 2022),
#' TU Dortmund University - Chair of Structural Concrete (since 2023)
#' @name Internal helper functions for MC_IS()
NULL


## ---------------------------------------------------------------
## Random number generator utilities
## ---------------------------------------------------------------

#' @title Initialize master RNG stream
#' @description Sets RNG type `"L'Ecuyer-CMRG"` and optionally a fixed seed.
#'
#' @param seed Optional integer seed; if `NULL`, a random base seed is generated once per call.
#' @param debug.level Integer controlling verbosity level.
#'   Values >= 1 produce diagnostic console output.
#' @return A master RNG stream object (numeric vector of length 7)
#'         as required by [parallel::clusterSetRNGStream()] or usable
#'         directly in [future.apply::future_lapply()].
#'
#' @keywords internal
init_rng_master <- function(seed = NULL, debug.level = 0) {
  RNGkind("L'Ecuyer-CMRG")

  if (!is.null(seed)) {
    # fixed user seed -> reproducible
    set.seed(as.integer(seed))
    if (debug.level >= 0L) message(sprintf("[init_rng_master] User seed set to %d", seed))
  } else {
    base_seed <- sample.int(.Machine$integer.max, 1L)
    set.seed(base_seed)
    if (debug.level >= 0L) message(sprintf("[init_rng_master] Random RNG seed drawn: %d", base_seed))
  }

  # produce full L'Ecuyer-CMRG stream and return it
  stream <- parallel::nextRNGStream(.Random.seed) # Vector length 7
  if (debug.level >= 1L) {
    message(sprintf("[init_rng_master] L'Ecuyer stream head: %d (use this for cluster RNG)", as.integer(stream[1])))
  }
  return(stream)
}

#' @title Set cluster RNG streams
#' @description Assigns identical RNG streams to each worker in an existing cluster.
#'
#' @param cl Cluster object created by `parallel::makeCluster()`.
#' @param seed Integer base seed used on the master process.
#'
#' @keywords internal
set_cluster_rng <- function(cl, seed) {
  ## Es muss kein neuer Seed erzeugt werden - wir benutzen denselben
  ## Basis-Seed, den `init_rng_master()` bereits gesetzt hat.
  parallel::clusterSetRNGStream(cl, iseed = seed)
}


## ---------------------------------------------------------------
## Recording / logging utilities for iterative MC loops
## ---------------------------------------------------------------

#' @title Create empty recorder vectors for MC iterations
#' @description Preallocates numeric vectors and time tracking data frames used to record intermediate results during simulation loops.
#'
#' @param n_iter Expected maximum number of iterations/batches in the simulation loop.
#'
#' @return A list containing numeric vectors (`pf`, `var`, `cov`, etc.) and a nested time data frame with user/system/elapsed times per iteration.
#'
#' @keywords internal
make_recorders <- function(n_iter) {
  list(
    pf = numeric(n_iter),
    var = numeric(n_iter),
    cov = numeric(n_iter),
    I_sum = numeric(n_iter),
    n_sim = numeric(n_iter),
    time = data.frame(
      user = numeric(n_iter),
      sys = numeric(n_iter),
      elapsed = numeric(n_iter),
      child_user = numeric(n_iter),
      child_sys = numeric(n_iter)
    )
  )
}


#' @title Internal Parallel Dispatcher for Importance Sampling
#'
#' @description
#' Internal helper function used by \code{MC_IS_single()} to distribute
#' Monte Carlo sample blocks across worker threads in a reproducible and
#' backend-independent manner.
#'
#' The function ensures worker-invariant execution by:
#' \itemize{
#'   \item Performing deterministic chunking of sample indices,
#'   \item Avoiding per-worker random number generation,
#'   \item Supporting both \code{"future"} and \code{"parallel"} backends.
#' }
#'
#' This abstraction isolates parallel execution from the statistical
#' core of the Importance Sampling algorithm.
#'
#' @param chunk_indices List of integer vectors defining sample indices
#'   assigned to each worker.
#'
#' @param worker_fun Function that evaluates a given subset of samples.
#'
#' @param u_mat Matrix of samples in standard normal space.
#'
#' @param use_threads Number of worker threads.
#'
#' @param backend Character string specifying backend
#'   (\code{"future"} or \code{"parallel"}).
#'
#' @param debug.level Verbosity level for diagnostic output.
#'
#' @return A list containing worker evaluation results.
#'
#' @keywords internal
parallel_dispatch <- function(chunk_indices,
                              worker_fun,
                              u_mat,
                              use_threads,
                              backend,
                              debug.level = 0) {
  # --------------------------------------------------
  # Single-thread fallback
  # --------------------------------------------------
  if (use_threads == 1) {
    if (debug.level >= 2) {
      message("[MC_IS_system] Using lapply (single thread).")
    }

    return(lapply(chunk_indices, worker_fun, u_mat))
  }

  # --------------------------------------------------
  # Unix-like systems (Linux/macOS)
  # --------------------------------------------------
  if (.Platform$OS.type != "windows") {
    if (debug.level >= 2) {
      message("[MC_IS_system] Using mclapply (fork).")
    }

    return(
      parallel::mclapply(
        chunk_indices,
        worker_fun,
        u_mat,
        mc.cores = use_threads
      )
    )
  }

  # --------------------------------------------------
  # Windows fallback -> future
  # --------------------------------------------------
  if (debug.level >= 2) {
    message("[MC_IS_system] Using future_lapply (Windows fallback).")
  }

  if (!requireNamespace("future.apply", quietly = TRUE)) {
    stop("Package 'future.apply' required for Windows parallel execution.")
  }

  return(
    future.apply::future_lapply(
      chunk_indices,
      worker_fun,
      u_mat,
      future.seed = FALSE,
      future.globals = FALSE
    )
  )
}


#' @title Internal Monte Carlo Importance Sampling Routine (Single Limit State Function)
#'
#' @description
#' Performs the core Monte Carlo simulation with importance sampling for a single limit state function.
#' This internal routine is called by the public interface [`MC_IS()`] when the input argument `lsf`
#' is a single function rather than a list of system functions.
#'
#' The algorithm uses the FORM design point as center of the sampling density in standard normal space.
#' Each random variable is transformed between physical and normal spaces using its distribution object
#' (`PROB_BASEVAR$getlDistr()`), which provides parameterized CDF and inverse-CDF functions:
#'
#' \deqn{u = \Phi^{-1}(F_X(x))}
#' \deqn{x = F_X^{-1}(\Phi(u))}
#'
#' Parallel execution is supported via either base R's **parallel** package or the **future** backend.
#'
#' @param lsf Objective (limit state) function in form \code{function(x) \{ ... \}}.
#' @param lDistr List of input variable distributions created from TesiproV base variable objects.
#' @param cov_user Target coefficient of variation for convergence criterion.
#' @param n_batch Number of samples per batch (recommended > 1000).
#' @param n_max Maximum total number of samples before forced stop.
#' @param use_threads Number of threads/workers to use for parallel computation.
#' @param dataRecord Logical; if `TRUE`, all intermediate results are recorded and returned as data frame.
#' @param densityType Character; type of sampling density ("origin" keeps original input distributions).
#' @param dps Optional vector of design points supplied by user instead of FORM result.
#' @param debug.level Integer verbosity level: 0 = silent, 1 = info, 2 = detailed output during calculation.
#' @param streams RNG stream object created by [`init_rng_master()`].
#' @param libPaths_local Local library paths passed to cluster workers for package availability.
#' @param seed Optional integer seed for reproducibility across parallel runs. If `NULL`, a random seed is used.
#' @param backend Character; either `"parallel"` or `"future"` backend for parallelization on multiple cores.
#' @param adaptive_alpha Logical; if \code{TRUE}, adaptive mixture weights are used
#'   (system case only).
#'
#' @param alpha_update_rate Damping factor \eqn{\lambda \in [0,1]} controlling
#'   adaptation speed of mixture weights.
#'
#' @param adaptive_batch Logical; if TRUE, the batch size is adapted dynamically
#'   based on the current effective sample size (ESS) and target RSE.
#'
#' @param batch_control List controlling adaptive batch behaviour:
#'   \describe{
#'     \item{n_min}{Minimum batch size}
#'     \item{n_max}{Maximum batch size}
#'     \item{K_future}{Desired number of remaining iterations}
#'   }
#'
#' @param min_adapt_samples Integer. Minimum total number of MC samples required before adaptive mixtures are updated.
#'    The adaption mechanism is activated only after this threshold has been exceeded to ensure statistically stable estimates.
#'
#' @param alpha_min Numeric scalar in \eqn{(0,1)}.
#'   Lower bound for mixture weights \eqn{\alpha_i} during adaptive
#'   updating. This prevents individual components of the importance
#'   sampling density from collapsing to zero and ensures numerical
#'   stability in multimodal system analyses.
#'
#' @param stability_mode Character string specifying the numerical
#'   stabilization strategy used for importance sampling weight
#'   accumulation.
#'
#'   \describe{
#'     \item{"robust"}{
#'       Default. Uses full log-domain accumulation of weights
#'       across Monte Carlo iterations. This improves numerical
#'       stability for high-dimensional systems and very small
#'       failure probabilities.
#'     }
#'     \item{"fast"}{
#'       Uses shifted exponentiation with batch-level scaling.
#'       Computationally equivalent under normal engineering
#'       conditions but may be slightly less stable in extreme
#'       rare-event scenarios.
#'     }
#'   }
#'
#'   Both modes are mathematically equivalent for typical
#'   reliability problems. The \code{"robust"} mode is recommended
#'   for research applications and extreme reliability levels.
#'
#' @keywords internal
#'
MC_IS_single <- function(lsf, lDistr,
                         cov_user,
                         n_batch,
                         n_max,
                         use_threads,
                         backend = "future",
                         dataRecord,
                         densityType = "norm",
                         dps = NULL,
                         debug.level,
                         streams,
                         libPaths_local,
                         seed = NULL,
                         adaptive_alpha = FALSE,
                         alpha_update_rate = 0.1,
                         adaptive_batch = FALSE,
                         batch_control = list(),
                         min_adapt_samples = 10000,
                         alpha_min = 0.02,
                         stability_mode) {
  if (debug.level >= 1) {
    cat("DEBUG LEVEL INSIDE MC_IS_single:", debug.level, "\n")
  }

  ## --- Input validation ------------------------------------------------------
  if (missing(debug.level) || is.null(debug.level)) {
    debug.level <- 0
  }
  if (!is.function(lsf)) {
    stop("MC_IS_single: 'lsf' must be a function.", call. = FALSE)
  }
  if (!is.list(lDistr) || length(lDistr) < 1) {
    stop("MC_IS_single: 'lDistr' must be a list of marginals.", call. = FALSE)
  }

  if (!is.numeric(n_batch) || length(n_batch) != 1L || n_batch <= 0L) {
    stop("`n_batch` must be a positive integer (>= 1).")
  }

  if (!is.numeric(n_max) || length(n_max) != 1L || n_max <= 0L) {
    stop("`n_max` must be a positive integer.", call. = FALSE)
  }
  if (!is.numeric(use_threads) || length(use_threads) != 1L || use_threads < 1L) {
    stop("`use_threads` must be an integer >= 1.", call. = FALSE)
  }

  debug.TAG <- "MC_IS_single"
  if (debug.level >= 1) message("[MC_IS_single] Starting single-LSF MC_IS...")

  ## -------------------------------------------------
  ## Deterministic RNG manager (thread-independent)
  ## -------------------------------------------------
  rng_mgr <- create_rng_manager(seed)

  ## -------------------------------------------------
  ## Persistent cluster setup (parallel backend)
  ## -------------------------------------------------
  cl <- NULL

  if (backend == "parallel" && use_threads > 1L) {
    cl <- make_parallel_cluster(
      use_threads = use_threads,
      seed = seed
    )

    on.exit(parallel::stopCluster(cl), add = TRUE)
  }

  # ## Master RNG / seed handling
  # RNGkind("L'Ecuyer-CMRG")

  # # Normalise seed for future.apply and cluster
  # seed_for_future <- NULL # integer or 7-length stream or NULL
  # streams_master <- NULL # 7-length stream (for parallel::clusterSetRNGStream)

  # if (!is.null(seed)) {
  #   if (is.numeric(seed) && length(seed) == 1L && !is.na(seed)) {
  #     # integer seed -> reproducible for future.apply
  #     set.seed(as.integer(seed))
  #     # derive a full CMRG stream for cluster backend if needed
  #     streams_master <- parallel::nextRNGStream(.Random.seed)
  #     seed_for_future <- as.integer(seed)
  #     if (debug.level >= 1) message(sprintf("[MC_IS_single] Fixed master seed set to %d", as.integer(seed)))
  #   } else if (is.numeric(seed) && length(seed) == 7L) {
  #     # user supplied full L'Ecuyer stream
  #     # assign(".Random.seed", seed, envir = .GlobalEnv)
  #     set.seed(seed)
  #     streams_master <- seed
  #     seed_for_future <- seed
  #     if (debug.level >= 1) message("[MC_IS_single] Using supplied L'Ecuyer-CMRG stream.")
  #   } else {
  #     stop("Seed must be a single integer or a L'Ecuyer-CMRG stream (length 7).", call. = FALSE)
  #   }
  # } else {
  #   # no fixed seed -> random behaviour; for future.apply use future.seed = TRUE (resolve_future_seed(NULL) -> TRUE)
  #   streams_master <- parallel::nextRNGStream(.Random.seed)
  #   seed_for_future <- NULL
  #   if (debug.level >= 1) message("[MC_IS_single] No fixed seed supplied; using random streams.")
  # }

  ## Gueltige density-Typen - origin bedeutet: die originale Eingabeverteilung
  ## (aus dem Paket, das im PROB_BASEVAR definiert ist) wird verwendet.
  allowed_density <- c(
    "norm", "lnorm", "weibull", "gamma", "beta",
    "exponential", "cauchy", "logis", "origin"
  )

  if (!densityType %in% allowed_density) {
    stop(
      sprintf(
        "`densityType = \"%s\"` is not supported. Supported types are: %s. ",
        densityType,
        paste(allowed_density[allowed_density != "origin"], collapse = ", ")
      ),
      "Use \"origin\" to keep the original input distribution."
    )
  }


  tic <- proc.time()

  n_vars <- length(lDistr)
  total_batch <- n_batch # use_threads * n_batch

  ### cat("total_batch =", total_batch, "\n")

  if (n_vars < 1) stop("[MC_IS_single] lDistr must contain at least one marginal.", call. = FALSE)
  if (n_batch > n_max) n_batch <- n_max

  # ---------------------------------------------------------------------------
  # Design Point calculation using FORM
  # ---------------------------------------------------------------------------

  if (is.numeric(dps)) {
    dp <- as.numeric(dps)
    if (length(dp) == 1L && n_vars > 1L) dp <- rep(dp, n_vars)
    if (length(dp) != n_vars) stop("[MC_IS_single] dps must have length 1 or n_vars", call. = FALSE)
    res_form <- list(beta = NA_real_, pf = NA_real_, x_points = dp)
    if (debug.level >= 1) message("[MC_IS_single] Using user-supplied design points")
  } else {
    # If FORM expects the wrapper (funlist + params), build it
    # assume lDistr_funlist is list of funlists (with $p,$q,...)
    # --- Ensure we have funlists (one per marginal) -------------------------
    # lDistr_funlist <- result of normalization
    lDistr_funlist <- lapply(lDistr, function(dd) {
      # dd kann funlist oder wrapper(list(funlist, params)) sein
      if (is.list(dd) && !is.null(dd$p) && is.function(dd$p)) {
        return(dd)
      } # already funlist
      if (is.list(dd) && length(dd) >= 1 && is.list(dd[[1]]) && !is.null(dd[[1]]$p)) {
        return(dd[[1]])
      } # wrapper -> funlist
      # fallback: search for element that is a funlist
      if (is.list(dd)) {
        for (el in dd) {
          if (is.list(el) && !is.null(el$p) && is.function(el$p)) {
            return(el)
          }
        }
      }
      stop("[MC_IS_single] Cannot extract funlist from lDistr element", call. = FALSE)
    })

    # --- Build wrapper list expected by FORM: list(funlist, params) ----------
    # lDistr_for_FORM <- wrapper for FORM
    lDistr_for_FORM <- lapply(lDistr_funlist, function(fun) {
      params <- c(NA_real_, NA_real_)
      if (!is.null(fun$mean)) params[1] <- as.numeric(fun$mean)
      if (!is.null(fun$Sd)) params[2] <- as.numeric(fun$Sd)
      list(fun, params)
    })

    # Now call FORM with wrapper:
    # res_form <- TesiproV::FORM(lsf, lDistr_for_FORM, n_optim = 20, loctol = 1e-4)

    # -------------------------------------------------
    # Check if LSF is valid at starting point
    # -------------------------------------------------

    x_test <- sapply(lDistr_for_FORM, function(d) d[[1]]$mean)

    g_test <- try(lsf(x_test), silent = TRUE)

    if (inherits(g_test, "try-error") || !is.finite(g_test)) {
      if (debug.level >= 1) {
        message("[MC_IS_single] FORM skipped: invalid LSF start value")
      }

      res_form <- list(
        beta = NA_real_,
        pf = NA_real_,
        x_points = x_test
      )
    } else {
      res_form <- TesiproV::FORM(
        lsf,
        lDistr_for_FORM,
        n_optim = 20,
        loctol = 1e-4
      )
    }

    if (is.null(res_form$x_points) || length(res_form$x_points) == 0L) {
      stop("[MC_IS_single] FORM did not return valid design point x_points.", call. = FALSE)
    }
    # res_form <- TesiproV::FORM(lsf, lDistr, n_optim = 20, loctol = 1e-4)
    # after building lDistr_funlist and lDistr_for_FORM and calling FORM
    # use funlist for MC:
    lDistr <- lDistr_funlist

    # Debug-Ausgabe zur Kontrolle der Struktur
    if (debug.level >= 1) {
      message("[MC_IS_single] Structure of FORM result:")
      str(res_form)
      message(sprintf(
        "Designpoint (x-space): %s",
        paste(round(as.numeric(unlist(res_form$x_points)), 3), collapse = ", ")
      ))
      message(sprintf("[MC_IS_single] Beta from FORM: %.3f", res_form$beta))
    }

    # transform FORM x* to u*
    x_vec <- as.numeric(unlist(res_form$x_points))
    if (length(x_vec) == 0L || any(is.na(x_vec))) {
      stop("[MC_IS_single] FORM returned no valid design point (x_points).", call. = FALSE)
    }
    # ensure length
    if (length(x_vec) < n_vars) {
      warning("FORM returned fewer x* components than variables; padding with first value.")
      x_vec <- rep(x_vec[1], n_vars)
    }
    dp <- numeric(n_vars)
    for (i in seq_len(n_vars)) {
      x_star_i <- as.numeric(x_vec[i])
      Fxi <- tryCatch(lDistr[[i]]$p(x_star_i), error = function(e) NA_real_)
      Fxi <- min(max(as.numeric(Fxi), 1e-14), 1 - 1e-14)
      dp[i] <- qnorm(Fxi)
      if (!is.function(lDistr[[i]]$p) || !is.function(lDistr[[i]]$q)) stop("[MC_IS_single] marginal missing p/q", call. = FALSE)
    }
    # check marginals or missing p or q
  }

  if (!is.numeric(dp) || length(dp) != n_vars) {
    stop("[MC_IS_single] 'dp' must be numeric vector of length n_vars", call. = FALSE)
  }
  if (debug.level >= 2) {
    message(sprintf("[MC_IS_single] Using design point (u-space): %s", paste(round(dp, 4), collapse = ", ")))
  }

  ## ---- Debug-Ausgabe des Designpunkts --------------------------------------
  if (debug.level >= 2) {
    message(sprintf(
      "[MC_IS_single] Using design point: %s",
      paste(round(dp, 3), collapse = ", ")
    ))
  }

  # ------------------------------------------------------------
  # Stabilized design-point shift
  # ------------------------------------------------------------

  ### To do: implement switch use_shift_damping, but default = FALSE

  ### shift_damping <- min(1, 5 / sqrt(length(dp)))

  ### dp <- shift_damping * dp

  dp <- dp

  # ---------------------------------------------------------------------------
  # Detect whether LSF supports vectorized matrix input
  # ---------------------------------------------------------------------------

  lsf_vectorized <- FALSE

  if (debug.level >= 1) {
    message("[MC_IS_single] Testing if LSF supports vectorized evaluation...")
  }

  # Create small test matrix (2 samples)
  # x-space design point
  x_star <- as.numeric(unlist(res_form$x_points))

  test_mat <- matrix(
    rep(x_star, 2),
    nrow = 2,
    byrow = TRUE
  )
  # test_mat <- matrix(dp, nrow = 2, ncol = n_vars, byrow = TRUE)

  test_result <- try(lsf(test_mat), silent = TRUE)

  if (!inherits(test_result, "try-error") &&
    is.numeric(test_result) &&
    length(test_result) == 2 &&
    all(is.finite(test_result))) {
    lsf_vectorized <- TRUE

    if (debug.level >= 1) {
      message("[MC_IS_single] LSF detected as vectorized.")
    }
  } else {
    if (debug.level >= 1) {
      message("[MC_IS_single] LSF is NOT vectorized - falling back to row-wise evaluation.")
    }
  }

  # ---------------------------------------------------------------------------
  # Local Monte Carlo routine executed on each worker/thread
  # ---------------------------------------------------------------------------
  mc_local <- function(u_mat) {
    n_local <- nrow(u_mat)

    # --------------------------------------------------
    # Transform u -> x (vectorized)
    # --------------------------------------------------

    # code is slow
    # p_mat <- pnorm(u_mat)
    # p_mat <- pmin(pmax(p_mat, 1e-14), 1 - 1e-14)

    # q_list <- lapply(lDistr, function(d) d$q)

    # x_mat <- matrix(NA_real_, n_local, n_vars)

    # for (j in seq_len(n_vars)) {
    #   x_mat[, j] <- q_list[[j]](p_mat[, j])
    # }

    # Fix
    q_list <- lapply(lDistr, function(d) d$q)

    x_cols <- vector("list", n_vars)

    for (j in seq_len(n_vars)) {
      p <- pnorm(u_mat[, j])
      p <- pmin(pmax(p, 1e-14), 1 - 1e-14)

      x_cols[[j]] <- q_list[[j]](p)
    }

    x_mat <- do.call(cbind, x_cols)

    # --------------------------------------------------
    # log phi(u)  (standard normal)
    # --------------------------------------------------

    log_phi <- -0.5 * rowSums(u_mat^2) -
      0.5 * n_vars * log(2 * pi)

    # # --------------------------------------------------
    # # log f(x) (joint density in x-space)
    # # --------------------------------------------------

    # d_list <- lapply(lDistr, function(d) d$d)

    # log_fx_mat <- matrix(0, n_local, n_vars)

    # for (j in seq_len(n_vars)) {
    #   dens_vals <- d_list[[j]](x_mat[, j])
    #   log_fx_mat[, j] <- log(pmax(dens_vals, .Machine$double.xmin))
    # }

    # log_fx <- rowSums(log_fx_mat)

    # --------------------------------------------------
    # log h(u)  (shifted normal N(dp, I))
    # --------------------------------------------------

    # broadcast dp to matrix form
    diff <- u_mat - matrix(dp,
      nrow = n_local,
      ncol = n_vars,
      byrow = TRUE
    )

    log_h <- -0.5 * rowSums(diff^2) -
      0.5 * n_vars * log(2 * pi)

    # --------------------------------------------------
    # log weights (Importance sampling weights)
    # --------------------------------------------------
    # log_w <- log_fx - log_h
    log_w <- log_phi - log_h
    # numerical safety
    log_w[!is.finite(log_w)] <- -Inf

    # cat("mean(exp(log_fx - log_phi)) =", mean(exp(log_fx - log_phi)), "\n")

    # debug print
    # cat("mean(log_w) =", mean(log_w), "\n")
    # cat("var(log_w)  =", var(log_w), "\n")
    # cat("mean(w) =", mean(exp(log_w)), "\n")

    # --------------------------------------------------
    # Vectorized LSF evaluation
    # --------------------------------------------------

    # IMPORTANT: lsf must accept matrix input
    # g_vals <- lsf(x_mat)
    # g_vals <- apply(x_mat, 1, lsf)
    if (lsf_vectorized) {
      g_vals <- lsf(x_mat)
    } else {
      g_vals <- apply(x_mat, 1, lsf)
    }

    I_vals <- as.numeric(g_vals <= 0)

    # Debug print
    # cat("mean(I*w) =", mean(exp(log_w) * I_vals), "\n")
    # cat("mean(u_mat) =", colMeans(u_mat), "\n")

    # list(
    #   log_w  = log_w,
    #   I_vals = I_vals
    # )

    w <- exp(log_w)

    sum_Iw <- sum(I_vals * w)
    sum_w <- sum(w)
    sum_w2 <- sum(w * w)

    list(
      sum_Iw = sum_Iw,
      sum_w  = sum_w,
      sum_w2 = sum_w2,
      n      = length(w)
    )
  }

  # ## ---- Cluster setup --------------------------------------------------------
  # cl <- NULL # Platzhalter, falls kein Cluster gebraucht wird

  # ## ---- 1a. Seed fuer den Cluster bestimmen -------------------------------
  # ## `streams` ist der Vektor, den `init_rng_master()` zurueckgegeben hat.
  # ## Das erste Element ist ein gueltiger Integer-Seed.
  # ## seed_used <- as.integer(streams[1])

  # # streams <- init_rng_master(NULL)    # NULL -> neuer zufaelliger Seed pro Aufruf
  # # seed_used <- as.integer(streams[1])

  # ## -------------------------------------------------------------------------
  # ## 2) Wahl des Parallel-Backends
  # ## -------------------------------------------------------------------------
  # if (backend == "parallel" && Sys.info()[["sysname"]] != "Windows") {
  #   ## --- RNG-Stream vorbereiten --------------------------------------------
  #   ## Fuer das Parallel-Backend brauchen wir den vollstaendigen CMRG-Stream
  #   if (is.null(streams_master)) streams_master <- parallel::nextRNGStream(.Random.seed)

  #   cl <- make_ready_cluster(
  #     use_threads = use_threads,
  #     libPaths_local = libPaths_local,
  #     seed = streams_master,
  #     export_objs = c(
  #       "libPaths_local", "mc_local", "lsf", "lDistr", "dp",
  #       "n_batch", "densityType", "n_vars"
  #     ),
  #     debug.level = debug.level
  #   )

  #   on.exit(parallel::stopCluster(cl), add = TRUE)
  # } else if (backend == "future") {
  #   if (!requireNamespace("future", quietly = TRUE)) {
  #     stop("Package 'future' needed for backend = 'future'", call. = FALSE)
  #   }

  #   if (!requireNamespace("future.apply", quietly = TRUE)) {
  #     stop("Package 'future.apply' needed for backend = 'future'", call. = FALSE)
  #   }

  #   # ------------------------------------------------------------
  #   # CRAN-conform future handling
  #   # ------------------------------------------------------------
  #   # Save current plan
  #   old_plan <- future::plan()

  #   # Restore original plan when function exits
  #   on.exit(
  #     {
  #       future::plan(old_plan)
  #     },
  #     add = TRUE
  #   )

  #   # Only set a parallel plan if the user did not already set one
  #   if (inherits(old_plan, "sequential") && use_threads > 1) {
  #     if (.Platform$OS.type == "windows") {
  #       future::plan(future::multisession,
  #         workers = use_threads
  #       )
  #     } else {
  #       future::plan(future::multicore,
  #         workers = use_threads
  #       )
  #     }

  #     if (debug.level >= 1) {
  #       message(sprintf(
  #         "[MC_IS_single] Temporary future plan set with %d workers.",
  #         use_threads
  #       ))
  #     }
  #   } else {
  #     if (debug.level >= 1) {
  #       message("[MC_IS_single] Using existing future plan (no override).")
  #     }
  #   }
  # } else {
  #   # stop("Unsupported backend")
  #   if (backend != "parallel" && backend != "future") stop("Unsupported backend", call. = FALSE)
  # }

  ## -------------------------------------------------------------------------
  ## 3. Monte-Carlo-Loop (identisch fuer beide Back-ends)
  ## -------------------------------------------------------------------------
  n_sim <- 0L
  pf_mc <- 0
  var_mc <- 0
  w_mc <- 0
  I_mc <- 0
  var <- NA_real_
  cov <- 1
  pf <- NA_real_

  # Statistic container
  total_sum_Iw <- 0
  total_sum_w <- 0
  total_sum_w2 <- 0
  total_I_unw <- 0
  total_n <- 0

  if (dataRecord) {
    # rec <- make_recorders(ceiling(n_max / (use_threads * n_batch)))
    rec <- make_recorders(ceiling(n_max / (n_batch)))
  }

  ### cat(
  ###   "Initial RNG head:",
  ###   paste(.Random.seed[1:5], collapse = ", "),
  ###   "\n"
  ### )

  iter <- 0L

  while (TRUE) {
    iter <- iter + 1L

    # ============================================================
    # 3a. Deterministic sampling block (RNG only in master)
    # ============================================================

    # 1) Generate ALL samples in one deterministic block
    # u_mat <- matrix(rnorm(total_batch * n_vars),
    #   ncol = n_vars
    # )
    if (debug.level >= 1) {
      t_sample_start <- proc.time()
    }
    ## -------------------------------------------------
    ## Deterministic sampling block (thread independent)
    ## -------------------------------------------------
    u_vec <- rng_mgr$generate_norm_block(total_batch * n_vars)

    u_mat <- matrix(u_vec, ncol = n_vars)

    if (debug.level >= 1) {
      t_sample <- (proc.time() - t_sample_start)[3]
    }

    ### cat(
    ###   "After sampling RNG head:",
    ###   paste(.Random.seed[1:5], collapse = ", "),
    ###   "\n"
    ### )

    u_mat <- sweep(u_mat, 2, dp, "+")

    # 2) Deterministic split into worker chunks
    # if (use_threads <= 1) {
    #   chunk_indices <- list(seq_len(total_batch))
    # } else {
    #   chunk_indices <- split(
    #     seq_len(total_batch),
    #     cut(seq_len(total_batch),
    #       breaks = use_threads,
    #       labels = FALSE
    #     )
    #   )
    # }

    if (use_threads == 1) {
      chunk_indices <- list(seq_len(total_batch))
    } else {
      chunk_indices <- split(
        seq_len(total_batch),
        cut(seq_len(total_batch), breaks = use_threads)
      )
    }

    ### cat(
    #   "Chunk sizes:",
    #   sapply(chunk_indices, length),
    #   "\n"
    # )

    # 3) Worker function (NO RNG inside!)
    worker_fun <- function(idx) {
      mc_local(u_mat[idx, , drop = FALSE])
    }

    # # 4) Parallel evaluation (RNG disabled!)
    # if (use_threads == 1) {
    #   res_list <- lapply(chunk_indices, worker_fun)
    # } else if (backend == "parallel") {
    #   if (.Platform$OS.type == "windows") {
    #     res_list <- lapply(chunk_indices, worker_fun)
    #   } else {
    #     res_list <- parallel::parLapply(
    #       cl,
    #       chunk_indices,
    #       worker_fun
    #     )
    #   }
    # } else if (backend == "future") {
    #   res_list <- future.apply::future_lapply(
    #     chunk_indices,
    #     worker_fun,
    #     future.seed = FALSE,
    #     future.globals = FALSE # CRITICAL
    #   )
    # }
    if (debug.level >= 1) {
      t_parallel_start <- proc.time()
    }
    ## -------------------------------------------------
    ## Unified parallel evaluation (no RNG in workers)
    ## -------------------------------------------------
    # res_list <- run_parallel(
    #   X = seq_len(use_threads),
    #   FUN = function(worker_id) {
    #     idx <- chunk_indices[[worker_id]]
    #     worker_fun(idx)
    #   },
    #   backend = backend,
    #   use_threads = use_threads,
    #   seed = NULL # IMPORTANT: workers must not use RNG
    # )
    res_list <- run_parallel(
      X = chunk_indices,
      FUN = function(idx) {
        mc_local(u_mat[idx, , drop = FALSE])
      },
      backend = backend,
      use_threads = use_threads,
      cl = cl
    )

    if (debug.level >= 1) {
      t_parallel <- (proc.time() - t_parallel_start)[3]
    }

    if (debug.level >= 1) {
      cat(sprintf(
        "Sampling: %.3f s | Parallel: %.3f s\n",
        t_sample, t_parallel
      ))
    }

    # --- 3b. Aggregation ---------------------------------------------------------
    # all_log_w <- unlist(lapply(res_list, `[[`, "log_w"))
    # all_I_vals <- unlist(lapply(res_list, `[[`, "I_vals"))
    sum_Iw_iter <- sum(sapply(res_list, `[[`, "sum_Iw"))
    sum_w_iter <- sum(sapply(res_list, `[[`, "sum_w"))
    sum_w2_iter <- sum(sapply(res_list, `[[`, "sum_w2"))
    n_iter <- sum(sapply(res_list, `[[`, "n"))

    # --------------------------------------------------------
    # FAST MODE
    # --------------------------------------------------------
    # if (stability_mode == "fast") {
    #   max_log_w <- max(all_log_w)
    #   w_all <- exp(all_log_w - max_log_w)

    #   # accumulate using base sum (fast, stable enough here)
    #   sum_w_iter <- sum(w_all)
    #   sum_Iw_iter <- sum(w_all[all_I_vals == 1])

    #   # second moment for ESS
    #   sum_w2_iter <- sum(w_all * w_all)
    # }

    # # --------------------------------------------------------
    # # ROBUST MODE (numerically stable weight accumulation)
    # # --------------------------------------------------------
    # if (stability_mode == "robust") {
    #   # Compute global shift (log-sum-exp reference)
    #   # This prevents overflow when exponentiating large log-weights
    #   max_log_w <- max(all_log_w)

    #   # Shift log-weights before exponentiation
    #   # This is mathematically equivalent to fast-mode scaling
    #   shifted_weights <- exp(all_log_w - max_log_w)

    #   # First moment: sum of weights
    #   sum_w_iter <- sum(shifted_weights)

    #   # Weighted failure indicator
    #   if (any(all_I_vals == 1, na.rm = TRUE)) {
    #     sum_Iw_iter <- sum(shifted_weights[all_I_vals == 1])
    #   } else {
    #     sum_Iw_iter <- 0
    #   }

    #   # Second moment: required for ESS computation
    #   sum_w2_iter <- sum(shifted_weights * shifted_weights)
    # }

    # global accumulation
    total_sum_w <- total_sum_w + sum_w_iter
    total_sum_Iw <- total_sum_Iw + sum_Iw_iter
    total_sum_w2 <- total_sum_w2 + sum_w2_iter

    # total_n <- total_n + length(all_log_w)
    total_n <- total_n + n_iter

    # --- 3c. Recording ------------------------------------------------------------
    # record totals for this outer iteration
    n_sim <- total_n
    if (dataRecord) {
      rec <- record_step(rec, iter, total_sum_Iw, total_sum_w2, total_I_unw, n_sim, tic)
    }

    # ------ 3d. Schaetzer berechnen ----------------------

    pf <- NA_real_
    var_est <- NA_real_
    cov <- Inf
    ESS <- 0

    if (total_n > 0) {
      # unbiased importance sampling estimator
      pf_hat <- total_sum_Iw / total_n

      # Effective Sample Size
      if (total_sum_w2 > 0) {
        ESS <- (total_sum_w^2) / total_sum_w2
      } else {
        ESS <- 0
      }

      # variance approximation
      if (ESS > 0 && pf_hat > 0) {
        var_est <- pf_hat * (1 - pf_hat) / ESS
        cov <- sqrt(var_est) / pf_hat
      } else {
        var_est <- NA_real_
        cov <- Inf
      }

      pf <- pf_hat
    } else {
      pf <- 0
      var_est <- 0
      cov <- Inf
    }

    # if (total_sum_w > 0 && total_n > 0) {
    #   pf_hat <- total_sum_Iw / total_sum_w

    #   S_w2 <- total_sum_w2
    #   S_w <- total_sum_w
    #   N <- total_n

    #   if (total_sum_w > 0) {
    #     pf_hat <- total_sum_Iw / total_sum_w

    #     # Effective Sample Size
    #     ESS <- (total_sum_w^2) / total_sum_w2

    #     # Stable ESS-based variance approximation
    #     if (ESS > 0 && pf_hat > 0) {
    #       var_est <- pf_hat * (1 - pf_hat) / ESS
    #       cov <- sqrt(var_est) / pf_hat
    #     } else {
    #       var_est <- NA_real_
    #       cov <- Inf
    #     }
    #   } else {
    #     pf_hat <- 0
    #     var_est <- NA_real_
    #     cov <- Inf
    #     ESS <- 0
    #   }

    #   pf <- pf_hat
    # } else {
    #   # no weight mass -> pf = 0
    #   pf <- 0
    #   var_est <- 0
    #   cov <- Inf
    # }

    # debug printing (optional)
    if (debug.level >= 1) {
      message(sprintf("[MC_IS_single] iter=%d, n_sim=%d, pf=%.12g, cov=%.4f", iter, n_sim, pf, cov))
    }
    # convergence checks
    if (!is.na(cov) && is.finite(cov) && cov < cov_user) {
      break
    }
    if (n_sim >= n_max) {
      break # maximum iterations reached
    }
  }


  ## -------------------------------------------------------------------------
  ## 4. Assemble results
  ## -------------------------------------------------------------------------
  duration <- proc.time() - tic

  df <- if (dataRecord) as_record_df(rec, iter) else data.frame()


  if (!is.numeric(pf) || is.na(pf)) pf <- NA_real_
  beta_val <- if (!is.na(pf) && pf > 0) -stats::qnorm(pf) else NA_real_

  output <- list(
    method = "MCIS_Single",
    beta = beta_val, #-stats::qnorm(pf),
    pf = pf,
    FORM_beta = res_form$beta,
    FORM_pf = res_form$pf,
    design_points = dp,
    var = var_est,
    cov_mc = cov,
    cov_user = cov_user,
    ESS = ESS,
    ESS_ratio = if (n_sim > 0) ESS / n_sim else NA_real_,
    n_mc = n_sim,
    n_max = n_max,
    data = df,
    runtime = duration[1:5]
  )

  if (debug.level >= 1) message(sprintf("[MC_IS_single] Monte-Carlo finished in [s]: %g", duration[1]))
  return(output)
}


#' @title Update one recorder entry during main loop iteration
#' @description
#' Records current cumulative sums and timing information at iteration index \code{idx}.
#'
#' @inheritParams make_recorders
#'
#' @param idx Current iteration index within main loop.
#' @param pf_mc,var_mc,I_mc,n_sim Current cumulative sums from Monte Carlo batches.
#' @param tic Start time reference from \code{proc.time()} at simulation start.
#'
#' @return Updated recorder list object with new values inserted at position \code{idx}.
#'
#' @keywords internal
record_step <- function(rec, idx, pf_mc, var_mc, I_mc, n_sim, tic) {
  rec$pf[idx] <- pf_mc
  rec$var[idx] <- var_mc
  rec$I_sum[idx] <- I_mc
  rec$n_sim[idx] <- n_sim
  p_stamp <- proc.time() - tic
  rec$time$user[idx] <- p_stamp[1]
  rec$time$sys[idx] <- p_stamp[2]
  rec$time$elapsed[idx] <- p_stamp[3]
  rec$time$child_user[idx] <- p_stamp[4]
  rec$time$child_sys[idx] <- p_stamp[5]
  rec
}


#' @title Convert recorder lists into tidy data frame after completion
#' @description#'
#' Converts all recorded variables up to iteration \code{k} into a compact table format suitable for plotting or result export.
#'
#' Used internally after convergence or stop criterion is reached in Monte Carlo simulations.
#'
#' @param rec Recorder list created by [make_recorders()]
#' @param k Number of completed iterations/batches to include in output table.
#' @return Data frame containing columns: n_sim,pf,var,cov,I_sum,time_user,time_sys,time_elapsed,...
#' @keywords internal
#'
as_record_df <- function(rec, k) {
  if (k <= 0) {
    return(data.frame())
  }
  data.frame(
    n_sim = rec$n_sim[1:k],
    pf = rec$pf[1:k],
    var = rec$var[1:k],
    cov = rec$cov[1:k],
    I_sum = rec$I_sum[1:k],
    time_user = rec$time$user[1:k],
    time_sys = rec$time$sys[1:k],
    time_elapsed = rec$time$elapsed[1:k],
    time_user_child = rec$time$child_user[1:k],
    time_sys_child = rec$time$child_sys[1:k]
  )
}


## ---------------------------------------------------------------
## Parallel cluster setup utility used by MC_IS_single()
## ---------------------------------------------------------------

#' @title Prepare parallel cluster environment for MC Importance Sampling
#' @description Creates an R parallel cluster with specified number of workers,
#' exports required objects/functions and initializes reproducible RNG streams on each worker.
#'
#' @param use_threads Number of worker cores/threads to create via \link[parallel]{makeCluster}.
#' @param libPaths_local Library paths propagated to each worker via `.libPaths()`.
#' @param seed Integer base seed used for reproducible random numbers across workers.
#' @param export_objs Character vector naming objects/functions that should be exported from parent environment.
#'
#' @return Ready-to-use parallel cluster object configured with correct libraries and RNG streams.
#'
#' @importFrom parallel makeCluster clusterSetRNGStream clusterExport clusterEvalQ
#' @keywords internal
make_ready_cluster <- function(use_threads,
                               libPaths_local,
                               seed = NULL,
                               export_objs = character(0),
                               debug.level = 0) {
  # -------------------------------------------------
  # 1) Validate input
  # -------------------------------------------------
  stopifnot(is.numeric(use_threads), use_threads >= 1)
  stopifnot(is.character(export_objs))

  # -------------------------------------------------
  # 2) Choose cluster type depending on OS
  # -------------------------------------------------
  cl_type <- if (.Platform$OS.type == "windows") "PSOCK" else "FORK"

  cl <- parallel::makeCluster(use_threads, type = cl_type)
  # cl <- parallel::makeCluster(use_threads)

  # Ensure cluster is stopped on error
  on.exit(
    {
      try(parallel::stopCluster(cl), silent = TRUE)
    },
    add = TRUE
  )

  # -------------------------------------------------
  # 3) RNG handling
  # -------------------------------------------------
  if (!is.null(seed)) {
    ## seed ist ein kompletter CMRG-Stream (Laenge 7)
    parallel::clusterSetRNGStream(cl, iseed = seed)
    if (debug.level >= 1) {
      message("[make_ready_cluster] Fixed CMRG stream set on workers (backend parallel)")
    }
  } else {
    ## zufaellige Sub-Streams erzeugen
    parallel::clusterSetRNGStream(cl) # erzeugt automatisch neue Streams
    if (debug.level >= 1) {
      message("[make_ready_cluster] Random sub-streams generated (backend parallel)")
    }
  }

  #   ## 2) Export aller benoetigten Objekte
  #   all_export <- unique(c(export_objs, "libPaths_local"))
  #   parallel::clusterExport(cl,
  #     varlist = all_export,
  #     envir = parent.frame()
  #   )

  #   ## 3) Packages auf den Workern laden
  #   parallel::clusterEvalQ(cl, {
  #     .libPaths(libPaths_local)
  #     suppressMessages({
  #       if (!requireNamespace("TesiproV", quietly = TRUE)) library(TesiproV)
  #       if (!requireNamespace("edfun", quietly = TRUE)) library(edfun)
  #       if (!requireNamespace("evd", quietly = TRUE)) library(evd)
  #     })
  #   })
  #   cl
  # }

  # -------------------------------------------------
  # 4) Build clean export environment
  # -------------------------------------------------
  export_env <- new.env(parent = emptyenv())

  # Always export libPaths_local explicitly
  assign("libPaths_local", libPaths_local, envir = export_env)

  # Export additional requested objects if they exist
  if (length(export_objs) > 0) {
    for (obj in export_objs) {
      if (exists(obj, envir = parent.frame(), inherits = TRUE)) {
        assign(obj,
          get(obj, envir = parent.frame(), inherits = TRUE),
          envir = export_env
        )
      } else {
        warning(sprintf(
          "[make_ready_cluster] Object '%s' not found in calling environment - not exported.",
          obj
        ))
      }
    }
  }

  # Perform export
  parallel::clusterExport(
    cl,
    varlist = ls(export_env),
    envir = export_env
  )

  # -------------------------------------------------
  # 5) Initialize worker environments
  # -------------------------------------------------
  parallel::clusterEvalQ(cl, {
    .libPaths(libPaths_local)

    suppressMessages({
      if (!requireNamespace("TesiproV", quietly = TRUE)) {
        library(TesiproV)
      }

      if (!requireNamespace("edfun", quietly = TRUE)) {
        library(edfun)
      }

      if (!requireNamespace("evd", quietly = TRUE)) {
        library(evd)
      }
    })

    NULL
  })

  # Remove automatic stop handler (cluster valid)
  on.exit(NULL, add = FALSE)

  return(cl)
}

#' @title Resolve and validate RNG seed for Future backend
#' @description
#' Internal helper to normalize the \code{seed} argument used in
#' \link[future.apply]{future_lapply} calls within Monte Carlo simulations.
#'
#' The function ensures that the supplied seed is of a valid type:
#' either a single integer (for deterministic runs),
#' a complete L'Ecuyer-CMRG RNG stream (as returned by \link[parallel]{nextRNGStream}),
#' or \code{NULL}/\code{TRUE} for fully random behavior.
#'
#' This avoids common errors such as:
#' \itemize{
#'   \item "Argument 'seed' must be L'Ecuyer-CMRG RNG seed ..." when passing incomplete streams,
#'   \item identical random numbers on all workers due to improper seeding,
#'   \item non-reproducible results across different backends.
#' }
#'
#' @param seed Optional seed value provided by user or upstream function.
#' @description
#' May be one of:
#' \itemize{
#'   \item A single numeric/integer value (e.g. 1234) - fixed reproducible run.
#'   \item A numeric vector of length 7 (L'Ecuyer-CMRG stream) - substreams per worker.
#'   \item NULL - automatically use fully random streams (\code{future.seed = TRUE}).
#'  }
#'
#' @return A normalized object suitable for direct use in the argument
#'         \code{future.seed = resolve_future_seed(seed)} inside
#'         [future.apply::future_lapply()].
#'
#' @seealso [TesiproV::MC_IS()], [parallel::nextRNGStream()], [future.apply::future_lapply()]
#'
#' @author (C) 2026 M. Ricker - TU Dortmund University - Chair of Structural Concrete
#'
#' @keywords internal
resolve_future_seed <- function(seed) {
  ## --- Case 0: zero-length or NULL -> random automatic seeding ---
  if (is.null(seed) || length(seed) == 0L) {
    return(TRUE)
  }

  ## --- Case 1: Single integer/numeric ----------------------------------------
  if (is.numeric(seed) && length(seed) == 1L && !is.na(seed)) {
    return(as.integer(seed))
  }

  ## --- Case 2: Full L'Ecuyer-CMRG RNG stream ---------------------------------
  if (is.numeric(seed) && length(seed) == 7L && all(is.finite(seed))) {
    return(seed)
  }

  ## --- Otherwise invalid -----------------------------------------------------
  stop(
    sprintf(
      "Invalid 'seed': must be NULL/TRUE, single integer or L'Ecuyer-CMRG stream.\nGot type=%s length=%d.",
      typeof(seed), length(seed)
    ),
    call. = FALSE
  )
}


# =====================================================================
# MC_IS_system.R - Importance Sampling for a system of LSFs
# =====================================================================

#' Monte-Carlo Importance Sampling for a system of limit-state functions
#'
#' @inheritParams MC_IS_single
#' @param sys_type  Either "parallel" (default) or "serial"
#' @param beta_l  Optional threshold for reliability index screening.
#'   Limit-state functions with \eqn{\beta > beta_l} may be excluded
#'   from mixture construction in system analysis.
#'
#' @keywords internal
#'
MC_IS_system <- function(lsf,
                         lDistr,
                         cov_user = 0.05,
                         n_batch,
                         n_max,
                         use_threads,
                         sys_type,
                         dataRecord = TRUE,
                         beta_l,
                         densityType,
                         dps = NULL,
                         debug.level = 0,
                         streams = NULL,
                         libPaths_local = .libPaths(),
                         seed = NULL,
                         backend = "future",
                         adaptive_alpha = FALSE,
                         alpha_update_rate = 0.1,
                         adaptive_batch = FALSE,
                         batch_control = list(),
                         min_adapt_samples = 10000,
                         alpha_min = 0.02,
                         stability_mode = c("robust", "fast")) {
  if (debug.level >= 1) {
    cat("DEBUG LEVEL INSIDE MC_IS_system:", debug.level, "\n")
  }

  # Basic input checks
  stability_mode <- match.arg(stability_mode)

  if (missing(debug.level) || is.null(debug.level)) {
    debug.level <- 0
  }

  if (!is.list(lsf) || length(lsf) == 0L) {
    stop("MC_IS_system: 'lsf' must be a non-empty list of functions.")
  }

  if (!is.list(lDistr) || length(lDistr) != length(lsf)) {
    stop("MC_IS_system: 'lDistr' must be a list with one entry per LSF.")
  }

  if (!sys_type %in% c("serial", "parallel")) {
    stop("MC_IS_system: 'sys_type' must be 'serial' or 'parallel'.")
  }

  debug.TAG <- "MC_IS_System"
  if (debug.level >= 1) message("[MC_IS_system] starting system importance sampling...")

  tic <- proc.time()


  if (length(batch_control) == 0) {
    batch_control <- list(
      batch_min = 2000,
      batch_max = 100000,
      K_future = 5
    )
  }

  n_lsfs <- length(lsf)

  n_sim <- 0L
  total_sum_w <- 0
  total_sum_Iw <- 0
  total_sum_w2 <- 0

  ESS_global <- 0
  pf_hat <- NA_real_
  var_est <- NA_real_
  cov <- Inf

  # --------------------------------------------------
  # Normalize LSF input: accept functions or SYS_LSF objects
  # --------------------------------------------------
  lsf_fun <- lapply(lsf, function(obj) {
    if (is.function(obj)) {
      return(obj)
    } else if (!is.null(obj$func) && is.function(obj$func)) {
      return(obj$func)
    } else {
      stop("Invalid LSF format: must be function or SYS_LSF object.")
    }
  })


  # seeds setzen
  # if (!is.null(seed)) {
  #   RNGkind("L'Ecuyer-CMRG")
  #   set.seed(as.integer(seed))
  # }
  rng_mgr <- create_rng_manager(seed)


  ## create parallel cluster
  cl <- NULL

  if (backend == "parallel" && use_threads > 1L) {
    cl <- make_parallel_cluster(
      use_threads = use_threads,
      seed = seed
    )
    on.exit(parallel::stopCluster(cl), add = TRUE)
  }

  ## ------------------------------------------------------------------
  ## Normalize lDistr -> ensure each marginal is a funlist (with $d/$p/$q/$r)
  ## ------------------------------------------------------------------
  normalize_marginal <- function(m) {
    if (is.list(m) && !is.null(m$p) && is.function(m$p)) {
      return(m)
    } # already funlist
    if (is.list(m) && length(m) >= 1 && is.list(m[[1]]) && !is.null(m[[1]]$p) && is.function(m[[1]]$p)) {
      return(m[[1]])
    }
    stop("MC_IS_system: lDistr contains unexpected marginal format", call. = FALSE)
  }

  lDistr <- lapply(lDistr, function(per_lsf) {
    lapply(per_lsf, normalize_marginal)
  })

  ## ------------------------------------------------------------------
  ## 1) Build unique variable list and mappings
  ## ------------------------------------------------------------------

  var.names <- character(0)
  vars <- list() # store funlist per unique variable
  var.map <- integer(0)
  var.per.lsf <- integer(n_lsfs)

  for (i in seq_len(n_lsfs)) {
    nd <- length(lDistr[[i]])
    var.per.lsf[i] <- nd

    for (j in seq_len(nd)) {
      funlist <- lDistr[[i]][[j]]
      if (!is.list(funlist) || is.null(funlist$p) || !is.function(funlist$p)) {
        stop(sprintf("MC_IS_system: lDistr[[%d]][[%d]] is not a valid marginal (funlist)", i, j), call. = FALSE)
      }
      name_j <- funlist$name
      if (is.null(name_j) || name_j == "") name_j <- paste0("var_", length(var.names) + 1L)
      idx <- match(name_j, var.names, nomatch = 0L)
      if (idx == 0L) {
        var.names <- c(var.names, name_j)
        vars[[length(var.names)]] <- funlist
        var.map <- c(var.map, length(var.names))
      } else {
        var.map <- c(var.map, idx)
      }
    }
  }

  n_vars_unique <- length(vars)

  # var.in.lsf
  var.in.lsf <- vector("list", n_lsfs)
  k <- 1L
  for (i in seq_len(n_lsfs)) {
    var.in.lsf[[i]] <- var.map[k:(k + var.per.lsf[i] - 1L)]
    k <- k + var.per.lsf[i]
  }

  # lsf.in.var
  lsf.in.var <- vector("list", n_vars_unique)
  for (p in seq_len(n_vars_unique)) lsf.in.var[[p]] <- which(vapply(var.in.lsf, function(v) p %in% v, logical(1)))

  ## ------------------------------------------------------------------
  ## 2) FORM analysis -> design points per LSF
  ## ------------------------------------------------------------------
  res_form <- vector("list", n_lsfs)
  res_form.beta <- rep(NA_real_, n_lsfs)
  res_form.dp <- vector("list", n_lsfs)
  dp_u <- vector("list", n_lsfs)

  for (i in seq_len(n_lsfs)) {
    if (is.numeric(dps) && length(dps) == n_lsfs) {
      # --- user supplied design point ---
      x_star <- as.numeric(dps[i])

      Fxi <- tryCatch(lDistr[[i]][[1]]$p(x_star),
        error = function(e) NA_real_
      )

      Fxi <- min(max(as.numeric(Fxi), 1e-12), 1 - 1e-12)

      dp_u[[i]] <- qnorm(Fxi)

      res_form[[i]] <- list(
        beta = NA_real_,
        pf = NA_real_,
        x_points = x_star
      )

      res_form.beta[i] <- NA_real_
      res_form.dp[[i]] <- as.numeric(x_star)
    } else {
      # --- Funlists extrahieren ---
      lDistr_funlist_i <- lapply(lDistr[[i]], function(dd) {
        if (is.list(dd) && !is.null(dd$p) && is.function(dd$p)) {
          return(dd)
        }
        if (is.list(dd) && length(dd) >= 1 &&
          is.list(dd[[1]]) && !is.null(dd[[1]]$p)) {
          return(dd[[1]])
        }
        stop("Invalid marginal format for FORM")
      })

      # --- Wrapper bauen ---
      lDistr_for_FORM_i <- lapply(lDistr_funlist_i, function(fun) {
        params <- c(NA_real_, NA_real_)
        if (!is.null(fun$mean)) params[1] <- as.numeric(fun$mean)
        if (!is.null(fun$Sd)) params[2] <- as.numeric(fun$Sd)
        list(fun, params)
      })

      # --- FORM aufrufen ---
      res_form[[i]] <- TesiproV::FORM(
        lsf_fun[[i]],
        lDistr_for_FORM_i,
        n_optim = 20,
        loctol  = 1e-4
      )

      res_form.beta[i] <- res_form[[i]]$beta
      res_form.dp[[i]] <- as.numeric(unlist(res_form[[i]]$x_points))

      # --- U-Transformation ---
      x_star_vec <- res_form.dp[[i]]

      dp_u[[i]] <- numeric(length(lDistr_funlist_i))

      for (kidx in seq_along(lDistr_funlist_i)) {
        xsk <- as.numeric(x_star_vec[kidx])

        Fxi <- tryCatch(
          lDistr_funlist_i[[kidx]]$p(xsk),
          error = function(e) NA_real_
        )

        Fxi <- min(max(as.numeric(Fxi), 1e-12), 1 - 1e-12)

        dp_u[[i]][kidx] <- qnorm(Fxi)
      }
    }
  }


  # ------------------------------------------------------------
  # Automatic vectorization detection for each LSF
  # ------------------------------------------------------------

  lsf_vectorized_list <- logical(n_lsfs)

  for (k in seq_len(n_lsfs)) {
    # small test matrix (2 samples, correct variable subset)
    idxs <- var.in.lsf[[k]]

    x_star <- res_form.dp[[k]]

    test_mat <- matrix(
      rep(x_star, 2),
      nrow = 2,
      byrow = TRUE
    )

    test_result <- try(lsf_fun[[k]](test_mat), silent = TRUE)

    if (!inherits(test_result, "try-error") &&
      is.numeric(test_result) &&
      length(test_result) == 2 &&
      all(is.finite(test_result))) {
      lsf_vectorized_list[k] <- TRUE

      if (debug.level >= 1) {
        message(sprintf("[MC_IS_system] LSF %d detected as vectorized.", k))
      }
    }
  }

  # --------------------------------------------------
  # Globale Einbettung der Designpunkte im U-Raum
  # --------------------------------------------------

  dp_u_global <- vector("list", n_lsfs)

  for (k in seq_len(n_lsfs)) {
    u_star <- numeric(n_vars_unique)
    idxs <- var.in.lsf[[k]]
    u_star[idxs] <- dp_u[[k]]
    dp_u_global[[k]] <- u_star
  }

  if (debug.level >= 2) {
    dp_glob_str <- paste(
      sapply(seq_along(dp_u_global), function(i) {
        paste0(
          "LSF", i, ": (",
          paste(round(dp_u_global[[i]], 4), collapse = ", "),
          ")"
        )
      }),
      collapse = " | "
    )

    message(sprintf("[MC_IS_system] Global U-shifts: %s", dp_glob_str))
  }


  # --------------------------------------------------
  # Geometric analysis of failure regions (U-space)
  # --------------------------------------------------

  compute_geometry <- function(dp_u_list, res_form_list) {
    n_lsfs <- length(dp_u_list)

    if (n_lsfs < 2) {
      return(list(multimodal = FALSE))
    }

    max_angle <- 0
    max_dist <- 0

    for (i in 1:(n_lsfs - 1)) {
      for (j in (i + 1):n_lsfs) {
        # --- Distance between design points ---
        dist_ij <- sqrt(sum((dp_u_list[[i]] - dp_u_list[[j]])^2))

        max_dist <- max(max_dist, dist_ij)

        # --- Angle between gradients (alpha vectors) ---
        alpha_i <- res_form_list[[i]]$dy
        alpha_i <- alpha_i / sqrt(sum(alpha_i^2))

        alpha_j <- res_form_list[[j]]$dy
        alpha_j <- alpha_j / sqrt(sum(alpha_j^2))

        cos_ij <- sum(alpha_i * alpha_j)

        max_angle <- min(max_angle, cos_ij)
      }
    }

    multimodal_flag <- !(max_angle > 0.85 && max_dist < 1.5)

    return(list(
      multimodal = multimodal_flag,
      max_cos = max_angle,
      max_dist = max_dist
    ))
  }

  ## decision multimodal distribution or single shift
  geom_info <- compute_geometry(dp_u_global, res_form)

  use_multimodal <- geom_info$multimodal

  if (debug.level >= 1) {
    message(sprintf(
      "[MC_IS_system] Geometry: cos=%.3f, dist=%.3f -> multimodal=%s",
      geom_info$max_cos,
      geom_info$max_dist,
      use_multimodal
    ))
  }


  # ------------------------------------------------------------------
  # Parallel system: compute system design point
  # ------------------------------------------------------------------
  if (sys_type == "parallel") {
    g_sys <- function(x) {
      vals <- numeric(n_lsfs)
      for (i in seq_len(n_lsfs)) {
        vals[i] <- lsf_fun[[i]](x[var.in.lsf[[i]]])
      }
      min(vals)
    }

    lDistr_global <- lapply(vars, function(fun) {
      list(fun, c(fun$mean, fun$Sd))
    })

    res_form_sys <- TesiproV::FORM(
      g_sys,
      lDistr_global,
      n_optim = 20,
      loctol = 1e-4
    )

    x_star_sys <- as.numeric(res_form_sys$x_points)

    dp_u_sys <- numeric(n_vars_unique)
    for (j in seq_len(n_vars_unique)) {
      Fxi <- vars[[j]]$p(x_star_sys[j])
      dp_u_sys[j] <- qnorm(Fxi)
    }
  }

  # ------------------------------------------------------------------
  # Initial mixture weights (serial only)
  # ------------------------------------------------------------------
  if (sys_type == "serial") {
    pf_single <- pnorm(-res_form.beta)
    # alpha_i <- pf_single / sum(pf_single)
    alpha_i <- rep(1 / n_lsfs, n_lsfs)
  }

  ## ------------------------------------------------------------------
  ## 3) Build dp matrix (global)
  ## ------------------------------------------------------------------
  dp <- matrix(NA_real_, nrow = n_vars_unique, ncol = n_lsfs)
  for (i in seq_len(n_lsfs)) {
    local_vars <- var.in.lsf[[i]]
    if (!is.null(res_form.dp[[i]])) {
      for (kl in seq_along(local_vars)) {
        global_id <- local_vars[kl]
        if (kl <= length(res_form.dp[[i]])) dp[global_id, i] <- res_form.dp[[i]][kl]
      }
    }
  }

  dp.means <- rowMeans(dp, na.rm = TRUE)

  ## ------------------------------------------------------------------
  ## 4) ai weights
  ## ------------------------------------------------------------------
  ai <- matrix(0, nrow = n_vars_unique, ncol = n_lsfs)
  for (p in seq_len(n_vars_unique)) {
    comps <- lsf.in.var[[p]]
    if (length(comps) == 0L) next
    betas <- res_form.beta[comps]
    betas[!is.finite(betas) | betas <= 0] <- Inf
    inv_b <- 1 / betas
    s <- sum(inv_b, na.rm = TRUE)
    if (!is.finite(s) || s <= 0) ai[p, comps] <- 1 / length(comps) else ai[p, comps] <- inv_b / s
  }

  ## ------------------------------------------------------------------
  ## 5) Build hv mixtures per global variable
  ## ------------------------------------------------------------------
  hv_list <- vector("list", n_vars_unique)
  for (p in seq_len(n_vars_unique)) {
    comps <- lsf.in.var[[p]]
    if (length(comps) == 0L) {
      hv_list[[p]] <- list(rfun = function(n) rep(NA_real_, n), dfun = function(x) rep(0, length(x)))
      next
    }
    orig_funlist <- vars[[p]] # funlist (d,p,q,r,mean,Sd,...)
    sd_p <- orig_funlist$Sd
    comp_dfun <- list()
    comp_rfun <- list()
    comp_probs <- numeric(length(comps))
    for (ci in seq_along(comps)) {
      j <- comps[ci]
      mu_j <- dp[p, j]
      sd_loc <- ifelse(is.finite(sd_p) && sd_p > 0, sd_p, max(1e-8, abs(mu_j) * 0.1, 1e-3))
      dist_type <- tolower(if (is.null(orig_funlist$DistributionType) || orig_funlist$DistributionType == "") "norm" else orig_funlist$DistributionType)
      if (dist_type %in% c("lnorm", "lt", "weibull")) {
        if (!is.finite(mu_j) || mu_j <= 0) {
          mn <- mu_j
          sn <- sd_loc
          comp_rfun[[ci]] <- function(n, m = mn, s = sn) stats::rnorm(n, mean = m, sd = s)
          comp_dfun[[ci]] <- function(x, m = mn, s = sn) stats::dnorm(x, mean = m, sd = s)
        } else {
          sd_rel <- sd_loc / mu_j
          sn <- sqrt(log(1 + sd_rel^2))
          mn <- log(mu_j / sqrt(1 + sd_rel^2))
          comp_rfun[[ci]] <- function(n, mn = mn, sn = sn) stats::rlnorm(n, meanlog = mn, sdlog = sn)
          comp_dfun[[ci]] <- function(x, mn = mn, sn = sn) stats::dlnorm(x, meanlog = mn, sdlog = sn)
        }
      } else {
        comp_rfun[[ci]] <- function(n, m = mu_j, s = sd_loc) stats::rnorm(n, mean = m, sd = s)
        comp_dfun[[ci]] <- function(x, m = mu_j, s = sd_loc) stats::dnorm(x, mean = m, sd = s)
      }
      comp_probs[ci] <- ai[p, j]
    }
    if (!is.finite(sum(comp_probs)) || sum(comp_probs) <= 0) comp_probs <- rep(1 / length(comp_probs), length(comp_probs)) else comp_probs <- comp_probs / sum(comp_probs)
    hv_dfun <- function(x) {
      res <- numeric(length(x))
      for (ci in seq_along(comp_dfun)) res <- res + comp_probs[ci] * comp_dfun[[ci]](x)
      res
    }
    hv_rfun <- function(n) {
      if (n <= 0) {
        return(numeric(0))
      }
      comps_drawn <- sample(seq_along(comp_probs), size = n, replace = TRUE, prob = comp_probs)
      res <- numeric(n)
      for (ci in unique(comps_drawn)) {
        idx <- which(comps_drawn == ci)
        res[idx] <- comp_rfun[[ci]](length(idx))
      }
      res
    }
    hv_list[[p]] <- list(rfun = hv_rfun, dfun = hv_dfun, comps = comps, probs = comp_probs)
  }

  ## ------------------------------------------------------------------
  ## 6) Monte Carlo Loop (single-sample per iter)
  ## ------------------------------------------------------------------
  I_sum <- 0L
  pf_acc <- 0
  var_acc <- 0
  n_sim <- 0L
  cov <- Inf


  if (dataRecord) {
    max_iter_est <- ceiling(n_max)
    rec_pf <- numeric(max_iter_est)
    rec_var <- numeric(max_iter_est)
    rec_cov <- numeric(max_iter_est)
    rec_nsim <- integer(max_iter_est)
    rec_I_sum <- integer(max_iter_est)
    rec_time_user <- numeric(max_iter_est)
    rec_time_sys <- numeric(max_iter_est)
    rec_time_elapsed <- numeric(max_iter_est)
  }

  orig_d_list <- lapply(vars, function(v) v$d)


  # --------------------------------------------------------------
  # Monte Carlo Loop (worker-invariant, SNIS, global stabilization)
  # --------------------------------------------------------------

  # --------------------------------------------------------------
  # Monte Carlo Loop - Multimodal, worker-invariant, global log-stable
  # --------------------------------------------------------------

  if (!exists("dp_u")) {
    stop("dp_u not constructed before Monte Carlo loop")
  }

  n_sim <- 0L
  total_sum_Iw <- 0
  total_sum_w <- 0
  total_sum_w2 <- 0

  ESS_modes <- NULL


  if (stability_mode == "robust") {
    log_total_sum_w <- -Inf
    log_total_sum_Iw <- -Inf
    log_total_sum_w2 <- -Inf
  }

  total_batch <- n_batch
  iter <- 0L

  if (sys_type == "serial") {
    if (all(is.finite(res_form.beta))) {
      pf_single <- pnorm(-res_form.beta)
      alpha_i <- pf_single / sum(pf_single)
    } else {
      alpha_i <- rep(1 / n_lsfs, n_lsfs)
    }
  }

  state <- list(
    total_sum_w = total_sum_w,
    total_sum_Iw = total_sum_Iw,
    total_sum_w2 = total_sum_w2,
    log_total_sum_w = if (stability_mode == "robust") -Inf else NULL,
    log_total_sum_Iw = if (stability_mode == "robust") -Inf else NULL,
    log_total_sum_w2 = if (stability_mode == "robust") -Inf else NULL,
    n_sim = n_sim
  )


  while (TRUE) {
    iter <- iter + 1L
    total_batch <- n_batch

    if (debug.level >= 1) {
      cat("total_batch:", total_batch, "\n")
    }

    # ==========================================================
    # Sampling Block
    # ==========================================================

    # if (sys_type == "serial") {
    #   # Serial: multimodal sampling
    #   mode_ids <- sample(seq_len(n_lsfs),
    #     size = total_batch,
    #     replace = TRUE,
    #     prob = alpha_i
    #   )

    #   u_mat <- matrix(0, nrow = total_batch, ncol = n_vars_unique)

    #   for (i in seq_len(total_batch)) {
    #     shift <- dp_u_global[[mode_ids[i]]]
    #     u_mat[i, ] <- rnorm(n_vars_unique,
    #       mean = shift,
    #       sd = 1
    #     )
    #   }
    # } else { # parallel

    #   # Parallel: single shift at system design point
    #   u_mat <- matrix(rnorm(total_batch * n_vars_unique),
    #     ncol = n_vars_unique
    #   )

    #   u_mat <- sweep(u_mat, 2, dp_u_sys, "+")

    #   # Dummy mode ids (needed for ESS calculation)
    #   mode_ids <- rep(1L, total_batch)
    # }

    ## -------------------------------------------------
    ## Deterministic sampling (master only)
    ## -------------------------------------------------
    if (debug.level >= 1) {
      t_sample_start <- proc.time()
    }

    u_vec <- rng_mgr$generate_norm_block(total_batch * n_vars_unique)
    u_mat <- matrix(u_vec, ncol = n_vars_unique)

    if (debug.level >= 1) {
      t_sample <- (proc.time() - t_sample_start)[3]
    }

    # Use alway a multimodal sampling density
    # if (sys_type == "serial" && use_multimodal) {
    if (use_multimodal) {
      # --- Multimodal sampling ---
      # mode_ids <- sample(
      #   seq_len(n_lsfs),
      #   size = total_batch,
      #   replace = TRUE,
      #   prob = alpha_i
      # )
      ## mode_ids <- rep(seq_len(n_lsfs), length.out = total_batch)
      n_per_mode <- floor(total_batch / n_lsfs)

      mode_ids <- rep(seq_len(n_lsfs), each = n_per_mode)

      if (length(mode_ids) < total_batch) {
        mode_ids <- c(
          mode_ids,
          sample(seq_len(n_lsfs),
            total_batch - length(mode_ids),
            replace = TRUE
          )
        )
      }

      for (i in seq_len(total_batch)) {
        shift <- dp_u_global[[mode_ids[i]]]
        u_mat[i, ] <- u_mat[i, ] + shift
      }
    } else {
      # --- Single global shift ---
      if (sys_type == "serial") {
        # Use system-level FORM shift
        shift <- rowMeans(do.call(cbind, dp_u_global))

        u_mat <- sweep(u_mat, 2, shift, "+")

        mode_ids <- rep(1L, total_batch)
      }
    }

    # if (sys_type == "serial") {
    #   ## deterministic mode selection
    #   mode_probs <- alpha_i
    #   mode_ids <- sample(
    #     seq_len(n_lsfs),
    #     size = total_batch,
    #     replace = TRUE,
    #     prob = mode_probs
    #   )

    #   for (i in seq_len(total_batch)) {
    #     shift <- dp_u_global[[mode_ids[i]]]
    #     u_mat[i, ] <- u_mat[i, ] + shift
    #   }
    # } else {
    #   u_mat <- sweep(u_mat, 2, dp_u_sys, "+")
    #   mode_ids <- rep(1L, total_batch)
    # }

    # ==========================================================
    # Parallel Evaluation Block
    # ==========================================================

    # chunk_ids <- rep(seq_len(use_threads),
    #   length.out = total_batch
    # )

    # chunk_indices <- split(seq_len(total_batch), chunk_ids)
    if (use_threads <= 1L) {
      chunk_indices <- list(seq_len(total_batch))
    } else {
      chunk_indices <- split(
        seq_len(total_batch),
        cut(seq_len(total_batch), breaks = use_threads)
      )
    }

    worker_fun <- function(idx) {
      mode_ids <- rep(NA_integer_, nrow(u_mat))

      u_block <- u_mat[idx, , drop = FALSE]
      n_local <- nrow(u_block)

      # ----------------------------
      # 1) Transform to physical space
      # ----------------------------

      # p_mat <- pnorm(u_block)
      # p_mat <- pmin(pmax(p_mat, 1e-14), 1 - 1e-14)

      # x_mat <- matrix(NA_real_, n_local, n_vars_unique)

      # for (j in seq_len(n_vars_unique)) {
      #   x_mat[, j] <- vars[[j]]$q(p_mat[, j])
      # }

      # ----------------------------
      # 2) Evaluate LSFs
      # ----------------------------

      # I_matrix <- matrix(0L, n_local, n_lsfs)

      # for (k in seq_len(n_lsfs)) {
      #   idxs <- var.in.lsf[[k]]
      #   x_sub <- x_mat[, idxs, drop = FALSE]

      #   fun_k <- lsf_fun[[k]]

      #   if (lsf_vectorized_list[k]) {
      #     vals <- fun_k(x_sub)
      #   } else {
      #     vals <- apply(x_sub, 1, fun_k)
      #   }

      #   I_matrix[, k] <- as.integer(vals < 0)
      # }

      # FIX: faster without matrix
      fail_count <- integer(n_local)

      # for (k in seq_len(n_lsfs)) {
      #   idxs <- var.in.lsf[[k]]
      #   x_sub <- x_mat[, idxs, drop = FALSE]

      # FIX: faster
      for (k in seq_len(n_lsfs)) {
        idxs <- var.in.lsf[[k]]
        n_loc_vars <- length(idxs)

        x_sub <- matrix(NA_real_, n_local, n_loc_vars)

        for (j in seq_len(n_loc_vars)) {
          v <- idxs[j]

          p <- pnorm(u_block[, v])
          p <- pmin(pmax(p, 1e-14), 1 - 1e-14)

          x_sub[, j] <- vars[[v]]$q(p)
        }

        fun_k <- lsf_fun[[k]]

        if (lsf_vectorized_list[k]) {
          vals <- fun_k(x_sub)
        } else {
          vals <- apply(x_sub, 1, fun_k)
        }

        fail_count <- fail_count + (vals < 0)
      }

      # if (sys_type == "serial") {
      #   I_vals <- as.integer(rowSums(I_matrix) > 0)
      # } else {
      #   I_vals <- as.integer(rowSums(I_matrix) == n_lsfs)
      # }

      # FIX: faster
      if (sys_type == "serial") {
        I_vals <- as.integer(fail_count > 0)
      } else {
        I_vals <- as.integer(fail_count == n_lsfs)
      }

      # ----------------------------
      # 3) log phi(u)
      # ----------------------------

      log_phi <- rowSums(dnorm(u_block, log = TRUE))

      # ----------------------------
      # 4) log h(u)
      # ----------------------------

      # if (sys_type == "serial") {
      #   log_h_mat <- matrix(NA_real_, n_local, n_lsfs)

      #   for (k in seq_len(n_lsfs)) {
      #     shift <- dp_u_global[[k]]
      #     shifted <- sweep(u_block, 2, shift, "-")

      #     log_h_mat[, k] <-
      #       log(alpha_i[k]) +
      #       rowSums(dnorm(shifted, log = TRUE))
      #   }

      #   max_log <- apply(log_h_mat, 1, max)

      #   log_h <- max_log +
      #     log(rowSums(exp(log_h_mat - max_log)))
      # } else {
      #   shifted <- sweep(u_block, 2, dp_u_sys, "-")
      #   log_h <- rowSums(dnorm(shifted, log = TRUE))
      # }

      shift <- rowMeans(do.call(cbind, dp_u_global))

      if (sys_type == "serial" && use_multimodal) {
        # --- Multimodal mixture density ---
        log_h_mat <- matrix(NA_real_, n_local, n_lsfs)

        for (k in seq_len(n_lsfs)) {
          shift_k <- dp_u_global[[k]]

          shifted <- sweep(u_block, 2, shift_k, "-")

          log_h_mat[, k] <-
            log(alpha_i[k]) +
            rowSums(dnorm(shifted, log = TRUE))
        }

        max_log <- apply(log_h_mat, 1, max)

        log_h <- max_log +
          log(rowSums(exp(log_h_mat - max_log)))
      } else {
        # --- Single shift density ---
        shifted <- sweep(u_block, 2, shift, "-")

        log_h <- rowSums(dnorm(shifted, log = TRUE))
      }

      log_w <- log_phi - log_h

      list(
        log_w = log_w,
        I_vals = I_vals,
        # I_matrix = I_matrix,
        mode_ids = mode_ids[idx]
      )
    }


    # --------------------------------------------------
    # Parallele Auswertung
    # --------------------------------------------------

    # if (backend == "future") {
    #   res_list <- future.apply::future_lapply(
    #     chunk_indices,
    #     worker_fun,
    #     future.seed = FALSE
    #   )
    # } else {
    #   res_list <- lapply(chunk_indices, worker_fun)
    # }
    if (debug.level >= 1) {
      t_parallel_start <- proc.time()
    }

    res_list <- run_parallel(
      X = chunk_indices,
      FUN = worker_fun,
      backend = backend,
      use_threads = use_threads,
      cl = cl
    )

    if (debug.level >= 1) {
      t_parallel <- (proc.time() - t_parallel_start)[3]
    }

    if (debug.level >= 1) {
      cat(sprintf(
        "[MC_IS_system]: Sampling: %.3f s | Parallel: %.3f s\n",
        t_sample, t_parallel
      ))
    }

    # ==========================================================
    # Weight Accumulation Block (refactored)
    # ==========================================================

    all_log_w <- unlist(lapply(res_list, `[[`, "log_w"))
    all_I_vals <- unlist(lapply(res_list, `[[`, "I_vals"))
    # all_I_matrix <- do.call(rbind, lapply(res_list, `[[`, "I_matrix"))
    all_mode_ids <- unlist(lapply(res_list, `[[`, "mode_ids"))

    if (debug.level >= 1) {
      cat("len(all_log_w):", length(all_log_w), "\n")
    }

    state <- compute_weight_update(
      all_log_w,
      all_I_vals,
      stability_mode,
      sys_type,
      state
    )

    total_sum_w <- state$total_sum_w
    total_sum_Iw <- state$total_sum_Iw
    total_sum_w2 <- state$total_sum_w2
    log_total_sum_w <- state$log_total_sum_w
    log_total_sum_Iw <- state$log_total_sum_Iw
    log_total_sum_w2 <- state$log_total_sum_w2
    n_sim <- state$n_sim

    # ==========================================================
    # Estimator Update Block
    # ==========================================================

    if (sys_type == "serial" && use_multimodal && stability_mode == "fast") {
      max_log_w <- max(all_log_w)
      shifted_weights <- exp(all_log_w - max_log_w)

      ESS_modes <- numeric(n_lsfs)

      for (k in seq_len(n_lsfs)) {
        idx_k <- which(all_mode_ids == k)

        if (length(idx_k) > 0) {
          w_k <- shifted_weights[idx_k]
          ESS_modes[k] <- (sum(w_k)^2) / sum(w_k^2)
        } else {
          ESS_modes[k] <- 0
        }
      }
    }
    # --------------------------------------------------
    # Adaptive batch size update
    # --------------------------------------------------
    if (adaptive_batch && ESS_global > 0) {
      RSE_target <- cov_user
      K_future <- batch_control$K_future
      n_min_batch <- batch_control$n_min
      n_max_batch <- batch_control$n_max

      # Estimate current efficiency factor
      kappa_hat <- ESS_global / n_sim

      if (kappa_hat > 0) {
        ESS_target <- 1 / (RSE_target^2)
        N_target <- ESS_target / kappa_hat
        N_remaining <- max(N_target - n_sim, 0)

        n_batch_new <- ceiling(N_remaining / K_future)

        # Ensure divisibility by threads
        n_batch_new <- ceiling(n_batch_new / use_threads) * use_threads

        # Apply bounds
        n_batch <- max(n_min_batch, min(n_batch_new, n_max_batch))

        if (debug.level >= 2) {
          message(sprintf(
            "[MC_IS_system] Adaptive batch updated to %d (ESS=%.2f)",
            n_batch, ESS_global
          ))
        }
      }
    }


    # ==========================================================
    # Estimator Update Block
    # ==========================================================

    # FAST MODE or SERIAL SYSTEM
    if (stability_mode == "fast" || sys_type == "serial") {
      if (total_sum_w > 0) {
        pf_hat <- total_sum_Iw / total_sum_w
        ESS_global <- (total_sum_w^2) / total_sum_w2
      } else {
        pf_hat <- 0
        ESS_global <- 0
      }
    }

    # ROBUST MODE (parallel only)
    if (stability_mode == "robust" && sys_type == "parallel") {
      if (log_total_sum_w > -Inf) {
        pf_hat <- exp(log_total_sum_Iw - log_total_sum_w)
        ESS_global <- exp(2 * log_total_sum_w - log_total_sum_w2)
      } else {
        pf_hat <- 0
        ESS_global <- 0
      }
    }

    # Variance (common)
    if (ESS_global > 0 && pf_hat > 0) {
      var_est <- pf_hat * (1 - pf_hat) / ESS_global
      cov <- sqrt(var_est) / pf_hat
    } else {
      var_est <- NA_real_
      cov <- Inf
    }

    if (ESS_global > 0) {
      RSE_est <- 1 / sqrt(ESS_global)
    } else {
      RSE_est <- Inf
    }


    if (adaptive_alpha && sys_type == "serial" && use_multimodal && n_sim > min_adapt_samples) {
      stop("adaptive_alpha currently requires I_matrix; disable adaptive_alpha or re-enable I_matrix.")

      # pf_i_hat <- numeric(n_lsfs)
      #
      # if (stability_mode == "fast") {
      #   pf_sys_hat <- sum(shifted_weights[all_I_vals == 1])
      #
      #   if (!is.na(pf_sys_hat) && is.finite(pf_sys_hat) && pf_sys_hat > 0) {
      #     for (k in seq_len(n_lsfs)) {
      #       pf_i_hat[k] <-
      #         sum((all_I_matrix[, k] * all_I_vals) *
      #           shifted_weights) / pf_sys_hat
      #     }
      #   }
      # }
      #
      # if (stability_mode == "robust") {
      #   # pf_sys_hat in log-domain
      #   pf_sys_hat <- exp(log_total_sum_Iw - log_total_sum_w)
      #
      #   if (!is.na(pf_sys_hat) && is.finite(pf_sys_hat) && pf_sys_hat > 0) {
      #     for (k in seq_len(n_lsfs)) {
      #       idx_k <- which(all_I_matrix[, k] == 1 &
      #         all_I_vals == 1)
      #
      #       if (length(idx_k) > 0) {
      #         log_contrib <- all_log_w[idx_k]
      #
      #         max_log_c <- max(log_contrib)
      #
      #         pf_i_hat[k] <-
      #           exp(
      #             max_log_c +
      #               log(sum(exp(log_contrib - max_log_c))) -
      #               log_total_sum_w
      #           )
      #       } else {
      #         pf_i_hat[k] <- 0
      #       }
      #     }
      #   }
      # }
      #
      # # update alpha
      # if (sum(pf_i_hat) > 0) {
      #   alpha_target <- pf_i_hat / sum(pf_i_hat)
      #
      #   alpha_i <- (1 - alpha_update_rate) * alpha_i +
      #     alpha_update_rate * alpha_target
      #
      #   alpha_i <- pmax(alpha_i, alpha_min)
      #   alpha_i <- alpha_i / sum(alpha_i)
      # }
    }


    if (debug.level >= 2) {
      if (adaptive_alpha) {
        print(alpha_i)
      }
    }

    # ==========================================================
    # Convergence Check
    # ==========================================================
    if (!exists("ESS_global")) ESS_global <- 0

    ESS_min <- if (sys_type == "parallel") 50 else 30
    n_min <- 5 * n_batch

    if (debug.level >= 1) {
      if (iter %% 1 == 0) {
        cat(
          "Iteration:", iter,
          "n_sim:", n_sim,
          "cov:", cov, "\n"
        )
      }

      cat(sprintf(
        "[MC_IS_system]: ESS: %.3f \n",
        ESS_global
      ))
    }

    stop_condition <- (
      !is.na(cov) &&
        is.finite(cov) &&
        cov < cov_user &&
        ESS_global > ESS_min &&
        n_sim > n_min
    )

    if (stop_condition) {
      break
    }

    if (n_sim >= n_max) {
      break
    }
  }

  beta_final <- if (pf_hat > 0) -qnorm(pf_hat) else NA_real_

  out <- list(
    method    = "MCIS_System",
    sys_type  = sys_type,
    beta      = beta_final,
    pf        = pf_hat,
    var       = var_est,
    cov_mc    = cov,
    RSE_est   = RSE_est,
    ESS       = ESS_global,
    ESS_modes = if (sys_type == "serial") ESS_modes else NULL,
    n_mc      = n_sim
  )

  if (sys_type == "parallel" && ESS_global < ESS_min) {
    warning(sprintf(
      "Parallel system: Effective sample size (ESS = %.1f) is low. Relative standard error ~= %.2f%%.",
      ESS_global,
      100 * RSE_est
    ))
  }

  if (debug.level >= 1) message("[MC_IS_system] finished.")

  return(out)
}


#' Monte Carlo Importance Sampling using smooth system aggregation
#'
#' Performs reliability estimation for systems of limit-state functions
#' by constructing a **smooth global system limit-state function**
#' and applying the existing single-LSF importance sampling algorithm.
#'
#' The system limit-state function is built using \code{build_gsys()},
#' which aggregates component limit states via smooth min/max operators.
#'
#' This approach converts a system reliability problem into a standard
#' single limit-state reliability problem.
#'
#' @param sys_input List of \code{SYS_LSF} objects defining the system.
#'
#' @param sys_structure System aggregation type: \code{"serial"},
#' \code{"parallel"}, or a custom function.
#'
#' @param smooth_kappa Smoothing parameter controlling the
#' approximation of min/max aggregation.
#'
#' @inheritParams MC_IS
#'
#' @return Result object returned by \code{MC_IS_single()} with additional
#' metadata describing the system configuration.
#'
#' @keywords internal
#'
MC_IS_gsys <- function(sys_input,
                       sys_structure = "serial",
                       smooth_kappa = 20,
                       cov_user = 0.05,
                       n_batch = 5000,
                       n_max = 1e6,
                       use_threads = 4,
                       backend = "future",
                       dataRecord = TRUE,
                       densityType = "norm",
                       dps = NULL,
                       debug.level = 0,
                       seed = NULL,
                       adaptive_alpha = FALSE,
                       alpha_update_rate = 0.1,
                       adaptive_batch = FALSE,
                       batch_control = list(),
                       min_adapt_samples = 10000,
                       alpha_min = 0.02,
                       stability_mode = "robust") {
  if (debug.level >= 1) {
    message("[MC_IS_gsys] building smooth system limit-state function...")
  }

  # -------------------------------------------------
  # Build system representation
  # -------------------------------------------------

  sys_obj <- build_gsys(
    sys_input = sys_input,
    sys_structure = sys_structure,
    smooth_kappa = smooth_kappa,
    debug.level = debug.level
  )

  # debug
  # print(body(sys_obj$g_sys))

  g_sys <- sys_obj$g_sys
  distr_flat <- sys_obj$distr_flat

  if (debug.level >= 1) {
    message(sprintf(
      "[MC_IS_gsys] system has %d LSFs and %d variables",
      sys_obj$n_lsfs,
      sys_obj$n_vars
    ))
  }

  # debug
  # message("kappa=", smooth_kappa)

  # -------------------------------------------------
  # Run importance sampling using existing routine
  # -------------------------------------------------

  res <- MC_IS_single(
    lsf = g_sys,
    lDistr = distr_flat,
    cov_user = cov_user,
    n_batch = n_batch,
    n_max = n_max,
    use_threads = use_threads,
    backend = backend,
    dataRecord = dataRecord,
    densityType = densityType,
    dps = dps,
    debug.level = debug.level,
    seed = seed,
    adaptive_alpha = adaptive_alpha,
    alpha_update_rate = alpha_update_rate,
    adaptive_batch = adaptive_batch,
    batch_control = batch_control,
    min_adapt_samples = min_adapt_samples,
    alpha_min = alpha_min,
    stability_mode = stability_mode
  )

  # -------------------------------------------------
  # Attach system metadata
  # -------------------------------------------------

  res$method <- "MCIS_System_gsys"
  res$sys_structure <- sys_structure
  res$smooth_kappa <- smooth_kappa
  res$n_lsfs <- sys_obj$n_lsfs
  res$n_vars <- sys_obj$n_vars

  if (debug.level >= 1) {
    message("[MC_IS_gsys] finished.")
  }

  return(res)
}


#' @title Internal weight accumulation for Monte Carlo Importance Sampling (System Case)
#'
#' @description
#' Internal helper used by `MC_IS_system()` to accumulate importance sampling
#' weights and update running Monte Carlo statistics.
#'
#' The function performs numerically stable accumulation of:
#' \itemize{
#'   \item \eqn{\sum w},
#'   \item \eqn{\sum I w},
#'   \item \eqn{\sum w^2},
#'   \item and (in robust mode) their log-domain equivalents.
#' }
#'
#' Two stabilization modes are supported:
#'
#' \describe{
#'   \item{\code{"fast"}}{
#'     Batch-level stabilization using a max-log shift:
#'     \deqn{w_i = \exp(\log w_i - \max(\log w))}
#'     Suitable for most engineering applications.
#'   }
#'   \item{\code{"robust"}}{
#'     Global log-sum-exp accumulation across iterations:
#'     \deqn{\log \sum w = m + \log \sum \exp(\log w - m)}
#'     Recommended for rare-event simulation or very small failure probabilities.
#'   }
#' }
#'
#' The function updates the running state object and returns it.
#'
#' @param all_log_w Numeric vector of log-weights for the current batch.
#' @param all_I_vals Numeric vector of failure indicators (0/1).
#' @param stability_mode Character string: `"fast"` or `"robust"`.
#' @param sys_type Character string: `"serial"` or `"parallel"`.
#' @param state List containing current accumulated sums:
#'   \describe{
#'     \item{total_sum_w}{Cumulative sum of weights.}
#'     \item{total_sum_Iw}{Cumulative weighted failure sum.}
#'     \item{total_sum_w2}{Cumulative sum of squared weights.}
#'     \item{log_total_sum_w}{Log-domain cumulative weight sum (robust mode).}
#'     \item{log_total_sum_Iw}{Log-domain cumulative failure-weight sum (robust mode).}
#'     \item{log_total_sum_w2}{Log-domain cumulative squared-weight sum (robust mode).}
#'     \item{n_sim}{Total number of samples processed.}
#'   }
#'
#' @return Updated `state` list with accumulated statistics.
#'
#' @details
#' This function is internal and not intended to be called directly by users.
#' It is part of the system-level importance sampling implementation
#' in `MC_IS_system()`.
#'
#' @keywords internal
compute_weight_update <- function(all_log_w,
                                  all_I_vals,
                                  stability_mode,
                                  sys_type,
                                  state) {
  if (stability_mode == "fast" || sys_type == "serial") {
    max_log_w <- max(all_log_w)
    shifted_weights <- exp(all_log_w - max_log_w)

    sum_w_iter <- sum(shifted_weights)
    sum_Iw_iter <- sum(shifted_weights[all_I_vals == 1])
    sum_w2_iter <- sum(shifted_weights^2)

    state$total_sum_w <- state$total_sum_w + sum_w_iter
    state$total_sum_Iw <- state$total_sum_Iw + sum_Iw_iter
    state$total_sum_w2 <- state$total_sum_w2 + sum_w2_iter
    state$n_sim <- state$n_sim + length(shifted_weights)

    return(state)
  }

  if (stability_mode == "robust" && sys_type == "parallel") {
    log_w_batch <- all_log_w

    max_log_w <- max(log_w_batch)
    log_sum_w_batch <- max_log_w +
      log(sum(exp(log_w_batch - max_log_w)))

    if (any(all_I_vals == 1, na.rm = TRUE)) {
      log_Iw_batch <- log_w_batch[all_I_vals == 1]
      max_log_Iw <- max(log_Iw_batch)
      log_sum_Iw_batch <- max_log_Iw +
        log(sum(exp(log_Iw_batch - max_log_Iw)))
    } else {
      log_sum_Iw_batch <- -Inf
    }

    log_w2_batch <- 2 * log_w_batch
    max_log_w2 <- max(log_w2_batch)
    log_sum_w2_batch <- max_log_w2 +
      log(sum(exp(log_w2_batch - max_log_w2)))

    max_global_w <- max(state$log_total_sum_w, log_sum_w_batch)
    state$log_total_sum_w <- max_global_w +
      log(exp(state$log_total_sum_w - max_global_w) +
        exp(log_sum_w_batch - max_global_w))

    max_global_Iw <- max(state$log_total_sum_Iw, log_sum_Iw_batch)
    state$log_total_sum_Iw <- max_global_Iw +
      log(exp(state$log_total_sum_Iw - max_global_Iw) +
        exp(log_sum_Iw_batch - max_global_Iw))

    max_global_w2 <- max(state$log_total_sum_w2, log_sum_w2_batch)
    state$log_total_sum_w2 <- max_global_w2 +
      log(exp(state$log_total_sum_w2 - max_global_w2) +
        exp(log_sum_w2_batch - max_global_w2))

    state$n_sim <- state$n_sim + length(log_w_batch)

    return(state)
  }

  return(state)
}


#' Build a smooth system limit-state function (g_sys)
#'
#' Constructs a **global system limit-state function** from a list of
#' individual limit-state functions (`SYS_LSF` objects). The resulting
#' function aggregates all component limit states using a **smooth
#' approximation** of the minimum or maximum operator.
#'
#' The smooth aggregation improves numerical stability and differentiability,
#' which is beneficial for gradient-based reliability methods such as
#' FORM, SORM, and importance sampling algorithms.
#'
#' For a serial system the aggregation is
#'
#' \deqn{
#' g_{\kappa}(x) =
#' -\frac{1}{\kappa}\log
#' \left(
#' \sum_{i=1}^{m} \exp(-\kappa g_i(x))
#' \right)
#' }
#'
#' which approximates
#'
#' \deqn{
#' \min_i g_i(x).
#' }
#'
#' For a parallel system the aggregation is
#'
#' \deqn{
#' g_{\kappa}(x) =
#' \frac{1}{\kappa}\log
#' \left(
#' \sum_{i=1}^{m} \exp(\kappa g_i(x))
#' \right)
#' }
#'
#' which approximates
#'
#' \deqn{
#' \max_i g_i(x).
#' }
#'
#' The approximation becomes exact for
#' \eqn{\kappa \rightarrow \infty}.
#'
#' In addition to the system function, this helper also constructs a
#' **flattened list of marginal distributions** and a **variable index map**
#' that links the global variable vector to the variables of each
#' individual limit-state function.
#'
#' @param sys_input List of \code{SYS_LSF} objects defining the component
#'   limit-state functions of the system.
#'
#' @param sys_structure System aggregation structure.
#'   Supported options are
#'   \describe{
#'     \item{\code{"serial"}}{System fails if any component fails.}
#'     \item{\code{"parallel"}}{System fails only if all components fail.}
#'     \item{\code{function(g_vals)}}{Custom aggregation function applied
#'     to the vector of component limit-state values.}
#'   }
#'
#' @param smooth_kappa Positive numeric value controlling the smoothness
#'   of the aggregation. Larger values approach the exact \code{min}
#'   or \code{max} operator.
#'
#' @param debug.level Optional verbosity level for diagnostic output.
#'
#' @return A list containing:
#' \describe{
#'   \item{\code{g_sys}}{Callable system limit-state function
#'   \code{function(x)} operating on the global variable vector.}
#'
#'   \item{\code{distr_flat}}{Flattened list of marginal distribution
#'   objects corresponding to all variables in the system.}
#'
#'   \item{\code{var_map}}{List mapping each limit-state function to the
#'   indices of its variables in the global vector.}
#'
#'   \item{\code{n_vars}}{Total number of basic variables in the system.}
#'
#'   \item{\code{n_lsfs}}{Number of limit-state functions.}
#' }
#'
#' @details
#' This function is primarily used internally by reliability algorithms
#' that operate on a **single limit-state function**, such as
#' importance sampling or crude Monte Carlo simulation.
#'
#' By constructing a smooth global limit-state function, the system
#' reliability problem can be treated as a standard single-LSF problem,
#' enabling the reuse of algorithms originally developed for
#' component reliability.
#'
#' @keywords internal
#'
#' @seealso
#' \code{smooth_min}, \code{smooth_max},
#' \code{\link{MC_IS}}, \code{\link{SYS_LSF}}
#'
#' @author
#' (C) 2021-2026 K. Nille-Hauf, T. Feiri, M. Ricker, T. Lux -
#' Hochschule Biberach (until 2022),
#' TU Dortmund University - Chair of Structural Concrete (since 2023)
#'
build_gsys <- function(sys_input,
                       sys_structure = "serial",
                       smooth_kappa = 20,
                       debug.level = 0) {
  n_lsfs <- length(sys_input)

  is_custom <- is.function(sys_structure)

  if (n_lsfs < 1) {
    stop("build_gsys: no limit state functions provided.")
  }

  # -------------------------------------------------
  # 1. Extract LSF functions
  # -------------------------------------------------

  lsfs <- lapply(sys_input, function(obj) obj$getLSF())

  # -------------------------------------------------
  # 2. Build distribution list + variable mapping
  # -------------------------------------------------

  distr_flat <- list()
  var_map <- list()

  idx <- 1

  for (i in seq_along(sys_input)) {
    vars_i <- sys_input[[i]]$vars
    n_vars_i <- length(vars_i)

    var_map[[i]] <- idx:(idx + n_vars_i - 1)

    subdistr <- lapply(vars_i, function(v) {
      d <- v$getlDistr()

      if (is.list(d) && !is.null(d[[1]]$p)) {
        return(d[[1]])
      }

      if (is.list(d) && !is.null(d$p)) {
        return(d)
      }

      stop("build_gsys: invalid distribution object.")
    })

    distr_flat <- c(distr_flat, subdistr)

    idx <- idx + n_vars_i
  }

  n_vars <- length(distr_flat)


  # eval_lsfs <- function(x) {
  #   if (is.vector(x)) {
  #     g_vals <- numeric(n_lsfs)

  #     for (i in seq_len(n_lsfs)) {
  #       idxs <- var_map[[i]]

  #       xi <- as.numeric(x[idxs])

  #       g_vals[i] <- lsfs[[i]](xi)
  #     }

  #     return(g_vals)
  #   }

  #   if (is.matrix(x)) {
  #     n_samples <- nrow(x)

  #     G <- matrix(NA_real_, nrow = n_samples, ncol = n_lsfs)

  #     for (i in seq_len(n_lsfs)) {
  #       idxs <- var_map[[i]]

  #       Xi <- x[, idxs, drop = FALSE]

  #       G[, i] <- lsfs[[i]](Xi)
  #     }

  #     return(G)
  #   }

  #   stop("Unsupported input type in eval_lsfs")
  # }
  eval_lsfs <- function(x) {
    if (is.vector(x)) {
      g_vals <- numeric(n_lsfs)

      for (i in seq_len(n_lsfs)) {
        xi <- as.numeric(x[var_map[[i]]])

        g_vals[i] <- lsfs[[i]](xi)
      }

      return(g_vals)
    }

    if (is.matrix(x)) {
      n_samples <- nrow(x)

      G <- matrix(NA_real_, n_samples, n_lsfs)

      for (i in seq_len(n_lsfs)) {
        Xi <- x[, var_map[[i]], drop = FALSE]

        G[, i] <- lsfs[[i]](Xi)
      }

      return(G)
    }

    stop("Unsupported input type in eval_lsfs")
  }

  # -------------------------------------------------
  # 3. Build system LSF
  # -------------------------------------------------

  g_sys <- function(x) {
    G <- eval_lsfs(x)

    if (is.null(dim(G))) {
      if (is_custom) {
        return(sys_structure(G))
      }

      if (sys_structure == "serial") {
        m <- min(G)

        return(
          m - log(sum(exp(-smooth_kappa * (G - m)))) / smooth_kappa
        )
      }

      if (sys_structure == "parallel") {
        m <- max(G)

        return(
          m + log(sum(exp(smooth_kappa * (G - m)))) / smooth_kappa
        )
      }

      stop("Unsupported system structure.")
    }

    if (is.matrix(G)) {
      if (is_custom) {
        return(apply(G, 1, sys_structure))
      }

      if (sys_structure == "serial") {
        # m <- G[, 1]

        # if (ncol(G) > 1) {
        #   for (j in 2:ncol(G)) {
        #     m <- pmin(m, G[, j])
        #   }
        # }

        m <- G[, 1]
        for (j in seq_len(ncol(G) - 1) + 1) {
          m <- pmin(m, G[, j])
        }

        E <- exp(pmax(-smooth_kappa * (G - m), -700))
        S <- rowSums(E)

        S <- pmax(S, .Machine$double.xmin)

        res <- m - log(S) / smooth_kappa

        bad <- !is.finite(res)
        if (any(bad)) res[bad] <- m[bad]

        return(res)
      }

      if (sys_structure == "parallel") {
        # m <- G[, 1]

        # if (ncol(G) > 1) {
        #   for (j in 2:ncol(G)) {
        #     m <- pmax(m, G[, j])
        #   }
        # }
        m <- G[, 1]
        for (j in seq_len(ncol(G) - 1) + 1) {
          m <- pmax(m, G[, j])
        }

        E <- exp(pmin(smooth_kappa * (G - m), 700))
        S <- rowSums(E)

        S <- pmax(S, .Machine$double.xmin)

        res <- m + log(S) / smooth_kappa

        bad <- !is.finite(res)
        if (any(bad)) res[bad] <- m[bad]

        return(res)
      }

      stop("Unsupported system structure.")
    }

    stop("Unsupported system structure.")
  }

  # -------------------------------------------------
  # 4. Return structure
  # -------------------------------------------------

  list(
    g_sys = g_sys,
    distr_flat = distr_flat,
    var_map = var_map,
    n_vars = n_vars,
    n_lsfs = n_lsfs
  )
}

library(testthat)
library(microbenchmark)

testthat::local_edition(3)



# -----------------------------------------------------------------
# Check RNG initialization and helper functions
# -----------------------------------------------------------------
test_that("init_rng_master creates a consistent stream", {
  s1 <- init_rng_master(1234)
  s2 <- init_rng_master(1234)
  expect_equal(s1, s2)

  # Verify that .Random.seed is actually set
  expect_true(!is.null(.Random.seed))
})


test_that("record_step inserts values correctly", {
  rec <- make_recorders(5)
  rec$cov <- numeric(5) # für completeness
  rec <- record_step(rec,
    idx = 2,
    pf_mc = 0.1, var_mc = 0.02,
    I_mc = 3, n_sim = 200,
    tic = proc.time()
  )
  expect_equal(rec$pf[2], 0.1)
  expect_equal(rec$n_sim[2], 200)
  expect_true(all(!is.na(rec$time$user[2])))
})


test_that("make_ready_cluster returns a functional cluster", {
  libPaths <- .libPaths()
  cl <- make_ready_cluster(
    use_threads = 2,
    libPaths_local = libPaths,
    seed = 123,
    export_objs = character(0)
  )
  expect_true(inherits(cl, "cluster"))

  # Test call: each worker should have the same initial seed
  out <- parallel::parLapply(cl, 1:2, function(i) .Random.seed[1])
  expect_equal(out[[1]], out[[2]]) # identical start seed per worker

  parallel::stopCluster(cl)
})


# -----------------------------------------------------------------
# Performance benchmark test for new vs old implementation
# -----------------------------------------------------------------
test_that("new implementation is not slower than old one", {
  set.seed(321)

  n_batch <- 2e3
  n_vars <- 4

  fx_i <- matrix(runif(n_batch * n_vars, 0.1, 2), nrow = n_batch)
  hv_i <- matrix(runif(n_batch * n_vars, 0.1, 2), nrow = n_batch)

  mb <- microbenchmark(
    alt = {
      fx_prod <- apply(fx_i, 1, prod)
      hv_prod <- apply(hv_i, 1, prod)
      qt <- fx_prod / hv_prod
    },
    neu = {
      fx_prod <- rowSums(log(fx_i))
      hv_prod <- rowSums(log(hv_i))
      qt <- exp(fx_prod - hv_prod)
    },
    times = 15
  )

  # Expectation: median(new) ≤ median(old)*0.9 (≈10% faster)
  med_alt <- median(mb$time[mb$expr == "alt"])
  med_neu <- median(mb$time[mb$expr == "neu"])
  expect_lte(med_neu, med_alt * 0.9)
})


# -----------------------------------------------------------------
# Example Ex.5.11 from Nowak & Collins (2000)
# -----------------------------------------------------------------
test_that("Ex. 5.11 Nowak & Collins (fast version)", {
  if (.Platform$OS.type == "windows") {
    backend <- "future"
  } else {
    backend <- "parallel"
  }
  use_threads <- 1L # CI stabil

  X1 <- PROB_BASEVAR(
    Id = 1, Name = "Z",
    DistributionType = "norm",
    Mean = 100, Cov = 0.04
  )

  X2 <- PROB_BASEVAR(
    Id = 2, Name = "Fy",
    DistributionType = "lnorm",
    Mean = 40, Cov = 0.1
  )

  X3 <- PROB_BASEVAR(
    Id = 3, Name = "M",
    DistributionType = "gumbel",
    Package = "evd",
    Mean = 2000, Cov = 0.1
  )

  lsf1 <- SYS_LSF(vars = list(X1, X2, X3))
  lsf1$func <- function(M, Fy, Z) Z * Fy - M
  lsf1$check()

  form_rf <- PROB_MACHINE("FORM", "FORM")
  sorm <- PROB_MACHINE("SORM", "SORM")

  mcis <- PROB_MACHINE("MCIS", "MC_IS",
    options = list(
      cov_user = 0.05,
      n_max = 200000, # statt Millionen
      n_batch = 20000,
      use_threads = use_threads,
      backend = backend,
      seed = 1234
    )
  )

  ps <- SYS_PROB(
    sys_input = list(lsf1),
    probMachines = list(form_rf, sorm, mcis)
  )

  ps$runMachines()

  expect_length(ps$beta_single, 3)
  expect_false(any(is.na(ps$beta_single)))

  # FORM == SORM bei linearer LSF
  expect_equal(ps$beta_single[1], ps$beta_single[2], tolerance = 1e-8)

  # Plausibilität: Beta zwischen 3 und 5
  expect_gt(ps$beta_single[3], 3)
  expect_lt(ps$beta_single[3], 5)
})


# -----------------------------------------------------------------
# Monte Carlo Importance Sampling with many input variables -------
# -----------------------------------------------------------------
test_that("MCIS and high number of inputs", {
  if (.Platform$OS.type == "windows") {
    backend <- "future"
  } else {
    backend <- "parallel"
  }
  # Helper function to create a normal variable ------------------
  make_norm_var <- function(id, mean, sd) {
    PROB_BASEVAR(
      Id = id,
      Name = paste0("u", id),
      DistributionType = "norm",
      Mean = mean,
      Sd = sd
    )
  }

  # Create a set of 26 variables – the first has a different mean ------
  vars <- lapply(1:26, function(i) make_norm_var(i, 0.2, 0.1))
  vars[[1]]$Mean <- 0.5

  # LSF -----------------------------------------------------------
  lsf4 <- SYS_LSF(vars = vars, name = "RackwitzProb.")
  lsf4$func <- function(x) {
    x[1] - sum((x[-1]^2) / seq_along(x)[-1])
  }

  lsf4$check()

  # MC–IS machine(no fixed seed → new RNG stream each run)
  mcis <- PROB_MACHINE(
    name = "MC IS",
    fCall = "MC_IS",
    options = list(
      cov_user = 0.08,
      n_max = 400000,
      seed = 1234,
      backend = backend
    )
  )

  ps <- SYS_PROB(
    sys_input = list(lsf4),
    probMachines = list(mcis),
    debug.level = 0
  )

  ps$runMachines()

  # --------------------------------------------------------------
  # Plausibility checks
  # --------------------------------------------------------------
  # (c) MC‑IS – maximum 2% relative deviation allowed

  rel_err_1 <- abs(ps$beta_single[1] - 3.337) / 3.337

  message("Check: MC-IS must not deviate more than 2% from the reference beta")
  expect_lt(rel_err_1, 0.07)

  message("Check: Calculated beta must be less than 3.5")
  expect_lt(ps$beta_single[1], 3.5)

  message("Check: Calculated beta should be greater than 3.2")
  expect_gt(ps$beta_single[1], 3.1)

  expect_false(any(is.na(ps$beta_single)),
    info = "beta_single darf keine NA-Werte enthalten"
  )

  expect_length(ps$beta_single, 1)
})


# -----------------------------------------------------------------
# Simple linear LSF using FORM/SORM/MCC comparison ----------------
# -----------------------------------------------------------------
test_that("Simple linear LSf FORM SORM and MCC", {
  if (.Platform$OS.type == "windows") {
    backend <- "future"
  } else {
    backend <- "parallel"
  }
  # Fallback for Windows
  if (.Platform$OS.type == "windows") {
    use_threads <- 1L
  } else {
    use_threads <- 16L
  }

  var1 <- PROB_BASEVAR(Name = "E", Description = "Effect", DistributionType = "norm", Mean = 10, Sd = 0.2)
  var2 <- PROB_BASEVAR(Name = "R", Description = "Resistance", DistributionType = "norm", Mean = 35, Sd = 0.5)


  lsf <- SYS_LSF(vars = list(var1, var2), name = "LSF XY")
  lsf$func <- function(E, R) {
    R - E^1.5
  }
  # you can run this check to see if all transformations of the variables worked, all needed data is available and the set of vars fits to the limit state function
  lsf$check()


  machine1 <- PROB_MACHINE(name = "FORM Rack.-Fieß.", fCall = "FORM", options = list("n_optim" = 20, "loctol" = 0.001, "optim_type" = "rackfies"))
  machine2 <- PROB_MACHINE(name = "MC CoV 0.05", fCall = "MC_CRUDE", options = list("n_max" = 1e7, "cov_user" = 0.05, "use_threads" = use_threads, "seed" = 1234, backend = backend))
  machine3 <- PROB_MACHINE(name = "SORM", fCall = "SORM")

  ps <- SYS_PROB(
    sys_input = list(lsf),
    probMachines = list(machine1, machine2, machine3),
    debug.level = 0
  )


  ps$runMachines()

  expect_equal(ps$beta_single[1], 3.11578032380995) # FORM

  if (.Platform$OS.type == "windows") {
    rel_err <- abs(ps$beta_single[2] - 3.11578032380995) / 3.11578032380995
    expect_lt(rel_err, 0.02) # < 2 % Abweichung erlaubt
  } else {
    rel_err <- abs(ps$beta_single[2] - 3.09023230616781) / 3.09023230616781
    expect_lt(rel_err, 0.02) # < 2 % Abweichung erlaubt
  }

  expect_equal(ps$beta_single[3], 3.11578032380995) # SORM
})


test_that("MC_IS works with empirical distribution (small test)", {
  if (.Platform$OS.type == "windows") {
    backend <- "future"
  } else {
    backend <- "parallel"
  }

  skip_if_not_installed("EnvStats")

  set.seed(1)

  data_small <- rnorm(5000, 100, 4)

  X <- PROB_BASEVAR(
    Id = 1,
    Name = "Z",
    DistributionType = "emp",
    Package = "EnvStats",
    DistributionParameters = data_small
  )

  lsf <- SYS_LSF(vars = list(X), name = "Simple test")
  lsf$func <- function(Z) Z - 95
  lsf$check()

  mcis <- PROB_MACHINE(
    name = "MC IS",
    fCall = "MC_IS",
    options = list(
      cov_user = 0.05,
      n_max = 50000,
      seed = 1234,
      backend = backend
    )
  )

  ps <- SYS_PROB(
    sys_input = list(lsf),
    probMachines = list(mcis)
  )

  ps$runMachines()

  expect_false(is.na(ps$beta_single))
  expect_true(is.finite(ps$beta_single))
})

#' Student t Distribution
#'
#' Density, distribution function, quantile function and random generation
#' for the Student t distribution.
#'
#' The cumulative distribution function is given by
#' \deqn{
#' F_X(x) =
#' F_{t_{\nu}}\left(
#' \frac{x-m}{s}
#' \sqrt{\frac{n}{n+1}}
#' \right),
#' }
#' where \eqn{F_{t_{\nu}}(\cdot)} denotes the cumulative distribution
#' function of a Student t distribution with \eqn{\nu} degrees of freedom.
#'
#' The Student t distribution arises as the marginal
#' distribution of a normal distributed variable with conjugate
#' normal-gamma prior uncertainty in \eqn{m} and \eqn{s}.
#'
#' The hyperparameter vector \code{hyper.param} is defined as
#' \deqn{(m, s, n, \nu)}
#' where
#' \itemize{
#'   \item \eqn{m} is the mean value of an equivalent sample of size \eqn{n},
#'   \item \eqn{s} is the empirical standard deviation of an equivalent sample
#'         of size \eqn{\nu + 1},
#'   \item \eqn{\nu} denotes the degrees of freedom of the Student t distribution.
#' }
#'
#' For details see:
#' \url{https://www.rdocumentation.org/packages/stats/versions/3.6.2/topics/TDist}
#'
#' @name StudentT
#' @aliases StudentT st
#'
#' @param x,q Vector of quantiles.
#' @param p Vector of probabilities.
#' @param n_vals Number of random values to generate.
#' @param hyper.param Numeric vector \eqn{(m, s, n, \nu)}.
#'
#' @return
#' \itemize{
#'   \item \code{dst()} returns the density.
#'   \item \code{pst()} returns the distribution function.
#'   \item \code{qst()} returns the quantile function.
#'   \item \code{rst()} generates random deviates.
#' }
#'
#' @examples
#' hp <- c(40, 4, 3, 10)
#' dst(0.5, hp)
#'
#' @author
#' (C) 2021-2026 K. Nille-Hauf, T. Feiri, M. Ricker, T. Lux --
#' Hochschule Biberach (until 2022), TU Dortmund University - Chair of Structural Concrete (since 2023)


#' @rdname StudentT
#' @export
dst <- function(x, hyper.param){
  m <- hyper.param[1]
  s <- hyper.param[2]
  n <- hyper.param[3]
  v <- hyper.param[4]

  s_trans <- s/sqrt(n/(n+1))
  stats::dt((x - m)/s_trans, df = v) / s_trans
}

#' @rdname StudentT
#' @export
pst <- function(q, hyper.param){
  m <- hyper.param[1]
  s <- hyper.param[2]
  n <- hyper.param[3]
  v <- hyper.param[4]

  s_trans <- s/sqrt(n/(n+1))
  stats::pt((q - m)/s_trans, df = v)
}

#' @rdname StudentT
#' @export
qst <- function(p, hyper.param){
  m <- hyper.param[1]
  s <- hyper.param[2]
  n <- hyper.param[3]
  v <- hyper.param[4]

  s_trans <- s/sqrt(n/(n+1))
  m + s_trans * stats::qt(p, df = v)
}

#' @rdname StudentT
#' @export
rst <- function(n_vals, hyper.param){
  m <- hyper.param[1]
  s <- hyper.param[2]
  n <- hyper.param[3]
  v <- hyper.param[4]

  s_trans <- s/sqrt(n/(n+1))
  m + s_trans * stats::rt(n_vals, df = v)
}

#' @name MVFOSM
#' @title Mean-Value First-Order Second-Moment (MVFOSM)
#' @description
#' Mean-Value First-Order Second-Moment (MVFOSM) method for the
#' approximation of failure probabilities in structural reliability analysis.
#'
#' The MVFOSM method linearises the limit-state function at the mean values
#' of the basic random variables and estimates the reliability index
#' and probability of failure based on first-order moment information.
#'
#' This classical approach provides a computationally efficient
#' approximation but may be inaccurate for strongly nonlinear
#' limit-state functions.
#'
#' @param lsf LSF Definition, can be Expression or Function. Defined by the FLAG isExpression (see below)
#' @param lDistr List of Distributions
#' @param h If isExpression is False, than Finite Difference Method is used for partial deviation. h is the Windowsize
#' @param isExpression Boolean, If TRUE lsf has to be typeof expression, otherwise lsf has to be type of function()
#' @param debug.level If 0 no additional info if 2 high output during calculation
#'
#' @return MVFOSM returns an object containing the following elements:
#' \itemize{
#'   \item \code{beta}: Estimated reliability index.
#'   \item \code{pf}: Estimated probability of failure.
#'   \item \code{design.point}: Design point in the original \eqn{x}-space.
#'   \item \code{alphas}: Direction cosines (importance factors) in the standard normal space.
#'   \item \code{runtime}: Total runtime of the algorithm.
#' }
#' @references
#' Freudenthal, A. M. (1956).
#' Safety and the probability of structural failure.
#' \emph{Transactions of the American Society of Civil Engineers},
#' 121, 1337-1397.
#' @author (C) 2021-2026 K. Nille-Hauf, T. Feiri, M. Ricker, T. Lux -- Hochschule Biberach (until 2022),
#' TU Dortmund University - Chair of Structural Concrete (since 2023)
#' @export

MVFOSM <- function(lsf,
                   lDistr,
                   h=0.0001,
                   isExpression=FALSE,
                   debug.level){

  debug.TAG <- "MVFOSM"
  debug.print(debug.level,debug.TAG,c(TRUE), msg="Mean Value First Order Second Moment Method started...")
  tic<-proc.time()

  n_vars <- length(lDistr)
  m <- vector("numeric",n_vars)
  sd <- vector("numeric",n_vars)
  dx <- vector("numeric", n_vars)

  # Calculate Sd
  for (i in 1:n_vars){
    m[i] <- lDistr[[i]][[2]][1]
    sd[i] <- lDistr[[i]][[2]][2]
  }

  if(isExpression){
    s <- 0
    for(i in 1:n_vars){
      assign(lDistr[[i]][[1]]$name,m[i])
      d <- stats::D(lsf,lDistr[[i]][[1]]$name)
      dx[i] <- (eval(d))
      s <- s + (dx[i]^2)*(sd[i]^2)
    }
    sum_sd <- sqrt(s)
    lsf_mean <- eval(lsf)


  }else{
    finite_diff <- function(lsf,x,h){
      for (i in 1:length(x)) {
        x_lowerbound <- x
        x_upperbound <- x
        x_lowerbound[i] <- x[i]-h
        x_upperbound[i] <- x[i]+h

        dx_low <- lsf(x_lowerbound)
        dx_high <- lsf(x_upperbound)
        dx[i] <- (dx_high - dx_low)/(2*h)
      }
      return(dx)
    }
    dx <- vector("numeric", n_vars)
    dx <- finite_diff(lsf,m,h)
    s<-0
    for (i in 1:n_vars) {
      s <- s + ((dx[i]^2) * (sd[i]^2))
    }
    sum_sd <- sqrt(s)

    # Evaluate LSF with means
    lsf_mean <- lsf(m)
  }

  # Estimate beta and pf
  beta <- lsf_mean/sum_sd
  pf <- stats::pnorm(-beta,0,1)

  alpha <- vector("numeric",n_vars)
  # Get alphas
  for(i in 1:n_vars)
  {
    alpha[i] <- -(dx[i]*sd[i])/sum_sd
  }

  designPoint <- vector("numeric",n_vars)
  # Get designpoints
  for(i in 1:n_vars){
    designPoint[i] <- m[i]-(alpha[i]*beta*sd[i])
  }

  cat("\n")
  duration<-proc.time()-tic

  output <- list(
    "method"="MVFOSM",
    "beta"=beta,
    "pf"=pf,
    "design.point.x"=designPoint,
    "alpha"=alpha,
    "runtime"=duration[1:5]
  )

  debug.print(debug.level,debug.TAG,c(duration), msg="Mean Value First Order Second Moment Method finished in: ")
  return(output)

  return(output)
}

---
title: "TesiproV – Reliability Analysis Methods for Structural Engineering"
author: "Konstantin Nille-Hauf, Tania Feiri, Marcus Ricker, Til Lux"
date: "`r format(Sys.Date(), '%Y-%m-%d')`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{TesiproV – Reliability Analysis Methods for Structural Engineering}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

This document describes how the TesiproV package can be used. The package was developed as part of the "TesiproV" research project (sponsored by BMWi, the German Federal Ministry for Economic Affairs and Energy) and is used for the probabilistic safety assessment of structural components and systems. The source code of the package is freely available as open source.


# Introduction

The **TesiproV** package provides a flexible and extensible framework for *probabilistic reliability analysis* of structural components and systems.  
It enables engineers and researchers to quantify the safety level of structures by evaluating the probability of failure under uncertain loads, material properties and geometric parameters.

In classical deterministic design procedures, reliability is ensured through partial safety factors that implicitly account for uncertainties.  
However, such approaches often lack transparency regarding the actual reliability achieved.  
Probabilistic methods overcome this limitation by explicitly modelling all relevant sources of uncertainty — e.g., variability in material strength, load effects or model uncertainties — as random variables with defined statistical distributions.

The fundamental concept behind reliability analysis is the *limit‑state function* $g(X) = 0$, which separates safe from failed states:

$$
g(X) = R(X) - E(X) = 0
$$

where $R(X)$ represents the resistance and $E(X)$ denotes the load effect.  
Failure occurs when $g(X) < 0$. The random vector $X = [X_1, X_2, …]$ contains all basic variables describing these uncertainties.

The probability of failure can then be expressed as:

$$
P_f = P[g(X) < 0]
$$

and converted into a *reliability index* β via

$$
β = -Φ^{-1}(P_f)
$$

where Φ denotes the cumulative standard normal distribution.  
Both the reliability index and the probability of failure are measures used to assess and compare reliability.
The calculated probability of failure differs from the actual probability of failure, as the governing factor for the failure of a component or structure — human error — is not captured in the analysis.

TesiproV implements several algorithms to estimate these quantities efficiently:

- **MVFOSM** — Mean Value First Order Second Moment method; fast linear approximation based on moment statistics.
- **FORM** — First Order Reliability Method; iterative search for design point on failure surface.
- **SORM** — Second Order Reliability Method; curvature correction around design point.
- **Monte Carlo simulations** — direct numerical evaluation using random sampling or importance/subset sampling schemes.

These methods allow both single limit‑state analyses and complex system evaluations involving multiple interacting components.  
In addition, TesiproV supports parametric studies where selected input parameters are systematically varied to investigate sensitivity or perform probabilistic optimisation.

Since version 0.9.5 the entire code base has been refactored to improve consistency between probabilistic objects (`PROB_BASEVAR`, `SYS_LSF`, `SYS_PROB`, …), error handling in limit‑state functions (`$check()`), transformations between mean/standard deviation and native distribution parameters (`$prepare()`), and parallel execution via the **future** package was added.

Beyond its computational capabilities, TesiproV aims to provide a transparent educational tool for understanding reliability theory in structural engineering practice.  
Its object‑oriented design makes it straightforward to define new limit‑state functions, integrate additional distributions or link external solvers, e.g., for advanced Monte Carlo techniques.

The project was originally initiated at Hochschule Biberach in cooperation with the Institue of Structural Concrete at RWTH Aachen University within a BMWi research program (*Allgemeingültiges Verfahren zur Herleitung von Teilsicherheitsbeiwerten im Massivbau auf Basis probabilistischer Verfahren anhand ausgewählter Versagensarten (TesiproV) – Erstellung eines Richtlinienentwurfs*, FKZ 03TNK003E) and is now maintained by the Chair of Structural Concrete at TU Dortmund University.

Source code repository: <https://gitlab.tu-dortmund.de/ls-massivbau/tesiprov>  
License: [MIT](https://opensource.org/licenses/MIT)


# Installation

Official releases of the R package TesiproV are available on CRAN (<https://CRAN.R-project.org/package=TesiproV>) and can be installed in R as: 

```{r, eval = FALSE}
install.packages("TesiproV")
```

You can install a pre-release version of TesiproV directly from the GitLab server at TU Dortmund University:

```{r, eval = FALSE}
remotes::install_gitlab("ls-massivbau/tesiprov", host = "gitlab.tu-dortmund.de")
```

If you need any permissions, please contact <tesiprov.lmb.ab@tu-dortmund.de>.

Make sure to initialize the TesiproV package before using it:

```{r}
library("TesiproV")
```

## Conceptual background

Reliability analysis quantifies how likely a structure will perform safely throughout its lifetime despite inherent uncertainties.
Typical sources of uncertainty include variations in geometry (e.g., effective depth), material properties (e.g., concrete strength),
loading conditions (permanent vs variable actions), environmental influences and modelling simplifications.
Within TesiproV every uncertain quantity is represented by an object-oriented random variable definition.
Transformation between mean value / standard deviation and native distribution parameters is handled automatically,
ensuring consistent probabilistic input even when empirical data are incomplete.
System reliability can be evaluated either analytically using bounds (“SimpleBounds”) or numerically via simulation-based approaches.
Parallelisation through the parallel (base R) and future backend allows efficient computation even for large-scale Monte-Carlo analyses on multi-core machines.


# Basic building blocks

## Random variable objects

A basic random variable describes an uncertain quantity used in a limit-state function.
It is defined by its statistical distribution type (including parameters) or its statistical moments (mean value (*Mean*) and standard deviation (*Sd*)).
Supported distributions include:

| Distribution | Identifier | Package | Description |
|---------------|-------------|----------|--------------|
| Normal | `"norm"` | stats | Gaussian distribution defined by mean and standard deviation |
| Lognormal | `"lnorm"` | stats | Distribution of a variable whose logarithm is normally distributed |
| Lognormal | `"slnorm"` | brms | shifted lognormal distribution |
| Gumbel (Type I extreme value)  | `"gumbel"`  | evd  | Used for modeling maxima or minima (extreme values) |
| Gamma  | `"gamma"`  | stats  | positive-valued model defined by shape and scale parameters |
| Exponential  | `"exp"` | stats | special case of gamma with shape=1 |
|"Beta"| `"beta"`| stats | bounded on [0,1]; often used for proportions or probabilities |
|"Weibull"| `"weibull"`|stats| flexible positive-valued model for material strength or lifetime data|
|"Student-t"|`"st"`|TesiproV internal| heavy-tailed continuous; useful for sensitivity analyses|
|"Log-Student-t"| `"lt"`|TesiproV internal| extended heavy-tailed log-scale distribution|
|"Empirical"| `"emp"`| evd | Empirical Distribution based on given Data|
|"Binomial"| `"binom"`|stats| discrete count data; placeholder type|
|"Uniform"|`"unif"`|stats| constant probability across an interval; placeholder type|


Additional placeholders exist for future extensions.

### Example – PROB_BASEVAR

```{r}
var_fck <- PROB_BASEVAR(
  Id = 1,
  Name = "f_ck",
  Description = "Concrete compressive strength",
  DistributionType = "lnorm",
  Mean = 47.35,
  Sd = 5.86
)
```


* Id - numeric, for identification purpose 
* Name - Name of the var, must be the same as in the function input
* Description - Description for the output protocol
* Package - Packagename, which provides the distribution function (dname,pname,qname,rname; name=DistributionType) 
* DistributionType - Name of the distributionfunction (without the d, p, q or r)
* Mean - Expected value (first statistical moment)
* Sd - standard deviation (square root of second statistical moment)
* Cov - cofficient of variation: Cov = Mean/Sd
* X0 - location parameter, used for shifted log-normal distribution "slnorm"

Internally TesiproV automatically derives all missing distribution parameters such as shape or scale parameters from the given expected value (mean) and standard deviation (square root of variance) if possible.

Basic variable using the example of concrete compressive strength.
As distribution function a logStudentT-distribution adjusted within the package is used. For this distribution a d-, p-, q- and rlt-function is available, therefore Package = "TesiproV".

```{r}
var.f_ck <- PROB_BASEVAR(Id = 1, Name = "f_ck", Description = "Char. Betondruckfestigkeit", Package = "TesiproV", DistributionType = "lt", DistributionParameters = c(3.85, 0.09, 3, 10))
```

## Parametric random variables

To perform parameter studies over varying input values (e.g., different mean strengths or load levels), use `PARAM_BASEVAR`. It extends PROB_BASEVAR with fields defining which property should vary (`ParamType`) and which values are tested (`ParamValues`).

### Example – PARAM_BASEVAR

```{r}
pvar_d <- PARAM_BASEVAR(
  Id = 1,
  Name = "d",
  Description = "Effective depth",
  DistributionType = "norm",
  Sd = 10,
  ParamType = "Mean",
  ParamValues = c(seq(200, 800, 50), seq(1000, 3000, 100)) + 10
)
```

During each iteration of a parametric study the corresponding value from `ParamValues` is assigned automatically.

## Determistic variables

If certain quantities are treated as deterministic constants rather than random variables, they can be modelled using `PROB_DETVAR`. Internally these are represented as normal distributions with infinitesimal standard deviation so that all algorithms remain compatible.

### Example PROB_DETVAR

```{r}
var.gamma_c <- PROB_DETVAR(Id = 1, Name = "gamma_c", Description = "PSF Concrete", Value = 1.5)
```

During each iteration of a parametric study the corresponding value from `ParamValues` is assigned automatically.


## Parametric deterministic variables

Frequently, deterministic variables in a parameter calculation also depend on the varying parameter size. In these cases, a parametric deterministic variable can be created. With each iteration of the parametric base variable, the next step of the parametric deterministic variable is also used.

In this case, a normal distribution with infinitesimal standard deviation is also used internally. 

Note: Parametric variables are only considered for the parametric system object (SYS_PARAM) while the normal system object (SYS_PROB) only uses the default value.

### Example PARAM_DETVAR

```{r}
var.V_Ed <- PARAM_DETVAR(Id = 1, Name = "V_Ed", Description = "applied shear force", ParamValues = c(1, 2, 3, 4, seq(5, 10, 0.5)))
```


## Limit-state function object (SYS_LSF)

A limit-state function encapsulates the mathematical definition of failure. It can be provided either as an explicit R function or as an arithmetic expression converted internally into a callable closure.

Variables are used as input values for the functions. The naming in the function header (`function(E, R)`) must correspond to its use in the function body (`function(E,R){R-E}`). The naming must also match the name of the variable itself. Therefore, a `list()` with all used `PROB_BASEVAR()` objects is passed to the SYS_LSF() object (note: the objects with their variable names, here var1 and var2, are to be passed, not the variable name itself). In the case of multiple limit state functions being used, the function can be given a name to make it easier to assign in printouts later.
An objective function() can then be defined via the `$func()` property. Alternatively, `$expr()` can be used to assign an `expression(R-E)`.
By means of `$check()` an internal check run is performed, which ensures that all the base variables are complete and suitable.

### Example SYS_LSF

```{r}
var1 <- PROB_BASEVAR(Name = "E", Description = "Effect", DistributionType = "norm", Mean = 10, Sd = 1)
var2 <- PROB_BASEVAR(Name = "R", Description = "Resistance", DistributionType = "norm", Mean = 15, Sd = 1.5)


lsf <- SYS_LSF(vars = list(var1, var2), name = "LSF XY")
lsf$func <- function(E, R) {
  R - E
}
# you can run this check to see if all transformations of the variables worked, all needed data is available and the set of vars fits to the limit state function
lsf$check()
```

The `$check()` method validates that all associated variables are consistent before analysis begins.

## Reliability algorithms (PROB_MACHINE)

Reliability assessment methods are implemented in modular solver objects called `PROB_MACHINE`. Each machine defines one algorithm via its field `$fCall` (e.g., `"FORM"`, `"SORM"`, `"MC_CRUDE"`) together with optional settings stored in `$options`.

* Level 1 (partial safety factors)
* Level 2 (MVFOSM, FORM, SORM)
* Level 3 (MC_CRUDE, MCIS, MC_SubSam)

Levels 2 and 3 can be implemented by means of this package (in order to validate Level 1 procedures). The accuracy of the calculated failure probabilities - and ultimately, reliability indexes - increases through the levels.

A solution-oriented approach is always modelled in a so-called `PROB_MACHINE()`, i.e., an object that stands for a solution algorithm. Over the property `fCall`, the following reliability assessment methods are available:

Level 2 Methods:
* Mean Value First-Order Second Moment (MVFOSM) 
* Advanced First-Order Reliability Method (FORM)
* Second-Order Reliability Method (SORM)

Level 3 Methods:
* Crude Monte-Carlo Simulation (MC_CRUDE)
* Monte-Carlo Simulation with Importance Sampling (MC_IS)
* Monte-Carlo Simulation with Subset Sampling (MC_SubSam)

The property `$name` serves again for better identification in the later output protocol. A list() object in the `$options` property can then be used to specify special options according to the methods.


### MVFOSM

Options:

* h - Accuracy of partial derivatives (finite difference method).
* isExpression - FALSE/TRUE MVFOSM can handle both expressions and functions.

Example:

```{r}
machine <- PROB_MACHINE(name = "MVFOSM", fCall = "MVFOSM", options = list("isExpression" = FALSE, "h" = 0.1))
```
### FORM

Options:

* n_optim - max. amount of evaulation runs
* loctol - local accuracy of the optimisation algorithm (if lower then this threshold, the optim stops)
* OptimType - Algorithms one can choose: Rackwitz-Fiessler Algorithmus "rackfies" (quite stable) or Augmented Lagrange Solver "auglag" (fast but sometimes difficult)

Example:

```{r}
machine <- PROB_MACHINE(name = "FORM Rack.-Fies.", fCall = "FORM", options = list("n_optim" = 20, "loctol" = 0.001, "optim_type" = "rackfies"))

machine <- PROB_MACHINE(name = "FORM Lagrange.", fCall = "FORM", options = list("optim_type" = "auglag"))
```
### SORM

Options:

* No settings possible

Example:

```{r}
machine <- PROB_MACHINE(name = "SORM", fCall = "SORM")
```
### MC_Crude

Options:

* cov_user=0.025 - abort criterion for Cov.
* n_batch=100 - evaulations each step.
* n_max=1e7 - abort cirterion for max cycles of evaluation
* use_threads=6 - amount of threads for parallel computing (1 = single core calculation)
* backend - parallel backend, eather "future" or "parallel", use "future" on Windows
* dataRecord=TRUE - if true, there is a list() with all the data per calculation step in the result (used for visualization of beta)
* debug.level=0 - Integer verbosity level (0 = silent, 1 = summary, 2 = detailed).
* seed - Optional integer seed for reproducible rondom numbers. 

Example:

```{r}
machine <- PROB_MACHINE(name = "MC CoV 0.05", fCall = "MC_CRUDE", options = list("n_max" = 1e6, "cov_user" = 0.05))
```

### MC_IS

Options:

* cov_user=0.025 - abort criterion for cov.
* n_batch=100 - evaulations each step.
* n_max=1e7 - abort criterion for maximum cycles of evaluation
* use_threads=4 - amount of cores for parallel computing (1 = single core calculation).
* backend=NULL - parallel backend, either "future" or "parallel", use "future" on Windows.
* sys_type="serial" - Character string, either "serial" or "parallel", specifying system configuration.
* dataRecord=TRUE - If true, there is a list() with all the data per calculation step in the result (used for visualization of beta).
* beta_l=100 - optional threshold; limit states with β > beta_l may be excluded in system analysis.
* density Type - Sampling density type (currently "norm" supported).
* dps - Optional vector of design points in physical space; if supplied, FORM analysis is skipped.
* debug.level - Integer verbosity level (0 = silent, 1 = summary, 2 = detailed).
* seed - Optional integer seed for reproducible rondom numbers. 

Further options can be found in the documentation of MC_IS. Changing the options is not recommended for inexperienced users.

```{r, eval = FALSE}
help("MC_IS")
```

Example:

```{r}
machine <- PROB_MACHINE(name = "MC IS", fCall = "MC_IS", options = list("cov_user" = 0.05, "n_max" = 300000))
```
### MC_SubSam

Options:

* Nsubset=1e5 - number of samples in each simulation level.
* p0=0.1 - level probability or conditional probability.
* MaxSubsets=10 - maximum number of simulation levels that are used to terminate the simulation procedure to avoid infinite loop when the target domain cannot be reached.
* Alpha=0.05 - confidence level.
* variance="uniform" - "gaussian" or "uniform".
* debug.level - Integer verbosity level (0 = silent, 1 = summary, 2 = detailed).
* seed - Optional integer seed for reproducible rondom numbers. 

Example:

```{r}
machine <- PROB_MACHINE(name = "MC Subset Sampling", fCall = "MC_SubSam", options = list("variance" = "uniform"))
```


## System objects

The system object combines all the previous objects and organises the correct calculation of the mathematical model.
Two main classes exist:

* `SYS_PROB` – classical probabilistic system analysis; may contain several LSFs,
* `SYS_PARAM` – parametric sweep over one LSF using varying input parameters.

The `SYS_LSF` object already contains a list of the required basic variables. However, the system object still requires a list of functions to be calculated and the list of algorithms to be used.
In addition, the `$sys_type` property can be used to specify whether the equations have a series `$sys_type="serial"` or parallel `$sys_type="parallel"` relationship. 

The calculation process is then triggered with `$run_machines()`. In the console, the user is informed about the current state of the calculation. After completing the calculation, the results can be retrieved via the `SYS_PROB` object (in the example below `ps`). For more information see result interpretation.

You can set a debug.level field of SYS_PROB to 1 or 2, to get more information during the calculation, by default that value is zero.


### Example – single problem

```{r, results="hide"}
var1 <- PROB_BASEVAR(Name = "E", Description = "Effect", DistributionType = "norm", Mean = 10, Sd = 1)
var2 <- PROB_BASEVAR(Name = "R", Description = "Resistance", DistributionType = "norm", Mean = 15, Sd = 1.5)


lsf <- SYS_LSF(vars = list(var1, var2), name = "LSF XY")
lsf$func <- function(E, R) {
  R - E
}
# you can run this check to see if all transformations of the variables worked, all needed data is available and the set of vars fits to the limit state function
lsf$check()


machine <- PROB_MACHINE(name = "FORM Rack.-Fieß.", fCall = "FORM", options = list("n_optim" = 20, "loctol" = 0.001, "optim_type" = "rackfies"))

ps <- SYS_PROB(
  sys_input = list(lsf),
  probMachines = list(machine),
  debug.level = 0
)

ps$runMachines()
```

Retrieve results

```{r}
ps$beta_single
```

### Example – system calculations

For systems consisting of multiple LSFs connected either in series or in parallel:

```{r}
# Systemberechnungn (System aus zwei gleichen LSF´s)
ps_sys <- SYS_PROB(
  sys_input = list(lsf, lsf),
  probMachines = list(machine),
  sys_type = "serial"
)

ps_sys$runMachines()
ps_sys$calculateSystemProbability()
```

Resulting system reliability indices:

```{r}
ps_sys$beta_sys
```

## Parametric studies (SYS_PARAM)

Parametric analyses use the class `SYS_PARAM`. This extends `SYS_PROB` to iterate over varying input parameters defined by one or more `PARAM_BASEVAR` objects.

Example:

```{r}
# Force in kN
pvar_F_mod <- PARAM_BASEVAR(
  Name = "Force", DistributionType = "gumbel", Sd = 2, Package = "evd",
  ParamType = "Mean", ParamValues = c(17, 18, 19, 20) 
)

# yield strength in kN, for Mean the mean of the unshifted distribution must be specified.
fy <- PROB_BASEVAR(
  Name = "fy", DistributionType = "slnorm", Package = "brms", Mean = 265000, Sd = 25000, x0 = 160000
)

# plastic moment of resistance in m³ - considered deterministic in this example
W_pl <- 2.5e-4 
# length of cantilever beam in m - considered deterministic in this example
l    <- 2     

lsf_param <- SYS_LSF(vars = list(pvar_F_mod, fy), name = "Spaethe (1992) - Beispiel 2")
lsf_param$func <- function(Force,fy) {
  return(W_pl*fy-l*Force)
}

machine_form <- PROB_MACHINE(name = "FORM", fCall = "FORM")

ps_param <- SYS_PARAM(
  sys_input = list(lsf_param),
  probMachines = list(machine_form)
)

ps_param$runMachines()
print(ps_param$beta_params)
```

Results per iteration can be easily visualised:

```{r}
para_values <- c(17, 18, 19, 20) 
beta_values <- ps_param$beta_params
plot(para_values, beta_values, type = "l", xlab = "Force F [kN]", ylab = "Reliability index β [-]")
```

## Parallelisation in TesiproV
### General
TesiproV supports parallel execution for Monte Carlo-based reliability methods, including `MC_IS()` and `MC_CRUDE()`. The package provides two parallel backends:

* "parallel" -- based on base R's parallel package  
* "future" -- based on the future ecosystem

TesiproV does not modify the active future plan. Users are responsible for configuring parallel execution explicitly.

### Reproducibility and Determinism
All Monte Carlo simulations in TesiproV are designed to be:

* Deterministic for a fixed seed  
* Independent of the number of worker threads  
* Independent of the selected backend

This means that:
```{r, eval=FALSE}
MC_IS(..., seed = 123, use_threads = 1)
MC_IS(..., seed = 123, use_threads = 8)
```

will produce identical results.

Determinism is guaranteed by performing all random number generation in the master process. Worker processes do not generate random numbers.

### Backend Overview
1. The "parallel" Backend

* Uses `parallel::mclapply()` on Linux/macOS  
* Uses PSOCK clusters on Windows  
* Suitable for high-performance Linux clusters  
* No dependency on `future`

Example:

```{r, eval=FALSE}
MC_IS(..., backend = "parallel", use_threads = 8)
```

2. The "future" Backend
   
Delegates parallelisation to the future framework.
TesiproV does not change the global future plan. Users must configure parallel execution explicitly.

Linux / macOS example:
```{r, eval=FALSE}
library(future)
plan(multicore, workers = 8)

MC_IS(..., backend = "future")
```

Windows example:
```{r, eval=FALSE}
library(future)
plan(multisession, workers = 8)

MC_IS(..., backend = "future")If no future plan is set, execution will be sequential.
```

### Recommended Parallel Settings

Parallel efficiency depends strongly on the computational cost of the limit-state function (LSF).

| Problem Type               | Recommended Threads | Backend Recommendation      | Suggested `n_batch`      |
|----------------------------|--------------------|-----------------------------|--------------------------|
| Very simple LSF           | 1-2                | `"future"` or `"parallel"`  | 50,000 - 100,000         |
| Moderately expensive LSF  | 4-8                | `future(multicore)`         | 100,000 - 250,000        |
| Expensive LSF             | 8-16               | `future(multicore)`         | 250,000 - 500,000        |
| Very expensive LSF        | ~16                | `future(multicore)`         | ~500,000                 |


Using more than 16 threads on large shared-memory systems often leads to memory-bandwidth saturation.

### Batch Size Selection
Performance improves significantly when n_batch is large enough to amortize parallel overhead.

Typical observations:

* Small batches -> overhead dominates
* Medium batches -> optimal performance
* Very large batches -> memory-bound regime

For HPC workloads, a batch size between:

```{r, eval=FALSE}
250,000 -- 500,000
```

often provides near-optimal performance.

### Performance Characteristics
Parallel scaling depends on:

* LSF computational complexity
* Memory bandwidth
* NUMA architecture
* Batch size

For CPU-intensive LSFs, typical scaling on modern multi-core systems is:

* 2 threads -> ~1.6x speedup  
* 4 threads -> ~2.8-3.5x  
* 8 threads -> ~4-6×  
* 16 threads -> ~6-9×  
* 32 threads -> often limited by memory bandwidth


### HPC Usage Notes
On large Linux clusters (e.g., NUMA systems):

* `future(multicore)` typically provides the best performance  
* Avoid very small n_batch values  
* Prefer 8-16 threads for optimal efficiency  
* Use deterministic seeds for reproducible studies

Example HPC workflow:
```{r, eval=FALSE}
library(future)
plan(multicore, workers = 16)

res <- MC_IS(
  lsf = my_lsf,
  lDistr = my_distr,
  n_batch = 300000,
  backend = "future",
  seed = 12345
)
```

## Interpretation of results and post processing

After computation each system object contains structured result lists:
For regular problems (`SYS_PROB`):

* Single calculations -> `$beta_single`, `$res_single`,
* System calculations -> `$beta_sys`, `$res_sys`.

Here beta_* is the beta values summarised while res_* is the full result output via list(). The list `$res_*` is structured as follows `ps$res_*[[lsf]][[machine]]$value`.

For parametric studies (`SYS_PARAM`):

* Parameter iterations -> `$beta_params`, `$res_params`.

where the first list level of `$res_params` corresponds to the parameter iterations. The data can be then processed individually.

In addition, the results can be printed to a file using the command `$printResults(filename)`, where filename corresponds to the file name and the file is stored in the current working directory (see `getwd()` or `setwd()`).

The project can also be stored differently to allow later postprocessing in other files. For this, the command `$saveProject(level,filename)` is used. The following stages describe a different level of detail:

* 1. Only the beta values are stored.
* 2. The `$res_*` object is saved.
* 3. The entire `ps` object is stored.
* 4. An image of the current working memory is created, which can be reloaded later.

Stages 1 to 3 can be then read back into working memory using the `readRDS(filename)` command, while Stage 4 is reloaded using `load(filename)`.

Example:

```{r, eval = FALSE}
ps <- SYS_PARAM(
  sys_input = list(lsf),
  probMachines = list(machine)
)

ps$runMachines()

local_beta <- ps$beta
local_runtime <- ps$res_single[[1]][[1]]$runtime

# define path with setwd("...")
ps$printResults("result_file") # Prints a report file in .txt format with all informations about the calculation
ps$saveProject(4, "project_file") # stores the project in a file
```

The results can be reloaded later via:

```{r, eval = FALSE}
readRDS("example_proj_res.rds")
load("example_proj.RData")
```

Debug output verbosity is controlled by field $debug.level (0=silent … 2=detailed).

# Examples

## Two normally distributed variables
A limit state equation 20-(x1-x2)^2-8*(x1+x2-4)^3 is to be investigated, where X1 and X2 are normally distributed around the expected value of 0.25 and the standard deviation of 1. The example is taken from the documentation of the UQLab software. 

Firstly, two basic variables are created and assigned to the R variables Var1 and Var2. The first variable is named as "X1" and the second as "X2". These names are then reused in the definition of the limit state function (i.e., to compare the input variables of the function). In addition, both variables receive the indicator "norm" as distribution type (DistributionType) and the mean value defined above or the corresponding standard deviation.

Subsequently, a limit state function object (SYS_LSF) is assigned to the R variable lsf. A list of variables (Var1,Var2) is passed to the object and a name is assigned for later identification. In addition, the corresponding limit state equation is defined via the object property "func".

The R variable "FORM" is assigned a solution method of the Level 2 (FORM) without any additional parameters.

Finally, the modules are merged in the SYS_PROB object and the R variable "ps" and the calculation is started by means of the $runMachines() method.
In the following result log the intermediate results of the individual iteration steps are shown. Using the "ps" object, different results (here, for example, the beta value) can be retrieved after a successful calculation.


```{r, results="hide"}
Var1 <- PROB_BASEVAR(Id = 1, Name = "X1", DistributionType = "norm", Mean = 0.25, Sd = 1) # kN/m²
Var2 <- PROB_BASEVAR(Id = 2, Name = "X2", DistributionType = "norm", Mean = 0.25, Sd = 1) # m

lsf <- SYS_LSF(vars = list(Var1, Var2), name = "UQLab 2d_hat")
lsf$func <- function(X1, X2) {
  return(20 - (X1 - X2)^2 - 8 * (X1 + X2 - 4)^3)
}

form <- PROB_MACHINE(name = "FORM Rack.-Fieß.", fCall = "FORM")

ps <- SYS_PROB(
  sys_input = list(lsf),
  probMachines = list(form)
)
ps$runMachines()
```

```{r}
ps$beta_single
```

## Non-normally distributed basic variable
To demonstrate non-normally distributed basic variables, an example from Nowak & Collins (Example 5.11, Page 127/128) is calculated. Here normally distributed, lognormally distributed as well as gumbel distributed variables are used. While "norm" and "lnorm" are still contained in the standard scope of R (package "stats" belongs to base R) the Gumbel distribution is to be considered on the basis of the EVD-package (EVD := extreme value distributions).

```{r, results ='hide'}
library("evd")
X1 <- PROB_BASEVAR(Id = 1, Name = "Z", DistributionType = "norm", Mean = 100, Cov = 0.04) # kN/m²
X2 <- PROB_BASEVAR(Id = 2, Name = "Fy", DistributionType = "lnorm", Mean = 40, Cov = 0.1) # m
X3 <- PROB_BASEVAR(Id = 2, Name = "M", DistributionType = "gumbel", Package = "evd", Mean = 2000, Cov = 0.1) # m

lsf <- SYS_LSF(vars = list(X1, X2, X3), name = "Nowak & Collins Exp5.11_1")
lsf$func <- function(M, Fy, Z) {
  return(Z * Fy - M)
}

form <- PROB_MACHINE(name = "FORM Rack.-Fieß.", fCall = "FORM")

ps <- SYS_PROB(
  sys_input = list(lsf),
  probMachines = list(form)
)
ps$runMachines()
```

```{r}
ps$beta_single
```



## Multiple limit state functions in one system

Basically, the "SYS_PROB" object can calculate multiple LSF's using both Level 2 and Level 3 methods. This can be used when a system consists of several limit state functions and when the system reliability is to be determined using SimpleBounds. In this case, the failure probability must first be determined for each individual limit state equation. These results are then used to calculate bounds for the system failure probability. On the other hand, the development of a limit state function often poses several variants in equations or variables, so that by the simultaneous calculation of several equations satisfactory comparability is achieved and a high usability remains granted.

However, if the system failure probability can not to be sufficiently estimated by means of the SimpleBounds, simulation methods (e.g., MC-Crude) can be used to calculate the system failure probability more exactly. Then, the system can also be constructed according to the following scheme.

In this example, the above examples are first determined individually as a system using a FORM calculation and then the system failure probability is determined using different methods.



```{r, results="hide"}
Var1 <- PROB_BASEVAR(Id = 1, Name = "X1", DistributionType = "norm", Mean = 0.25, Sd = 1) # kN/m²
Var2 <- PROB_BASEVAR(Id = 2, Name = "X2", DistributionType = "norm", Mean = 0.25, Sd = 1) # m

lsf1 <- SYS_LSF(vars = list(Var1, Var2), name = "UQLab 2d_hat")
lsf1$func <- function(X1, X2) {
  return(20 - (X1 - X2)^2 - 8 * (X1 + X2 - 4)^3)
}


X1 <- PROB_BASEVAR(Id = 4, Name = "Z", DistributionType = "norm", Mean = 100, Cov = 0.04) # kN/m²
X2 <- PROB_BASEVAR(Id = 5, Name = "Fy", DistributionType = "lnorm", Mean = 40, Cov = 0.1) # m
X3 <- PROB_BASEVAR(Id = 6, Name = "M", DistributionType = "gumbel", Package = "evd", Mean = 2000, Cov = 0.1) # m

lsf2 <- SYS_LSF(vars = list(X1, X2, X3), name = "Nowak & Collins Exp5.11_1")
lsf2$func <- function(M, Fy, Z) {
  return(Z * Fy - M)
}

form <- PROB_MACHINE(name = "FORM Rack.-Fieß.", fCall = "FORM")

ps <- SYS_PROB(
  sys_input = list(lsf1, lsf2),
  sys_type = "serial",
  probMachines = list(form)
)
ps$runMachines()
```

```{r}
ps$beta_single

ps$calculateSystemProbability()
ps$beta_sys
```


The following lines are not executed because the calculation time is several minutes depending on the machine.
Results in the range of the above interval are to be expected. The results can be retrieved after the calculation using "$beta_sys".
```{r, eval = FALSE}
params <- list("cov_user" = 0.05, "n_max" = 50000)
ps$calculateSystemProbability("MC", params)

params <- list("cov_user" = 0.05, "n_max" = 50000)
ps$calculateSystemProbability("MCIS", params)

ps$calculateSystemProbability("MCSUS", params)
```

## Parameterstudy

* Punching according to DIN EN 1992-1 + NA(D)
* Effective depth (var.d) is parameterised (cf. vector at "ParamValues")
* Additional parametric deterministic variable "var.V_Ed" which depends on "pre.d
* Calculation by FORM and postprocessing see below.


```{r, results="hide", warning=FALSE}
var.d <- PARAM_BASEVAR(Id = 1, Name = "d", Description = "Stat. Nutzhöhe", DistributionType = "norm", Sd = 10, ParamType = "Mean", ParamValues = c(seq(200, 800, 50), seq(1000, 3000, 100)) + 10)
pre.d <- c(seq(200, 800, 50), seq(1000, 3000, 100))

var.f_ck <- PROB_BASEVAR(Id = 2, Name = "f_ck", Description = "Char.Betondruckfestigkeit", Package = "TesiproV", DistributionType = "lt", DistributionParameters = c(3.85, 0.09, 3, 10))
pre.f_ck <- 35

var.f_ywk <- PROB_BASEVAR(Id = 3, Name = "f_ywk", Description = "Zugfestigkeit Bügelbewehrung", DistributionType = "norm", Mean = 560, Cov = 0.02)
pre.f_ywk <- 500

var.f_yk <- PROB_BASEVAR(Id = 4, Name = "f_yk", Description = "Zugfestigkeit Längsbewehrung", DistributionType = "norm", Mean = 560, Sd = 30)
pre.f_yk <- 500

var.rho_l <- PROB_BASEVAR(Id = 5, Name = "rho_l", Description = "Biegebewehrungsgrad", DistributionType = "norm", Mean = 0.01, Cov = 0.02)
pre.rho_l <- var.rho_l$Mean

var.c_D <- PROB_BASEVAR(Id = 6, Name = "c_D", Description = "Stützendurchmesser", DistributionType = "norm", Mean = 500 + 0.003 * 500, Sd = 4 + 0.006 * 500)
pre.c_D <- 500

var.theta_c <- PROB_BASEVAR(Id = 7, Name = "theta_c", Description = "Modellunsicherheit c nach DIBT", DistributionType = "lnorm", Mean = 1.0644, Cov = 0.1679)

var.gamma_c <- PROB_DETVAR(Id = 8, Name = "gamma_c", Description = "TSB Beton", Value = 1.5)
var.gamma_s <- PROB_DETVAR(Id = 9, Name = "gamma_s", Description = "TSB Stahl", Value = 1.15)
var.alpha_cc <- PROB_DETVAR(Id = 10, Name = "alpha_cc", Description = "Langzeitfaktor AlphaCC", Value = 0.85)

V_Ed <- c(
  0.6412447, 0.8760515, 1.1424185, 1.4399554, 1.7554079, 2.0188587, 2.3003489, 2.5997038, 2.9167825, 3.2514671, 3.6036567, 3.9732630,
  4.3602076, 6.0800504, 7.0424326, 8.0726216, 9.1702994, 10.3351829, 11.5670173, 12.8655715, 14.2306347, 15.6620132, 17.1595283, 18.7230145,
  20.3523175, 22.0472935, 23.8078076, 25.6337332, 27.5249511, 29.4813487, 31.5028193, 33.5892621, 35.7405811, 37.9566849
)
var.V_Ed <- PARAM_DETVAR(Id = 11, Name = "V_Ed", Description = "Einwirkende Querkraft", ParamValues = V_Ed)

lsf1 <- SYS_LSF(
  vars = list(
    var.d, var.f_ck, var.f_yk, var.f_ywk, var.rho_l, var.c_D,
    var.theta_c, var.V_Ed, var.gamma_c, var.gamma_s, var.alpha_cc
  ),
  name = "Durchstanzwiderstand ohne Bewehrung"
)
lsf1$func <- function(d, f_ck, f_yk, f_ywk, rho_l, c_D, theta_c, V_Ed, gamma_c, gamma_s, alpha_cc) {
  f_cd <- f_ck * alpha_cc / gamma_c
  f_yd <- f_yk / gamma_s
  f_ywd <- f_ywk / gamma_s

  u_0 <- c_D * pi
  u_1 <- (c_D / 2 + 2 * d) * 2 * pi
  u_0.5 <- (c_D / 2 + 0.5 * d) * 2 * pi

  k <- min(1 + sqrt(200 / d), 2)

  if ((u_0 / d) < 4) {
    C_Rd_c <- 0.18 * (0.1 * u_0 / d + 0.6)
  } else {
    C_Rd_c <- 0.18
  }
  rho_l <- min(rho_l, 0.5 * f_cd / f_yd, 0.02)
  C_Rd_c1 <- C_Rd_c * k * (100 * rho_l * f_ck)^(1 / 3)

  v_Rd_c <- C_Rd_c1
  V_Rd_c <- v_Rd_c * d * u_1 * 1e-6 * theta_c
  return(V_Rd_c - V_Ed)
}

form <- PROB_MACHINE(name = "FORM RF.", fCall = "FORM", options = list("n_optim" = 20, "loctol" = 0.0001, "optim_type" = "rackfies"))

ps <- SYS_PARAM(
  sys_input = list(lsf1),
  probMachines = list(form)
)
ps$runMachines()
```

```{r, results="hide", warning=FALSE}
d <- pre.d
beta <- unlist(unname(ps$beta_params))
plot(x = d, y = beta, type = "l")
```


# References

1. JOINT COMMITTEE ON STRUCTURAL SAFETY, JCSS Probabilistic Model Code. URL: https://www.jcss-lc.org/, 2001. 

2. Marelli & Sudret (2014): UQLab: A Framework for Uncertainty Quantification in MATLAB, ICVRAM 2014 Conference Proceedings. DOI:10.1061/9780784413609.257

3. NOWAK, A. S.; COLLINS, K. R. Reliability of structures. CRC Press, 2012.

4. DIN EN 1992-1-1: 2011-01: DIN EN 1992-1-1: Design of concrete structures – Part 1-1: General rules and rules for buildings; German version EN 1992-1-1:2004 + AC:2010. 

5. DIN EN 1992-1-1/NA: 2013-04: National Annex – Nationally determined parameters – DIN EN 1992-1-1: Design of concrete structures – Part 1-1: General rules and rules for buildings. 

# Aknowledgement 

The TesiproV package was developed within the framework of the research project “TesiproV”, funded by the German Federal Ministry for Economic Affairs and Energy (BMWi), and the research project “Qualitätskontrollen und Bauwerkszuverlässigkeit im Bauwesen”, funded by the German Institute for Building Technology (DIBt). The authors gratefully acknowledge the financial support provided by these institutions.

# Contact & License

For questions, feedback or collaboration requests please contact us at:

**TU Dortmund University – Chair of Structural Engineering**  
August-Schmidt-Strasse 8 · D‑44227 Dortmund · Germany  

Email: [tesiprov.lmb.ab@tu-dortmund.de](mailto:tesiprov.lmb.ab@tu-dortmund.de)  

Additional contacts:  
* [nille-hauf@hochschule-bc.de](mailto:nille-hauf@hochschule-bc.de) ([konstantin.nille-hauf@googlemail.com](mailto:konstantin.nille-hauf@googlemail.com))  
* [tania.feiri@tu-dortmund.de](mailto:tania.feiri@tu-dortmund.de)  
* [marcus.ricker@tu-dortmund.de](mailto:marcus.ricker@tu-dortmund.de)  
* [til.lux@tu-dortmund.de](mailto:til.lux@tu-dortmund.de)

---

© 2021–2026 K. Nille-Hauf, T. Feiri, M. Ricker, T. Lux -- Hochschule Biberach (until 2022), TU Dortmund University – Chair of Structural Concrete (since 2023)

TesiproV is released under the [MIT License](https://opensource.org/licenses/MIT).  
You are free to use, modify and redistribute the software provided that this copyright notice and license information remain included.

For official releases see the [CRAN package TesiproV](https://CRAN.R-project.org/package=TesiproV).  
Source code repository: <https://gitlab.tu-dortmund.de/ls-massivbau/tesiprov>

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parallel_helpers.R
\name{create_rng_manager}
\alias{create_rng_manager}
\title{Deterministic RNG Block Manager (L'Ecuyer-CMRG)}
\usage{
create_rng_manager(seed = NULL)
}
\arguments{
\item{seed}{Optional integer for reproducibility.}
}
\value{
An RNG manager object (environment).
}
\description{
Provides deterministic generation of random number blocks that are
independent of the number of worker threads.

The RNG state is advanced only by the requested block size.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_debugPrint.R
\name{is_missing}
\alias{is_missing}
\title{Helper: is a value missing (NULL, NA, zero-length, or empty string)
Used inside validateArgs() to avoid the missing value where TRUE/FALSE needed
problem.}
\usage{
is_missing(x)
}
\arguments{
\item{x}{any R object}
}
\value{
TRUE if x should be treated as missing, FALSE otherwise
}
\description{
Helper: is a value missing (NULL, NA, zero-length, or empty string)
Used inside validateArgs() to avoid the missing value where TRUE/FALSE needed
problem.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clsModelContainer.R
\docType{class}
\name{SYS_LSF-class}
\alias{SYS_LSF-class}
\alias{SYS_LSF}
\title{System Limit-State Function (SYS_LSF)}
\description{
A reference-class that stores a user-defined limit-state function.
The function can be written either as an **expression** (`expr`) or
as a **regular R function** (`func`).  All random variables that appear
in the function must be supplied in the `vars` list as `PROB_BASEVAR`
objects.
}
\details{
The method `getLSF()` returns a **closure** that expects a numeric vector
`x` (ordered exactly like `vars`) and evaluates the limit-state function.

For consistency across FORM and Monte Carlo reliability methods,
limit-state functions must either accept a single numeric vector
argument or explicitly named scalar arguments.
The use of \code{function(...)} is not supported.
}
\section{Fields}{

\describe{
\item{\code{name}}{Optional human-readable name of the limit-state function.}

\item{\code{expr}}{Symbolic expression (e.g. `expression(f_ck - d_nom)`) that is
later converted to a callable function via `ExpressionToFunction()`.}

\item{\code{func}}{Actual R function implementing the limit-state equation.
May be `NULL` until defined by user or converted from `expr`.}

\item{\code{vars}}{List of `PROB_BASEVAR` objects that provide the random variables.}
}}

\section{Methods}{

\describe{
\item{\code{ExpressionToFunction()}}{Transforms a valid expression into a objective function. Need the set of Variables with correct spelled names and IDs}

\item{\code{check()}}{Checks all variables. You dont need to execute this, since the system object will do anyway.}
}}

\section{Limit-State Function (LSF) API}{


A limit-state function must follow one of the two supported
interface patterns:

\describe{
  \item{\code{function(x)}}{
    A single numeric vector argument \code{x} containing all
    basic variables in the order specified in \code{vars}.

    Example:
    \preformatted{
    lsf$func <- function(x) {
      x[1] - sum((x[-1]^2) / seq_along(x)[-1])
    }
    }
  }

  \item{\code{function(Z, Fy, M, ...)}}{
    Explicitly named scalar arguments corresponding to the
    variable names defined in \code{vars}.

    Example:
    \preformatted{
    lsf$func <- function(Z, Fy, M) {
      Z * Fy - M
    }
    }
  }
}

The following form is \strong{not supported}:

\preformatted{
function(...)
}

Using \code{...} as the sole argument leads to inconsistent
behaviour between FORM-based and Monte Carlo-based reliability
methods and is therefore prohibited.

For maximum robustness and compatibility across all reliability
algorithms, the use of the vector form \code{function(x)} is
recommended.
}

\examples{
vars <- list(
  PROB_BASEVAR(Name = "f_ck", DistributionType = "norm", Mean = 30, Sd = 1),
  PROB_BASEVAR(Name = "d_nom", DistributionType = "norm", Mean = 0.2, Sd = 0.01)
)
lsf <- SYS_LSF(name = "Bending resistance", vars = vars)
lsf$ExpressionToFunction()

}
\author{
(C) 2021-2026 K. Nille-Hauf, T. Feiri, M. Ricker, T. Lux -- Hochschule Biberach (until 2022),
TU Dortmund University - Chair of Structural Concrete (since 2023)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MC_CRUDE.R
\name{MC_CRUDE}
\alias{MC_CRUDE}
\title{Crude Monte Carlo Simulation}
\usage{
MC_CRUDE(
  lsf,
  lDistr,
  cov_user = 0.05,
  n_batch = 10000,
  n_max = 1e+07,
  use_threads = parallel::detectCores(),
  backend = c("future", "parallel"),
  dataRecord = TRUE,
  debug.level = 0,
  seed = NULL
)
}
\arguments{
\item{lsf}{Limit-state function. Must follow the LSF interface
specification described in the "Limit-State Function (LSF) Interface"
section below.}

\item{lDistr}{List of distribution objects as returned by \code{PROB_BASEVAR$getlDistr()}.}

\item{cov_user}{Target coefficient of variation to be achieved.}

\item{n_batch}{Batch size per iteration (used for parallel computation).}

\item{n_max}{Maximum number of Monte Carlo samples (stopping criterion).}

\item{use_threads}{Number of parallel threads. Set to 1 for single-core
execution. Parallel execution is not supported on Windows.}

\item{backend}{Parallel backend, either \code{"future"} (default) or \code{"parallel"}.}

\item{dataRecord}{Logical; if \code{TRUE}, intermediate results are recorded.}

\item{debug.level}{Integer controlling verbosity (0 = silent, 2 = detailed output).}

\item{seed}{Optional integer value for reproducible random numbers.}
}
\value{
MC_CRUDE returns an object containing the following elements:
\itemize{
  \item \code{method}: Character string identifying the method ("MCC").
  \item \code{beta}: Estimated reliability index \eqn{\beta = -\Phi^{-1}(p_f)}.
  \item \code{pf}: Estimated probability of failure.
  \item \code{var}: Estimated variance of the Monte Carlo estimator.
  \item \code{cov_mc}: Estimated coefficient of variation of \code{pf}.
  \item \code{cov_user}: Target coefficient of variation specified by the user.
  \item \code{n_mc}: Total number of Monte Carlo samples generated.
  \item \code{n_max}: Maximum allowed number of samples.
  \item \code{n_batch}: Batch size per iteration.
  \item \code{n_threads}: Number of threads used.
  \item \code{runtime}: CPU time returned by \code{proc.time()}.
  \item \code{data}: Optional data frame containing intermediate results
        (only if \code{dataRecord = TRUE}).
}
}
\description{
Crude Monte Carlo simulation (MC_CRUDE) method for the estimation of failure
probabilities in structural reliability analysis.

The crude Monte Carlo method estimates the probability of failure
by direct sampling of the basic random variables and evaluation
of the limit-state function.

The failure probability is estimated as
\deqn{
  \hat{p}_f = \frac{1}{N}
  \sum_{i=1}^{N}
  \mathbf{1}_{\{g(\mathbf{X}_i) \le 0\}}.
}

Although conceptually simple and unbiased, the method becomes
computationally inefficient for very small failure probabilities,
as a large number of samples is required to obtain stable estimates.
}
\section{Limit-State Function (LSF) Interface}{


The argument \code{lsf} must define a valid limit-state function
using one of the following supported signatures:

\describe{
  \item{\code{function(x)}}{
    A single numeric vector \code{x} containing all basic variables
    in the order defined in \code{lDistr}.

    Example:
    \preformatted{
    lsf <- function(x) {
      x[1] - sum((x[-1]^2) / seq_along(x)[-1])
    }
    }
  }

  \item{\code{function(Z, Fy, M, ...)}}{
    Explicitly named scalar arguments corresponding to the variable
    names defined in the probabilistic model.

    Example:
    \preformatted{
    lsf <- function(Z, Fy, M) {
      Z * Fy - M
    }
    }
  }
}

The following form is \strong{not supported}:

\preformatted{
function(...)
}

Using \code{...} as the sole argument may lead to inconsistent
behaviour across different reliability algorithms and is therefore
prohibited.

For maximum robustness and cross-method compatibility,
the vector form \code{function(x)} is recommended.
}

\section{Windows users}{


On Windows systems, it is recommended to use
\code{backend = "future"} for multi-core parallelization.
The \code{"parallel"} backend does not support forking on Windows.
}

\references{
Spaethe, G. (1991).
\emph{Die Sicherheit tragender Baukonstruktionen}.
Springer.'
}
\author{
(C) 2021-2026 K. Nille-Hauf, T. Feiri, M. Ricker, T. Lux -- Hochschule Biberach (until 2022),
TU Dortmund University - Chair of Structural Concrete (since 2023)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clsParam.R
\docType{class}
\name{PARAM_BASEVAR-class}
\alias{PARAM_BASEVAR-class}
\alias{PARAM_BASEVAR}
\title{Parametric Random Variable (PARAM_BASEVAR)}
\description{
Extends `PROB_BASEVAR` so that one of the distribution parameters (Mean,
Sd or DistributionType) can be **swept** over a user-defined vector of values.
}
\details{
The method `nextParam()` advances `pos` (circularly) and updates the basic
fields (`Mean`, `Sd`, `DistributionType`, `DistributionParameters`) before
calling `prepare()` again.  `getCurrentParam()` returns the *last* value that
was used (convenient for readable run names).
}
\section{Fields}{

\describe{
\item{\code{ParamValues}}{Numeric vector holding the values that will be assigned
during the sweep.}

\item{\code{ParamType}}{Which parameter is varied: `"Mean"`, `"Sd"` or
`"DistributionType"`.}

\item{\code{pos}}{Current index inside `ParamValues` (used internally).}
}}


\examples{
pvar <- PARAM_BASEVAR(
  Name = "E_mod",
  DistributionType = "norm",
  ParamType = "Mean",
  ParamValues = c(30e3, 35e3, 40e3)
)
pvar$nextParam() # sets Mean = 30.000 and prepares the distribution
pvar$getCurrentParam()

}
\author{
(C) 2021-2026 K. Nille-Hauf, T. Feiri, M. Ricker, T. Lux -- Hochschule Biberach (until 2022),
TU Dortmund University - Chair of Structural Concrete (since 2023)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clsModelContainer.R
\docType{class}
\name{SYS_PROB-class}
\alias{SYS_PROB-class}
\alias{SYS_PROB}
\title{System Probability Solution Object (SYS_PROB)}
\description{
The \code{SYS_PROB} class represents a probabilistic system
consisting of one or more limit-state functions (\code{SYS_LSF})
and a set of reliability algorithms (\code{PROB_MACHINE}).

It allows the execution of reliability analyses for
serial or parallel systems, aggregation of single
limit-state results, and system-level probability evaluation.

Each limit-state function may include multiple random variables
(\code{PROB_BASEVAR}) with different probability distributions.
The transformation between mean/standard deviation and the
native distribution parameters is handled automatically.

\itemize{

  \item Normal ("norm"):
  mean = \eqn{\mu}, standard deviation = \eqn{\sigma}.
  Parameters: mean, sd.

  \item Lognormal ("lnorm"):
  defined by mean \eqn{m} and sd \eqn{s}.
  Parameters:
  \eqn{\mu = \log\left( \frac{m}{\sqrt{1 + (s/m)^2}} \right)},
  \eqn{\sigma = \sqrt{ \log\left( 1 + (s/m)^2 \right) }}.

  \item shifted Lognormal ("slnorm"):
  defined by Mean \eqn{m}, Sd \eqn{s} and \eqn{x0}.
  Parameters:
  \eqn{\mu = \log\left( \frac{m-x0}{\sqrt{1 + (s/(m-x0))^2}} \right)},
  \eqn{\sigma = \sqrt{ \log\left( 1 + (s/(m-x0))^2 \right) }}.

  \item Gumbel ("gumbel", package "evd"):
  location-scale type I extreme value distribution.
  Parameters:
  location = \eqn{\mu - \gamma \cdot \beta},
  where \eqn{\gamma} is the Euler-Mascheroni constant,
  and
  \eqn{\beta = \frac{\sigma \sqrt{6}}{\pi}}.

  \item Gamma ("gamma"):
  shape-scale parameterization.
  Parameters:
  \eqn{k = \frac{\mu^2}{\sigma^2}},
  \eqn{\theta = \frac{\sigma^2}{\mu}}.

  \item Exponential ("exp"):
  rate \eqn{\lambda} or scale \eqn{\theta = 1/\lambda}.
  Parameter:
  \eqn{\lambda = 1/\mu}.

  \item Beta ("beta"):
  bounded on \eqn{[0,1]}.
  Parameters:
  \eqn{\alpha = \mu t},
  \eqn{\beta = (1 - \mu) t},
  with
  \eqn{t = \frac{\mu (1-\mu)}{\sigma^2} - 1}.

  \item Weibull ("weibull"):
  shape \eqn{k} and scale \eqn{\lambda}.
  Parameters:
  empirical approximation
  \eqn{k \approx \left( \frac{\sigma}{\mu} \right)^{-1.086}},
  \eqn{\lambda = \frac{\mu}{\Gamma\left(1 + \frac{1}{k}\right)}}.

  \item Empirical ("emp"):
  non-parametric distribution defined by observed sample.
  Parameter: \code{obs} = numeric vector of observations.
  The empirical CDF is constructed from the ordered sample.
  Duplicate values are allowed but may trigger a warning.

  \item Student t ("st"):
  marginal density of a normal-gamma model describing uncertainty
  in \eqn{\mu} and \eqn{\sigma}.
  Hyperparameters:
  \eqn{(m, s, n, \nu)}.

  \item Log-Student t ("lt"):
  marginal density of a log-normal-gamma model describing uncertainty
  in \eqn{\mu} and \eqn{\sigma}.
  Hyperparameters:
  \eqn{(m, s, n, \nu)}.
}

The types `"binom"` and `"unif"` are reserved for future extensions.
The package automatically checks consistency between mean, standard deviation and coefficient of variation.

Methods available in this class:

* `$runMachines()` - executes all solution algorithms for each limit state function.
* `$calculateSystemProbability()` - computes system reliability using bounds or Monte Carlo based methods (`MC_IS`, `MC_CRUDE`, `MC_SubSam`).
* `$printResults()` - prints a detailed report to console or file.
* `$saveProject(level)` - saves results at different detail levels.
* `$plotGraph(plotType)` - experimental plotting routine for simulation performance.
}
\details{
The class is built on \code{setRefClass} and therefore uses mutable fields.
All objects (`PROB_BASEVAR`, `SYS_LSF`, ...) are expected to be already
**checked** (via the `check()` method) before they are passed to `SYS_PROB`.
}
\section{Fields}{

\describe{
\item{\code{sys_input}}{List of `SYS_LSF` objects - the individual limit-state functions.}

\item{\code{sys_type}}{Character string that defines whether the system
is *serial* or *parallel* (not yet implemented).}

\item{\code{probMachines}}{List of `PROB_MACHINE` objects - the analysis methods.}

\item{\code{res_single}}{List of results per machine for each limit-state function.}

\item{\code{res_sys}}{List of results for the system-level reliability calculation.}

\item{\code{beta_single}}{Matrix of \eqn{\beta}-values obtained from the single-problem analyses.}

\item{\code{beta_sys}}{Matrix of \eqn{\beta}-values obtained from the system reliability calculations.}

\item{\code{params_sys}}{List that stores the parameter sets used in Monte-Carlo runs.}

\item{\code{debug.level}}{Numeric verbosity level (0 = silent, 1 = basic, 2 = detailed).}
}}

\section{Methods}{

\describe{
\item{\code{calculateSystemProbability(calcType = "simpleBounds", params = list())}}{Calculates the system probablity if more than one lsf is given and a system_type (serial or parallel) is set.
If calcType is empty (or simpleBounds), only simpleBounds are applied to further calculation of single soultions.
If calcType is MCIS, than a Monte Carlo Importance Sampling Method is used (only for parallel systems available).
If calcType is MCC, than a Crude Monte Carlo Simulation is used.
If calcType is MCSUS, than the Subset Sampling Algorithm ll be used.
You can pass arguments to methods via the params field, while the argument has to be a named list (for example check the vignette).}

\item{\code{plotGraph(plotType = "sim.performance")}}{not finally implemented. Do not use.}

\item{\code{printResults(path = "")}}{TesiproV can create a report file with all the necessary data for you. If you provide a path (or filename, without ending) it will store
the data there, otherwise it will report to the console. Set the path via setwd() or check it via getwd().}

\item{\code{runMachines()}}{Starts solving all given problems (sys_input) with all given algorithms (probMachines).
First checks that all analysis methods are loaded before validating limit-state functions.}

\item{\code{saveProject(level, filename = "tesiprov_project")}}{You can save your calculation project with saveProject().
There are four different levels of detail to save
1st Level: Only the beta values
2nd Level: The result Objects of single or systemcalculation
3th Level: All The Probablity System Object, including limit state functions, machines and solutions
4th Level: An image of your entire workspace}
}}

\examples{
ps <- SYS_PROB(
  sys_input = list(SYS_LSF(), SYS_LSF()),
  probMachines = list(PROB_MACHINE()),
  sys_type = "serial"
)
\dontrun{
ps$runMachines()
ps$beta_sys
ps$res_sys
ps$printResults("example_1")
ps$saveProject(4, "example_1")
}

}
\author{
(C) 2021-2026 K. Nille-Hauf, T. Feiri, M. Ricker, T. Lux -- Hochschule Biberach (until 2022),
TU Dortmund University - Chair of Structural Concrete (since 2023)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parallel_helpers.R
\name{run_parallel}
\alias{run_parallel}
\title{Run a parallel computation using either "parallel" or "future"}
\usage{
run_parallel(
  X,
  FUN,
  backend = c("parallel", "future"),
  use_threads = NULL,
  cl = NULL
)
}
\arguments{
\item{X}{List (or vector) of inputs that will be split among the workers.}

\item{FUN}{Function that will be executed on each chunk.}

\item{backend}{Either "parallel" or "future".}

\item{use_threads}{Number of workers (relevant for both back‑ends).}

\item{cl}{Optional cluster object created with \code{parallel::makeCluster()}.
Only required when using the \code{"parallel"} backend on Windows, where
\code{parallel::parLapply()} is used. Ignored on Unix-like systems where
\code{parallel::mclapply()} is available.}
}
\value{
List of results (identical to the output of `parLapply` /
  `future_lapply`). The list is **named** – the names are taken from the
  split index (`split_idx`).
}
\description{
Small wrapper that hides the branching between the two back‑ends.
  It always returns a **named list** whose names are identical to the names
  of the input split list. This guarantees that `unlist()` produces the same
  named vector for both back‑ends, which is required by the unit tests.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clsParam.R
\docType{class}
\name{SYS_PARAM-class}
\alias{SYS_PARAM-class}
\alias{SYS_PARAM}
\title{Object for parametric Studies (SYS_PARAM)}
\description{
Extends `SYS_PROB` to perform a **parametric sweep** over one or more
`PARAM_BASEVAR` objects.  For each combination of parameter values a full
reliability analysis is executed and the results are stored in
`beta_params` and `res_params`.
}
\details{
The method `runMachines()` is almost identical to the one in `SYS_PROB`,
but it loops over **all** parameter values, creates a readable run name,
and stores the whole result hierarchy.
}
\section{Fields}{

\describe{
\item{\code{beta_params}}{List of \eqn{\beta}-matrices (one matrix per parameter run).}

\item{\code{res_params}}{List of detailed result objects (one per run).}

\item{\code{nParams}}{Numeric vector: number of parameter values for each
`PARAM_BASEVAR`.}

\item{\code{nLsfs}}{Number of limit-state functions in the system.}

\item{\code{nMachines}}{Number of analysis machines (methods) used.}
}}

\section{Methods}{

\describe{
\item{\code{printResults(path = "")}}{TesiproV can create a report file with all the necessary data for you. If you provide a path (or filename, without ending) it will store
the data there, otherwise it will report to the console. Set the path via setwd() or check it via getwd().}

\item{\code{runMachines()}}{Starts solving all given problems (sys_input) with all given algorithms (probMachines).
First checks that all analysis methods are loaded before validating limit-state functions.}
}}

\examples{
# Simple 2-parameter sweep ------------------------------------------------
# two random variables that will be varied
pv1 <- PARAM_BASEVAR(
  Name = "E_mod",
  DistributionType = "norm",
  ParamType = "Mean",
  ParamValues = c(30e3, 35e3, 40e3)
)
pv2 <- PARAM_BASEVAR(
  Name = "f_ck",
  DistributionType = "norm",
  ParamType = "Mean",
  ParamValues = c(30, 35, 40)
)

lsf <- SYS_LSF(vars = list(pv1, pv2), name = "Bending")
ps <- SYS_PARAM(
  sys_input = list(lsf),
  probMachines = list(PROB_MACHINE(name = "FORM", fCall = "FORM")),
  sys_type = "serial"
)

\dontrun{
ps$runMachines()
ps$beta_params # matrix of \eqn{\beta}-values for every run
}

}
\author{
(C) 2021-2026 K. Nille-Hauf, T. Feiri, M. Ricker, T. Lux -- Hochschule Biberach (until 2022),
TU Dortmund University - Chair of Structural Concrete (since 2023)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_debugPrint.R
\name{is_empty}
\alias{is_empty}
\title{Test whether an object is *empty*}
\usage{
is_empty(x)
}
\arguments{
\item{x}{Any R object.}
}
\value{
A logical scalar (`TRUE` if `x` is empty, `FALSE` otherwise).
}
\description{
This lightweight helper mimics the behaviour of `pracma::isempty()` but
depends only on base R.  An object is considered *empty* when it contains
no data:
  * `NULL`
  * an atomic vector, matrix, array or factor of length 0
  * a list (including named lists) of length 0
  * a data frame with zero rows (`nrow == 0`)
}
\details{
The function is deliberately strict - environments, functions, S4/R6
objects, etc. are **not** treated as empty.  If you need additional
classes to be regarded as empty you can extend the internal checks.
}
\examples{
\dontrun{
is_empty(NULL) # TRUE
is_empty(numeric(0)) # TRUE
is_empty(list()) # TRUE
is_empty(data.frame()) # TRUE
is_empty(1:5) # FALSE
is_empty(list(a = 1)) # FALSE
is_empty(data.frame(x = 1)) # FALSE
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MCIS_helpers.R
\name{make_ready_cluster}
\alias{make_ready_cluster}
\title{Prepare parallel cluster environment for MC Importance Sampling}
\usage{
make_ready_cluster(
  use_threads,
  libPaths_local,
  seed = NULL,
  export_objs = character(0),
  debug.level = 0
)
}
\arguments{
\item{use_threads}{Number of worker cores/threads to create via \link[parallel]{makeCluster}.}

\item{libPaths_local}{Library paths propagated to each worker via `.libPaths()`.}

\item{seed}{Integer base seed used for reproducible random numbers across workers.}

\item{export_objs}{Character vector naming objects/functions that should be exported from parent environment.}
}
\value{
Ready-to-use parallel cluster object configured with correct libraries and RNG streams.
}
\description{
Creates an R parallel cluster with specified number of workers,
exports required objects/functions and initializes reproducible RNG streams on each worker.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MCIS_helpers.R
\name{Internal helper functions for MC_IS()}
\alias{Internal helper functions for MC_IS()}
\title{Internal helper functions for Monte Carlo Importance Sampling (MC_IS)}
\description{
This file contains internal helper functions used by the main function
[`MC_IS()`] to perform Monte Carlo simulations with importance sampling.

The helpers defined here support:
* random number generation and cluster setup,
* creation of recorder objects for iterative logging,
* single limit state simulation routine (`MC_IS_single()`),
* conversion of recorded data to data frames after completion.

Additional helper sets will be provided for system calculations
(serial and parallel systems) in separate files.
}
\details{
These functions are not intended to be called directly by users.
They are exported internally to allow modular access within the package.
}
\seealso{
[TesiproV::MC_IS()] for the public interface that calls these helpers.
}
\author{
(C) 2021-2026 K. Nille-Hauf, T. Feiri, M. Ricker, T. Lux -- Hochschule Biberach (until 2022),
TU Dortmund University - Chair of Structural Concrete (since 2023)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_debugPrint.R
\name{debug.print}
\alias{debug.print}
\title{internal  Helper function to debug more easy}
\usage{
debug.print(
  infoLevel,
  flag = "",
  values = character(0),
  msg = "",
  type = "INFO",
  step_interval = 10,
  iter = NA
)
}
\arguments{
\item{infoLevel}{If 0, no Output (just Errors), if 1 little output, if 2 bigger output}

\item{flag}{Parse additonal info}

\item{values}{If you check variables then post this into values}

\item{msg}{here add some extra msg}

\item{type}{Type can be "INFO" or "ERROR"}

\item{step_interval}{Optional integer specifying
the interval at which debug messages are printed.}

\item{iter}{Current iteration index used for
conditional debug output.}
}
\description{
internal  Helper function to debug more easy
}
\author{
(C) 2021-2026 K. Nille-Hauf, T. Feiri, M. Ricker, T. Lux -- Hochschule Biberach (until 2022),
TU Dortmund University - Chair of Structural Concrete (since 2023)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MCIS_helpers.R
\name{build_gsys}
\alias{build_gsys}
\title{Build a smooth system limit-state function (g_sys)}
\usage{
build_gsys(
  sys_input,
  sys_structure = "serial",
  smooth_kappa = 20,
  debug.level = 0
)
}
\arguments{
\item{sys_input}{List of \code{SYS_LSF} objects defining the component
limit-state functions of the system.}

\item{sys_structure}{System aggregation structure.
Supported options are
\describe{
  \item{\code{"serial"}}{System fails if any component fails.}
  \item{\code{"parallel"}}{System fails only if all components fail.}
  \item{\code{function(g_vals)}}{Custom aggregation function applied
  to the vector of component limit-state values.}
}}

\item{smooth_kappa}{Positive numeric value controlling the smoothness
of the aggregation. Larger values approach the exact \code{min}
or \code{max} operator.}

\item{debug.level}{Optional verbosity level for diagnostic output.}
}
\value{
A list containing:
\describe{
  \item{\code{g_sys}}{Callable system limit-state function
  \code{function(x)} operating on the global variable vector.}

  \item{\code{distr_flat}}{Flattened list of marginal distribution
  objects corresponding to all variables in the system.}

  \item{\code{var_map}}{List mapping each limit-state function to the
  indices of its variables in the global vector.}

  \item{\code{n_vars}}{Total number of basic variables in the system.}

  \item{\code{n_lsfs}}{Number of limit-state functions.}
}
}
\description{
Constructs a **global system limit-state function** from a list of
individual limit-state functions (`SYS_LSF` objects). The resulting
function aggregates all component limit states using a **smooth
approximation** of the minimum or maximum operator.
}
\details{
The smooth aggregation improves numerical stability and differentiability,
which is beneficial for gradient-based reliability methods such as
FORM, SORM, and importance sampling algorithms.

For a serial system the aggregation is

\deqn{
g_{\kappa}(x) =
-\frac{1}{\kappa}\log
\left(
\sum_{i=1}^{m} \exp(-\kappa g_i(x))
\right)
}

which approximates

\deqn{
\min_i g_i(x).
}

For a parallel system the aggregation is

\deqn{
g_{\kappa}(x) =
\frac{1}{\kappa}\log
\left(
\sum_{i=1}^{m} \exp(\kappa g_i(x))
\right)
}

which approximates

\deqn{
\max_i g_i(x).
}

The approximation becomes exact for
\eqn{\kappa \rightarrow \infty}.

In addition to the system function, this helper also constructs a
**flattened list of marginal distributions** and a **variable index map**
that links the global variable vector to the variables of each
individual limit-state function.


This function is primarily used internally by reliability algorithms
that operate on a **single limit-state function**, such as
importance sampling or crude Monte Carlo simulation.

By constructing a smooth global limit-state function, the system
reliability problem can be treated as a standard single-LSF problem,
enabling the reuse of algorithms originally developed for
component reliability.
}
\seealso{
\code{smooth_min}, \code{smooth_max},
\code{\link{MC_IS}}, \code{\link{SYS_LSF}}
}
\author{
(C) 2021-2026 K. Nille-Hauf, T. Feiri, M. Ricker, T. Lux -
Hochschule Biberach (until 2022),
TU Dortmund University - Chair of Structural Concrete (since 2023)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clsParam.R
\docType{class}
\name{PARAM_LSF-class}
\alias{PARAM_LSF-class}
\alias{PARAM_LSF}
\title{Parametric Limit-State Function (PARAM_LSF)}
\description{
Identical to `SYS_LSF` but inherits from it to make the class hierarchy clear
when a parametric study is performed.  No additional fields or methods are
required.
}

\author{
(C) 2021-2026 K. Nille-Hauf, T. Feiri, M. Ricker, T. Lux -- Hochschule Biberach (until 2022),
TU Dortmund University - Chair of Structural Concrete (since 2023)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clsModelContainer.R
\name{ensure_pkg}
\alias{ensure_pkg}
\title{Ensure that a required package is available}
\usage{
ensure_pkg(pkg)
}
\arguments{
\item{pkg}{Character string with the package name}
}
\value{
Invisible TRUE (used only for its side-effect)
}
\description{
Ensure that a required package is available
}
\note{
The function stops with an informative error if the required
  package is not installed.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MCIS_helpers.R
\name{set_cluster_rng}
\alias{set_cluster_rng}
\title{Set cluster RNG streams}
\usage{
set_cluster_rng(cl, seed)
}
\arguments{
\item{cl}{Cluster object created by `parallel::makeCluster()`.}

\item{seed}{Integer base seed used on the master process.}
}
\description{
Assigns identical RNG streams to each worker in an existing cluster.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MCIS_helpers.R
\name{MC_IS_system}
\alias{MC_IS_system}
\title{Monte-Carlo Importance Sampling for a system of limit-state functions}
\usage{
MC_IS_system(
  lsf,
  lDistr,
  cov_user = 0.05,
  n_batch,
  n_max,
  use_threads,
  sys_type,
  dataRecord = TRUE,
  beta_l,
  densityType,
  dps = NULL,
  debug.level = 0,
  streams = NULL,
  libPaths_local = .libPaths(),
  seed = NULL,
  backend = "future",
  adaptive_alpha = FALSE,
  alpha_update_rate = 0.1,
  adaptive_batch = FALSE,
  batch_control = list(),
  min_adapt_samples = 10000,
  alpha_min = 0.02,
  stability_mode = c("robust", "fast")
)
}
\arguments{
\item{lsf}{Objective (limit state) function in form \code{function(x) \{ ... \}}.}

\item{lDistr}{List of input variable distributions created from TesiproV base variable objects.}

\item{cov_user}{Target coefficient of variation for convergence criterion.}

\item{n_batch}{Number of samples per batch (recommended > 1000).}

\item{n_max}{Maximum total number of samples before forced stop.}

\item{use_threads}{Number of threads/workers to use for parallel computation.}

\item{sys_type}{Either "parallel" (default) or "serial"}

\item{dataRecord}{Logical; if `TRUE`, all intermediate results are recorded and returned as data frame.}

\item{beta_l}{Optional threshold for reliability index screening.
Limit-state functions with \eqn{\beta > beta_l} may be excluded
from mixture construction in system analysis.}

\item{densityType}{Character; type of sampling density ("origin" keeps original input distributions).}

\item{dps}{Optional vector of design points supplied by user instead of FORM result.}

\item{debug.level}{Integer verbosity level: 0 = silent, 1 = info, 2 = detailed output during calculation.}

\item{streams}{RNG stream object created by [`init_rng_master()`].}

\item{libPaths_local}{Local library paths passed to cluster workers for package availability.}

\item{seed}{Optional integer seed for reproducibility across parallel runs. If `NULL`, a random seed is used.}

\item{backend}{Character; either `"parallel"` or `"future"` backend for parallelization on multiple cores.}

\item{adaptive_alpha}{Logical; if \code{TRUE}, adaptive mixture weights are used
(system case only).}

\item{alpha_update_rate}{Damping factor \eqn{\lambda \in [0,1]} controlling
adaptation speed of mixture weights.}

\item{adaptive_batch}{Logical; if TRUE, the batch size is adapted dynamically
based on the current effective sample size (ESS) and target RSE.}

\item{batch_control}{List controlling adaptive batch behaviour:
\describe{
  \item{n_min}{Minimum batch size}
  \item{n_max}{Maximum batch size}
  \item{K_future}{Desired number of remaining iterations}
}}

\item{min_adapt_samples}{Integer. Minimum total number of MC samples required before adaptive mixtures are updated.
The adaption mechanism is activated only after this threshold has been exceeded to ensure statistically stable estimates.}

\item{alpha_min}{Numeric scalar in \eqn{(0,1)}.
Lower bound for mixture weights \eqn{\alpha_i} during adaptive
updating. This prevents individual components of the importance
sampling density from collapsing to zero and ensures numerical
stability in multimodal system analyses.}

\item{stability_mode}{Character string specifying the numerical
  stabilization strategy used for importance sampling weight
  accumulation.

  \describe{
    \item{"robust"}{
      Default. Uses full log-domain accumulation of weights
      across Monte Carlo iterations. This improves numerical
      stability for high-dimensional systems and very small
      failure probabilities.
    }
    \item{"fast"}{
      Uses shifted exponentiation with batch-level scaling.
      Computationally equivalent under normal engineering
      conditions but may be slightly less stable in extreme
      rare-event scenarios.
    }
  }

  Both modes are mathematically equivalent for typical
  reliability problems. The \code{"robust"} mode is recommended
  for research applications and extreme reliability levels.}
}
\description{
Monte-Carlo Importance Sampling for a system of limit-state functions
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MCIS_helpers.R
\name{init_rng_master}
\alias{init_rng_master}
\title{Initialize master RNG stream}
\usage{
init_rng_master(seed = NULL, debug.level = 0)
}
\arguments{
\item{seed}{Optional integer seed; if `NULL`, a random base seed is generated once per call.}

\item{debug.level}{Integer controlling verbosity level.
Values >= 1 produce diagnostic console output.}
}
\value{
A master RNG stream object (numeric vector of length 7)
        as required by [parallel::clusterSetRNGStream()] or usable
        directly in [future.apply::future_lapply()].
}
\description{
Sets RNG type `"L'Ecuyer-CMRG"` and optionally a fixed seed.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clsModelContainer.R
\docType{class}
\name{PROB_DETVAR-class}
\alias{PROB_DETVAR-class}
\alias{PROB_DETVAR}
\title{Object to store a deterministic parameters (PROB_DETVAR)}
\description{
A convenience subclass of `PROB_BASEVAR` that represents a deterministic
value (i.e. a variable with practically zero variance).  The object is
automatically converted to a normal distribution with an infinitesimal
standard deviation (`Sd = Mean / 1e7`).
}

\section{Inherited fields}{

\describe{
Inherits all fields of `PROB_BASEVAR` (see that documentation for a full list).
}
}

\section{Additional fields}{

\describe{
  \item{\code{Value}}{
    The deterministic value used internally as \code{Mean}.
  }
}
}

\examples{
det <- PROB_DETVAR(Name = "E_mod", Value = 30e3)
det$prepare()

}
\author{
(C) 2021-2026 K. Nille-Hauf, T. Feiri, M. Ricker, T. Lux -- Hochschule Biberach (until 2022),
TU Dortmund University - Chair of Structural Concrete (since 2023)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MCIS_helpers.R
\name{MC_IS_single}
\alias{MC_IS_single}
\title{Internal Monte Carlo Importance Sampling Routine (Single Limit State Function)}
\usage{
MC_IS_single(
  lsf,
  lDistr,
  cov_user,
  n_batch,
  n_max,
  use_threads,
  backend = "future",
  dataRecord,
  densityType = "norm",
  dps = NULL,
  debug.level,
  streams,
  libPaths_local,
  seed = NULL,
  adaptive_alpha = FALSE,
  alpha_update_rate = 0.1,
  adaptive_batch = FALSE,
  batch_control = list(),
  min_adapt_samples = 10000,
  alpha_min = 0.02,
  stability_mode
)
}
\arguments{
\item{lsf}{Objective (limit state) function in form \code{function(x) \{ ... \}}.}

\item{lDistr}{List of input variable distributions created from TesiproV base variable objects.}

\item{cov_user}{Target coefficient of variation for convergence criterion.}

\item{n_batch}{Number of samples per batch (recommended > 1000).}

\item{n_max}{Maximum total number of samples before forced stop.}

\item{use_threads}{Number of threads/workers to use for parallel computation.}

\item{backend}{Character; either `"parallel"` or `"future"` backend for parallelization on multiple cores.}

\item{dataRecord}{Logical; if `TRUE`, all intermediate results are recorded and returned as data frame.}

\item{densityType}{Character; type of sampling density ("origin" keeps original input distributions).}

\item{dps}{Optional vector of design points supplied by user instead of FORM result.}

\item{debug.level}{Integer verbosity level: 0 = silent, 1 = info, 2 = detailed output during calculation.}

\item{streams}{RNG stream object created by [`init_rng_master()`].}

\item{libPaths_local}{Local library paths passed to cluster workers for package availability.}

\item{seed}{Optional integer seed for reproducibility across parallel runs. If `NULL`, a random seed is used.}

\item{adaptive_alpha}{Logical; if \code{TRUE}, adaptive mixture weights are used
(system case only).}

\item{alpha_update_rate}{Damping factor \eqn{\lambda \in [0,1]} controlling
adaptation speed of mixture weights.}

\item{adaptive_batch}{Logical; if TRUE, the batch size is adapted dynamically
based on the current effective sample size (ESS) and target RSE.}

\item{batch_control}{List controlling adaptive batch behaviour:
\describe{
  \item{n_min}{Minimum batch size}
  \item{n_max}{Maximum batch size}
  \item{K_future}{Desired number of remaining iterations}
}}

\item{min_adapt_samples}{Integer. Minimum total number of MC samples required before adaptive mixtures are updated.
The adaption mechanism is activated only after this threshold has been exceeded to ensure statistically stable estimates.}

\item{alpha_min}{Numeric scalar in \eqn{(0,1)}.
Lower bound for mixture weights \eqn{\alpha_i} during adaptive
updating. This prevents individual components of the importance
sampling density from collapsing to zero and ensures numerical
stability in multimodal system analyses.}

\item{stability_mode}{Character string specifying the numerical
  stabilization strategy used for importance sampling weight
  accumulation.

  \describe{
    \item{"robust"}{
      Default. Uses full log-domain accumulation of weights
      across Monte Carlo iterations. This improves numerical
      stability for high-dimensional systems and very small
      failure probabilities.
    }
    \item{"fast"}{
      Uses shifted exponentiation with batch-level scaling.
      Computationally equivalent under normal engineering
      conditions but may be slightly less stable in extreme
      rare-event scenarios.
    }
  }

  Both modes are mathematically equivalent for typical
  reliability problems. The \code{"robust"} mode is recommended
  for research applications and extreme reliability levels.}
}
\description{
Performs the core Monte Carlo simulation with importance sampling for a single limit state function.
This internal routine is called by the public interface [`MC_IS()`] when the input argument `lsf`
is a single function rather than a list of system functions.

The algorithm uses the FORM design point as center of the sampling density in standard normal space.
Each random variable is transformed between physical and normal spaces using its distribution object
(`PROB_BASEVAR$getlDistr()`), which provides parameterized CDF and inverse-CDF functions:

\deqn{u = \Phi^{-1}(F_X(x))}
\deqn{x = F_X^{-1}(\Phi(u))}

Parallel execution is supported via either base R's **parallel** package or the **future** backend.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parallel_helpers.R
\name{make_parallel_cluster}
\alias{make_parallel_cluster}
\title{Create a parallel cluster (fork on *nix, PSOCK on Windows)}
\usage{
make_parallel_cluster(use_threads, seed = NULL)
}
\arguments{
\item{use_threads}{Integer \>=1 – number of workers.}

\item{seed}{Optional integer or L'Ecuyer‑CMRG stream (length 7) for reproducibility.}
}
\value{
A cluster object (or NULL when the backend is "future").
}
\description{
Returns a `parallel` cluster object that already has the
  RNG streams (L'Ecuyer‑CMRG) and required packages loaded.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clsParam.R
\docType{class}
\name{PARAM_DETVAR-class}
\alias{PARAM_DETVAR-class}
\alias{PARAM_DETVAR}
\title{Parametric Deterministic Variable (PARAM_DETVAR)}
\description{
Like `PROB_DETVAR` but the deterministic value itself can be swept.
The class inherits from `PROB_BASEVAR` and therefore re-uses the same
transformation machinery.
}
\section{Fields}{

\describe{
\item{\code{ParamValues}}{Vector of deterministic values that will be assigned
sequentially in a parametric sweep.}

\item{\code{pos}}{Current index (used internally).}
}}


\examples{
pdet <- PARAM_DETVAR(
  Name = "E_mod",
  ParamValues = c(30e3, 35e3, 40e3)
)
pdet$nextParam() # sets Mean = 30.000, Sd = Mean/1e7 and prepares
pdet$getCurrentParam()

}
\author{
(C) 2021-2026 K. Nille-Hauf, T. Feiri, M. Ricker, T. Lux -- Hochschule Biberach (until 2022),
TU Dortmund University - Chair of Structural Concrete (since 2023)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clsModelContainer.R
\docType{class}
\name{PROB_BASEVAR-class}
\alias{PROB_BASEVAR-class}
\alias{PROB_BASEVAR}
\title{Object to store the distribution model for basic variables (PROB_BASEVAR)}
\description{
Stores the distribution model for a basic random variable that can be used
in any limit-state function.  The class knows how to transform between
*mean / standard deviation* and the native distribution parameters.
}
\details{
The method `prepare()` calculates missing parameters and checks for consistency.
`getlDistr()` returns a list containing *four* functions (`d`, `p`, `q`, `r`)
together with meta-information; the result is cached to avoid repeated
`loadNamespace()` calls.
}
\section{Fields}{

\describe{
\item{\code{Id}}{Integer identifier (position inside the LSF vector).}

\item{\code{Name}}{Human-readable name (e.g. `"f_ck"`).}

\item{\code{Description}}{Short textual description.}

\item{\code{DistributionType}}{Distribution identifier (`"norm"`, `"lnorm"`, `"slnorm"`, `"gumbel"`,
`"gamma"`, `"exp"`, `"beta"`, `"st"`, `"binom"`,`"emp"`,
`"unif"`, `"weibull"`, `"lt"`).}

\item{\code{Mean}}{Mean value of the variable.}

\item{\code{Sd}}{Standard deviation.}

\item{\code{Cov}}{Coefficient of variation (`Sd / Mean`).}

\item{\code{x0}}{Shifting parameter (used for shifted log-normal distribution `"slnorm"`).}

\item{\code{Package}}{Name of the R package that provides the density
functions (`"stats"`, `"evd"`, `"EnvStats"` or `"brms"`).}

\item{\code{DistributionParameters}}{Numeric vector with the *native* distribution
parameters (e.g. `c(mu, sigma)` for a normal).}

\item{\code{.cache}}{Private environment that stores cached density
functions (used internally, **do not** access
from user code).}
}}

\section{Methods}{

\describe{
\item{\code{prepare()}}{Performs transformations between mean/sd and distribution parameters.
Ensures that Mean, Sd and Cov are consistent and valid.}
}}

\examples{
## Normal random variable --------------------------------------------------
var1 <- PROB_BASEVAR(
  Name = "f_ck", DistributionType = "norm",
  Mean = 30, Sd = 1
)
var1$prepare()

## Gumbel random variable (package `evd`) ----------------------------------
var2 <- PROB_BASEVAR(
  Name = "M", DistributionType = "gumbel",
  Package = "evd", Mean = 2000, Sd = 200
)
var2$prepare()

}
\author{
(C) 2021-2026 K. Nille-Hauf, T. Feiri, M. Ricker, T. Lux -- Hochschule Biberach (until 2022),
TU Dortmund University - Chair of Structural Concrete (since 2023)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MVFOSM.R
\name{MVFOSM}
\alias{MVFOSM}
\title{Mean-Value First-Order Second-Moment (MVFOSM)}
\usage{
MVFOSM(lsf, lDistr, h = 1e-04, isExpression = FALSE, debug.level)
}
\arguments{
\item{lsf}{LSF Definition, can be Expression or Function. Defined by the FLAG isExpression (see below)}

\item{lDistr}{List of Distributions}

\item{h}{If isExpression is False, than Finite Difference Method is used for partial deviation. h is the Windowsize}

\item{isExpression}{Boolean, If TRUE lsf has to be typeof expression, otherwise lsf has to be type of function()}

\item{debug.level}{If 0 no additional info if 2 high output during calculation}
}
\value{
MVFOSM returns an object containing the following elements:
\itemize{
  \item \code{beta}: Estimated reliability index.
  \item \code{pf}: Estimated probability of failure.
  \item \code{design.point}: Design point in the original \eqn{x}-space.
  \item \code{alphas}: Direction cosines (importance factors) in the standard normal space.
  \item \code{runtime}: Total runtime of the algorithm.
}
}
\description{
Mean-Value First-Order Second-Moment (MVFOSM) method for the
approximation of failure probabilities in structural reliability analysis.

The MVFOSM method linearises the limit-state function at the mean values
of the basic random variables and estimates the reliability index
and probability of failure based on first-order moment information.

This classical approach provides a computationally efficient
approximation but may be inaccurate for strongly nonlinear
limit-state functions.
}
\references{
Freudenthal, A. M. (1956).
Safety and the probability of structural failure.
\emph{Transactions of the American Society of Civil Engineers},
121, 1337-1397.
}
\author{
(C) 2021-2026 K. Nille-Hauf, T. Feiri, M. Ricker, T. Lux -- Hochschule Biberach (until 2022),
TU Dortmund University - Chair of Structural Concrete (since 2023)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MCIS_helpers.R
\name{as_record_df}
\alias{as_record_df}
\title{Convert recorder lists into tidy data frame after completion}
\usage{
as_record_df(rec, k)
}
\arguments{
\item{rec}{Recorder list created by [make_recorders()]}

\item{k}{Number of completed iterations/batches to include in output table.}
}
\value{
Data frame containing columns: n_sim,pf,var,cov,I_sum,time_user,time_sys,time_elapsed,...
}
\description{
#'
Converts all recorded variables up to iteration \code{k} into a compact table format suitable for plotting or result export.

Used internally after convergence or stop criterion is reached in Monte Carlo simulations.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MCIS_helpers.R
\name{record_step}
\alias{record_step}
\title{Update one recorder entry during main loop iteration}
\usage{
record_step(rec, idx, pf_mc, var_mc, I_mc, n_sim, tic)
}
\arguments{
\item{idx}{Current iteration index within main loop.}

\item{pf_mc, var_mc, I_mc, n_sim}{Current cumulative sums from Monte Carlo batches.}

\item{tic}{Start time reference from \code{proc.time()} at simulation start.}
}
\value{
Updated recorder list object with new values inserted at position \code{idx}.
}
\description{
Records current cumulative sums and timing information at iteration index \code{idx}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SORM.R
\name{SORM}
\alias{SORM}
\title{Second-Order Reliability Method (SORM)}
\usage{
SORM(lsf, lDistr, debug.level = 0)
}
\arguments{
\item{lsf}{objective function with limit state function in form of
\code{function(x) \{ x[1] + x[2] + ... \}}.}

\item{lDistr}{list ob distribiutions regarding the distribution object of TesiproV}

\item{debug.level}{If 0 no additional info if 2 high output during calculation}
}
\value{
SORM returns an object containing the following elements:
\itemize{
  \item \code{beta}: Hasofer-Lind reliability index.
  \item \code{pf}: Estimated probability of failure.
  \item \code{u_points}: Design point(s) in the standard normal space (\eqn{u}-space).
  \item \code{dy}: Gradient(s) of the limit-state function at the design point.
}
}
\description{
Second-Order Reliability Method (SORM) for the approximation
of failure probabilities in structural reliability analysis.

The SORM extends the First-Order Reliability Method (FORM)
by incorporating second-order information of the limit-state function
at the design point. By accounting for local curvature effects,
SORM generally provides more accurate failure probability estimates
for nonlinear limit-state functions.

The method evaluates the principal curvatures of the limit-state surface
in standard normal space and applies an asymptotic correction
to the FORM probability of failure.
}
\references{
Breitung, K. (1989).
Asymptotic approximations for probability integrals.
\emph{Probabilistic Engineering Mechanics}, 4(4), 187-190.

Cai, G. Q., & Elishakoff, I. (1994).
Refined second-order reliability analysis.
\emph{Structural Safety}, 14(4), 267-276.

Hohenbichler, M., Gollwitzer, S., Kruse, W., & Rackwitz, R. (1987).
New light on first- and second-order reliability methods.
\emph{Structural Safety}, 4, 267-284.

Tvedt, L. (1990).
Distribution of quadratic forms in normal space -
Applications to structural reliability.
\emph{Journal of Engineering Mechanics}, 116(6), 1183-1197.

Marelli, S., & Sudret, B. (2014).
UQLab: A framework for uncertainty quantification in Matlab.
In \emph{Proceedings of the 2nd International Conference on
Vulnerability, Risk Analysis and Management (ICVRAM2014)},
Liverpool, United Kingdom, 2554-2563.

Lacaze, S., & Missoum, S. (2015).
CODES: A toolbox for computational design (Version 1.0).
Retrieved from \url{http://www.codes.arizona.edu/toolbox/}

Wu, X. Z. (2017).
Implementing statistical fitting and reliability analysis
for geotechnical engineering problems in R.
\emph{Georisk: Assessment and Management of Risk for Engineered Systems and Geohazards},
11(2), 173-188.
}
\author{
(C) 2021-2026 K. Nille-Hauf, T. Feiri, M. Ricker, T. Lux -- Hochschule Biberach (until 2022),
TU Dortmund University - Chair of Structural Concrete (since 2023)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FORM.R
\name{FORM}
\alias{FORM}
\title{First-Order Reliability Method (FORM)}
\usage{
FORM(
  lsf,
  lDistr,
  n_optim = 10,
  loctol = 0.01,
  optim_type = "rackfies",
  debug.level = 0
)
}
\arguments{
\item{lsf}{Objective function representing the limit-state, e.g. \code{function(R,E){R-E}}.
Supplied automatically by a SYS_ object - do not provide manually.}

\item{lDistr}{List of distribution objects created by TesiproV. Supplied automatically by
a SYS_ object - do not provide manually.}

\item{n_optim}{Number of optimization cycles (not required for Lagrangian algorithms).}

\item{loctol}{Local tolerance for convergence of the solver algorithm.}

\item{optim_type}{Optimization type: `"auglag"` (Augmented Lagrangian) or `"rackfies"`
(Rackwitz-Fiessler iterative scheme).}

\item{debug.level}{Verbosity level: 0 = silent, 1 = basic info, 2 = detailed output.}
}
\value{
A list containing:
  * `beta` - Hasofer-Lind reliability index
  * `pf` - probability of failure
  * `x_points` - design point in physical space
  * `dy` - gradient vector at design point
}
\description{
First-Order Reliability Method (FORM) for the approximation
of failure probabilities in structural reliability analysis.

The FORM estimates the probability of failure by transforming
the basic random variables into standard normal space
and approximating the limit-state function by a first-order
(linear) Taylor expansion at the design point.

The reliability index is defined as the minimum distance
from the origin to the limit-state surface in standard normal space.
The probability of failure is then approximated using
the standard normal cumulative distribution function.
}
\references{
Hasofer, A. M., & Lind, N. C. (1974).
An exact and invariant first-order reliability format.
\emph{Journal of the Engineering Mechanics Division, ASCE},
100(1), 111-121.

Rackwitz, R., & Fieszler, B. (1978).
Structural reliability under combined random load sequences.
\emph{Computers & Structures}, 9(5), 489-494.
}
\author{
(C) 2021-2026 K. Nille-Hauf, J.P. Schulze-Ardey, T. Feiri, M. Ricker, T. Lux -- Hochschule Biberach (until 2022),
TU Dortmund University - Chair of Structural Concrete (since 2023)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zzz.R
\name{.onLoad}
\alias{.onLoad}
\title{Package startup - register default parallelisation preferences}
\usage{
.onLoad(libname, pkgname)
}
\arguments{
\item{libname}{Ignored. Path to the library where the package is installed
(standard argument for `.onLoad`).}

\item{pkgname}{Ignored. Name of the package (standard argument for `.onLoad`).}
}
\description{
This internal function is executed automatically when the
**TesiproV** package is loaded (via `library(TesiproV)`).

Setting package-specific default options that can later be used by TesiproV
functions to configure parallel
execution in a controlled and CRAN-compliant way.

Two global options are defined:

* **`TesiproV.future.default`** - the default parallel backend.
  The default is `"multisession"`, which works on all operating systems
  including Windows.

* **`TesiproV.max_workers.default`** - the default upper limit for the
  number of parallel workers. The default is `4L`.

These values serve as internal fallback settings and may be overridden
explicitly by user-level function arguments or environment variables
within TesiproV functions.

The function returns `invisible(NULL)` and is **not exported**. It is
intended solely for internal package initialisation.
}
\details{
The function performs two steps:

1. It checks whether the namespace of the package `future` is available
   and aborts with a clear error message if it is missing.
2. It registers internal default options via `options()` without modifying
   the active `future` plan.

By not calling `future::plan()` during package startup, TesiproV avoids
altering the user's global parallel configuration. This follows CRAN
recommendations that packages should not change global execution plans
on load.
}
\examples{
## Load the package (initialises default options)
library(TesiproV)

}
\author{
(C) 2021-2026 K. Nille-Hauf, T. Feiri, M. Ricker, T. Lux -- Hochschule Biberach (until 2022),
TU Dortmund University - Chair of Structural Concrete (since 2023)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MCIS_helpers.R
\name{MC_IS_gsys}
\alias{MC_IS_gsys}
\title{Monte Carlo Importance Sampling using smooth system aggregation}
\usage{
MC_IS_gsys(
  sys_input,
  sys_structure = "serial",
  smooth_kappa = 20,
  cov_user = 0.05,
  n_batch = 5000,
  n_max = 1e+06,
  use_threads = 4,
  backend = "future",
  dataRecord = TRUE,
  densityType = "norm",
  dps = NULL,
  debug.level = 0,
  seed = NULL,
  adaptive_alpha = FALSE,
  alpha_update_rate = 0.1,
  adaptive_batch = FALSE,
  batch_control = list(),
  min_adapt_samples = 10000,
  alpha_min = 0.02,
  stability_mode = "robust"
)
}
\arguments{
\item{sys_input}{List of \code{SYS_LSF} objects defining the system.}

\item{sys_structure}{System aggregation type: \code{"serial"},
\code{"parallel"}, or a custom function.}

\item{smooth_kappa}{Smoothing parameter controlling the
approximation of min/max aggregation.}

\item{cov_user}{Target coefficient of variation (CoV) for the Monte Carlo
estimator. Simulation stops once this threshold is reached.}

\item{n_batch}{Number of samples generated per iteration.}

\item{n_max}{Maximum total number of samples (upper stopping limit).}

\item{use_threads}{Number of worker threads (for parallel execution).}

\item{backend}{Parallel backend, either \code{"future"} or \code{"parallel"}.}

\item{dataRecord}{Logical; if \code{TRUE}, intermediate results per iteration
are stored and returned.}

\item{densityType}{Sampling density type (currently \code{"norm"} supported).}

\item{dps}{Optional vector of design points in physical space; if supplied,
FORM analysis is skipped.}

\item{debug.level}{Integer verbosity level (0 = silent, 1 = summary, 2 = detailed).}

\item{seed}{Optional integer seed for reproducible random numbers.}

\item{adaptive_alpha}{Logical; if \code{TRUE}, adaptive mixture weights are used
(system case only).}

\item{alpha_update_rate}{Damping factor \eqn{\lambda \in [0,1]} controlling
adaptation speed of mixture weights.}

\item{adaptive_batch}{Logical; if TRUE, the batch size is adapted dynamically
based on the current effective sample size (ESS) and target RSE.}

\item{batch_control}{List controlling adaptive batch behaviour:
\describe{
  \item{n_min}{Minimum batch size}
  \item{n_max}{Maximum batch size}
  \item{K_future}{Desired number of remaining iterations}
}}

\item{min_adapt_samples}{Integer. Minimum total number of MC samples required before adaptive mixtures are updated.
The adaption mechanism is activated only after this threshold has been exceeded to ensure statistically stable estimates.}

\item{alpha_min}{Numeric scalar in \eqn{(0,1)}.
Lower bound for mixture weights \eqn{\alpha_i} during adaptive
updating. This prevents individual components of the importance
sampling density from collapsing to zero and ensures numerical
stability in multimodal system analyses.}

\item{stability_mode}{Character string specifying the numerical
  stabilization strategy used for importance sampling weight
  accumulation.

  \describe{
    \item{"robust"}{
      Default. Uses full log-domain accumulation of weights
      across Monte Carlo iterations. This improves numerical
      stability for high-dimensional systems and very small
      failure probabilities.
    }
    \item{"fast"}{
      Uses shifted exponentiation with batch-level scaling.
      Computationally equivalent under normal engineering
      conditions but may be slightly less stable in extreme
      rare-event scenarios.
    }
  }

  Both modes are mathematically equivalent for typical
  reliability problems. The \code{"robust"} mode is recommended
  for research applications and extreme reliability levels.}
}
\value{
Result object returned by \code{MC_IS_single()} with additional
metadata describing the system configuration.
}
\description{
Performs reliability estimation for systems of limit-state functions
by constructing a **smooth global system limit-state function**
and applying the existing single-LSF importance sampling algorithm.
}
\details{
The system limit-state function is built using \code{build_gsys()},
which aggregates component limit states via smooth min/max operators.

This approach converts a system reliability problem into a standard
single limit-state reliability problem.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clsModelContainer.R
\docType{class}
\name{PROB_MACHINE-class}
\alias{PROB_MACHINE-class}
\alias{PROB_MACHINE}
\title{Reliability Algorithm (PROB_MACHINE)}
\description{
Stores the definition of a reliability method (FORM, SORM, Monte-Carlo,
etc.) that can be executed by `SYS_PROB` or `SYS_PARAM`.  The actual
algorithm is called via the character string `fCall` (e.g. `"FORM"` or
`"MC_IS"`).  Optional arguments are passed through the list `options`.
}
\details{
The helper method `getMethodLevel()` returns an integer that indicates
the *complexity* of the method (1 = first-order, 2 = second-order, 3 = Monte-Carlo).
}
\section{Fields}{

\describe{
\item{\code{name}}{Descriptive name of the machine (appears in reports).}

\item{\code{fCall}}{Function name that implements the algorithm (must be
available in the package namespace, e.g. `FORM`,
`SORM`, `MC_CRUDE`, `MC_IS`, `MC_SubSam`).}

\item{\code{options}}{Additional options for the algorithm (list, e.g.
`list(n_max = 1e6, cov_user = 0.05)`).}
}}


\examples{
## FORM machine ------------------------------------------------------------
form_rf <- PROB_MACHINE(
  name = "FORM - Rack-Fies",
  fCall = "FORM",
  options = list(n_optim = 20, loctol = 1e-3, optim_type = "rackfies")
)

## Monte-Carlo importance sampling -----------------------------------------
mcis <- PROB_MACHINE(
  name = "MC-IS",
  fCall = "MC_IS",
  options = list(cov_user = 0.05, n_max = 3e5, seed = 1234)
)

}
\author{
(C) 2021-2026 K. Nille-Hauf, T. Feiri, M. Ricker, T. Lux -- Hochschule Biberach (until 2022),
TU Dortmund University - Chair of Structural Concrete (since 2023)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/logStudentT.R
\name{LogStudentT}
\alias{LogStudentT}
\alias{dlt}
\alias{logStudentT}
\alias{lt}
\alias{plt}
\alias{qlt}
\alias{rlt}
\title{Log-Student t Distribution}
\usage{
dlt(x, hyper.param)

plt(q, hyper.param)

qlt(p, hyper.param)

rlt(n_vals, hyper.param)
}
\arguments{
\item{x, q}{Vector of quantiles.}

\item{hyper.param}{Numeric vector \eqn{(m, s, n, \nu)}.}

\item{p}{Vector of probabilities.}

\item{n_vals}{Number of random values to generate.}
}
\value{
\itemize{
  \item \code{dlt()} returns the density.
  \item \code{plt()} returns the distribution function.
  \item \code{qlt()} returns the quantile function.
  \item \code{rlt()} generates random deviates.
}
}
\description{
Density, distribution function, quantile function and random generation
for the log-Student t distribution.
}
\details{
The cumulative distribution function is given by
\deqn{
F_X(x) =
F_{t_{\nu}}\left(
\frac{\ln(X/m)}{s}
\sqrt{\frac{n}{n+1}}
\right),
}
where \eqn{F_{t_{\nu}}(\cdot)} denotes the cumulative distribution
function of a log-Student t distribution with \eqn{\nu} degrees of freedom.

The log-Student t distribution arises as the marginal
distribution of a log-normal distributed variable with conjugate
log-normal-gamma prior uncertainty in \eqn{m} and \eqn{s}.

The hyperparameter vector \code{hyper.param} is defined as
\deqn{(m, s, n, \nu)}
where
\itemize{
  \item \eqn{m} is the mean value of an equivalent sample of size \eqn{n},
  \item \eqn{s} is the empirical standard deviation of an equivalent sample
        of size \eqn{\nu + 1},
  \item \eqn{\nu} denotes the degrees of freedom of the Student t distribution.
}

For details see:
\url{https://www.jcss-lc.org/publications/jcsspmc/concrete.pdf}
}
\examples{
hp <- c(3.4, 0.14, 3, 10)
#' dlt(0.5, hp)

}
\author{
(C) 2021-2026 K. Nille-Hauf, T. Feiri, M. Ricker, T. Lux --
Hochschule Biberach (until 2022), TU Dortmund University - Chair of Structural Concrete (since 2023)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MC_SubSam.R
\name{MC_SubSam}
\alias{MC_SubSam}
\title{Monte Carlo Simulation with Subset-Sampling}
\usage{
MC_SubSam(
  lsf,
  lDistr,
  Nsubset = 1e+05,
  p0 = 0.1,
  MaxSubsets = 10,
  Alpha = 0.05,
  variance = "uniform",
  debug.level = 0,
  seed = NULL
)
}
\arguments{
\item{lsf}{limit-state function}

\item{lDistr}{list of basevariables in input space}

\item{Nsubset}{number of samples in each simulation level}

\item{p0}{level probability or conditional probability}

\item{MaxSubsets}{maximum number of simulation levels that are used to terminate the simulation procedure to avoid infinite loop when the target domain cannot be reached}

\item{Alpha}{confidence level}

\item{variance}{gaussian, uniform}

\item{debug.level}{If 0 no additional info if 2 high output during calculation}

\item{seed}{Optional integer value. If specified, sets a fixed seed for reproducible random numbers (useful for testing).}
}
\value{
MC_SubSam returns an object containing the following elements:
\itemize{
  \item \code{beta}: Estimated reliability index.
  \item \code{pf}: Estimated probability of failure.
  \item \code{betaCI}: Confidence interval of the reliability index.
  \item \code{pfCI}: Confidence interval of the probability of failure.
  \item \code{CoV}: Coefficient of variation of the estimator.
  \item \code{NumOfSubsets}: Number of Markov chains (subsets).
  \item \code{NumOfEvalLSF_nom}: Nominal number of limit-state function evaluations
        (Markov chains / iterations).
  \item \code{NumOfEvalLSF_eff}: Effective number of limit-state function evaluations.
  \item \code{runtime}: Total runtime of the function.
}
}
\description{
MonteCarlo with Subset-Sampling (MC_SubSam) method for the
estimation of small failure probabilities in structural reliability analysis.

Subset Simulation is an advanced Monte Carlo technique that
expresses a rare failure event as a sequence of intermediate
conditional events with higher probabilities.

The method combines conditional sampling with Markov Chain Monte Carlo (MCMC)
techniques to efficiently estimate very small probabilities of failure,
which would be computationally expensive to obtain using crude Monte Carlo simulation.
}
\references{
Au, S. K., & Beck, J. L. (2001).
Estimation of small failure probabilities in high dimensions
by subset simulation.
\emph{Probabilistic Engineering Mechanics}, 16(4), 263-277.

Marelli, S., & Sudret, B. (2014).
UQLab: A framework for uncertainty quantification in Matlab.
In \emph{Proceedings of the 2nd International Conference on
Vulnerability, Risk Analysis and Management (ICVRAM2014)},
Liverpool, United Kingdom, 2554-2563.
}
\author{
(C) 2021-2026 K. Nille-Hauf, T. Feiri, M. Ricker, T. Lux -- Hochschule Biberach (until 2022),
TU Dortmund University - Chair of Structural Concrete (since 2023)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MCIS_helpers.R
\name{parallel_dispatch}
\alias{parallel_dispatch}
\title{Internal Parallel Dispatcher for Importance Sampling}
\usage{
parallel_dispatch(
  chunk_indices,
  worker_fun,
  u_mat,
  use_threads,
  backend,
  debug.level = 0
)
}
\arguments{
\item{chunk_indices}{List of integer vectors defining sample indices
assigned to each worker.}

\item{worker_fun}{Function that evaluates a given subset of samples.}

\item{u_mat}{Matrix of samples in standard normal space.}

\item{use_threads}{Number of worker threads.}

\item{backend}{Character string specifying backend
(\code{"future"} or \code{"parallel"}).}

\item{debug.level}{Verbosity level for diagnostic output.}
}
\value{
A list containing worker evaluation results.
}
\description{
Internal helper function used by \code{MC_IS_single()} to distribute
Monte Carlo sample blocks across worker threads in a reproducible and
backend-independent manner.

The function ensures worker-invariant execution by:
\itemize{
  \item Performing deterministic chunking of sample indices,
  \item Avoiding per-worker random number generation,
  \item Supporting both \code{"future"} and \code{"parallel"} backends.
}

This abstraction isolates parallel execution from the statistical
core of the Importance Sampling algorithm.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MCIS_helpers.R
\name{compute_weight_update}
\alias{compute_weight_update}
\title{Internal weight accumulation for Monte Carlo Importance Sampling (System Case)}
\usage{
compute_weight_update(all_log_w, all_I_vals, stability_mode, sys_type, state)
}
\arguments{
\item{all_log_w}{Numeric vector of log-weights for the current batch.}

\item{all_I_vals}{Numeric vector of failure indicators (0/1).}

\item{stability_mode}{Character string: `"fast"` or `"robust"`.}

\item{sys_type}{Character string: `"serial"` or `"parallel"`.}

\item{state}{List containing current accumulated sums:
\describe{
  \item{total_sum_w}{Cumulative sum of weights.}
  \item{total_sum_Iw}{Cumulative weighted failure sum.}
  \item{total_sum_w2}{Cumulative sum of squared weights.}
  \item{log_total_sum_w}{Log-domain cumulative weight sum (robust mode).}
  \item{log_total_sum_Iw}{Log-domain cumulative failure-weight sum (robust mode).}
  \item{log_total_sum_w2}{Log-domain cumulative squared-weight sum (robust mode).}
  \item{n_sim}{Total number of samples processed.}
}}
}
\value{
Updated `state` list with accumulated statistics.
}
\description{
Internal helper used by `MC_IS_system()` to accumulate importance sampling
weights and update running Monte Carlo statistics.

The function performs numerically stable accumulation of:
\itemize{
  \item \eqn{\sum w},
  \item \eqn{\sum I w},
  \item \eqn{\sum w^2},
  \item and (in robust mode) their log-domain equivalents.
}

Two stabilization modes are supported:

\describe{
  \item{\code{"fast"}}{
    Batch-level stabilization using a max-log shift:
    \deqn{w_i = \exp(\log w_i - \max(\log w))}
    Suitable for most engineering applications.
  }
  \item{\code{"robust"}}{
    Global log-sum-exp accumulation across iterations:
    \deqn{\log \sum w = m + \log \sum \exp(\log w - m)}
    Recommended for rare-event simulation or very small failure probabilities.
  }
}

The function updates the running state object and returns it.
}
\details{
This function is internal and not intended to be called directly by users.
It is part of the system-level importance sampling implementation
in `MC_IS_system()`.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MCIS_helpers.R
\name{make_recorders}
\alias{make_recorders}
\title{Create empty recorder vectors for MC iterations}
\usage{
make_recorders(n_iter)
}
\arguments{
\item{n_iter}{Expected maximum number of iterations/batches in the simulation loop.}
}
\value{
A list containing numeric vectors (`pf`, `var`, `cov`, etc.) and a nested time data frame with user/system/elapsed times per iteration.
}
\description{
Preallocates numeric vectors and time tracking data frames used to record intermediate results during simulation loops.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TesiproV.R
\docType{package}
\name{TesiproV-package}
\alias{TesiproV}
\alias{TesiproV-package}
\title{TesiproV: A package for the calculation of reliability and failure probability in civil engineering}
\description{
The Package provides three main types of objects:
\enumerate{
\item Objects for modeling base variables
\item Objects for modeling limit state functions and systems of them
\item Objects for modeling solving algorithms
}
}
\details{
By creating and combining those objects, one is able to model quite complex problems in terms of structural reliablity calculation.
For normally distributed variables there might be an workflow to calculate correlated problems (but no systems then).
There is also implemented a new distribution (logStudentT, often used for conrete compression strength) to show how one can implement
your very own or maybe combined multi modal distribution and use it with TesiproV.
}
\section{Objects for base variables}{

\code{PROB_BASEVAR}, \code{PROB_DETVAR}, \code{PARAM_BASEVAR}, \code{PARAM_DETVAR}
}

\section{Limit state functions}{

\code{SYS_LSF}, \code{PROB_SYS}, \code{PARAM_SYS}
}

\section{Solving algorithms}{

\code{PROB_MACHINE}
}

\seealso{
Useful links:
\itemize{
  \item \url{https://mvb.ab.tu-dortmund.de/}
}

}
\author{
(C) 2021-2026 K. Nille-Hauf, T. Feiri, M. Ricker, T. Lux -- Hochschule Biberach (until 2022),
TU Dortmund University - Chair of Structural Concrete (since 2023)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MCIS_helpers.R
\name{resolve_future_seed}
\alias{resolve_future_seed}
\title{Resolve and validate RNG seed for Future backend}
\usage{
resolve_future_seed(seed)
}
\arguments{
\item{seed}{Optional seed value provided by user or upstream function.}
}
\value{
A normalized object suitable for direct use in the argument
        \code{future.seed = resolve_future_seed(seed)} inside
        [future.apply::future_lapply()].
}
\description{
Internal helper to normalize the \code{seed} argument used in
\link[future.apply]{future_lapply} calls within Monte Carlo simulations.

The function ensures that the supplied seed is of a valid type:
either a single integer (for deterministic runs),
a complete L'Ecuyer-CMRG RNG stream (as returned by \link[parallel]{nextRNGStream}),
or \code{NULL}/\code{TRUE} for fully random behavior.

This avoids common errors such as:
\itemize{
  \item "Argument 'seed' must be L'Ecuyer-CMRG RNG seed ..." when passing incomplete streams,
  \item identical random numbers on all workers due to improper seeding,
  \item non-reproducible results across different backends.
}

May be one of:
\itemize{
  \item A single numeric/integer value (e.g. 1234) - fixed reproducible run.
  \item A numeric vector of length 7 (L'Ecuyer-CMRG stream) - substreams per worker.
  \item NULL - automatically use fully random streams (\code{future.seed = TRUE}).
 }
}
\seealso{
[TesiproV::MC_IS()], [parallel::nextRNGStream()], [future.apply::future_lapply()]
}
\author{
(C) 2026 M. Ricker - TU Dortmund University - Chair of Structural Concrete
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/StudentT.R
\name{StudentT}
\alias{StudentT}
\alias{dst}
\alias{st}
\alias{pst}
\alias{qst}
\alias{rst}
\title{Student t Distribution}
\usage{
dst(x, hyper.param)

pst(q, hyper.param)

qst(p, hyper.param)

rst(n_vals, hyper.param)
}
\arguments{
\item{x, q}{Vector of quantiles.}

\item{hyper.param}{Numeric vector \eqn{(m, s, n, \nu)}.}

\item{p}{Vector of probabilities.}

\item{n_vals}{Number of random values to generate.}
}
\value{
\itemize{
  \item \code{dst()} returns the density.
  \item \code{pst()} returns the distribution function.
  \item \code{qst()} returns the quantile function.
  \item \code{rst()} generates random deviates.
}
}
\description{
Density, distribution function, quantile function and random generation
for the Student t distribution.
}
\details{
The cumulative distribution function is given by
\deqn{
F_X(x) =
F_{t_{\nu}}\left(
\frac{x-m}{s}
\sqrt{\frac{n}{n+1}}
\right),
}
where \eqn{F_{t_{\nu}}(\cdot)} denotes the cumulative distribution
function of a Student t distribution with \eqn{\nu} degrees of freedom.

The Student t distribution arises as the marginal
distribution of a normal distributed variable with conjugate
normal-gamma prior uncertainty in \eqn{m} and \eqn{s}.

The hyperparameter vector \code{hyper.param} is defined as
\deqn{(m, s, n, \nu)}
where
\itemize{
  \item \eqn{m} is the mean value of an equivalent sample of size \eqn{n},
  \item \eqn{s} is the empirical standard deviation of an equivalent sample
        of size \eqn{\nu + 1},
  \item \eqn{\nu} denotes the degrees of freedom of the Student t distribution.
}

For details see:
\url{https://www.rdocumentation.org/packages/stats/versions/3.6.2/topics/TDist}
}
\examples{
hp <- c(40, 4, 3, 10)
dst(0.5, hp)

}
\author{
(C) 2021-2026 K. Nille-Hauf, T. Feiri, M. Ricker, T. Lux --
Hochschule Biberach (until 2022), TU Dortmund University - Chair of Structural Concrete (since 2023)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MC_IS.R
\name{MC_IS}
\alias{MC_IS}
\title{Monte Carlo Simulation with Importance Sampling}
\usage{
MC_IS(
  lsf,
  lDistr,
  cov_user = 0.05,
  n_batch = 5000,
  n_max = 1e+06,
  use_threads = 4,
  backend = NULL,
  sys_type = "serial",
  dataRecord = TRUE,
  beta_l = 100,
  densityType = "norm",
  dps = NULL,
  debug.level = 0,
  seed = NULL,
  adaptive_alpha = FALSE,
  alpha_update_rate = 0.1,
  adaptive_batch = FALSE,
  batch_control = list(),
  min_adapt_samples = 10000,
  alpha_min = 0.02,
  stability_mode = c("robust", "fast")
)
}
\arguments{
\item{lsf}{A single limit-state function \code{function(x)} or a list of
such functions for system reliability. Must follow the LSF interface
specification described below.}

\item{lDistr}{List of marginal distribution objects (funlists with \code{$d},
\code{$p}, \code{$q}, \code{$r}) corresponding to input variables.}

\item{cov_user}{Target coefficient of variation (CoV) for the Monte Carlo
estimator. Simulation stops once this threshold is reached.}

\item{n_batch}{Number of samples generated per iteration.}

\item{n_max}{Maximum total number of samples (upper stopping limit).}

\item{use_threads}{Number of worker threads (for parallel execution).}

\item{backend}{Parallel backend, either \code{"future"} or \code{"parallel"}.}

\item{sys_type}{Character string, either \code{"serial"} or \code{"parallel"},
specifying system configuration.}

\item{dataRecord}{Logical; if \code{TRUE}, intermediate results per iteration
are stored and returned.}

\item{beta_l}{Optional threshold; limit states with \eqn{\beta > beta_l}
may be excluded in system analysis.}

\item{densityType}{Sampling density type (currently \code{"norm"} supported).}

\item{dps}{Optional vector of design points in physical space; if supplied,
FORM analysis is skipped.}

\item{debug.level}{Integer verbosity level (0 = silent, 1 = summary, 2 = detailed).}

\item{seed}{Optional integer seed for reproducible random numbers.}

\item{adaptive_alpha}{Logical; if \code{TRUE}, adaptive mixture weights are used
(system case only).}

\item{alpha_update_rate}{Damping factor \eqn{\lambda \in [0,1]} controlling
adaptation speed of mixture weights.}

\item{adaptive_batch}{Logical; if TRUE, the batch size is adapted dynamically
based on the current effective sample size (ESS) and target RSE.}

\item{batch_control}{List controlling adaptive batch behaviour:
\describe{
  \item{n_min}{Minimum batch size}
  \item{n_max}{Maximum batch size}
  \item{K_future}{Desired number of remaining iterations}
}}

\item{min_adapt_samples}{Integer. Minimum total number of MC samples required before adaptive mixtures are updated.
The adaption mechanism is activated only after this threshold has been exceeded to ensure statistically stable estimates.}

\item{alpha_min}{Numeric scalar in \eqn{(0,1)}.
Lower bound for mixture weights \eqn{\alpha_i} during adaptive
updating. This prevents individual components of the importance
sampling density from collapsing to zero and ensures numerical
stability in multimodal system analyses.}

\item{stability_mode}{Character string specifying the numerical
  stabilization strategy used for importance sampling weight
  accumulation.

  \describe{
    \item{"robust"}{
      Default. Uses full log-domain accumulation of weights
      across Monte Carlo iterations. This improves numerical
      stability for high-dimensional systems and very small
      failure probabilities.
    }
    \item{"fast"}{
      Uses shifted exponentiation with batch-level scaling.
      Computationally equivalent under normal engineering
      conditions but may be slightly less stable in extreme
      rare-event scenarios.
    }
  }

  Both modes are mathematically equivalent for typical
  reliability problems. The \code{"robust"} mode is recommended
  for research applications and extreme reliability levels.}
}
\value{
MC_IS returns an object containing the following elements:
\itemize{
  \item \code{method} - method identifier
  \item \code{pf} - estimated failure probability
  \item \code{beta} - reliability index \eqn{\beta = -\Phi^{-1}(P_f)}
  \item \code{var} - variance estimate
  \item \code{cov_mc} - coefficient of variation
  \item \code{n_mc} - number of samples used
  \item \code{ESS} - effective sample size
  \item \code{ESS_modes} - effective sample size per mixture component (system case)
  \item \code{data} - optional iteration history (if \code{dataRecord = TRUE})
}
}
\description{
Performs Monte Carlo simulation with Importance Sampling (IS) to estimate
structural failure probabilities for single limit state functions or
systems of limit state functions.

The method reduces the variance of crude Monte Carlo by sampling from a
shifted (or multimodal) density in standard normal space and correcting
via likelihood ratios. For systems, a multimodal sampling density based on
FORM design points is used (cf. Melchers & Beck, Chapter 5).

The implementation supports:
\itemize{
  \item Single limit state functions (FORM-based IS shift)
  \item Serial and parallel systems (union / intersection events)
  \item Multimodal sampling densities for system reliability
  \item Optional adaptive mixture weights (adaptive \eqn{\alpha_i})
  \item Parallel execution (via \pkg{future} or \pkg{parallel})
  \item Deterministic reproducibility via fixed seeds
  \item Automatic detection of vectorized limit-state functions
}
}
\details{
\strong{Single limit state case:}
The importance sampling density is constructed in standard normal space
using the FORM design point \eqn{u^*}. Samples are generated from
\eqn{u \sim \mathcal{N}(u^*, I)} and transformed to physical space.


\strong{System case:}
For a system with limit states \eqn{g_i}, the failure event is
\eqn{F = \cup_i F_i} (serial system) or
\eqn{F = \cap_i F_i} (parallel system).

The sampling density follows Melchers & Beck (Eq. 5.26):
\deqn{
  h(u) = \sum_{i=1}^m \alpha_i \, \phi(u - u_i^*)
}
where \eqn{u_i^*} are the FORM design points in standard normal space and
\eqn{\alpha_i} are mixture weights.

By default, mixture weights are proportional to individual failure
probabilities:
\deqn{
  \alpha_i \propto P(F_i).
}

If \code{adaptive_alpha = TRUE}, mixture weights are updated iteratively
based on weighted failure frequency estimates:
\deqn{
  \alpha_i^{new} =
  (1-\lambda)\alpha_i^{old} +
  \lambda \frac{\hat{P}(F_i)}{\sum_k \hat{P}(F_k)}
}
where \eqn{\lambda} is given by \code{alpha_update_rate}.

The failure probability is estimated using a self-normalized IS estimator:
\deqn{
  \hat{P}_f =
  \frac{\sum I(u) w(u)}{\sum w(u)}
}
with likelihood ratio
\deqn{
  w(u) = \frac{\phi(u)}{h(u)}.
}

Numerical stability is ensured via global log-weight stabilization.


\strong{Stopping Criterion}

The simulation terminates when both:
\itemize{
  \item The estimated coefficient of variation (CoV) falls below \code{cov_user}, and
  \item The effective sample size (ESS) exceeds a minimum threshold.
}

For parallel systems with very small failure probabilities, the ESS-based
safeguard prevents premature termination due to unstable variance estimates.


\strong{Numerical stability}

The default \code{"robust"} mode performs global log-domain
accumulation of importance sampling weights using a
log-sum-exp formulation. This avoids overflow and underflow
effects in high-dimensional or rare-event settings.

The alternative \code{"fast"} mode uses shifted exponentiation
at the batch level and is computationally equivalent for most
engineering reliability problems.


\strong{Automatic detection of vectorized limit-state functions}

For performance reasons, the algorithm automatically checks whether the
supplied limit-state function (LSF) supports vectorized evaluation.

A short internal test is performed before the simulation starts. If the LSF
accepts a matrix of input samples and returns a numeric vector of matching
length, it is evaluated in fully vectorized form:

\deqn{g(X_{1:n}) \rightarrow \{g(x_1), \dots, g(x_n)\}}

Otherwise, the algorithm falls back to row-wise evaluation using
\code{apply()}.

This mechanism is fully automatic and backward compatible.
Users do not need to modify existing scalar LSF definitions.

For computationally expensive LSFs (e.g., nonlinear models or surrogate FEM
models), vectorized evaluation can significantly improve performance.


\strong{Parallel execution and future plan handling}

When \code{backend = "future"} is used, the function respects the currently
active \pkg{future} plan.

If no parallel plan has been set by the user (i.e., the active plan is
sequential), a temporary plan is created internally using:

\itemize{
  \item \code{multicore} on Unix-like systems,
  \item \code{multisession} on Windows.
}

The original plan is automatically restored after completion of the
simulation.

If the user has already defined a parallel strategy via
\code{future::plan()}, it will not be modified.

This design ensures CRAN compliance while preserving full flexibility
for advanced users.


\strong{Reproducibility}

Reproducible results across different numbers of cores are ensured through:

\itemize{
  \item L'Ecuyer-CMRG random number streams,
  \item deterministic master sampling,
  \item worker-invariant parallel chunking,
  \item controlled seed propagation.
}

If \code{seed} is supplied, results are reproducible regardless of the
number of worker threads.
}
\section{Limit-State Function (LSF) Interface}{


The argument \code{lsf} must define a valid limit-state function
using one of the following supported signatures:

\describe{
  \item{\code{function(x)}}{
    A single numeric vector \code{x} containing all basic variables
    in the order defined in \code{lDistr}.

    This form is fully compatible with vectorized evaluation
    in Monte Carlo and importance sampling algorithms.
  }

  \item{\code{function(Z, Fy, M, ...)}}{
    Explicitly named scalar arguments corresponding to the variable
    names defined in the probabilistic model.
  }
}

The use of \code{function(...)} as the sole argument is not supported.
Such usage may cause inconsistent behaviour between FORM-based and
Monte Carlo-based reliability methods.

The vector form \code{function(x)} is recommended for
best performance and consistency.
}

\section{Windows users}{


On Windows systems, it is recommended to use
\code{backend = "future"} for multi-core parallelization.
The \code{"parallel"} backend does not support forking on Windows.
}

\references{
Ditlevsen, O., & Madsen, H. O. (1996).
\emph{Structural Reliability Methods}.
Wiley.

Melchers, R. E., & Beck, A. T. (2018).
\emph{Structural Reliability Analysis and Prediction}.
Wiley.

Spaethe, G. (1991).
\emph{Die Sicherheit tragender Baukonstruktionen}.
Springer.
}
\author{
(C) 2021-2026 K. Nille-Hauf, T. Feiri, M. Ricker, T. Lux -- Hochschule Biberach (until 2022),
TU Dortmund University - Chair of Structural Concrete (since 2023)
}
